#
# Copyright (C) 1996-1998 by the Board of Trustees
#    of Leland Stanford Junior University.
# 
# This file is part of the SimOS distribution. 
# See LICENSE file for terms of the license. 
#

###
### modes.tcl
###

FileSourced modes.tcl


if ![info exists modesByPID] {
    set modesByPID 0
}


timing create modes
log [timing fields]

proc modesSwitch args {
    global PROCESS PID CPU CELL M RUNNING_HIVE PARAM
    global modesByPID
    if $modesByPID {
        set args [linsert $args 0 PID$PID($CPU)]
    }

    if [info exists RUNNING_HIVE] {
        eval timing switch modes ${CELL}_$PID($CPU) cell$CELL $args
    } elseif { $PARAM(MACHINE.Count) == 1 } {
        eval timing switch modes $PID($CPU) $args
    } else {
        eval timing switch modes $M($CPU)_$PID($CPU) machine$M($CPU) $args
    }
}

###
### Initialization based on current state
###

annotation set simos enter {
    if [inUserMode] {
        modesSwitch
    } elseif {$SIMOS(ISA)=="ALPHA"} {
        if {$PROCESS($CPU) == "kernel idle"} {
            modesSwitch kernel idle
        } elseif {$pc & 1} {
            modesSwitch pal
        } else {
            modesSwitch kernel
        }
    } elseif {$PROCESS($CPU) == "idle"} {
        modesSwitch kernel idle
    } else {
        modesSwitch kernel
    }
}


###
### The Annotations
###

if {$SIMOS(ISA)=="ALPHA"} { 
    annotation set osEvent switchIn {
        modesSwitch kernel
    }
} else { 
    annotation set osEvent switchIn {
        modesSwitch kernel
    }
}


annotation set osEvent procexit {
    if [info exists RUNNING_HIVE] {
        timing terminate modes ${CELL}_$PID($CPU)
    } elseif { $PARAM(MACHINE.Count) == 1 } {
        timing terminate modes $PID($CPU)
    } else {
        timing terminate modes $M($CPU)_$PID($CPU)
    }
}

annotation set osEvent startKernel {
    timing start modes kernel
}

annotation set osEvent endKernel {
    timing end modes kernel
}

if {$SIMOS(ISA)=="ALPHA"} { 
    annotation set osEvent startPal {
        timing start modes pal
    }
    
    annotation set osEvent endPal {
        timing end modes pal
    }
}

annotation set osEvent startIdle {
    if { [timing current modes] != "idle"} {
	timing start modes idle
    }
}

annotation set osEvent endIdle {
    timing end modes idle
}

annotation set osEvent startSync {
    timing start modes sync
}

annotation set osEvent endSync {
    timing end modes sync
}

if [info exists RUNNING_HIVE] {
    annotation set osEvent startRpc {
        timing start modes rpcwait
    }

    annotation set osEvent endRpc {
        timing end modes rpcwait
    }
}

###
### Need to dump to stats 
###

if [info exists modesDumpPeriod] {
    annotation set periodic $modesDumpPeriod -tag modes {
        log [timing dump modes]
    }
} else {
    annotation set simos periodic -tag modes {
        log [timing dump modes]
    }
}

annotation set simos exit {
    timing exit modes
    log [timing dump modes]
}

