#
# Copyright (C) 1998 by the Board of Trustees
#    of Leland Stanford Junior University.
# Copyright (C) 1998 Digital Equipment Corporation
#
# This file is part of the SimOS distribution.
# See LICENSE file for terms of the license.
#

#
# Copyright (C) 1996-1998 by the Board of Trustees
#    of Leland Stanford Junior University.
# 
# This file is part of the SimOS distribution. 
# See LICENSE file for terms of the license. 
#

###
### This file is responsible for raising OS level annotations and tracking
### the current process.
###
###

####
#### init this module
####

FileSourced osf/os.tcl

source syscalls.tcl
source osfpal.tcl


# The endUser annotation fires twice before you get
# two startUser annotations. This is because of the
# mode bit is not changed for a tlb miss in palcode
# and we can take a second tlb miss - verghese
annotation type osEvent enum {
    startUser   endUser
    startKernel endKernel
    startPal    endPal
    startIdle   endIdle
    startSync   endSync
    switchIn    switchOut
    procstart   procexit   procexec  procwait
}

# must agree with cpus-alpha/alpha-shared/ev5.c
annotation type exc enum {
    reset mchk arith interrupt ndtb_miss pdtb_miss unalign dtb_fault itb_miss itb_acv opdec fen pal
}

proc inUserMode {} {
    global IPR
    set PS $IPR(0x10f)
    return [expr ($PS & 24) != 0]
}

####
#### the annotations 
####

##
## process and idle tracking
##

annotation set simos enter {
    set PCBB($CPU) $IPR(0x157)
    if {$PCBB($CPU)==0} {
	### only if booting
	### hack to make it unique across processors
	###
	set PCBB($CPU) [expr -1 * $CPU]
    }
    console "OS.TCl: entering for cpu $CPU. pc=$pc \n"
    set PID($CPU) $PCBB($CPU)
    if [catch {currentProcess}] {
        # The procedures called check exceptions.
        set PROCESS($CPU) [currentProcess]
        set OSPID($CPU) [currentPID]
        set OSTID($CPU) [currentTID]
    } elseif {(   ([hex $pc] >= [symbol read vmunix::idle_thread:START])
               && ([hex $pc] <= ([symbol read vmunix::idle_thread:END] + 4)))} {
        # idle
        set PROCESS($CPU) "kernel idle"
        ###set OSPID($CPU) idle$CPU
        set OSPID($CPU) [currentPID]
        set OSTID($CPU) [currentTID]
    } elseif {(   (![catch {symbol read vmunix::checkRunq:START}])
               && ([hex $pc] >= [symbol read vmunix::checkRunq:START])
               && ([hex $pc] <= ([symbol read vmunix::checkRunq:END] + 4)))} {
        # runq
        set PROCESS($CPU) "runq"
        ###set OSPID($CPU) runq$CPU
        set OSPID($CPU) [currentPID]
        set OSTID($CPU) [currentTID]
    } else {
        set PROCESS($CPU) [currentProcess]
        set OSPID($CPU) [currentPID]
        set OSTID($CPU) [currentTID]
    }

    console "OS.TCL entering for cpu=$CPU with pid=$OSPID($CPU) tid=$OSTID($CPU) pcbb=$PCBB($CPU) process=$PROCESS($CPU) \n"
}


proc newproc { pid tid process comment } {
    global CPU PROCESS PID PCBB OSPID OSTID CYCLES
    if {($PCBB($CPU) != $PID($CPU)) || ($tid != $OSTID($CPU)) || ($pid != $OSPID($CPU)) || ($process != $PROCESS($CPU))} { 
	if {$PROCESS($CPU) == "kernel idle"} {
	    annotation exec osEvent endIdle
	}
        annotation exec osEvent switchOut
        if {$PCBB($CPU) == $PID($CPU)} {
            ###console "XXX newproc.$CPU Switching from ($OSPID($CPU),$OSTID($CPU),$PROCESS($CPU)) to ($pid,$tid,$process) with same PCBB=$PCBB($CPU) $comment \n"               
        }
        set PROCESS($CPU) $process
        set OSPID($CPU) $pid
        set OSTID($CPU) $tid
        set PID($CPU) $PCBB($CPU)
        #console "XXX newproc.$CPU $comment $pid $tid ($PCBB($CPU)) $process \n"
        annotation exec osEvent switchIn
	if {$PROCESS($CPU) == "kernel idle"} {
	    annotation exec osEvent startIdle
	}
    } else {
        ###console "XXX newproc.$CPU Switching to the same (pid,tid,process)=($OSPID($CPU),$OSTID($CPU),$PROCESS($CPU)) and same PCBB=$PCBB($CPU) comment=$comment \n"               
    }
}

proc currentProcess {} {
    global CPU SIMOS
    if [catch {set currProc [currentProcessByCpu $CPU]}] {
        if {$SIMOS(RestoringCpt)==0} {
            # booting
            set currProc "boot"
        } else {
            set currProc "someprocess"
        }
    }
    return $currProc
}

proc currentProcessByCpu { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES
    set thread [symbol read "vmunix::active_threads<$cpu>"]
    ###
    ### The following were traversing swapable structures.
    ### Find a non-swapable path instead to avoid exceptions.
    ###
    ### set stack_layout [symbol read "vmunix::((struct thread *)$thread)->stack"]
    ### set task  [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.utask"]
    ### set newPROCESS [symbol read "vmunix::((struct utask *)$task)->uu_comm"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set newPROCESS [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_comm"]
    return $newPROCESS
}

proc currentPID {} {
    global CPU SIMOS
    if [catch {set currPID [currentPIDByCpu $CPU]}] {
        if {$SIMOS(RestoringCpt)==0} {
            # booting
            set currPID "boot$CPU"
        } else {
            set currPID "someprocess$CPU"
        }
    }
    return $currPID
}

proc currentPIDByCpu { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    ###
    ### The following were traversing swapable structures.
    ### Find a non-swapable path instead to avoid exceptions.
    ###
    ### set stack_layout [symbol read "vmunix::((struct thread *)$thread)->stack"]
    ### set pproc [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.proc"] 
    ### set newPID [symbol read "vmunix::((struct proc *)$pproc)->p_pid"] 
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set newPID [symbol read "vmunix::((struct super_task *)$super_task)->proc.p_pid"] 
    return $newPID
}

proc currentTID {} {
    global CPU SIMOS
    if [catch {set currTID [currentTIDByCpu $CPU]}] {
        if {$SIMOS(RestoringCpt)==0} {
            # booting
            set currTID "boot$CPU"
        } else {
            set currTID "someprocess$CPU"
        }
    }
    return $currTID
}

proc currentTIDByCpu { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES KSEG_MASK

    set newTID [symbol read "vmunix::active_threads<$cpu>"]
    return [hex [expr $newTID & $KSEG_MASK]]
}

proc currentPPID {} {
    global CPU SIMOS
    if [catch {set currPPID [currentPPIDByCpu $CPU]}] {
        if {$SIMOS(RestoringCpt)==0} {
            # booting
            set currPPID "boot$CPU"
        } else {
            set currPPID "someprocess$CPU"
        }
    }
    return $currPPID
}

proc currentPPIDByCpu { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    ###
    ### The following were traversing swapable structures.
    ### Find a non-swapable path instead to avoid exceptions.
    ###
    ### set stack_layout [symbol read "vmunix::((struct thread *)$thread)->stack"]
    ### set pproc [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.proc"] 
    ### set ppid [symbol read "vmunix::((struct proc *)$pproc)->p_ppid"] 
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set ppid [symbol read "vmunix::((struct super_task *)$super_task)->proc.p_ppid"] 
    return $ppid
}

proc ContextSwitch { ptr} { 
   
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CPU CYCLES
   
    set stack_layout [hex $ptr]
 
    ##
    ## find base of get_sp()->uthread
    ## 

    set task  [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.utask"]
    set pproc [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.proc"] 
    
    set newPROCESS [symbol read "vmunix::((struct utask *)$task)->uu_comm"]
    set newPID [symbol read "vmunix::((struct proc *)$pproc)->p_pid"] 
    set newTID [symbol read "vmunix::((struct stack_layout *)$stack_layout)->uthread.thread"] 

    if {($newPROCESS != $PROCESS($CPU)) || ($newPID != $OSPID($CPU)) || ($newTID != $OSTID($CPU))} { 
	newproc  $newPID $newTID $newPROCESS ContextSwitch
    }
}

######################################################
####  Annotations 
######################################################

### PALENTRY(0x6c01) == callpal_swpctx

annotation set pc 0x6c01 {
    #console "CALL_PAL swpctxt $a0 \n"
    set PCBB($CPU) $a0
}
    
annotation set pc vmunix::idle_thread:START {
    #newproc  idle$CPU [currentTID] "idle" idle_thread_START
    ###set OSPID($CPU) idle$CPU
    set OSPID($CPU) [currentPID]
    set OSTID($CPU) [currentTID]
    set PROCESS($CPU) "kernel idle"
    annotation exec osEvent startIdle
}

annotation set pc vmunix::idle_thread:END {
    if {$PROCESS($CPU) != "kernel idle"} {
        # this can happen when we take an interrupt on idle end
        console "OS.TCL: ignoring end idle when process != idle $CYCLES\n"
    } else {
        annotation exec osEvent endIdle
        ###newproc  runq$CPU [currentTID] "runq" idle_threadEND
        newproc  [currentPID] [currentTID] "runq" idle_threadEND
    }
}

annotation set pc vmunix::swtch:START {
    ###newproc  runq$CPU [currentTID] "runq" swtchSTART
    newproc  [currentPID] [currentTID] "runq" swtchSTART
}

annotation set pc vmunix::swtch:END {
    newproc  [currentPID] [currentTID] [currentProcess] swtchEND
}

annotation set pc vmunix::swtch_pri:START {
    ###newproc  runq$CPU [currentTID] "runq" swtch_pri_START
    newproc  [currentPID] [currentTID] "runq" swtch_pri_START
}

annotation set pc vmunix::swtch_pri:END {
    newproc  [currentPID] [currentTID] [currentProcess] swtch_pri_END
}

####
#### This is the key annotation. Occurs just after the callpal swpctx
####
annotation  set pc vmunix::thread_continue {
    newproc [currentPID] [currentTID] [currentProcess] thread_continue
}

console "@@@ adding annotation to vmunix::newproc:END\n"

annotation set pc vmunix::newproc:END {
    global a1 sp v0
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CPU CYCLES CHILD_PID

    if {$v0 != 0} {
	set th $v0
	set ptask [symbol read "vmunix::((struct thread *)$th)->task"]
	set newPID [symbol read "vmunix::((struct super_task *)$ptask)->proc.p_pid"]
	# child process of the fork
	log "newproc in os.tcl: current PID is $OSPID($CPU), new pid is $newPID\n"
	set CHILD_OSPID($CPU) $newPID
	annotation exec osEvent procstart
	# does not cause a process switch
    }
}

annotation set pc vmunix::waitf:END {
    ## process join
    global CHILD_PID CPU a3 v0
    set CHILD_OSPID($CPU) 0
    if {$v0 == 0} {
	catch {set CHILD_OSPID($CPU) [Double $a3]} err
    }
    annotation exec osEvent procwait
}

annotation set pc vmunix::common_exec:END {
    annotation exec osEvent switchOut
    set PROCESS($CPU) [currentProcess]
    annotation exec osEvent procexec
    annotation exec osEvent switchIn
}

annotation set pc vmunix::exit:START {
    global a0 a1 sp v0 ra
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CPU CYCLES

    set pproc $a0
    set pid [symbol read "vmunix::((struct proc *)$pproc)->p_pid"]
#    console "exit:START: current PID is $OSPID($CPU), exiting pid is $pid (ra=$ra)\n"
    set sym [symbol find vmunix $ra]
#    console "\t ra is in $sym\n"
#    annotation exec osEvent procexit
}

if {$detailLevel >= 2} {

    ##
    ## kernel - user transitions
    ##



    set KENTRY([symbol read vmunix::_XentInt:START]) Intr
    set KENTRY([symbol read vmunix::_XentArith:START]) Arith_FAULT 
    set KENTRY([symbol read vmunix::_XentMM:START]) MM_FAULT 
    set KENTRY([symbol read vmunix::_XentIF:START]) Instr_FAULT 
    set KENTRY([symbol read vmunix::_XentUna:START]) Align_FAULT 
    set KENTRY([symbol read vmunix::_Xsyscall:START]) SysCall 

    foreach i [array names KENTRY] {
        console "kernel entry at $i for $KENTRY($i)\n"
	annotation set pc $i {
            annotation exec osEvent startKernel
        }
    }
    
    annotation set exc {
        if [catch {set currentPalEntry($CPU) $PALENTRY($pc)}] {
            set currentPalEntry($CPU) "callpal_$pc"
        }

        ##  PAL_retsys           PAL_rti
        if {($pc==0x6f41) || ($pc==0x6fc1)} { 
            annotation exec osEvent endKernel
        }
        if {($pc==0x6241) && $a2==0x9} {
            console "OS.TCL:: cServe jToPal. Pusing extra startPal state\n"
            annotation exec osEvent startPal
        }
       
	annotation exec osEvent startPal
        if {[inUserMode]} {
            annotation exec osEvent endUser
        }
    }
         
    annotation set inst rfe {
        annotation exec osEvent endPal
        if {[inUserMode]} {
            annotation exec osEvent startUser
        }
    }
        

    
    ##
    ## synchronization stuff
    ##

    set syncEnd(1) "A"
    set syncEnd(2) "B"
    set syncEnd(3) "C"
    set syncEnd(4) "D"

    # set lockmode [symbol read vmunix::lockmode]
    # console "os.tcl-> Lock mode is $lockmode\n"
    #assert {$lockmode == 2}

    # ugly hardcoded constants for now
    # completely kernel dependant
    # eventually need labels or have the init routine set globals
    # Sync code is complicated because it is copied
    # at startup depending on kernel type
    # also has multiple exit points

    set simple_lock [symbol read vmunix::simple_lock]
    ## The sync stuff is hard coded and should 
    ## Only be active for the right kernel
    if {$simple_lock == 0xfffffc00004e5710 } {

        ## vmunix 4.0

        set kernelMatch 1

        console "Kernel matches, enabling sync annotations!!!\n"
        set simple_lock_end 0xfffffc00004e572c
        # this goes to simple_lock_miss
        # that is a complicated routine and should be 
        # probably included in kern not sync
        # therefore we do an endSync here
        set simple_lock_end_long_path 0xfffffc00004e5744
        # used in kernel_stats and sync_stats
        set simple_lock_miss_end 0xfffffc00004e5d78

#         set syncProcStart [symbol read vmunix::simple_lock_$syncEnd($lockmode):START]
#         set syncProcEnd [symbol read vmunix::simple_lock_$syncEnd($lockmode):END]
#         set simple_lock_end [expr $simple_lock - $syncProcStart + $syncProcEnd]

        set simple_lock_try [symbol read vmunix::simple_lock_try]
        set simple_lock_try_end_1 0xfffffc00004e5a98
        set simple_lock_loop 0xfffffc00004e5a9c
        set simple_lock_try_end_0 0xfffffc00004e5aa4

#         set syncProcStart [symbol read vmunix::simple_unlock_$syncEnd($lockmode):START]
#         set syncProcEnd [symbol read vmunix::simple_unlock_$syncEnd($lockmode):END]
#         set simple_unlock_end [expr $simple_unlock - $syncProcStart + $syncProcEnd]

        set simple_unlock [symbol read vmunix::simple_unlock]
        set simple_unlock_end 0xfffffc00004e58ac

    }
    
    if {$simple_lock == 0xfffffc0000521d20  } {

        ## vmunix 4.0d

        set kernelMatch 1

        console "Kernel matches, enabling sync annotations!!!\n"
        set simple_lock_end 0xfffffc0000521d3c
        # this goes to simple_lock_miss
        # that is a complicated routine and should be 
        # probably included in kern not sync
        # therefore we do an endSync here
        set simple_lock_end_long_path 0xfffffc0000521d54
        # used in kernel_stats and sync_stats
        set simple_lock_miss_end 0xfffffc00005223e8
        #         set syncProcStart [symbol read vmunix::simple_lock_$syncEnd($lockmode):START]
        #         set syncProcEnd [symbol read vmunix::simple_lock_$syncEnd($lockmode):END]
        #         set simple_lock_end [expr $simple_lock - $syncProcStart + $syncProcEnd]

        set simple_lock_try [symbol read vmunix::simple_lock_try]

        set simple_lock_try_end_1 0xfffffc00005220d8
        set simple_lock_loop 0xfffffc00005220dc
        set simple_lock_try_end_0 0xfffffc00005220e4 

        #         set syncProcStart [symbol read vmunix::simple_lock_try_$syncEnd($lockmode):START]
        #         set syncProcEnd [symbol read vmunix::simple_lock_try_$syncEnd($lockmode):END]
        #         set simple_lock_try_end [expr $simple_lock_try - $syncProcStart + $syncProcEnd]
        
        set simple_unlock [symbol read vmunix::simple_unlock]
        set simple_unlock_end 0xfffffc0000521eec

        #         set syncProcStart [symbol read vmunix::simple_unlock_$syncEnd($lockmode):START]
        #         set syncProcEnd [symbol read vmunix::simple_unlock_$syncEnd($lockmode):END]
        #         set simple_unlock_end [expr $simple_unlock - $syncProcStart + $syncProcEnd]
    }

    if {$kernelMatch == 1} {
        
        console "kernelMatch = 1\n"

        annotation set pc $simple_lock {
            #console "SYNC $CYCLES SL_S $CPU\n"
            annotation exec osEvent startSync
        }
        annotation set pc $simple_lock_end {
            ##console "SYNC $CYCLES SL_E $CPU\n"
            annotation exec osEvent endSync
        }
        annotation set pc $simple_lock_end_long_path {
            #console "SYNC $CYCLES SL_ELP $CPU\n"
            annotation exec osEvent endSync
        }
        for {set i 0} {$i < 32} {incr i} {
            set syncTryLoop($i) 0
        }
        # more ugliness (yeach)
        # there is a loop in this function back to the first instr
        # must not fire startSync in the loop case
        annotation set pc $simple_lock_try {
            #console "SYNC $CYCLES SLT_S $CPU\n"
            if {$syncTryLoop($CPU) == 0} {
                annotation exec osEvent startSync
                set syncTryLoop($CPU) 1
            }
        }
        annotation set pc $simple_lock_try_end_1 {
            #console "SYNC $CYCLES SLT_E1 $CPU\n"
            annotation exec osEvent endSync
            set syncTryLoop($CPU) 0
        }
        annotation set pc $simple_lock_try_end_0 {
            #console "SYNC $CYCLES SLT_E0 $CPU\n"
            annotation exec osEvent endSync
            set syncTryLoop($CPU) 0
        }
        annotation set pc $simple_unlock {
            #console "SYNC $CYCLES SU_S $CPU\n"
            annotation exec osEvent startSync
        }
        annotation set pc $simple_unlock_end {
            #console "SYNC $CYCLES SU_E $CPU\n"
            annotation exec osEvent endSync
        }
    }

    if {0} {

	# MP kernel
	#console "INFO: DigitalUnix multiprocessor\n"
	
	annotation set pc vmunix::fiolock:START {
	    annotation exec osEvent startSync
	}
	annotation set pc vmunix::fiolock:END {
	    annotation exec osEvent endSync
	}

	annotation set pc vmunix::fiounlock:START {
	    annotation exec osEvent startSync
	}
	annotation set pc vmunix::fiounlock:END {
	    annotation exec osEvent endSync
	}

	
    }
}

