########################################################################
#
#      ABSTRACT:  Simple procs for vspace breakdown of a thread
#
#      AUTHOR:    nikos@eng.pko.dec.com
#
#      $Id: thread_vspace.tcl,v 1.1 1999/06/02 22:51:44 nikos Exp $
#      $Author: nikos $
#      $Date: 1999/06/02 22:51:44 $
#
########################################################################

FileSourced osf/thread_vspace.tcl

###
### SESL returns the string if the type of the variable is caddr_t, instead of the pointer value.
### To get around this when trying to read uu_text_start, uu_data_start, uu_stack_start, uu_stack_end,
### calculate the offsets from the start of super_task and read directly the simulated memory.
###
### ((unsigned long)&active_threads[0]->stack->uthread->utask.uu_text_start) - ((unsigned long)active_threads[0]->task)
### ((unsigned long)&active_threads[0]->stack->uthread->utask.uu_data_start) - ((unsigned long)active_threads[0]->task)
### ((unsigned long)&active_threads[0]->stack->uthread->utask.uu_stack_start) - ((unsigned long)active_threads[0]->task)
### ((unsigned long)&active_threads[0]->stack->uthread->utask.uu_stack_end) - ((unsigned long)active_threads[0]->task)
### sizeof(active_threads[0]->stack->pad)
###
set UU_TEXT_START_OFFSET_IN_SUPER_TASK    1080
set UU_DATA_START_OFFSET_IN_SUPER_TASK    1088
set UU_STACK_START_OFFSET_IN_SUPER_TASK   1096
set UU_STACK_END_OFFSET_IN_SUPER_TASK     1104
set K_STACK_SIZE                         14904

proc currentThreadTextStart { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES UU_TEXT_START_OFFSET_IN_SUPER_TASK

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    #set textStart [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_text_start"] 
    set textStart [Double [expr $super_task + $UU_TEXT_START_OFFSET_IN_SUPER_TASK]]
    return $textStart
}

proc currentThreadTextSize { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES PGSHIFT

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set textPages [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_tsize"] 
    set textSize [expr $textPages << $PGSHIFT]
    return $textSize
}

proc currentThreadTextEnd { cpu } {
    set textStart [currentThreadTextStart $cpu]
    set textSize  [currentThreadTextSize $cpu]
    return [hex [expr $textStart + $textSize]]
}

proc currentThreadBSSStart { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES UU_DATA_START_OFFSET_IN_SUPER_TASK

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    #set BSSStart [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_data_start"] 
    set BSSStart [Double [expr $super_task + $UU_DATA_START_OFFSET_IN_SUPER_TASK]]
    return $BSSStart
}

proc currentThreadBSSSize { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES PGSHIFT

    set thread     [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set BSSPages   [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_dsize"] 
    set BSSSize    [expr $BSSPages << $PGSHIFT]
    return $BSSSize
}

proc currentThreadBSSEnd { cpu } {
    set BSSStart [currentThreadBSSStart $cpu]
    set BSSSize  [currentThreadBSSSize $cpu]
    return [hex [expr $BSSStart + $BSSSize]]
}

proc currentThreadUStackStart { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES UU_STACK_START_OFFSET_IN_SUPER_TASK

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    #set UStackStart [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_stack_start"] 
    set UStackStart [Double [expr $super_task + $UU_STACK_START_OFFSET_IN_SUPER_TASK]]
    return $UStackStart
}

proc currentThreadUStackCurrSize { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES PGSHIFT

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set UStackCurrPages [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_ssize"] 
    set UStackCurrSize [expr $UStackCurrPages << $PGSHIFT]
    return $UStackCurrSize
}

proc currentThreadUStackEnd { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES UU_STACK_END_OFFSET_IN_SUPER_TASK

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    #set UStackEnd [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_stack_end"] 
    set UStackEnd [Double [expr $super_task + $UU_STACK_END_OFFSET_IN_SUPER_TASK]]
    return $UStackEnd
}

proc IsCurrentThreadUStackGrowingUp { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES PGSHIFT

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set super_task [symbol read "vmunix::((struct thread *)$thread)->task"]
    set UStackGrowUp [symbol read "vmunix::((struct super_task *)$super_task)->utask.uu_stack_grows_up"] 
    return $UStackGrowUp
}

proc currentThreadKStackStart { cpu } {
    global sp
    global MEMORY PROCESS PID  UU_COMM  UTASK_OFFSET UPROC_OFFSET a0 CYCLES 

    set thread [symbol read "vmunix::active_threads<$cpu>"]
    set stack_layout [symbol read "vmunix::((struct thread *)$thread)->stack"]
    return [hex $stack_layout]
}

proc currentThreadKStackEnd { cpu } {
    global K_STACK_SIZE
    return [hex [expr [currentThreadKStackStart $cpu] + $K_STACK_SIZE]]
}


###
### XXX
### Hardcoded values for now.
### Should read them from simulated /usr/shlib/so_locations
###
proc LibLoaderTextStart { } {
    return 0x3ff80000000
}

proc LibLoaderTextEnd { } {
    return 0x3ff80080000
}

proc LibTextStart { } {
    return 0x3ff80080000
}

proc LibTextEnd { } {
    return 0x3ffc0000000
}

proc LibLoaderBSSStart { } {
    return 0x3ffc0000000
}

proc LibLoaderBSSEnd { } {
    return 0x3ffc0080000
}

proc LibBSSStart { } {
    return 0x3ffc0080000
}

proc LibBSSEnd { } {
    return 0x40000000000
}
