/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*
 * syslimits.h -- maximum configurable sizes
 *
 * These are system maximums, the individual interfaces define their own
 * where appropriate (such as the max number of disks per hd interface)
 */

#ifndef SYSLIMITS_H
#define SYSLIMITS_H

/* maximum number of cpus */
/* check to see if defined in makedefs, set to 32 if not */
#ifndef SIM_MAXCPUS
#define SIM_MAXCPUS       32
#endif

/* JeffG: Fix of LARGE_SIMULATION hack */
#if (SIM_MAXCPUS > 32)
# define LARGE_SIMULATION
#else
# undef  LARGE_SIMULATION
#endif

/* maximum number of PCI slots on a node */
#define SIM_MAXSLOTS      4

/* maximum number of host cpus */
#define SIM_MAXHOSTCPUS   64

/* maximum number of local memory regions on any cpu.
 * A local memory region is an area in which a group of cpus share
 * a different memory bank than the global memory bank.
 */
#define SIM_MAXLOCALREGIONS 4

/* maximum number of local regions on the system
 */
#define SIM_SYSMAXLOCALREGIONS 8
 
#endif /* SYSLIMITS_H */
