/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * tcl_init.c
 * 
 * $Author: stets $
 * $Date: 2000/03/21 06:23:51 $
 *****************************************************************/
#include <sys/signal.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/mman.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netdb.h>
#include <file_formats.h>
#include <signal.h>
#include <assert.h>

#include "syslimits.h"
#include "sim_error.h"
#include "cpu_interface.h"
#include "simtypes.h"
#include "machine_params.h"
#include "tcl_init.h"
#include "params.h"
#include "visual.h"
#include <sys/stat.h>
#include "profile.h"
#include "simutil.h"
#include "hw_events.h"
#include "arch_specifics.h"

static void MiscInit(Tcl_Interp *);
static void MachineStateInit(Tcl_Interp *);
#ifdef USE_FLASHLITE
extern int restoringCpt;
static void FlashliteTclInit(Tcl_Interp *);
#endif

extern void SimOSVarsInit(Tcl_Interp *);
extern void AnnotationInit(Tcl_Interp *);
extern void SelectorInit(Tcl_Interp *);
extern void DetailInit(Tcl_Interp *);
extern void SymInit(Tcl_Interp *);
extern void StatisticsInit(Tcl_Interp *);
extern void TimingInit(Tcl_Interp *);
extern void TimingInit(Tcl_Interp *);
extern void StackTclInit(Tcl_Interp *);
extern void MipsyTclInit(Tcl_Interp *);
#ifndef NO_EMBRA
extern void EmbraTclInit(Tcl_Interp *);
#endif
extern void BinaryInit(Tcl_Interp *);
extern void ScacheAnnInit(Tcl_Interp *);
extern void ScacheAnnLateInit(Tcl_Interp *);
extern void ExpectInit(Tcl_Interp *);
extern void ParamInit(Tcl_Interp *);
extern void SoloTclInit(Tcl_Interp *);
#ifdef HWBCOPY
extern void HwbcopyAnnInit(Tcl_Interp *);
#endif
extern void ClusterTclInit(Tcl_Interp *interp);
extern void FaultInit(Tcl_Interp *);
extern FILE* Simrmt_RemoteFileOpen(char *pathName);
extern int HiveCmd(ClientData, Tcl_Interp *, int, char **);
extern void SimErrorLateInit(Tcl_Interp *);

#if defined(SIM_ALPHA)
extern void AlphaTclInit(Tcl_Interp *);
extern void PrintInstsInit(Tcl_Interp *);
extern void GammaTclInit(Tcl_Interp *);
extern void DeltaTclInit(Tcl_Interp *);
extern void EpsilonTclInit(Tcl_Interp *);
#if defined(EXTERNAL_TRACE)
extern void ExternalTraceTclInit(Tcl_Interp *interp);
#endif
#ifdef USE_VISTA_SDF
extern void SDF_Tcl_Init(Tcl_Interp *);
#endif
static void (*initFunc[])(Tcl_Interp *) = {
   SimOSVarsInit,
   ParamInit,
   AnnotationInit,
   SelectorInit,
   DetailInit,
   SymInit,
   MachineStateInit,
   StatisticsInit,
   MiscInit,
   BinaryInit,
   TimingInit,
   ExpectInit,
   HWEventsEarlyInit,
   PrintInstsInit,
   AlphaTclInit,
   GammaTclInit,
   DeltaTclInit,
   EpsilonTclInit,
   ScacheAnnInit,
#if defined(EXTERNAL_TRACE)
   ExternalTraceTclInit,
#endif
#ifdef USE_VISTA_SDF
   SDF_Tcl_Init
#endif
};
#endif

#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
static void (*initFunc[])(Tcl_Interp *) = {
   SimOSVarsInit,
   ParamInit,
   AnnotationInit,
   SelectorInit,
   DetailInit,
   SymInit,
   MachineStateInit,
   StatisticsInit,
   MiscInit,
   TimingInit,
   StackTclInit,
   MipsyTclInit,
#ifndef NO_EMBRA
   EmbraTclInit,
#endif
   BinaryInit,
   ScacheAnnInit,
   FaultInit,
#ifdef HWBCOPY
   HwbcopyAnnInit,
#endif
   ExpectInit,
#ifdef USE_FLASHLITE
   FlashliteTclInit,
#endif
   SoloTclInit,
   HWEventsEarlyInit
};

#endif 

#if defined(SIM_X86)
static void (*initFunc[])(Tcl_Interp *) = {
   SimOSVarsInit,
   ParamInit,
   AnnotationInit,
   SelectorInit,
   DetailInit,
   SymInit,
   MachineStateInit,
   StatisticsInit,
   MiscInit,
   TimingInit,
   StackTclInit,
   BinaryInit,
   ScacheAnnInit,
   ClusterTclInit,
   ExpectInit,
   HWEventsEarlyInit
};
#endif 

#define NUM_INIT_FUNCS  (sizeof(initFunc)/sizeof(void *))


#if defined(SIM_ALPHA)
/* Hack to invalidate access checks when doing address translation 
 * Else accesses to kernel address when the annotation was in user
 * mode causes translation failures.
 */
extern int MemrefFromSimos;
#define MEMREF_DISABLE_TLB_CHECK (MemrefFromSimos = 1)
#define MEMREF_ENABLE_TLB_CHECK (MemrefFromSimos = 0)
#else
#define MEMREF_DISABLE_TLB_CHECK
#define MEMREF_ENABLE_TLB_CHECK
#endif

static char *GetTclErrorString(int code)
{
   char *pszErr=NULL;

   switch (code) {

      case CPT_DISKIO_IN_PROGRESS:
         pszErr = "Disk I/O in progress.";
         break;

      case CPT_BITMAP_ALLOC_FAILURE:
         pszErr = "Bitmap Alloc failure.";
         break;
            
      case CPT_NO_CALLBACK:
         pszErr = "Internal error: No checkpoint callback routine.";
         break;

      case CPT_MISC_ERROR:
         pszErr = "Internal error.";
         break;

      case CPT_BAD_DIRECTORY:
         pszErr = "Bad directory specification.";
         break;
         
      default:
         pszErr = "Unknown error.";
         break;
   }

   return pszErr;
}

/*
 * Digital CC chokes on type a[] = {};
 */

static void (*lateInitFunc[])(Tcl_Interp *) = {
#if !defined(SIM_ALPHA)
   ScacheAnnLateInit,
#else
   NULL,
#endif
   SimErrorLateInit
};
#if !defined(SIM_ALPHA)
#define NUM_LATE_INIT_FUNCS  (sizeof(lateInitFunc)/sizeof(void*))
#else
#define NUM_LATE_INIT_FUNCS  (0)
#endif

Tcl_Interp *TCLInterp = NULL;

static char *errorHandlingScript = NULL;

int TclIniting = 1;

extern int machineParamVec[MAX_MACHINES];
extern MachineScopeType machineScope;


void TCLInit(char *initFile)
{
   Tcl_DString result;
   struct stat buf;
   char *exname;
   int i;
   
   TCLInterp = Tcl_CreateInterp();

   if (Tcl_Init(TCLInterp) != TCL_OK ) {
      CPUWarning("SIMOS: ERROR initing Tcl: %s\n", TCLInterp->result);
      exit(1);
   }

   for (i=0; i<NUM_INIT_FUNCS; i++) {
      ASSERT(initFunc[i]);
      (*initFunc[i])(TCLInterp);
   }

   if ((exname = Tcl_TildeSubst(TCLInterp, initFile, &result)) == NULL) {
      CPUWarning("SIMOS: ERROR: bad init file name: %s\n", initFile);
      exit(1);
   }

   if (stat(exname, &buf)) {
      CPUWarning("SIMOS: init.simos file required\n");
      exit(1);
   }
   
   Tcl_DStringFree(&result);
   
   if (Tcl_EvalFile(TCLInterp, initFile) != TCL_OK) {
      char *errorStr = Tcl_GetVar(TCLInterp, "errorInfo", TCL_GLOBAL_ONLY);
      CPUWarning("TCL error while initializing.\n%s\n", errorStr);
      exit(1);
   }

   TclIniting = 0;

#if NUM_LATE_INIT_FUNCS > 0   
   for (i=0; i<NUM_LATE_INIT_FUNCS; i++) {
      if ( lateInitFunc[i] ) {
	(*lateInitFunc[i])(TCLInterp);
      }
   }
#endif   
   
   MachineSettingsComplete();

   SetAdditionalParams(TCLInterp);
}


/*
 * command dispatch
 */

static void usage(Tcl_Interp *interp, char *cmdName, tclcmd *cmds)
{
   int i;
   
   Tcl_AppendResult(interp, cmdName, " usage:\n\r", NULL);

   for (i=0; cmds[i].name; i++) {
      Tcl_AppendResult(interp, cmdName, cmds[i].usage, "\n\r", NULL);
   }
}

int DispatchCmd(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
   tclcmd *cmds = (tclcmd*)data;
   int i, c;
   size_t length;
   char *buf;

   if (argc == 1) {
      usage(interp, argv[0], cmds);
      return TCL_OK;
   }

   c = argv[1][0];
   length = strlen(argv[1]);

   for (i=0; cmds[i].name; i++) {
      if ((c == cmds[i].name[0]) && (strncmp(argv[1], cmds[i].name, length) == 0)) {
         if ((cmds[i].numargs == -1) || (cmds[i].numargs == argc)) {
            ASSERT(cmds[i].cmd);
            return (*cmds[i].cmd)(interp, argc, argv); 
         } else {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], cmds[i].usage, "\"", NULL);
            return TCL_ERROR;
         }
      }
   }
   length = 1; 
   for (i=0; i<argc; i++) {
      length += strlen(argv[i]);
      length += 1; /* room for space */
   }
   buf = malloc(length);
   /* cons up a procedure call */
   sprintf(buf, "%s_%s", argv[0], argv[1]);
   for (i=2; i<argc; i++) {
      strcat(buf, " ");
      strcat(buf, argv[i]);
   }

   if (Tcl_Eval(interp, buf) != TCL_OK) {
      free(buf);
      return TCL_ERROR;
   }
   free(buf);
   return TCL_OK;
}


/*
 * translate virtual addresses using knowledge of the OS
 */

Reg TclTranslateVirtual(int cpuNum, VA vAddr)
{
   char buf[128];       /* Ugly fixed size hack gets you in trouble bv */
   Reg pAddr = 0;
   
   if (sizeof(vAddr) == sizeof(int)) 
      sprintf(buf, "TclTranslateVirtual %d 0x%x", cpuNum, (int)vAddr);
   else { 
#ifdef __alpha
      sprintf(buf, "TclTranslateVirtual %d 0x%lx", cpuNum, (Reg64)vAddr);
#else
#ifdef linux
      sprintf(buf, "TclTranslateVirtual %d 0x%x", cpuNum, (int)vAddr);
#else
      sprintf(buf, "TclTranslateVirtual %d 0x%llx", cpuNum, (Reg64)vAddr);
#endif
#endif
   }
   if (Tcl_Eval(TCLInterp, buf) != TCL_OK) {
      CPUWarning("ERROR: in %s: %s\n", buf, TCLInterp->result);
      Tcl_ResetResult(TCLInterp);
      return 0;
   }

   if (String2Reg(TCLInterp, TCLInterp->result, &pAddr) != TCL_OK) {
      CPUWarning("ERROR: TclTranslateVirtual returned non-int \"%s\"\n",
                 TCLInterp->result);
      Tcl_ResetResult(TCLInterp);
      return 0;
   }

   Tcl_ResetResult(TCLInterp);
   return pAddr;
}


/*
 * Generate disk name based on cell/controller/disk information.
 * Returns NULL if no DiskFileName procedure has been defined in Tcl.
 */

char* TclDiskFileName(int machine, int node, int ctrl, int unit)
{
  char  buf[64];
  char* res;

  sprintf(buf, "DiskFileName %d %d %d %d", machine, node, ctrl, unit);
  if (Tcl_Eval(TCLInterp, buf) != TCL_OK) {
    /* probably no Tcl procedure defined */
    return (char*) NULL;
  }
  /* must return copy of string */
  res = (char*) malloc(strlen(TCLInterp->result)+1);
  ASSERT(res);
  strcpy(res, TCLInterp->result);
  Tcl_ResetResult(TCLInterp);

  return res;
}


void
TclDoCheckpoint(char* name)
{
  char buf[256];

  sprintf(buf, "DoCheckpointProc %s", name);
  if (Tcl_Eval(TCLInterp, buf) == TCL_OK) {
    Tcl_ResetResult(TCLInterp);
  }
}

void
TclRestoreCheckpoint(char* name)
{
  char buf[256];

  sprintf(buf, "RestoreCheckpointProc %s", name);
  if (Tcl_Eval(TCLInterp, buf) == TCL_OK) {
    Tcl_ResetResult(TCLInterp);
  }
}


/*
 * misc. tcl commands (log, console, debug)
 */

#ifdef OPTION_AS_HASHTABLE
static Tcl_HashTable opts;
#else
static struct OptionBinding {
#ifndef DIGITAL_UNIX_CC    
   const
#endif
   char *name;
   const char *value;
} *optionBindings = NULL;
static int nOptionBindings = 0;
#endif /* OPTION_AS_HASHTABLE */

void TCLDefineOption(char *opt, char *val)
{
   fprintf(stderr, "Defining option %s with value %s\n", opt, val);
   fprintf(stderr, "TCLInterp=%p\n", TCLInterp);

   {
      int i = nOptionBindings;
      nOptionBindings++;
      optionBindings = realloc(optionBindings, nOptionBindings * sizeof(struct OptionBinding));
      optionBindings[i].name = SaveString(opt);
      optionBindings[i].value = SaveString(val);
   }
}

static void TCLOptionInit(Tcl_Interp *interp)
{
   int i;
   for (i = 0; i < nOptionBindings; i++) {
      Tcl_SetVar2(TCLInterp, "OPTION", (char *)optionBindings[i].name, (char *)optionBindings[i].value,
                  TCL_GLOBAL_ONLY);
   }
}


/*
 *   Search for the file name filename on the simosPath. Result is the
 *   full pathname of the (possibly remote) file. The result is only
 *   good untill the next call to SearchForFile and may be the string
 *   that was passed in.
 */

char fullFileName[2048];

char *SearchForFile(char *filename)
{
   int listArgc, j;
   char **listArgv;
   Tcl_DString buffer;
   FILE *fp = NULL;
   char *fname;
   char *path;
   
   fullFileName[0] = '\0';
   
   if (strchr(filename, ':')) {
      return filename;
   }

   if ((filename[0] == '~') || strchr(filename, '/')) {
      return filename;
   }
      
   if ((path = Tcl_GetVar(TCLInterp, "simosPath", TCL_GLOBAL_ONLY)) == NULL) {
      return NULL;
   }

   if (Tcl_SplitList(TCLInterp, path, &listArgc, &listArgv) != TCL_OK) {
      CPUWarning("SIMOS: can't parse simosPath list\n");
      return NULL;
   }
   
   for (j = 0; j < listArgc; j++) {
      ASSERT((strlen(listArgv[j]) + strlen(filename) + 2) < 2048);

      if (strchr(listArgv[j], ':')) {
         /* Going remote... */
         sprintf(fullFileName, "%s/%s", listArgv[j], filename);
         
         fp = Simrmt_RemoteFileOpen(fullFileName);
         if (fp) { 
            fclose(fp);
            return fullFileName;
         }
         
      } else {
         /* Actual local file */
         sprintf(fullFileName,"%s/%s", listArgv[j], filename);

         fname = Tcl_TildeSubst(TCLInterp, fullFileName, &buffer);
         if (fname) { 
            if (fullFileName != fname) {
               ASSERT(strlen(fname) < 2048);
               strcpy(fullFileName, fname);
            }
            Tcl_DStringFree(&buffer);
            
            fp = fopen(fullFileName, "r");
            if (fp) { 
               fclose(fp);
               return fullFileName;
            }
         }
      }
   }
   
   return NULL;
}


/* Evaluate a file on a remote server. Code stolen from
 TclUnixUtil.c  Note that I don't modify tcl's concept
 of the current script, since that's only available as
 an internal slot.*/

static int 
EvalRemote(Tcl_Interp *interp, char *name) {
   FILE *fp = NULL;
   Tcl_DString buffer;
   size_t amountRead = 0;
   int result;
   Tcl_DStringInit(&buffer);
   
   if ((fp = Simrmt_RemoteFileOpen(name)) == NULL) {
      return TCL_ERROR;
   }
           
   do {
      char cmdBuffer[8192];
      amountRead = fread(cmdBuffer, 1, 8192, fp);
      Tcl_DStringAppend(&buffer, cmdBuffer, (int) amountRead);
   } while (amountRead == 8192);

   /* Ready to evaluate. */
   result = Tcl_Eval(interp, Tcl_DStringValue(&buffer));
   if (result == TCL_RETURN) {
      result = TCL_OK;
   }
   if (result == TCL_ERROR) {
      char msg[200];
      sprintf(msg, "\n    (file \"%.150s\" line %d)", 
              Tcl_DStringValue(&buffer),
              interp->errorLine);
      Tcl_AddErrorInfo(interp, msg);
   }
   Tcl_DStringFree(&buffer);
   fclose(fp);
   return result;
}

static int
SourceCmd(ClientData clientData, Tcl_Interp *interp,
                 int argc, char *argv[])
{
   char *givenFile;
   char *filename;
   
   if (argc != 2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], " scriptName\"", NULL);
      return TCL_ERROR;
   }

   givenFile = argv[1];

   filename = SearchForFile(givenFile);
   if (filename) { 
      if (strchr(filename,':')) {
         return EvalRemote(interp, filename);
      } else {
         return Tcl_EvalFile(interp, filename);
      }
   }

   Tcl_AppendResult(interp, "can't find file to source", NULL);
   return TCL_ERROR;
}


static int
RandomCmd(ClientData clientData, Tcl_Interp *interp,
                 int argc, char *argv[])
{
   int value;
   int limit;
   char buf[32];
#if !defined(__alpha) && !defined(linux)
   extern long random(void);
#endif
#ifdef some_platform_maybe_sun
   extern int srandom(unsigned seed);
#endif
   
   if (argc < 2) {
      goto error;
   }
   
   if (!strcmp(argv[1], "seed")) {
      if (argc == 2) {
         value = time(0);

      } else if (argc == 3) {
         if (Tcl_GetInt(interp, argv[2], &value) != TCL_OK) {
            return TCL_ERROR;
         }

      } else {
         goto error;
      }

      srandom(value);
      
   } else {
      if (argc != 2) {
         goto error;
      }
      
      if (Tcl_GetInt(interp, argv[1], &limit) != TCL_OK) {
         return TCL_ERROR;
      }

      sprintf(buf, "%d", (int)(random() % limit));

      Tcl_AppendResult(interp, buf, NULL);
   }

   return TCL_OK;

error:
   Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " limit | seed ?value?\"", NULL);
   return TCL_ERROR;
}


/*
 * Tcl Error Handler
 */

int HandlerCmd(ClientData clientData, Tcl_Interp *interp,
               int argc, char *argv[])
{
   errorHandlingScript = SaveString(argv[1]);
   return TCL_OK;
}

void HandleError(Tcl_Interp *interp, char *script)
{
   char buf[64];

   if (errorHandlingScript) {
      sprintf(buf, "\n    (script line %d)", interp->errorLine);
      Tcl_AddErrorInfo(interp, buf);
      
      Tcl_AddErrorInfo(interp, "\n    invoked from within\n{");
      Tcl_AddErrorInfo(interp, (char*)script);
      Tcl_AddErrorInfo(interp, "}");
      
      if (Tcl_GlobalEval(interp, errorHandlingScript) != TCL_OK) {
         CPUWarning("Tcl ERROR in error handler: %s\n", interp->result);
      }
      
   } else {
      CPUWarning("Tcl ERROR: %s\n", interp->result);
   }
}

static int 
MachineSelectCmd(ClientData clientData, Tcl_Interp *interp,
                   int argc, char *argv[])
{
   int listArgc, count, retVal = TCL_OK;
   Tcl_IntValueType listVal;
   char **listArgv;
   int machineVecBackup[MAX_MACHINES];
   MachineScopeType machineScopeBackup;

   if (argc != 3) {
      interp->result = "wrong # args";
      return TCL_ERROR;
   }
   if (Tcl_SplitList(interp, argv[1], &listArgc, &listArgv) != TCL_OK) {
      return TCL_ERROR;
   }

   /* Save the vector for later */
   for (count = 0; count < MAX_MACHINES; count++) {
      machineVecBackup[count] = machineParamVec[count];
      machineParamVec[count] = FALSE;
   }
   machineScopeBackup = machineScope;
   machineScope = MACHINE_BLOCK;

   /* Update the vector for the current machine block */
   for (count = 0; (retVal == TCL_OK) && (count < listArgc); count++) {
      retVal = Tcl_ExprLong(interp, listArgv[count], (long *)&listVal);
      if ((retVal == TCL_OK) && (listVal < 0 || listVal >= MAX_MACHINES)) {
         interp->result = "machine number out of range";
         retVal = TCL_ERROR;
      }
      if ((retVal == TCL_OK) && !machineVecBackup[listVal]) {
         interp->result = "can only write to machines active in current machine scope";
         retVal = TCL_ERROR;
      }
      if (retVal == TCL_OK)
         machineParamVec[listVal] = TRUE;
   }

   if (retVal == TCL_OK)
      retVal = Tcl_Eval(interp, argv[2]);

   /* Restore the vector (even if there was an error above) */
   for (count = 0; count < MAX_MACHINES; count++) {
      machineParamVec[count] = machineVecBackup[count];
   }
   machineScope = machineScopeBackup;

   return retVal;
}  

static int 
MachineValueCmd(ClientData clientData, Tcl_Interp *interp,
                   int argc, char *argv[])
{
   int count, retVal = TCL_OK;
   Tcl_IntValueType machineVal;
   int machineVecBackup[MAX_MACHINES];
   MachineScopeType machineScopeBackup;
   char *string;

   if (argc != 3) {
      interp->result = "wrong # args";
      return TCL_ERROR;
   }

   /* Update the vector for the current machine block */
   if (Tcl_ExprLong(interp, argv[1], (long *)&machineVal) != TCL_OK)
      return TCL_ERROR;
   if ((retVal == TCL_OK) && (machineVal < 0 || machineVal >= MAX_MACHINES)) {
      interp->result = "machine number out of range";
      return TCL_ERROR;
   }

   /* Save the vector for later */
   for (count = 0; count < MAX_MACHINES; count++) {
      machineVecBackup[count] = machineParamVec[count];
      machineParamVec[count] = (count==machineVal)?TRUE:FALSE;
   }
   machineScopeBackup = machineScope;
   machineScope = MACHINE_VAL;

   string = Tcl_GetVar(interp, argv[2], TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
   if (string == NULL)
      retVal = TCL_ERROR;
   else
      interp->result = string;

   /* Restore the vector (even if there was an error above) */
   for (count = 0; count < MAX_MACHINES; count++) {
      machineParamVec[count] = machineVecBackup[count];
   }
   machineScope = machineScopeBackup;

   return retVal;
}

static int
LogCmd(ClientData clientData, Tcl_Interp *interp,
              int argc, char *argv[])
{
   int i;
   int cnxt = VisualGetContext();
   
   if (cnxt > 0) { 
      Tcl_SetResult(interp, "",TCL_STATIC );

      for (i=1; i<argc; i++) {
         Tcl_AppendResult(interp, argv[i], 0);
      }
      VisualPushString(interp->result);
      Tcl_SetResult(interp,"OK",TCL_STATIC);

   } else {
      for (i=1; i<argc; i++) {
         CPUPrint("%s", argv[i]);
      }
   }
   
   return TCL_OK;
}

static int
ConsoleCmd(ClientData clientData, Tcl_Interp *interp,
                  int argc, char *argv[])
{
   int i;
   char *mark;
   
   for (i=1; i<argc; i++) {
      CPUPrint("%s", argv[i]);
      
      mark = argv[i];
      while (*mark) {
         fputc(*mark, stderr);
         if (*mark == '\n') {
            fputc('\r', stderr);
         }
         mark++;
      }
   }

   return TCL_OK;
}

static int
DebugCmd(ClientData clientData, Tcl_Interp *interp,
                int argc, char *argv[])
{
   int cpuid;
   if (argc == 1) {
      cpuid = -1;      /* code for any cpu */
   } else {
      if (Tcl_GetInt(interp, argv[1], &cpuid) != TCL_OK) {
         Tcl_AppendResult(interp, "Bad cpu number passed to debug command",
                          NULL);
         return TCL_ERROR;
      }

      ASSERT(TOTAL_CPUS);
      if (cpuid < 0 || cpuid >= TOTAL_CPUS) {
         Tcl_AppendResult(interp, "Bad cpu number passed to debug command",
                          NULL);
         return TCL_ERROR;
      }
   }
   
   CPUVec.Handle_Debug_Signal(cpuid, 0);
   return TCL_OK;
}


static int
AssertCmd(ClientData clientData, Tcl_Interp *interp,
                 int argc, char *argv[])
{
   int result, code;
   
   if (argc == 1) {
      CPUError("Tcl assert failing (looping)\n");
   
   } else if (argc == 2) {
      if ((code = Tcl_ExprBoolean(interp, argv[1], &result)) != TCL_OK) {
         return code;
      }

      if (!result) {
         CPUError("Tcl assertion failed (looping): %s\n", argv[1]);
      }
      
   } else {
      Tcl_AppendResult(interp, "Bad number of args, should be: assert expr", NULL);
      return TCL_ERROR;
   }

   return TCL_OK;
}

static int
CPUEnterCmd(ClientData clientData, Tcl_Interp *interp,
                   int argc, char *argv[])
{
#if defined(SIM_MIPS32) || defined(SIM_MIPS64) || defined(SIM_ALPHA)
   CPUType newSimulator = NO_CPU;
   char buf[64];

   if ((argc < 2) || (argc > 3)) {
      Tcl_AppendResult(interp, "Wrong number of args, should be: "
                       "cpuEnter cpuModel ?cpuNum?\n", NULL);
      return TCL_ERROR;
   }
   /* Three argument version -- switch a specific CPU between MIPSY and MXS */
   if (argc == 3) {
#if defined(SIM_ALPHA)
   ASSERT(0);
#else
      /* RPB hack -- calling directly into mipsy/MXS from common */
      extern void MXSSwitch(int cpu, int enterMXS);

      if (Tcl_GetInt(interp, argv[2], &cpuNum) != TCL_OK) {
         Tcl_AppendResult(interp, "Can't parse cpu number '%s' for cpuEnter\n",
                          argv[2], NULL);
         return TCL_ERROR;
      }

      ASSERT(TOTAL_CPUS);
      if ((cpuNum < 0) || (cpuNum >= TOTAL_CPUS)) {
         Tcl_AppendResult(interp, "Cpu number %s out of bounds for cpuEnter\n",
                          argv[2], NULL);
         return TCL_ERROR;
      }
      if (!strcmp(argv[1], "mipsy"))
         MXSSwitch(cpuNum, FALSE);
      else if (!strcmp(argv[1], "mxs"))
         MXSSwitch(cpuNum, TRUE);
      else {
         Tcl_AppendResult(interp, "Can't enter cpu type: %s\n", argv[1], NULL);
         return TCL_ERROR;
      }
#endif
   }
   /* Two argument version -- switch entire simulator between CPU models */
   else {
      int cpuType;
      for (cpuType = BASE; cpuType <= NO_CPU; cpuType++) {
         if (strcmp(argv[1], simName[cpuType]) == 0) {
         newSimulator = (CPUType) cpuType;
         break;
         }
      }
      if (cpuType >= NO_CPU) {
         sprintf(buf, "Can't enter cpu type: %s\n", argv[1]);
         Tcl_AppendResult(interp, buf, NULL);
         return TCL_ERROR;
      }

      free(machines.CpuModel);
      machines.CpuModel = SaveString(argv[1]);
      
      CPUVec.ExitSimulator(newSimulator);
   }
   return TCL_OK;

#endif
}

static int
SimosExitCmd(ClientData clientData, Tcl_Interp *interp,
                   int argc, char *argv[])
{
   CPUType newSimulator;

   if (argc != 1) {
      Tcl_AppendResult(interp, "Wrong number of args, should be: ", argv[0], NULL);
      return TCL_ERROR;
   }
   
   newSimulator = NO_CPU;
   CPUVec.ExitSimulator(newSimulator);

   return TCL_OK;
}

static int
ProfileCmd(ClientData clientData, Tcl_Interp *interp,
                  int argc, char *argv[])
{
   if (argc != 2) {
      Tcl_AppendResult(interp, "Wrong number of args, should be: profile start|stop\n",
                       NULL);
      return TCL_ERROR;
   }
   if (!strcmp(argv[1], "start")) {
      StartProfile();
   } else if (!strcmp(argv[1], "stop")) {
      StopProfile();
   } else {
      Tcl_AppendResult(interp, "Bad arg, should be: profile start|stop\n",
                       NULL);
      return TCL_ERROR;
   }
   return TCL_OK;
}

extern char *CheckpointID;

static int
DoCheckpointCmd(ClientData clientData, Tcl_Interp *interp,
                       int argc, char *argv[])
{
   int  result;

   ASSERT(CPUVec.DoCheckpoint);
   if (argc == 2) 
      CheckpointID = argv[1];
   result = CPUVec.DoCheckpoint(0);

   if (0 != result) {

      Tcl_SetErrorCode(interp, "CHECKPOINT", "SAVE_ERROR", 
                       GetTclErrorString(result), (char *)NULL);
      Tcl_AddErrorInfo(interp, "doCheckpoint");
      result = TCL_ERROR;
   }
   else
      result = TCL_OK;

   return result;
}

static int
DisableCPUCmd(ClientData clientData, Tcl_Interp *interp,
                     int argc, char *argv[])
{
   int cpuNum;

   if (argc != 2) {
      Tcl_AppendResult(interp, "Bad arg, should be: disableCPU cpuNum\n",
                       NULL);
      return TCL_ERROR;  
   }
   cpuNum = atoi(argv[1]);
   ASSERT (CPUVec.FaultInject);
   CPUVec.FaultInject(cpuNum,SIMFAULT_DISABLECPU);
   return TCL_OK;
}

static int
ExitUserProcCmd(ClientData clientData, Tcl_Interp *interp,
                       int argc, char *argv[])
{
   int cpuNum;

   if (argc != 2) {
      Tcl_AppendResult(interp, "Bad arg, should be: exitUserProc cpuNum\n",
                       NULL);
      return TCL_ERROR;  
   }
   cpuNum = atoi(argv[1]);
   ASSERT(CPUVec.ProcMakeProcExit);
   CPUVec.ProcMakeProcExit(cpuNum);
   return TCL_OK;
}

void 
MiscInit(Tcl_Interp *interp)
{
   Tcl_CreateCommand(interp, "handler", HandlerCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

#ifdef OPTION_AS_HASHTABLE
   Tcl_SetVar2(interp, "OPTION", "need_to_set_so_I_can_trace", "",
               TCL_GLOBAL_ONLY);

   Tcl_TraceVar(interp, "OPTION",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                OptionAccess, (ClientData) NULL);
#else
   TCLOptionInit(interp);
#endif /* OPTION_AS_HASHTABLE */


   Tcl_CreateCommand(interp, "random", RandomCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "source", SourceCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "log", LogCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "console", ConsoleCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "debug", DebugCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "assert", AssertCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "cpuEnter", CPUEnterCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "simosExit", SimosExitCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "profile", ProfileCmd, 
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "doCheckpoint", DoCheckpointCmd, 
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "do_checkpoint", DoCheckpointCmd, 
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "disableCPU", DisableCPUCmd, 
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "exitUserProc", ExitUserProcCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "machine", MachineSelectCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp, "machine_val", MachineValueCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);

#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
   Tcl_CreateCommand(interp, "hive", HiveCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);
#endif
}

#ifdef USE_FLASHLITE
extern void FlashliteConfigDefaults(void);
static void
FlashliteTclInit(Tcl_Interp *interp)
{
   FlashliteConfigDefaults();
}
#endif


/*
 *   Debug stub for access to machine state
 */

Tcl_HashTable regTable;

#define NUM_REGS    ((sizeof reg_names)/sizeof(char*))

char* reg_names[] = REGISTER_NAMES;


#define FIRST_FP_REG                   32
#define IS_FPREG(_regnum)              (((_regnum) >= FIRST_FP_REG) && ((_regnum) < 64))

#define FIRST_DFP_REG                  (84)
#define IS_DFPREG(_regnum)             ((_regnum) >= FIRST_DFP_REG)
#ifdef SIM_MIPS64
#define DFPREGNUM_TO_FPREGNUM(_regnum) (((_regnum) - FIRST_DFP_REG) + FIRST_FP_REG)
#else
#define DFPREGNUM_TO_FPREGNUM(_regnum) (((_regnum) - FIRST_DFP_REG)*2 + FIRST_FP_REG)
#endif

static char *RegisterAccess(ClientData clientData, Tcl_Interp *interp,
                            char *name1, char *name2, int flags)
{
   Tcl_HashEntry *entryPtr;
   int regnum = 0;
   char buf[32];
   char *valstr;
   Reg regValue = 0;
   int cpunum;
   int argc;
   char **argv;
   
   union {
      double  fpr;
#ifdef SIM_MIPS64
      float fprs;
      Reg fgr;
#else
      struct {
         Reg  fgr1;
         Reg  fgr0;
      } fgr;
#endif
   } fpreg;
   
   if ((!CPUVec.GetRegister) || (!CPUVec.CurrentCpuNum)) {
      CPUWarning("TCL: WARNING - Too early to read registers. Returning 0.\n");
      Tcl_SetVar2(interp, name1, name2, "0", 0);
      return NULL;
   }

   cpunum = CPUVec.CurrentCpuNum();

   if (!strcmp(name1,"REGISTER")) {
      if (Tcl_SplitList(interp, name2, &argc, &argv) != TCL_OK) {
         return "cannot parse register number";
      }

      if ((argc != 1) && (argc != 2)) {
         free(argv);
         return "bad register list";
      }
      
      entryPtr = Tcl_FindHashEntry(&regTable, argv[0]);
      if (entryPtr == NULL) {
         free(argv);
         return "bad register name";
      }
      
      regnum = (int)PTR_TO_UINT64(Tcl_GetHashValue(entryPtr));

      if (argc == 2) {
         if (Tcl_GetInt(interp, argv[1], &cpunum) != TCL_OK) {
            free(argv);
            return "bad cpu number";
         }
      }
      free(argv);
      
   } else {
      entryPtr = Tcl_FindHashEntry(&regTable, name1);
      if (entryPtr == NULL) {
         return "bad register name";
      }
      
      regnum = (int)PTR_TO_UINT64(Tcl_GetHashValue(entryPtr));
   }
   
   
   if (flags & TCL_TRACE_READS) {
      if (IS_DFPREG(regnum)) {
         regnum = DFPREGNUM_TO_FPREGNUM(regnum);
#ifdef SIM_MIPS64
         CPUVec.GetRegister(cpunum, regnum, &fpreg.fgr);
#else
         CPUVec.GetRegister(cpunum, regnum, &fpreg.fgr.fgr0);
         CPUVec.GetRegister(cpunum, regnum+1, &fpreg.fgr.fgr1);
#endif
         sprintf(buf, "%23.20f", fpreg.fpr);
         
      } else {
         CPUVec.GetRegister(cpunum, regnum, &regValue);
         if (IS_FPREG(regnum)) {
            sprintf(buf, "%23.20f", *(float*)&regValue);
         } else if (sizeof(Reg) == sizeof(int)) {
            sprintf(buf, "0x%x", (int)regValue);
         } else { 
            PrintLLX(buf,regValue);
         }
      }
      
      Tcl_SetVar2(interp, name1, name2, buf, 0);

   } else if (flags & TCL_TRACE_WRITES) {
      if ((valstr = Tcl_GetVar2(interp, name1, name2, 0)) == NULL) {
         return "bad reg";
      }
      if (IS_DFPREG(regnum)) {
         regnum = DFPREGNUM_TO_FPREGNUM(regnum);
         if (Tcl_GetDouble(interp, valstr, &(fpreg.fpr)) != TCL_OK) {
            return "bad value";
         }
#ifdef SIM_MIPS64
         CPUVec.PutRegister(cpunum, regnum, fpreg.fgr);
#else
         CPUVec.PutRegister(cpunum, regnum, fpreg.fgr.fgr0);
         CPUVec.PutRegister(cpunum, regnum+1, fpreg.fgr.fgr1);
#endif         
      } else if (IS_FPREG(regnum)) {
         double dtmp;
         if (Tcl_GetDouble(interp, valstr, &dtmp) != TCL_OK) {
            return "bad value";
         } else {
#ifdef SIM_MIPS64
            fpreg.fprs = (float)dtmp;
#else
	    fpreg.fpr = dtmp;
#endif
	    
         }
#ifdef SIM_MIPS64
         CPUVec.PutRegister(cpunum, regnum, fpreg.fgr);
#else
         CPUVec.PutRegister(cpunum, regnum, fpreg.fgr.fgr0);
         CPUVec.PutRegister(cpunum, regnum+1, fpreg.fgr.fgr1);
#endif         
      } else {
	 Reg value=0;
	 if (String2Reg(interp,valstr,&value)!=TCL_OK) 
	    return "bad value";
	 
         CPUVec.PutRegister(cpunum, regnum, value);
      }
   }
   
   return NULL;
}

static char *MemoryAccess(ClientData clientData, Tcl_Interp *interp,
                          char *name1, char *name2, int flags)
{
   Reg addr = 0;
   char buf[32];
   char *valstr;
   int value = 0;
   int cpunum;

   if ((!CPUVec.PutMemory) || (!CPUVec.GetMemory) || (!CPUVec.CurrentCpuNum)) {
      CPUWarning("TCL: WARNING - Too early to read MEMORY. Returning 0.\n");
      Tcl_SetVar2(interp, name1, name2, "0", 0);
      return NULL;
   }

   cpunum = CPUVec.CurrentCpuNum();
   
   /* find addr (should take symbols) */
   if ((name2 == NULL)
       || (String2Reg(interp, name2, &addr) != TCL_OK)) {
      return "bad addr";
   }
   
   if (flags & TCL_TRACE_READS) {
      enum result_t res;

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.GetMemory(cpunum, addr, 4, (char *)&value);
      MEMREF_ENABLE_TLB_CHECK;

      if (res != SUCCESS) {
         return "not mapped";
      }
      value = MO2HO_4(value);
      sprintf(buf, "0x%x", value);
      Tcl_SetVar2(interp, name1, name2, buf, 0);
      
   } else if (flags & TCL_TRACE_WRITES) {
      enum result_t res;

      if ((valstr = Tcl_GetVar2(interp, name1, name2, 0)) == NULL) {
         return "bad addr";
      }
      if (Tcl_GetInt(interp, valstr, &value) != TCL_OK) {
         return "bad addr";
      }

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.PutMemory(cpunum, addr, 4, (char *)&value);
      MEMREF_ENABLE_TLB_CHECK;
      if (res != SUCCESS) {
         return "not mapped";
      }
   }
   
   return NULL;
}


static char *PhysicalAccess(ClientData clientData, Tcl_Interp *interp,
                          char *name1, char *name2, int flags)
{
   Reg addr = 0;
   char buf[32];
   PA value = 0;
   int cpunum;

   if ((!CPUVec.TranslateVirtualNoSE) || (!CPUVec.CurrentCpuNum)) {
      return "Too early to read PHYSICAL.";
   }

   cpunum = CPUVec.CurrentCpuNum();
   
   /* find addr (should take symbols) */
   if ((name2 == NULL)
       || (String2Reg(interp, name2, &addr) != TCL_OK)) {
      return "bad addr";
   }
   
   if (flags & TCL_TRACE_READS) {
      if (CPUVec.TranslateVirtualNoSE(cpunum, addr, &value) != SUCCESS) {
         return "not mapped";
      }
#ifdef __alpha
      sprintf(buf, "0x%lx", (uint64)value);
#else
      sprintf(buf,"0x%llx", (uint64)value);
#endif
      Tcl_SetVar2(interp, name1, name2, buf, 0);
      
   } else if (flags & TCL_TRACE_WRITES) {
      return "can't write PHYSICAL";
   }
   
   return NULL;
}


static char *CPUAccess(ClientData clientData, Tcl_Interp *interp,
                       char *name1, char *name2, int flags)
{
   char buf[32];

   if (!CPUVec.CurrentCpuNum) {
      Tcl_SetVar(interp, name1, "0", 0);
      return NULL;
   }
   
   if (flags & TCL_TRACE_READS) {
      /* set its value */
      sprintf(buf, "%d", (CPUVec.CurrentCpuNum) ? CPUVec.CurrentCpuNum() : 0);
      Tcl_SetVar(interp, name1, buf, 0);

   } else if (flags & TCL_TRACE_WRITES) {
      return "Writes to CPU ignored";
   }
   
   return NULL;
}

static char *CellAccess(ClientData clientData, Tcl_Interp *interp,
                        char *name1, char *name2, int flags)
{
   char buf[32];

   if (!CPUVec.CurrentCpuNum) {
      Tcl_SetVar(interp, name1, "0", 0);
      return NULL;
   }
   
   if (flags & TCL_TRACE_READS) {
      int cpu = CPUVec.CurrentCpuNum();
      /* set its value */
      sprintf(buf, "%d", cpu / CPUS_PER_CELL(M_FROM_CPU(cpu)));
      Tcl_SetVar(interp, name1, buf, 0);

   } else if (flags & TCL_TRACE_WRITES) {
      return "Writes to CELL ignored";
   }
   
   return NULL;
}

static char *CycleAccess(ClientData clientData, Tcl_Interp *interp,
                         char *name1, char *name2, int flags) 
{
   int cpunum;
   char buf[32];

   if ((!CPUVec.CycleCount) || (!CPUVec.CurrentCpuNum)) {
      CPUWarning("TCL: WARNING - Too early to read CYCLES. Returning 0.\n");
      Tcl_SetVar(interp, name1, "0", 0);
      return NULL;
   }
   if (flags & TCL_TRACE_READS) {
      cpunum = CPUVec.CurrentCpuNum();
      PrintLLD(buf,CPUVec.CycleCount(cpunum));
      Tcl_SetVar(interp, name1, buf, 0);
   } else if (flags & TCL_TRACE_WRITES) {
      return "Writes to CYCLES ignored";
   }
   return NULL;
}

static char *InstsAccess(ClientData clientData, Tcl_Interp *interp,
                         char *name1, char *name2, int flags) 
{
   int cpunum;
   char buf[32];

   if ((!CPUVec.InstrCount) || (!CPUVec.CurrentCpuNum)) {
      CPUWarning("TCL: WARNING - Too early to read INSTS. Returning 0.\n");
      Tcl_SetVar(interp, name1, "0", 0);
      return NULL;
   }

   ASSERT(CPUVec.CurrentCpuNum);
   ASSERT(CPUVec.InstrCount);

   cpunum = CPUVec.CurrentCpuNum();

   if (flags & TCL_TRACE_READS) {
      PrintLLD(buf, CPUVec.InstrCount(cpunum));
      Tcl_SetVar(interp, name1, buf, 0);
   } else if (flags & TCL_TRACE_WRITES) {
      return "Writes to INSTS ignored";
   }
   return NULL;
}

static int ReadStringCmd(ClientData clientData, Tcl_Interp *interp,
                         int argc, char *argv[])
{
   Tcl_DString dstr;
   char *addrStr;
   VA addr=0;
   char c[2];
   int cpunum;
   
   if ((!CPUVec.PutMemory) || (!CPUVec.GetMemory) || (!CPUVec.CurrentCpuNum)) {
      Tcl_AppendResult(interp, "too early to read memory", 0);
      return TCL_ERROR;
   }

   cpunum = CPUVec.CurrentCpuNum();

   if (argc != 2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], " addr\"", NULL);
      return TCL_ERROR;
   }

   addrStr = argv[1];
   if (String2Reg(interp,addrStr,&addr) != TCL_OK) {
      return TCL_ERROR;
    }

   Tcl_DStringInit(&dstr);

   c[1] = 0;
   while (1) {
      enum result_t res;

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.GetMemory(cpunum, addr, 1, c);
      MEMREF_ENABLE_TLB_CHECK;
      if (res != SUCCESS) {
         break;
      }
      if (!c[0]) {
         break;
      }
      Tcl_DStringAppend(&dstr, c, 1);
      addr++;
   }

   Tcl_AppendResult(interp, Tcl_DStringValue(&dstr), NULL);
   Tcl_DStringFree(&dstr);

   return TCL_OK;
}

static int WriteStringCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
  char *addrStr;
  char *strStr;
  VA addr=0;
  char c;
  int cpunum;
   
  if ((!CPUVec.PutMemory) || (!CPUVec.GetMemory) || (!CPUVec.CurrentCpuNum)) {
    Tcl_AppendResult(interp, "too early to read memory", 0);
    return TCL_ERROR;
  }

  cpunum = CPUVec.CurrentCpuNum();

  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], " addr\"", NULL);
    return TCL_ERROR;
  }

  addrStr = argv[1];
  strStr = argv[2];
  if (String2Reg(interp,addrStr,&addr) != TCL_OK) {
    return TCL_ERROR;
  }

#if 0
  fprintf(stderr, "WriteStringCmd: addrStr=%p strStr=%p\n", addrStr, strStr);
  fprintf(stderr, "                addrStr=%s strStr=%s\n", addrStr, strStr);
#endif
  do {
     enum result_t res;
     c = *strStr++;

     MEMREF_DISABLE_TLB_CHECK;
     res = CPUVec.PutMemory(cpunum, addr, 1, &c);
     MEMREF_ENABLE_TLB_CHECK;
     if (res != SUCCESS) {
      break;
    }
    addr++;
  } while (c);

  return TCL_OK;
}

static int ReadWideStringCmd(ClientData clientData, Tcl_Interp *interp,
                             int argc, char *argv[])
{
   Tcl_DString dstr;
   char *addrStr;
   VA addr=0;
   char c[2];
   int cpunum;
   
   if ((!CPUVec.PutMemory) || (!CPUVec.GetMemory) || (!CPUVec.CurrentCpuNum)) {
      Tcl_AppendResult(interp, "too early to read memory", 0);
      return TCL_ERROR;
   }

   cpunum = CPUVec.CurrentCpuNum();

   if (argc != 2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
                       argv[0], " addr\"", NULL);
      return TCL_ERROR;
   }

   addrStr = argv[1];
   if (String2Reg(interp,addrStr,&addr) != TCL_OK) {
      return TCL_ERROR;
    }

   Tcl_DStringInit(&dstr);

   while (1) {
      enum result_t res;

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.GetMemory(cpunum, addr, 2, c);
      MEMREF_ENABLE_TLB_CHECK;
      if (res != SUCCESS) {
         break;
      }
      c[1] = 0;
      if (!c[0]) {
         break;
      }
      Tcl_DStringAppend(&dstr, c, 1);
      addr+=2;
   }

   Tcl_AppendResult(interp, Tcl_DStringValue(&dstr), NULL);
   Tcl_DStringFree(&dstr);

   return TCL_OK;
}

void MachineStateInit(Tcl_Interp *interp)
{
   Tcl_HashEntry *entryPtr;
   int i, dummy;
   char buf[8];
   
   /*
    *  Initialize register name -> number hash table
    */

   Tcl_InitHashTable(&regTable, TCL_STRING_KEYS);

   for (i=0; i<NUM_REGS; i++) {
      entryPtr = Tcl_CreateHashEntry(&regTable, reg_names[i], &dummy);
      Tcl_SetHashValue(entryPtr, UINT_TO_PTRSIZE(i));
   }

   for (i=0; i<31; i++) {
      sprintf(buf, "r%d", i);
      entryPtr = Tcl_CreateHashEntry(&regTable, buf, &dummy);
      Tcl_SetHashValue(entryPtr, UINT_TO_PTRSIZE(i));
   }
   
   /* John & others desired register access through common symbols
    * ($zero, $at, $r0, $r1, $r2, ...)
    *  So I'm adding that support here...
    */
   for (i = 0; i < NUM_REGS; i++) {
      Tcl_SetVar(interp, reg_names[i], "0", TCL_GLOBAL_ONLY);
      Tcl_TraceVar(interp, reg_names[i],
                   TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                   RegisterAccess, (ClientData) NULL);
   }

   for (i = 0; i < 32; i++) {
      char tbuf[5];
      sprintf(tbuf,"r%d",i);
      Tcl_SetVar(interp, tbuf, "0", TCL_GLOBAL_ONLY);
      Tcl_TraceVar(interp, tbuf,
                   TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                   RegisterAccess, (ClientData) NULL);
   }

   /*
    *  Must define 1 element of the array so TCL thinks its defined
    */
   
   Tcl_SetVar2(interp, "REGISTER", "zero", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "REGISTER",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                RegisterAccess, (ClientData) NULL);

   Tcl_SetVar2(interp, "MEMORY", "0", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "MEMORY",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                MemoryAccess, (ClientData) NULL);

   Tcl_SetVar2(interp, "PHYSICAL", "0", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "PHYSICAL",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                PhysicalAccess, (ClientData) NULL);

   Tcl_SetVar(interp, "CPU", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "CPU",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                CPUAccess, (ClientData) NULL);

   Tcl_SetVar(interp, "CELL", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "CELL",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                CellAccess, (ClientData) NULL);

   Tcl_SetVar(interp, "CYCLES", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "CYCLES",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                CycleAccess, (ClientData) NULL);

   Tcl_SetVar(interp, "INSTS", "0", TCL_GLOBAL_ONLY);
   Tcl_TraceVar(interp, "INSTS",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
                InstsAccess, (ClientData) NULL);

   Tcl_CreateCommand(interp, "readString", ReadStringCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp, "writeString", WriteStringCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp, "readWideString", ReadWideStringCmd,
                     (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);
}



/* 
 * Tcl support for writing binary files.
 * Currently Max files is 8
 */
#define ANN_MAX_BIN_FILES 8
#define ANN_MAX_BIN_ITEMS 16
FILE *binaryOpenFiles[ANN_MAX_BIN_FILES];
Tcl_DString binaryImmediates[ANN_MAX_BIN_FILES];
int         binaryImmediatePtr[ANN_MAX_BIN_FILES];

int 
BinaryOpenCmd(Tcl_Interp *interp, int argc, char *argv[])
{
   int i;
   FILE *fp;

   if ((fp = fopen(argv[2], argv[3])) == NULL) {
      Tcl_AppendResult(interp, "Could not open binary trace file: ", argv[2], NULL);
      return TCL_ERROR;
   } 

   for (i=0; i<ANN_MAX_BIN_FILES; i++) {
      if ((binaryOpenFiles[i] == NULL) && (binaryImmediatePtr[i] == -1)) {
         break;
      }
   }

   if (i == ANN_MAX_BIN_FILES) {
      Tcl_AppendResult(interp, "Too many open binary files", NULL);
      return TCL_ERROR;
   } 

   binaryOpenFiles[i] = fp;
   sprintf(interp->result, "%d", i);
   return TCL_OK;
}

int 
BinaryImmediateCmd(Tcl_Interp *interp, int argc, char *argv[])
{
   int i;

   for (i=0; i<ANN_MAX_BIN_FILES; i++) {
      if ((binaryOpenFiles[i] == NULL) && (binaryImmediatePtr[i] == -1)) {
         break;
      }
   }

   if (i == ANN_MAX_BIN_FILES) {
      Tcl_AppendResult(interp, "Too many open binary files", NULL);
      return TCL_ERROR;
   } 

   Tcl_DStringAppend(&binaryImmediates[i], argv[2], -1);
   binaryImmediatePtr[i] = 0;

   sprintf(interp->result, "%d", i);
   return TCL_OK;
}


int 
BinaryCloseCmd(Tcl_Interp *interp, int argc, char *argv[])
{

   int i;
   FILE *fp;
   Tcl_DString* dsPtr;

   i = atoi(argv[2]);

   if ((i<0) || (i>=ANN_MAX_BIN_FILES)) {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                       argv[2], NULL);
      return TCL_ERROR;
   }

   if ((fp = binaryOpenFiles[i]) != NULL) {
      fclose (fp);
      binaryOpenFiles[i] = NULL;
      return TCL_OK;
   }

   dsPtr = &binaryImmediates[i];
   if (Tcl_DStringLength(dsPtr) != 0) {
      Tcl_DStringFree(dsPtr);
      binaryImmediatePtr[i] = -1;
      return TCL_OK;
   }

   Tcl_AppendResult(interp, "Invalid binary-file descriptor: ", argv[2], NULL);
   return TCL_ERROR;
}

int 
BinarySeekCmd(Tcl_Interp *interp, int argc, char *argv[])
{

   int i;
   FILE *fp;

   i = atoi(argv[2]);

   if ((i<0) || (i>=ANN_MAX_BIN_FILES)) {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                       argv[2], NULL);
      return TCL_ERROR;
   }

   if ((fp = binaryOpenFiles[i]) != NULL) {
      long offset = atol(argv[3]);
      int whence  = atoi(argv[4]);
      fseek(fp, offset, whence);
      return TCL_OK;
   }

   Tcl_AppendResult(interp, "Invalid binary-file descriptor: ", argv[2], NULL);
   return TCL_ERROR;
}

int 
BinaryTellCmd(Tcl_Interp *interp, int argc, char *argv[])
{

   int i;
   FILE *fp;

   i = atoi(argv[2]);

   if ((i<0) || (i>=ANN_MAX_BIN_FILES)) {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                       argv[2], NULL);
      return TCL_ERROR;
   }

   if ((fp = binaryOpenFiles[i]) != NULL) {
      sprintf(interp->result, "%ld", ftell(fp));
      return TCL_OK;
   }

   Tcl_AppendResult(interp, "Invalid binary-file descriptor: ", argv[2], NULL);
   return TCL_ERROR;
}

int
BinaryWriteCmd(Tcl_Interp *interp, int argc, char *argv[])
{

   int i, j;
   FILE *fp;
   unsigned long buf[ANN_MAX_BIN_ITEMS];

   i = atoi(argv[2]);
   if ((i >= 0) && (i<ANN_MAX_BIN_FILES) && ((fp = binaryOpenFiles[i]) != NULL)) {
      argc -= 3;
      j = 3;

      while (argc) {
         for (i=0; argc && i<ANN_MAX_BIN_ITEMS; i++, argc--, j++) {
            buf[i] = strtoul(argv[j], (char **)NULL, 0);
         }
         
         if (fwrite((void *)buf, sizeof(buf[0]), i, fp) != i) {
            Tcl_AppendResult(interp, "Could not write to binary file for descriptor",
                             argv[2], NULL);
            return TCL_ERROR;
         }
      }
      return TCL_OK;
      
   } else {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ", argv[2], NULL);
      return TCL_ERROR;
   }
}

char putmemBuf[4096];

int
BinaryPutmemCmd(Tcl_Interp *interp, int argc, char *argv[])
{
   int i = 0;
   FILE *fp;
   Tcl_DString* dsPtr;
   VA vaddr=0;
   uint  len = 0, strl = 0;
   int cpunum;

   ASSERT(CPUVec.CurrentCpuNum);

   cpunum = CPUVec.CurrentCpuNum();

   if (Tcl_GetInt(interp, argv[2], (int*)&i) != TCL_OK) {
      return TCL_ERROR;
   }

   if ((i<0) || (i>=ANN_MAX_BIN_FILES)) {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                       argv[2], NULL);
      return TCL_ERROR;
   }

   if (String2Reg(interp, argv[3], &vaddr) != TCL_OK) {
      return TCL_ERROR;
   }

   if (Tcl_GetInt(interp, argv[4], (int*)&len) != TCL_OK) {
      return TCL_ERROR;
   }

   ASSERT(len <= 4096); /* size of putmemBuf */
   if ((fp = binaryOpenFiles[i]) != NULL) {
      enum result_t res;

      if (fread((void *)putmemBuf, 1, len, fp) != len) {
         Tcl_AppendResult(interp, "Could not read from binary file for descriptor",
                          argv[2], NULL);
         return TCL_ERROR;
      }

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.PutMemory(cpunum, vaddr, len, putmemBuf);
      MEMREF_ENABLE_TLB_CHECK;
      if (res != SUCCESS) {
         Tcl_AppendResult(interp, "addr not mapped", NULL);
         return TCL_ERROR;
      }

      return TCL_OK;
   }

   dsPtr = &binaryImmediates[i];
   
   if ((strl = Tcl_DStringLength(dsPtr)) != 0) {
      enum result_t res;
      int ip = binaryImmediatePtr[i];
      if ((ip + len) > strl) {
         Tcl_AppendResult(interp, "Too many bytes requested from binary immediate string",
                          argv[2], NULL);
         return TCL_ERROR;
      }

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.PutMemory(cpunum, vaddr, len, ip + Tcl_DStringValue(dsPtr));
      MEMREF_ENABLE_TLB_CHECK;
      if (res != SUCCESS) {
         Tcl_AppendResult(interp, "addr not mapped", NULL);
         return TCL_ERROR;
      }
      binaryImmediatePtr[i] = ip+len;
      return TCL_OK;
   }

   Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                    argv[2], NULL);
   return TCL_ERROR;
}

int
BinaryGetmemCmd(Tcl_Interp *interp, int argc, char *argv[])
{
   int i = 0;
   FILE *fp;
   uint vaddr = 0, len = 0;
   int cpunum;

   ASSERT(CPUVec.CurrentCpuNum);

   cpunum = CPUVec.CurrentCpuNum();

   if (Tcl_GetInt(interp, argv[2], (int*)&i) != TCL_OK) {
      return TCL_ERROR;
   }

   if ((i<0) || (i>=ANN_MAX_BIN_FILES)) {
      Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                       argv[2], NULL);
      return TCL_ERROR;
   }

   if (Tcl_GetInt(interp, argv[3], (int*)&vaddr) != TCL_OK) {
      return TCL_ERROR;
   }

   if (Tcl_GetInt(interp, argv[4], (int*)&len) != TCL_OK) {
      return TCL_ERROR;
   }

   ASSERT(len <= 4096); /* size of putmemBuf */
   if ((fp = binaryOpenFiles[i]) != NULL) {
      enum result_t res;

      MEMREF_DISABLE_TLB_CHECK;
      res = CPUVec.GetMemory(cpunum, vaddr, len, putmemBuf);
      MEMREF_ENABLE_TLB_CHECK;

      if (res != SUCCESS) {
         Tcl_AppendResult(interp, "addr not mapped", NULL);
         return TCL_ERROR;
      }

      if (fwrite((void *)putmemBuf, 1, len, fp) != len) {
         Tcl_AppendResult(interp, "Could not write to binary file for descriptor",
                          argv[2], NULL);
         return TCL_ERROR;
      }

      return TCL_OK;
   }

   Tcl_AppendResult(interp, "Invalid binary-file descriptor: ",
                    argv[2], NULL);
   return TCL_ERROR;
}

static tclcmd binCmds[] = {
{   "open",        4, BinaryOpenCmd,      " open pathname mode"},
{   "close",       3, BinaryCloseCmd,     " close filedesc"},
{   "write",      -1, BinaryWriteCmd,     " write filedesc ?word1 ...?"},
{   "putmem",      5, BinaryPutmemCmd,    " putmem filedesc vaddr len"},
{   "getmem",      5, BinaryGetmemCmd,    " getmem filedesc vaddr len"},
{   "immediate",   3, BinaryImmediateCmd, " immediate string"},
{   "seek",        5, BinarySeekCmd,      " seek filedesc offset whence"},
{   "tell",        3, BinaryTellCmd,      " tell filedesc"},
{   NULL,            0,  NULL,               NULL}
};



void
BinaryInit(Tcl_Interp *interp)
{
   int i;

   Tcl_CreateCommand(interp, "binary", DispatchCmd,
                     (ClientData)binCmds, (Tcl_CmdDeleteProc*)NULL);

   for (i=0; i<ANN_MAX_BIN_FILES; i++) {
      binaryOpenFiles[i] = NULL;
      Tcl_DStringInit(&binaryImmediates[i]);
      binaryImmediatePtr[i] = -1;
   }
}

int
String2Reg(Tcl_Interp *interp, char *string, Reg *val)
{
  uint64 value=0;
  if (sizeof(Reg)==sizeof(int)) {
    return Tcl_GetInt(interp, string, (int*)val);
  }
  ASSERT (sizeof(Reg)==sizeof(uint64));

#ifdef __alpha
   if (string[0] == '0' && string[1]== 'x') { 
      if (sscanf(string,"0x%lx",&value)!=1) {
	 return TCL_ERROR;
      } 
   } else {
      if (sscanf(string,"%ld",&value)!=1) { 
	 return TCL_ERROR;
      }
   }
#else  
   if (string[0] == '0' && string[1]== 'x') { 
      if (sscanf(string,"0x%llx",&value)!=1) {
	 return TCL_ERROR;
      } 
   } else {
      if (sscanf(string,"%lld",&value)!=1) { 
	 return TCL_ERROR;
      }
   }
#endif
   *val = value;
   return TCL_OK;

} /* String2Reg () */



