/***********************************************************************
 *
 *	ABSTRACT:    External Trace Routine Macros For SimOS
 *
 *	$Id: external_trace.h,v 1.8 1999/11/09 22:49:47 nikos Exp $
 *	$Author: nikos $
 *	$Date: 1999/11/09 22:49:47 $
 *
 ***********************************************************************/

#ifndef EXTERNAL_TRACE_H
#define EXTERNAL_TRACE_H

#ifdef EXTERNAL_TRACE

#   ifndef SIMOS
#      define SIMOS
#      define SIMOS_DEFINED_IN_EXTERNAL_TRACE
#   endif /* SIMOS */

#   include "alpha.h"

/* Macros to help with the external trace macro substitution */
#   define EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(a,b)     EXTERNAL_TRACE_MACRO_PRO_CPP(a, b)
#   define EXTERNAL_TRACE_MACRO_PRO_CPP(a, b)        EXTERNAL_TRACE_MACRO_CPP(a, b)
#   define EXTERNAL_TRACE_MACRO_CPP(a, b)            a##b

/* Macros to include the external trace-specific header file */
#   define EXTERNAL_TRACE_INCLUDE_FILE               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _HEADER)
#   define FILE_INCLUDE_MACRO_PRO_CPP(a)             FILE_INCLUDE_MACRO_CPP(a)
#   define FILE_INCLUDE_MACRO_CPP(a)                 #a

#   include FILE_INCLUDE_MACRO_PRO_CPP(EXTERNAL_TRACE_INCLUDE_FILE)

#   ifdef SIMOS_DEFINED_IN_EXTERNAL_TRACE
#      undef SIMOS
#      undef SIMOS_DEFINED_IN_EXTERNAL_TRACE
#   endif /* SIMOS_DEFINED_IN_EXTERNAL_TRACE */

#endif /* EXTERNAL_TRACE */

/*
 * Say you want to create trace macros of the form SECRET_NAME_dump_cache(), SECRET_NAME_set_vea(),
 * etc, etc, but you don't want the name SECRET_NAME to appear anywhere else,
 * except your own trace-specific source files. To do that, follow the steps below:
 * 
 * 1. Create your source files x.c, x.h, y.c, y.h, ... You should have in these files
 *    the definitions for the macros SECRET_NAME_yyy(), etc, etc.
 *    The Makefile currently requires these source files to be in cpus-alpha/alpha-shared.
 *    This can be easily changed.
 *
 * 2. In the simos code, instead of inserting macros with names SECRET_NAME_yyy(),
 *    insert macros with names EXTERNAL_TRACE_yyy().
 *
 * 3. For every trace macro of the form EXTERNAL_TRACE_yyy(a, b) inserted in simos code,
 *    insert in this file the following macro definitions :
 *    #ifdef EXTERNAL_TRACE
 *    #define EXTERNAL_TRACE_yyy           EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _yyy)
 *    #else
 *    #define EXTERNAL_TRACE_yyy(_a, _b)
 *    #endif
 *
 * 4. Compile with :
 *    EXTERNAL_TRACE=SECRET_NAME
 *    SECRET_NAME_INC_DIR=/x/y      (include path)
 *    SECRET_NAME_LIB=/x/y/libz.a   (tracing lib)
 *    SECRET_NAME_OBJS=x.o y.o      (objects holding secret trace functions)
 *    SECRET_NAME_HEADER=w.h        (header file to include in SimOS code)
 */

#ifdef EXTERNAL_TRACE /*********************************************/

#define EXTERNAL_TRACE_SET_VEA                    EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_VEA)
#define EXTERNAL_TRACE_INSTRUCTION                EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _INSTRUCTION)
#define EXTERNAL_TRACE_PE                         EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _PE)
#define EXTERNAL_TRACE_PS                         EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _PS)
#define EXTERNAL_TRACE_DTBCM                      EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DTBCM)
#define EXTERNAL_TRACE_MCSR                       EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _MCSR)
#define EXTERNAL_TRACE_ICSR                       EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _ICSR)
#define EXTERNAL_TRACE_ASN                        EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _ASN)
#define EXTERNAL_TRACE_INTERRUPT                  EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _INTERRUPT)
#define EXTERNAL_TRACE_PTE                        EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _PTE)
#define EXTERNAL_TRACE_PTE_FROM_TBE               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _PTE_FROM_TBE)
#define EXTERNAL_TRACE_DTBCM                      EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DTBCM)
#define EXTERNAL_TRACE_SET_TBHOOK                 EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_TBHOOK)
#define EXTERNAL_TRACE_GET_TBHOOK                 EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _GET_TBHOOK)
#define EXTERNAL_TRACE_ICACHE_MISS                EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _ICACHE_MISS)
#define EXTERNAL_TRACE_DCACHE_RMISS               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DCACHE_RMISS)
#define EXTERNAL_TRACE_DCACHE_WMISS               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DCACHE_WMISS)
#define EXTERNAL_TRACE_DMA_REQ                    EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DMA_REQ)
#define EXTERNAL_TRACE_SET_BR_ACTION              EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_BR_ACTION)
#define EXTERNAL_TRACE_CLEAR_BR_ACTION            EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _CLEAR_BR_ACTION)
#define EXTERNAL_TRACE_SET_BR_PC                  EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_BR_PC)
#define EXTERNAL_TRACE_SET_BR_NEXTPC              EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_BR_NEXTPC)
#define EXTERNAL_TRACE_EMIT_BR_ACTION             EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _EMIT_BR_ACTION)
#define EXTERNAL_TRACE_ICACHE_FLUSH               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _ICACHE_FLUSH)
#define EXTERNAL_TRACE_DCACHE_FLUSH               EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DCACHE_FLUSH)
#define EXTERNAL_TRACE_SET_DEST_REG_PTR           EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _SET_DEST_REG_PTR)
#define EXTERNAL_TRACE_DEST_REGISTER_VALUE        EXTERNAL_TRACE_MACRO_PRO_PRO_CPP(EXTERNAL_TRACE, _DEST_REGISTER_VALUE)

#else /* ~EXTERNAL_TRACE*/ /*********************************************/

#define EXTERNAL_TRACE_SET_VEA(_a,_b)
#define EXTERNAL_TRACE_INSTRUCTION(_a,_b,_c)
#define EXTERNAL_TRACE_PE(_a)
#define EXTERNAL_TRACE_PS(_a)
#define EXTERNAL_TRACE_DTBCM(_a)
#define EXTERNAL_TRACE_MCSR(_a)
#define EXTERNAL_TRACE_ICSR(_a)
#define EXTERNAL_TRACE_ASN(_a,_b)
#define EXTERNAL_TRACE_INTERRUPT(_a)
#define EXTERNAL_TRACE_PTE(_a,_b,_c)
#define EXTERNAL_TRACE_PTE_FROM_TBE(_a,_b)
#define EXTERNAL_TRACE_DTBCM(_a)
#define EXTERNAL_TRACE_SET_TBHOOK(_a)
#define EXTERNAL_TRACE_GET_TBHOOK()
#define EXTERNAL_TRACE_ICACHE_MISS()
#define EXTERNAL_TRACE_DCACHE_RMISS()
#define EXTERNAL_TRACE_DCACHE_WMISS()
#define EXTERNAL_TRACE_DMA_REQ(_cpu, _pa, _len, _write)
#define EXTERNAL_TRACE_SET_BR_ACTION(_cpu, _a)
#define EXTERNAL_TRACE_CLEAR_BR_ACTION(_cpu)
#define EXTERNAL_TRACE_SET_BR_PC(_cpu, _pc)
#define EXTERNAL_TRACE_SET_BR_NEXTPC(_cpu, _pc)
#define EXTERNAL_TRACE_EMIT_BR_ACTION(_cpu, _pc)
#define EXTERNAL_TRACE_ICACHE_FLUSH(_cpu, _writeback, _retain)
#define EXTERNAL_TRACE_DCACHE_FLUSH(_cpu, _writeback, _retain)
#define EXTERNAL_TRACE_SET_DEST_REG_PTR(_value_ptr)
#define EXTERNAL_TRACE_DEST_REGISTER_VALUE(_cpu)

#endif /* EXTERNAL_TRACE*/ /*********************************************/

#ifndef MP_SIM_CYCLE_GRAIN
#    ifdef EXTERNAL_TRACE
#        define MP_SIM_CYCLE_GRAIN      64
#    else /* ~EXTERNAL_TRACE */
#        define MP_SIM_CYCLE_GRAIN       1
#    endif /* EXTERNAL_TRACE */
#endif /* ~MP_SIM_CYCLE_GRAIN */

#endif /* EXTERNAL_TRACE_H */
