
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <alpha/inst.h>


#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "params.h"

#include "gamma.h"
#include "epsilon.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "disassembler.h"
#include "alpha_params.h"
#include "alpha_regs.h"
#include "alpha_trace.h"

#include "epsilon-scheduler.h"



unsigned long SIMOS_fetch_next_instr(int cpu, unsigned long predicted_pc, 
                                     VA *actualVPC, PA *actualPPC, union alpha_instruction *instr,
                                     void *instrDesc)
{
   AlphaState *pe = curPE = PE[cpu]; /* this statement is here to update curPE. -Jamey 4/30/98 */
   long inum = EpsilonFetchInstruction(cpu, predicted_pc);
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   *actualVPC = iis->PC;
   *actualPPC = iis->ppc;
   *instr = iis->instr;
   if (epsilonParams.debugScheduler)
      CPUWarning("  SIMOS_fetch_next_instr: inum=%ld pc=%lx nextpc=%lx epe->nextpc=%lx\n", 
                 inum, iis->PC, iis->nextpc, epe->nextpc);
   if (instrDesc != NULL) {
      iis->schedulerInstrDesc = instrDesc;
#if 0
check this out
      if (schedulerOps->pendingTrap[cpu].pending) {

         schedulerOps->notifyTrap(cpu, iis->schedulerInstrDesc, 
                                  schedulerOps->trapReasonMap[schedulerOps->pendingTrap[cpu].trapType], iis->PC);
         schedulerOps->pendingTrap[cpu].pending = 0;
      }
#endif
   }
   /* maybe stall fetch */
   return inum;
}

void SIMOS_set_scheduler_instr_desc(int cpu, unsigned long inum, void *instrDesc)
{
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   iis->schedulerInstrDesc = instrDesc;
#if 0
   handle pending trap here?
#endif
}

void SIMOS_fetch_instruction2(int cpu, unsigned long inum)
{
   curPE = PE[cpu];
   EpsilonFetchInstruction2(cpu, inum);
}

void SIMOS_map(int cpu, unsigned long inum)
{
   curPE = PE[cpu];
   EpsilonMapInstruction(cpu, inum);
}

unsigned long SIMOS_issue_memop(int cpu, unsigned long inum)
{
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   curPE = PE[cpu];
   EpsilonIssueInstruction(cpu, inum);
   return iis->eaddr;
}

unsigned long SIMOS_issue_cbranch(int cpu, unsigned long inum, unsigned long *pnextpc)
{
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   VA PC = iis->PC;
   VA nextpc;
   curPE = PE[cpu];
   EpsilonIssueInstruction(cpu, instid);
   nextpc = iis->nextpc;
   if (0)
   CPUWarning("SIMOS_issue_cbranch: cpu=%d inum=%ld pc=%lx taken=%d\n",
              cpu, inum, iis->PC, (nextpc != (PC+4)));
   *pnextpc = nextpc;
   return nextpc != (PC+4);  
}

unsigned long SIMOS_issue_jump(int cpu, unsigned long inum)
{
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   curPE = PE[cpu];
   EpsilonIssueInstruction(cpu, instid);
   if (0)
   CPUWarning("SIMOS_issue_jump: cpu=%d inum=%d pc=%lx nextpc=%lx\n",
              cpu, inum, iis->PC, iis->nextpc);
   return iis->nextpc ;  
}

void SIMOS_issue_other(int cpu, unsigned long inum)
{
   curPE = PE[cpu];
   EpsilonIssueInstruction(cpu, inum);
}

void SIMOS_do_read(int cpu, unsigned long inum)
{
   curPE = PE[cpu];
   EpsilonRequestMemop(cpu, inum);
   EpsilonCompleteMemop(cpu, inum);
}

void SIMOS_do_write(int cpu, unsigned long inum)
{
   curPE = PE[cpu];
   EpsilonRequestMemop(cpu, inum);
   EpsilonCompleteMemop(cpu, inum);
}

void SIMOS_fail_stc(int cpu, unsigned long inum)
{
   /* what to do here? */
   UNIMPLEMENTED();
}

void SIMOS_commit_instr(int cpu, unsigned long inum)
{
   Instid instid = inumInstid(inum);
   curPE = PE[cpu];
   if (EpsilonInstructionIsStore(cpu, instid)) {
      SIMOS_do_write(cpu, inum);
   };
   EpsilonCommitInstruction(cpu, inum);
}


void SIMOS_kill_instrs(int cpu, unsigned long causeInum, int killNext)
{
   AlphaState *pe = PE[cpu];
   EpsilonState *epe = EPE[cpu];
   InflightInstructionState *iis = &epe->iqueue[inumInstid(causeInum)];
   unsigned long abortInum = causeInum;
   curPE = PE[cpu];
   /* if the reason is a mispredict, then we start trapping at the next instruction */
   if (epsilonParams.debugScheduler)
      CPUWarning("SIMOS_kill_instrs: causeInum=%ld killNext=%d\n", causeInum, killNext);
   if (killNext) {
      abortInum = iis->inorderNext;
   }
   ASSERT(epe->trapPending);
#if 0   
   ASSERT(epe->trapCauseInum == causeInum);
   ASSERT(epe->trapAbortInum == abortInum);
#else
   if (epe->trapCauseInum != causeInum && epsilonParams.debugScheduler)
      CPUWarning("SIMOS_kill_instrs: @@Mismatch@@ epe->trapCauseInum=%ld epe->trapAbortInum=%ld "
                 "causeInum=%ld killNext=%ld epe->trapType=%d:%s\n",
                 epe->trapCauseInum, epe->trapAbortInum,
                 causeInum, killNext, epe->trapType, EpsilonTTName(epe->trapType));
#endif
   EpsilonProcessPendingTrap(pe, epe);
}

/* this is a callback from the scheduler, telling us it has detected a trap */
void SIMOS_notify_trap(int cpu, unsigned long trapCauseInum, int epsilonTrapType)
{
   if (!epsilonTrapNotificationInProgress) {
      schedulerOps->trapNotificationInProgress = 1;
      /* check for recursion */
      if (epsilonParams.debugScheduler)
         CPUWarning("SIMOS_notify_trap: cpu=%d trapCauseInum=%ld schedulerTrapType=%d\n",
                    cpu, trapCauseInum, epsilonTrapType);
      EpsilonNotifyTrap(cpu, trapCauseInum, epsilonTrapType);
      schedulerOps->trapNotificationInProgress = 0;
   }
}

unsigned long SIMOS_get_next_fetch_pc(int cpu)
{
   EpsilonState *epe = EPE[cpu];
   Inum inum = epe->lastInumUntrapped;
   InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
   if (epsilonParams.debugScheduler)
      CPUWarning("SIMOS_get_next_fetch_pc: cpu=%d lastInumUntrapped=%ld nextpc=%lx\n",
                 cpu, epe->lastInumFetched, iis->nextpc);
   return iis->nextpc;
}

unsigned long SIMOS_get_next_pc(int cpu, unsigned long inum)
{
   EpsilonState *epe = EPE[cpu];
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   return iis->nextpc;
}


unsigned long SIMOS_get_next_inum(int cpu)
{
   UNIMPLEMENTED();
   return 0;
}

unsigned long SIMOS_set_branch_prediction(int cpu, int taken)
{
   VA nextpc;
   EpsilonState *epe = EPE[cpu];
   Inum inum = epe->lastInumUntrapped;
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   if (taken) {
      long disp = iis->instr.b_format.branch_displacement;
      nextpc = iis->PC + (disp <<2)+4;
   } else {
      nextpc = iis->PC + 4;
   }
   epe->nextpc = iis->nextpc = nextpc;
   if (epsilonParams.debugScheduler)
      CPUWarning("SIMOS_set_branch_prediction: cpu=%d inum=%ld nextpc=%lx\n",
                 cpu, epe->lastInumFetched, epe->nextpc);
   return nextpc;
}

unsigned long SIMOS_set_jump_prediction(int cpu, unsigned long predicted_jump_target)
{
   EpsilonState *epe = EPE[cpu];
   Inum inum = epe->lastInumUntrapped;
   Instid instid = inumInstid(inum);
   InflightInstructionState *iis = &epe->iqueue[instid];
   epe->nextpc = iis->nextpc = predicted_jump_target;
   if (epsilonParams.debugScheduler)
      CPUWarning("SIMOS_set_jump_prediction: cpu=%d inum=%ld nextpc=%lx\n",
                 cpu, epe->lastInumFetched, epe->nextpc);
   return predicted_jump_target;
}

unsigned int SIMOS_need_to_abort(int cpu, unsigned long inum)
{
   return 0;
}



SimContext SIMOS_simContext = {
   NULL,
   SIMOS_fetch_next_instr,
   SIMOS_fetch_instruction2,
   SIMOS_set_scheduler_instr_desc,
   SIMOS_map,
   SIMOS_issue_memop,
   SIMOS_issue_cbranch,
   SIMOS_issue_jump,
   SIMOS_issue_other,
   SIMOS_do_read,
   SIMOS_do_write,
   SIMOS_fail_stc,
   SIMOS_commit_instr,
   SIMOS_kill_instrs,
   SIMOS_notify_trap,
   SIMOS_get_next_fetch_pc,
   SIMOS_get_next_pc,
   SIMOS_get_next_inum,
   SIMOS_set_branch_prediction,
   SIMOS_set_jump_prediction,
   SIMOS_need_to_abort
};

