#ifndef EPSILON_SCHEDULER_H
#define EPSILON_SCHEDULER_H

typedef  enum { 
  EpsilonTTExitSimulator, 
  EpsilonTTRegisterChangedExternally, 
  EpsilonTTInterrupt, 
  EpsilonTTFetchStalled, 
  EpsilonTTIMMUException, 
  EpsilonTTIllegalOpcode,
  EpsilonTTFEN,
  EpsilonTTArith,
  EpsilonTTDMMUHWReadException, 
  EpsilonTTDMMUReadException, 
  EpsilonTTDMMUWriteException,
  EpsilonTTSchedulerException, /* signaled by the scheduler */
  EpsilonTTMemTrap,
  /* these abort later instructions */
  EpsilonTTMispredict,
  EpsilonTTPalCall,
  EpsilonTTHwRei,
  EpsilonTTNoTrap,
  EpsilonTTNumTypes
} EpsilonTrapType;

/**
 * SimContext provides callbacks to Epsilon that may be used by the scheduler.
 * A structure is used so that the scheduler can be written as a shared-object with no other 
 * interface to SimOS, simplifying linking.
 */
typedef struct SimContext {
   FILE *output;
   unsigned long (*fetch_next_instr)(int cpu, unsigned long predicted_pc,
                                     unsigned long *actualVPC, unsigned long *actualPPC, 
                                     union alpha_instruction *instr,
                                     void *instr_desc);
   void (*fetch_instr2)(int cpu, unsigned long inum);
   void (*set_evsim_instr_desc)(int cpu, unsigned long inum, void *instr_desc);
   void (*map)(int cpu, unsigned long inum);
   unsigned long (*issue_memop)(int cpu, unsigned long inum); 
   unsigned long (*issue_cbranch)(int cpu, unsigned long inum, unsigned long *pnextpc);
   unsigned long (*issue_jump)(int cpu, unsigned long inum);
   void (*issue_other)(int cpu, unsigned long inum);
   void (*do_read)(int cpu, unsigned long inum);
   void (*do_write)(int cpu, unsigned long inum);
   void (*fail_stc)(int cpu, unsigned long inum);
   void (*commit_instr)(int cpu, unsigned long inum);
   void (*kill_instrs)(int cpu, unsigned long inum, int kill_next);
   void (*notify_trap)(int cpu, unsigned long trapCauseInum, int trapType);
   unsigned long (*get_next_fetch_pc)(int cpu);
   unsigned long (*get_next_pc)(int cpu, unsigned long inum);
   unsigned long (*get_next_inum)(int cpu);
   unsigned long (*set_branch_prediction)(int cpu, int taken);
   unsigned long (*set_jump_prediction)(int cpu, unsigned long predicted_jump_target);
   unsigned int (*need_to_abort)(int cpu, unsigned long inum);
} SimContext;

/** 
 * Each scheduler provides an instance of SchedulerOps to Epsilon,
 * implementing the methods and fields required by Epsilon.  
 */
typedef struct SchedulerOps {
   void (*tclInit)(void *interp);
   void (*init)(SimContext *simContext, int nthreads);
   void (*preCycle)();
   void (*cycle)();
   void (*postCycle)();
   void (*term)();
   void (*notifyTrap)(int cpu, unsigned long trapCauseInum, unsigned long trapAbortInum, 
                      void *schedulerInstrDesc, int trapType);
   int trapNotificationInProgress;
   int debug;
} SchedulerOps; 

extern SimContext SIMOS_simContext;
extern SchedulerOps *schedulerOps; /* this points to the scheduler currently in use */

extern void SIMOS_SchedulerInit();

#endif /* EPSILON_SCHEDULER_H */
