/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


#include <stdio.h>
#include <alpha/inst.h>

#include "sim_error.h"

#include "gamma.h"
#include "alpha_regs.h"
#include "alpha_insts.h"

#ifdef GAMMA_ASM_EXTERNS
 
OperateFunction gammaOperateTable[1<<13];
MemoryFunction  gammaMemoryTable[1<<6];
FPFunction      gammaFPTable[(1<<17)];

uint            gammaOperateFlags[1<<13];
uint            gammaFPFlags[1<<17];


#define GAMMA_OPTFMT(_opcode,_func,NAME)   \
    extern void NAME ## _ff(Reg,Reg,const Reg *rcp,Reg *rdestp);

#define GAMMA_C_OPTFMT       GAMMA_OPTFMT
#define GAMMA_CANTRAP_OPTFMT GAMMA_OPTFMT
#define GAMMA_COMPLEX_OPTFMT GAMMA_OPTFMT
#define GAMMA_2ARG_OPTFMT    GAMMA_OPTFMT

#define GAMMA_MEM_LOAD(NAME) \
extern void NAME ## _ff(Reg *, const Reg *, Reg*, PA, VA);
#define GAMMA_MEM_STORE   GAMMA_MEM_LOAD
#define GAMMA_MEM_FPLOAD  GAMMA_MEM_LOAD
#define GAMMA_MEM_FPSTORE GAMMA_MEM_LOAD

#define GAMMA_FP_OPER(_o,_f,NAME) extern void NAME ## _ff(const double *a, const double* b, const double *c, double *rdest, Reg *);
#define GAMMA_CANTRAP_FP_OPER(_o,_f,NAME) extern void NAME ## _ff(const double *a, const double* b, const double *c, double *rdest, Reg *);
#define GAMMA_SFP_OPER(_o,_f,NAME) extern void NAME ## _ff(const double *a, const double *B, const double *c, double *rdest, Reg *);
#define GAMMA_CANTRAP_SFP_OPER(_o,_f,NAME) extern void NAME ## _ff(const double *a, const double *B, const double *c, double *rdest, Reg *);
#define GAMMA_CVTFP_OPER    GAMMA_FP_OPER
#define GAMMA_CANTRAP_CVTFP_OPER    GAMMA_CANTRAP_FP_OPER
#define GAMMA_EMPTY_FP_OPER GAMMA_FP_OPER
#define GAMMA_FPC_OPER      GAMMA_FP_OPER

#define GAMMA_COMPLEX_FP(_o,NAME) extern void NAME ## _ff(const double *a, const double* b, const double *c, double *rdest, Reg *);

#else /* GAMMA_ASM_EXTERNS */

#ifdef GAMMA_ASM_TABLE
#undef GAMMA_OPTFMT
#undef GAMMA_2ARG_OPTFMT
#undef GAMMA_C_OPTFMT
#undef GAMMA_CANTRAP_OPTFMT
#undef GAMMA_COMPLEX_OPTFMT
#undef GAMMA_MEM_LOAD
#undef GAMMA_MEM_FPLOAD
#undef GAMMA_MEM_FPSTORE
#undef GAMMA_FP_OPER
#undef GAMMA_CANTRAP_FP_OPER
#undef GAMMA_FPC_OPER
#undef GAMMA_SFP_OPER
#undef GAMMA_CANTRAP_SFP_OPER
#undef GAMMA_COMPLEX_FP
#undef GAMMA_CVTFP_OPER
#undef GAMMA_CANTRAP_CVTFP_OPER
#undef GAMMA_EMPTY_FP_OPER

#define GAMMA_OPTFMT(_opcode,_func,NAME) \
    gammaOperateTable[(_opcode <<7) + _func] = NAME ## _ff; \
    gammaOperateFlags[(_opcode <<7) + _func] = AXPTYPE_INT;

#define GAMMA_2ARG_OPTFMT(_opcode,_func,NAME) \
    gammaOperateTable[(_opcode <<7) + _func] = NAME ## _ff; \
    gammaOperateFlags[(_opcode <<7) + _func] = AXPTYPE_INT;

#define GAMMA_C_OPTFMT(_opcode,_func,NAME) \
    gammaOperateTable[(_opcode <<7) + _func] = NAME ## _ff; \
    gammaOperateFlags[(_opcode <<7) + _func] = (AXPTYPE_INT|AXPTYPE_INT_CONDITIONAL);

#define GAMMA_CANTRAP_OPTFMT(_opcode,_func,NAME) \
    gammaOperateTable[(_opcode <<7) + _func] = NAME ## _ff; \
    gammaOperateFlags[(_opcode <<7) + _func] = (AXPTYPE_INT|AXPTYPE_CANTRAP);

#define GAMMA_COMPLEX_OPTFMT(_opcode,_func,NAME) \
    gammaOperateTable[(_opcode <<7) + _func] = NAME ## _ff; \
    gammaOperateFlags[(_opcode <<7) + _func] = (AXPTYPE_INT|AXPTYPE_GAMMA);


#define GAMMA_MEM_LOAD(NAME) \
    gammaMemoryTable[op_ ## NAME] = NAME ## _ff;

#define GAMMA_MEM_STORE  GAMMA_MEM_LOAD
#define GAMMA_MEM_FPLOAD GAMMA_MEM_LOAD
#define GAMMA_MEM_FPSTORE GAMMA_MEM_LOAD

#define GAMMA_FP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP);

#define GAMMA_CANTRAP_FP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_CANTRAP);

#define GAMMA_SFP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_SINGLE);

#define GAMMA_CANTRAP_SFP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_SINGLE|AXPTYPE_CANTRAP);

#define GAMMA_CVTFP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_CVT);

#define GAMMA_CANTRAP_CVTFP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_CVT|AXPTYPE_CANTRAP);

#define GAMMA_EMPTY_FP_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_EMPTY);

#define GAMMA_FPC_OPER(opcode,func,NAME) \
    gammaFPTable[(opcode<<11)+ func] = NAME ## _ff;\
    gammaFPFlags[(opcode<<11)+ func] = (AXPTYPE_FP|AXPTYPE_FP_CONDITIONAL);

#define GAMMA_COMPLEX_FP(o_,NAME) \
    gammaFPTable[(op_ ## o_ <<11)+ o_ ## _ ## NAME] = NAME ## _ff;\
    gammaFPFlags[(op_ ## o_ <<11)+ o_ ## _ ## NAME] = (AXPTYPE_FP);

#else /* GAMMA_ASM_TABLE */

#ifndef __GNUC__
#include <c_asm.h>
#endif

#ifdef __GNUC__
#define GAMMA_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    asm(#NAME " %1, %2, %0"  : "&=r" (lv1): "r" (a) , "r" (b)); \
   *rdestp = lv1; \
}
#else
#define GAMMA_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    lv1 = asm(#NAME " %a0,%a1,%v0", a, b); \
   *rdestp = lv1; \
}
#endif

#ifdef __GNUC__
#define GAMMA_2ARG_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    asm(#NAME " %1, %0"  : "&=r" (lv1): "r" (b)); \
   *rdestp = lv1; \
}
#else
#define GAMMA_2ARG_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    lv1 = asm(#NAME " %a0,%v0", b); \
   *rdestp = lv1; \
}
#endif


/* ************************************************************888
 * XXXXX XXXX extremely brittle code
 * make sure that the read spec for lv1 matches the number of the write spec for lv1
 * so that GCC will put both uses in the same register. -Jamey Hicks
 * 
 * *****************/
int never = 0;
#ifdef __GNUC__
#define GAMMA_C_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(long a, long b, const long *c, long *rdestp) \
{ \
    long lv1 = *c;  \
    asm(#NAME " %1, %2, %0"  : "&=r" (lv1): "r" (a) , "r" (b), "0" (lv1)); \
   *rdestp = lv1; \
}
#else
#define GAMMA_C_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(long a, long b, const long *c, long *rdestp) \
{ \
    long lv1 = *c;  \
    lv1 = asm("or %a2,%a2,%v0; " #NAME " %a0,%a1,%v0", a, b, lv1); \
   *rdestp = lv1; \
}
#endif

#ifdef __GNUC__
#define GAMMA_CANTRAP_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    asm("trapb; " #NAME " %1, %2, %0; trapb"  : "&=r" (lv1): "r" (a) , "r" (b)); \
   *rdestp = lv1; \
}
#else
#define GAMMA_CANTRAP_OPTFMT(_opcode,_func,NAME) \
void NAME ## _ff(Reg a, Reg b, Reg *c, Reg *rdestp) \
{ \
    Reg lv1;  \
    lv1 = asm("trapb; " #NAME " %a0, %a1, %v0; trapb", a, b); \
   *rdestp = lv1; \
}
#endif

/* implementation is in gamma.c */
#define GAMMA_COMPLEX_OPTFMT(_opcode,_func,NAME)

#ifdef __GNUC__
#define GAMMA_MEM_LOAD(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *raPtr, Reg *mAddr, PA pAddr, VA foo) \
{ \
     Reg lv;  \
     asm(#NAME " %0, 0(%1)" : "&=r" (lv) :  "r" (mAddr)) ; \
     *rdestp = lv;  \
}
#else
#define GAMMA_MEM_LOAD(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *raPtr, Reg *mAddr, PA pAddr, VA foo) \
{ \
     Reg lv;  \
     lv = asm(#NAME " %v0, 0(%a0)", mAddr) ; \
     *rdestp = lv;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_MEM_STORE(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *a, Reg *mAddr, PA pAddr, VA foo) \
{ \
     asm(#NAME " %1, %0" : "=o" (*mAddr) :  "r" (*a)) ; \
}
#else
#define GAMMA_MEM_STORE(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *a, Reg *mAddr, PA pAddr, VA foo) \
{ \
     asm(#NAME " %a0, 0(%a1)", *a, mAddr) ; \
}
#endif

#ifdef __GNUC__
#define GAMMA_MEM_FPLOAD(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *raPtr, Reg *mAddr, PA pAddr, VA foo) \
{ \
     double lv;  \
     asm(#NAME " %0, 0(%1)" : "&=f" (lv) :  "r" (mAddr)) ; \
     *((double*)rdestp) = lv;  \
}
#else
#define GAMMA_MEM_FPLOAD(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *raPtr, Reg *mAddr, PA pAddr, VA foo) \
{ \
     double lv;  \
     lv = dasm(#NAME " %f0, 0(%a0)", mAddr) ; \
     *((double*)rdestp) = lv;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_MEM_FPSTORE(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *a, Reg *mAddr, PA pAddr, VA foo) \
{ \
     double rv = *(double*)a; \
     asm(#NAME " %1, %0" : "=o" (*mAddr) :  "f" (rv)) ; \
}
#else
#define GAMMA_MEM_FPSTORE(NAME) \
void NAME ##_ff (Reg *rdestp, const Reg *a, Reg *mAddr, PA pAddr, VA foo) \
{ \
     double rv = *(double*)a; \
     dasm(#NAME " %f16, 0(%a1)", rv, mAddr) ; \
}
#endif

#ifdef __GNUC__
#define GAMMA_FP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double *)cr; \
  asm("mt_fpcr %4; trapb; " #NAME " %2 %3 %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fav) , "f" (fbv), "f" (fpcr)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_FP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double *)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; " #NAME " %f16, %f17, %f0; trapb", fav, fbv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_CANTRAP_FP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double *)cr; \
  asm("mt_fpcr %4; trapb; " #NAME " %2 %3 %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fav) , "f" (fbv), "f" (fpcr)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_CANTRAP_FP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double *)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; " #NAME " %f16,%f17,%f0; trapb ", fav, fbv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_FPC_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  asm("mt_fpcr %4; trapb; " #NAME " %2 %3 %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fav) , "f" (fbv), "f" (fpcr), "0" (fcv)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_FPC_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; fmov %f18, %f0; " #NAME " %f16,%f17,%f0; trapb;", fav, fbv, fcv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif


#ifdef __GNUC__
#define GAMMA_CVTFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  asm("mt_fpcr %3; trapb; " #NAME " %2  %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fbv), "f" (fpcr)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_CVTFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; " #NAME " %f16,%f0; trapb ", fbv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_CANTRAP_CVTFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  asm("mt_fpcr %3; trapb; " #NAME " %2 %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fbv), "f" (fpcr)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_CANTRAP_CVTFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; " #NAME " %f16, %f0; trapb ", fbv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif

#ifdef __GNUC__
#define GAMMA_SFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  asm("mt_fpcr %4; trapb; " #NAME " %2 %3 %0; trapb; mf_fpcr %1 " \
  : "=f" (fcv) , "=f" (fpcr) \
  : "f" (fav) , "f" (fbv), "f" (fpcr)); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#else
#define GAMMA_SFP_OPER(_o,_f,NAME) \
void NAME ## _ff(const double *a, const double *b, const double *c, double *rdestp, Reg *cr) \
{ \
  register double fav = *a; \
  register double fbv = *b; \
  register double fcv = *c; \
  register double fpcr = *(double*)cr; \
  dasm("mt_fpcr %f16", fpcr); \
  fcv = dasm("trapb; " #NAME " %f16,%f17,%f0; trapb ", fav, fbv); \
  fpcr = dasm("mf_fpcr %f0"); \
  *rdestp  = fcv; *(double*)cr = fpcr;  \
}
#endif


#define GAMMA_EMPTY_FP_OPER(_o,_f,NAME)  \
void NAME ## _ff(const double *a, const double *b, double *rdest, Reg *cr) \
{ \
     CPUWarning("Unimplemented FP opcode %s \n", #NAME);  \
}

#define GAMMA_COMPLEX_FP(o_,NAME)

Reg gammaReadFPCR()
{
  double fpcr;
  Reg result;
#ifdef __GNUC__
  asm("mf_fpcr %0, %0, %0\n" : "=f" (fpcr));
#else
  fpcr = dasm("mf_fpcr %f0\n");
#endif
  result = *(Reg*)&fpcr;
  return result;  
}


void gammaWriteFPCR(Reg newfpcr)
{
  double input = *(double *)&newfpcr;
#ifdef __GNUC__
  asm("mt_fpcr %0, %0, %0\n"
      : : "f" (input));
#else
  dasm("mt_fpcr %f16\n", input);
#endif
}


Reg gammaExchangeFPCR(Reg newfpcr)
{
  double input = *(double *)&newfpcr;
  double oldfpcr;
  Reg result;
#ifdef __GNUC__
  asm("mf_fpcr %0, %0, %0\n"
      "mt_fpcr %1, %1, %1\n"
      : "=f" (oldfpcr) : "f" (input));
#else
  oldfpcr = dasm("mf_fpcr %f0\n");
  asm("mt_fpcr %f16\n", input);
#endif
  result = *(Reg*)&oldfpcr;
  return result;  
}

/** @@@@ continue here */

#endif /*GAMMA_ASM_TABLE */
#endif /*GAMMA_ASM_EXTERNS */



#define GAMMA_INTA(NAME) GAMMA_OPTFMT(op_inta,inta_ ##NAME, NAME)
#define GAMMA_INTA_CANTRAP(NAME) GAMMA_CANTRAP_OPTFMT(op_inta,inta_ ##NAME, NAME)

#define GAMMA_INTL(NAME) GAMMA_OPTFMT(op_intl,intl_ ##NAME, NAME)
#define GAMMA_INTL_C(NAME) GAMMA_C_OPTFMT(op_intl,intl_ ##NAME, NAME)
#define GAMMA_INTL_CANTRAP(NAME) GAMMA_CANTRAP_OPTFMT(op_intl,intl_ ##NAME, NAME)
#define GAMMA_INTL_COMPLEX(NAME) GAMMA_COMPLEX_OPTFMT(op_intl,intl_ ##NAME, NAME)

#define GAMMA_INTS(NAME) GAMMA_OPTFMT(op_ints,ints_ ##NAME, NAME)

#define GAMMA_INTM(NAME) GAMMA_OPTFMT(op_intm,intm_ ##NAME, NAME)
#define GAMMA_INTM_CANTRAP(NAME) GAMMA_CANTRAP_OPTFMT(op_intm,intm_ ##NAME, NAME)

#define GAMMA_INTMISC(NAME) GAMMA_2ARG_OPTFMT(op_intmisc,intmisc_ ##NAME, NAME)

#define GAMMA_FPI(NAME)  GAMMA_FP_OPER(op_flti,flti_ ##NAME, NAME)
#define GAMMA_CANTRAP_FPI(NAME)  GAMMA_CANTRAP_FP_OPER(op_flti,flti_ ##NAME, NAME)
#define GAMMA_CVTI(NAME) GAMMA_CVTFP_OPER(op_flti,flti_ ##NAME, NAME)
#define GAMMA_CANTRAP_CVTI(NAME) GAMMA_CANTRAP_CVTFP_OPER(op_flti,flti_ ##NAME, NAME)
#define GAMMA_CVTL(NAME) GAMMA_CVTFP_OPER(op_fltl,fltl_ ##NAME,NAME)
#define GAMMA_CANTRAP_CVTL(NAME) GAMMA_CANTRAP_CVTFP_OPER(op_fltl,fltl_ ##NAME,NAME)
#define GAMMA_FPL(NAME)  GAMMA_FP_OPER(op_fltl,fltl_ ##NAME, NAME)
#define GAMMA_FPL_C(NAME)  GAMMA_FPC_OPER(op_fltl,fltl_ ##NAME, NAME)
#define GAMMA_SFP(NAME)  GAMMA_SFP_OPER(op_flti,flti_ ##NAME,NAME)
#define GAMMA_EMPTY_FPL(NAME) GAMMA_EMPTY_FP_OPER(op_fltl,fltl_ ##NAME,NAME)

GAMMA_INTA (addl)
GAMMA_INTA (s4addl)
GAMMA_INTA (subl)
GAMMA_INTA (s4subl)
GAMMA_INTA (cmpbge)
GAMMA_INTA (s8addl)
GAMMA_INTA (s8subl)
GAMMA_INTA (cmpult)
GAMMA_INTA (addq)
GAMMA_INTA (s4addq)
GAMMA_INTA (subq)
GAMMA_INTA (s4subq)
GAMMA_INTA (cmpeq)
GAMMA_INTA (s8addq)
GAMMA_INTA (s8subq)
GAMMA_INTA (cmpule)
GAMMA_INTA_CANTRAP (addlv)
GAMMA_INTA_CANTRAP (sublv)
GAMMA_INTA (cmplt)
GAMMA_INTA_CANTRAP (addqv)
GAMMA_INTA_CANTRAP (subqv)
GAMMA_INTA (cmple)

/* Caution: the cmov are fixed. -Jamey Hicks */

GAMMA_INTL   (and)
GAMMA_INTL   (bic)
GAMMA_INTL_C (cmovlbs)
GAMMA_INTL_C (cmovlbc)
GAMMA_INTL   (bis)
GAMMA_INTL_C (cmoveq)
GAMMA_INTL_C (cmovne)
GAMMA_INTL   (ornot)
GAMMA_INTL   (xor)
GAMMA_INTL_C (cmovlt)
GAMMA_INTL_C (cmovge)
GAMMA_INTL_CANTRAP (eqv)  /* bugnion XXX !!! */
GAMMA_INTL_COMPLEX(amask)
GAMMA_INTL_C (cmovle)
GAMMA_INTL_C (cmovgt)
GAMMA_INTL_COMPLEX(implver)

GAMMA_INTS (mskbl)
GAMMA_INTS (extbl)
GAMMA_INTS (insbl)
GAMMA_INTS (mskwl)
GAMMA_INTS (extwl)
GAMMA_INTS (inswl)
GAMMA_INTS (mskll)
GAMMA_INTS (extll)
GAMMA_INTS (insll)
GAMMA_INTS (zap)
GAMMA_INTS (zapnot)
GAMMA_INTS (mskql)
GAMMA_INTS (srl)
GAMMA_INTS (extql)
GAMMA_INTS (sll)
GAMMA_INTS (insql)
GAMMA_INTS (sra)
GAMMA_INTS (mskwh)
GAMMA_INTS (inswh)
GAMMA_INTS (extwh)
GAMMA_INTS (msklh)
GAMMA_INTS (inslh)
GAMMA_INTS (extlh)
GAMMA_INTS (mskqh)
GAMMA_INTS (insqh)
GAMMA_INTS (extqh)

GAMMA_INTM (mull)
GAMMA_INTM (mulq)
GAMMA_INTM (umulh)
GAMMA_INTM_CANTRAP (mullv)
GAMMA_INTM_CANTRAP (mulqv)

GAMMA_MEM_LOAD  (ldq_u)
GAMMA_MEM_LOAD  (ldl)
GAMMA_MEM_LOAD  (ldq)

GAMMA_MEM_STORE (stq_u)
GAMMA_MEM_STORE (stl)
GAMMA_MEM_STORE (stq)

GAMMA_INTMISC (sextb)
GAMMA_INTMISC (sextw)

/*
 * memops can get memory exceptions, but no FP exceptions
 */
GAMMA_MEM_FPLOAD (ldf)
GAMMA_MEM_FPLOAD (ldg)
GAMMA_MEM_FPLOAD (lds)
GAMMA_MEM_FPLOAD (ldt)

GAMMA_MEM_FPSTORE (stf)
GAMMA_MEM_FPSTORE (stg)
GAMMA_MEM_FPSTORE (sts)
GAMMA_MEM_FPSTORE (stt)

/* VAX opcodes */
#ifdef later
GAMMA_FP_OPER (addfc)
GAMMA_FP_OPER (subfc)
GAMMA_FP_OPER (mulfc)
GAMMA_FP_OPER (divfc)
GAMMA_FP_OPER (addgc)
GAMMA_FP_OPER (subgc)
GAMMA_FP_OPER (mulgc)
GAMMA_FP_OPER (divgc)
GAMMA_FP_OPER (addf)
GAMMA_FP_OPER (subf)
GAMMA_FP_OPER (mulf)
GAMMA_FP_OPER (divf)
GAMMA_FP_OPER (addg)
GAMMA_FP_OPER (subg)
GAMMA_FP_OPER (mulg)
GAMMA_FP_OPER (divg)
GAMMA_FP_OPER (cmpgeq)
GAMMA_FP_OPER (cmpglt)
GAMMA_FP_OPER (cmpgle)
GAMMA_FP_OPER (addfuc)
GAMMA_FP_OPER (subfuc)
GAMMA_FP_OPER (mulfuc)
GAMMA_FP_OPER (divfuc)
GAMMA_FP_OPER (addguc)
GAMMA_FP_OPER (subguc)
GAMMA_FP_OPER (mulguc)
GAMMA_FP_OPER (divguc)
GAMMA_FP_OPER (addfu)
GAMMA_FP_OPER (subfu)
GAMMA_FP_OPER (mulfu)
GAMMA_FP_OPER (divfu)
GAMMA_FP_OPER (addgu)
GAMMA_FP_OPER (subgu)
GAMMA_FP_OPER (mulgu)
GAMMA_FP_OPER (divgu)
GAMMA_FP_OPER (addfsc)
GAMMA_FP_OPER (subfsc)
GAMMA_FP_OPER (mulfsc)
GAMMA_FP_OPER (divfsc)
GAMMA_FP_OPER (addgsc)
GAMMA_FP_OPER (subgsc)
GAMMA_FP_OPER (mulgsc)
GAMMA_FP_OPER (divgsc)
GAMMA_FP_OPER (addfs)
GAMMA_FP_OPER (subfs)
GAMMA_FP_OPER (mulfs)
GAMMA_FP_OPER (divfs)
GAMMA_FP_OPER (addgs)
GAMMA_FP_OPER (subgs)
GAMMA_FP_OPER (mulgs)
GAMMA_FP_OPER (divgs)
GAMMA_FP_OPER (cmpgeqs)
GAMMA_FP_OPER (cmpglts)
GAMMA_FP_OPER (cmpgles)
GAMMA_FP_OPER (addfsuc)
GAMMA_FP_OPER (subfsuc)
GAMMA_FP_OPER (mulfsuc)
GAMMA_FP_OPER (divfsuc)
GAMMA_FP_OPER (addgsuc)
GAMMA_FP_OPER (subgsuc)
GAMMA_FP_OPER (mulgsuc)
GAMMA_FP_OPER (divgsuc)
GAMMA_FP_OPER (addfsu)
GAMMA_FP_OPER (subfsu)
GAMMA_FP_OPER (mulfsu)
GAMMA_FP_OPER (divfsu)
GAMMA_FP_OPER (addgsu)
GAMMA_FP_OPER (subgsu)
GAMMA_FP_OPER (mulgsu)
GAMMA_FP_OPER (divgsu) 
#endif /* later: VAX */
GAMMA_CANTRAP_FPI (addsc)
GAMMA_CANTRAP_FPI (subsc)
GAMMA_CANTRAP_FPI (mulsc)
GAMMA_CANTRAP_FPI (divsc)
GAMMA_CANTRAP_FPI (addtc)
GAMMA_CANTRAP_FPI (subtc)
GAMMA_CANTRAP_FPI (multc)
GAMMA_CANTRAP_FPI (divtc)
GAMMA_CANTRAP_FPI (addsm)
GAMMA_CANTRAP_FPI (subsm)
GAMMA_CANTRAP_FPI (mulsm)
GAMMA_CANTRAP_FPI (divsm)
GAMMA_CANTRAP_FPI (addtm)
GAMMA_CANTRAP_FPI (subtm)
GAMMA_CANTRAP_FPI (multm)
GAMMA_CANTRAP_FPI (divtm)

GAMMA_COMPLEX_FP(fltl,mf_fpcr)
GAMMA_COMPLEX_FP(fltl,mt_fpcr)

/*
 * single precision
 */

GAMMA_SFP(adds)
GAMMA_SFP(subs)
GAMMA_SFP(muls)
GAMMA_SFP(divs)


GAMMA_CANTRAP_FPI (addt)
GAMMA_CANTRAP_FPI (subt)
GAMMA_CANTRAP_FPI (mult)
GAMMA_CANTRAP_FPI (divt)
GAMMA_CANTRAP_FPI (cmptun)
GAMMA_CANTRAP_FPI (cmpteq)
GAMMA_CANTRAP_FPI (cmptlt)
GAMMA_CANTRAP_FPI (cmptle)
GAMMA_CANTRAP_FPI (addsd)
GAMMA_CANTRAP_FPI (subsd)
GAMMA_CANTRAP_FPI (mulsd)
GAMMA_CANTRAP_FPI (divsd)
GAMMA_CANTRAP_FPI (addtd)
GAMMA_CANTRAP_FPI (subtd)
GAMMA_CANTRAP_FPI (multd)
GAMMA_CANTRAP_FPI (divtd)
GAMMA_CANTRAP_FPI (addsuc)
GAMMA_CANTRAP_FPI (subsuc)
GAMMA_CANTRAP_FPI (mulsuc)
GAMMA_CANTRAP_FPI (divsuc)
GAMMA_CANTRAP_FPI (addtuc)
GAMMA_CANTRAP_FPI (subtuc)
GAMMA_CANTRAP_FPI (multuc)
GAMMA_CANTRAP_FPI (divtuc)
GAMMA_CANTRAP_FPI (addsum)
GAMMA_CANTRAP_FPI (subsum)
GAMMA_CANTRAP_FPI (mulsum)
GAMMA_CANTRAP_FPI (divsum)
GAMMA_CANTRAP_FPI (addtum)
GAMMA_CANTRAP_FPI (subtum)
GAMMA_CANTRAP_FPI (multum)
GAMMA_CANTRAP_FPI (divtum)
GAMMA_CANTRAP_FPI (addsu)
GAMMA_CANTRAP_FPI (subsu)
GAMMA_CANTRAP_FPI (mulsu)
GAMMA_CANTRAP_FPI (divsu)
GAMMA_CANTRAP_FPI (addtu)
GAMMA_CANTRAP_FPI (subtu)
GAMMA_CANTRAP_FPI (multu)
GAMMA_CANTRAP_FPI (divtu)
GAMMA_CANTRAP_FPI (addsud)
GAMMA_CANTRAP_FPI (subsud)
GAMMA_CANTRAP_FPI (mulsud)
GAMMA_CANTRAP_FPI (divsud)
GAMMA_CANTRAP_FPI (addtud)
GAMMA_CANTRAP_FPI (subtud)
GAMMA_CANTRAP_FPI (multud)
GAMMA_CANTRAP_FPI (divtud)
GAMMA_CANTRAP_FPI (addssuc)
GAMMA_CANTRAP_FPI (subssuc)
GAMMA_CANTRAP_FPI (mulssuc)
GAMMA_CANTRAP_FPI (divssuc)
GAMMA_CANTRAP_FPI (addtsuc)
GAMMA_CANTRAP_FPI (subtsuc)
GAMMA_CANTRAP_FPI (multsuc)
GAMMA_CANTRAP_FPI (divtsuc)
GAMMA_CANTRAP_FPI (addssum)
GAMMA_CANTRAP_FPI (subssum)
GAMMA_CANTRAP_FPI (mulssum)
GAMMA_CANTRAP_FPI (divssum)
GAMMA_CANTRAP_FPI (addtsum)
GAMMA_CANTRAP_FPI (subtsum)
GAMMA_CANTRAP_FPI (multsum)
GAMMA_CANTRAP_FPI (divtsum)
GAMMA_CANTRAP_FPI (addssu)
GAMMA_CANTRAP_FPI (subssu)
GAMMA_CANTRAP_FPI (mulssu)
GAMMA_CANTRAP_FPI (divssu)
GAMMA_CANTRAP_FPI (addtsu)
GAMMA_CANTRAP_FPI (subtsu)
GAMMA_CANTRAP_FPI (multsu)
GAMMA_CANTRAP_FPI (divtsu)
GAMMA_CANTRAP_FPI (cmptunsu)
GAMMA_CANTRAP_FPI (cmpteqsu)
GAMMA_CANTRAP_FPI (cmptltsu)
GAMMA_CANTRAP_FPI (cmptlesu)
GAMMA_CANTRAP_FPI (addssud)
GAMMA_CANTRAP_FPI (subssud)
GAMMA_CANTRAP_FPI (mulssud)
GAMMA_CANTRAP_FPI (divssud)
GAMMA_CANTRAP_FPI (addtsud)
GAMMA_CANTRAP_FPI (subtsud)
GAMMA_CANTRAP_FPI (multsud)
GAMMA_CANTRAP_FPI (divtsud)
GAMMA_CANTRAP_FPI (addssuic)
GAMMA_CANTRAP_FPI (subssuic)
GAMMA_CANTRAP_FPI (mulssuic)
GAMMA_CANTRAP_FPI (divssuic)
GAMMA_CANTRAP_FPI (addtsuic)
GAMMA_CANTRAP_FPI (subtsuic)
GAMMA_CANTRAP_FPI (multsuic)
GAMMA_CANTRAP_FPI (divtsuic)
GAMMA_CANTRAP_FPI (addssuim)
GAMMA_CANTRAP_FPI (subssuim)
GAMMA_CANTRAP_FPI (mulssuim)
GAMMA_CANTRAP_FPI (divssuim)
GAMMA_CANTRAP_FPI (addtsuim)
GAMMA_CANTRAP_FPI (subtsuim)
GAMMA_CANTRAP_FPI (multsuim)
GAMMA_CANTRAP_FPI (divtsuim)
GAMMA_CANTRAP_FPI (addssui)
GAMMA_CANTRAP_FPI (subssui)
GAMMA_CANTRAP_FPI (mulssui)
GAMMA_CANTRAP_FPI (divssui)
GAMMA_CANTRAP_FPI (addtsui)
GAMMA_CANTRAP_FPI (subtsui)
GAMMA_CANTRAP_FPI (multsui)
GAMMA_CANTRAP_FPI (divtsui)
GAMMA_CANTRAP_FPI (addssuid)
GAMMA_CANTRAP_FPI (subssuid)
GAMMA_CANTRAP_FPI (mulssuid)
GAMMA_CANTRAP_FPI (divssuid)
GAMMA_CANTRAP_FPI (addtsuid)
GAMMA_CANTRAP_FPI (subtsuid)
GAMMA_CANTRAP_FPI (multsuid)
GAMMA_CANTRAP_FPI (divtsuid)

    /*
     * FPL FPL_C cannot generate exceptions
     */
    
GAMMA_FPL (cpys)
GAMMA_FPL (cpysn)
GAMMA_FPL (cpyse)

GAMMA_FPL_C(fcmoveq)
GAMMA_FPL_C(fcmovne)
GAMMA_FPL_C(fcmovlt)
GAMMA_FPL_C(fcmovge)
GAMMA_FPL_C(fcmovle)
GAMMA_FPL_C(fcmovgt)


#ifdef later_vms
FP_CVT (cvtdgc)
FP_CVT (cvtgfc)
FP_CVT (cvtgdc)
FP_CVT (cvtgqc)
FP_CVT (cvtqfc)
FP_CVT (cvtqgc)
FP_CVT (cvtdg)
FP_CVT (cvtgf)
FP_CVT (cvtgd)
FP_CVT (cvtgq)
FP_CVT (cvtqf)
FP_CVT (cvtqg)
FP_CVT (cvtdguc)
FP_CVT (cvtgfuc)
FP_CVT (cvtgduc)
FP_CVT (cvtgqvc)
FP_CVT (cvtdgu)
xFP_CVT (cvtgfu)
FP_CVT (cvtgdu)
FP_CVT (cvtgqv)
FP_CVT (cvtdgsc)
FP_CVT (cvtgfsc)
FP_CVT (cvtgdsc)
FP_CVT (cvtgqsc)
FP_CVT (cvtdgs)
FP_CVT (cvtgfs)
FP_CVT (cvtgds)
FP_CVT (cvtgqs)
FP_CVT (cvtdgsuc)
FP_CVT (cvtgfsuc)
FP_CVT (cvtgdsuc)
FP_CVT (cvtgqsvc)
FP_CVT (cvtdgsu)
FP_CVT (cvtgfsu)
FP_CVT (cvtgdsu)
FP_CVT (cvtgqsv)

#endif /*later_vax */
    
GAMMA_CANTRAP_CVTI (cvttsc)
GAMMA_CANTRAP_CVTI (cvttqc)
GAMMA_CANTRAP_CVTI (cvtqsc)
GAMMA_CANTRAP_CVTI (cvtqtc)
GAMMA_CANTRAP_CVTI (cvttsm)
GAMMA_CANTRAP_CVTI (cvttqm)
GAMMA_CANTRAP_CVTI (cvtqsm)
GAMMA_CANTRAP_CVTI (cvtqtm)
GAMMA_CANTRAP_CVTI (cvtts)
GAMMA_CANTRAP_CVTI (cvttq)
GAMMA_CANTRAP_CVTI (cvtqs)
GAMMA_CANTRAP_CVTI (cvtqt)
GAMMA_CANTRAP_CVTI (cvttsd)
GAMMA_CANTRAP_CVTI (cvttqd)
GAMMA_CANTRAP_CVTI (cvtqsd)
GAMMA_CANTRAP_CVTI (cvtqtd)
GAMMA_CANTRAP_CVTI (cvttsuc)
GAMMA_CANTRAP_CVTI (cvttqvc)
GAMMA_CANTRAP_CVTI (cvttsum)
GAMMA_CANTRAP_CVTI (cvttqvm)
GAMMA_CANTRAP_CVTI (cvttsu)
GAMMA_CANTRAP_CVTI (cvttqv)
GAMMA_CANTRAP_CVTI (cvttsud)
GAMMA_CANTRAP_CVTI (cvttqvd)
GAMMA_CANTRAP_CVTI (cvtst)
GAMMA_CANTRAP_CVTI (cvtsts)
GAMMA_CANTRAP_CVTI (cvttssuc)
GAMMA_CANTRAP_CVTI (cvttqsvc)
GAMMA_CANTRAP_CVTI (cvttssum)
GAMMA_CANTRAP_CVTI (cvttqsvm)
GAMMA_CANTRAP_CVTI (cvttssu)
GAMMA_CANTRAP_CVTI (cvttqsv)
GAMMA_CANTRAP_CVTI (cvttssud)
GAMMA_CANTRAP_CVTI (cvttqsvd)
/* GAMMA_CANTRAP_CVTI(cvttss) - gcc claims no such operation */
GAMMA_CANTRAP_CVTI (cvttssuic)
GAMMA_CANTRAP_CVTI (cvttqsvic)
GAMMA_CANTRAP_CVTI (cvtqssuic)
GAMMA_CANTRAP_CVTI (cvtqtsuic)
GAMMA_CANTRAP_CVTI (cvttssuim)
GAMMA_CANTRAP_CVTI (cvttqsvim)
GAMMA_CANTRAP_CVTI (cvtqssuim)
GAMMA_CANTRAP_CVTI (cvtqtsuim)
GAMMA_CANTRAP_CVTI (cvttssui)
GAMMA_CANTRAP_CVTI (cvttqsvi)
GAMMA_CANTRAP_CVTI (cvtqssui)
GAMMA_CANTRAP_CVTI (cvtqtsui)
GAMMA_CANTRAP_CVTI (cvttssuid)
GAMMA_CANTRAP_CVTI (cvttqsvid)
GAMMA_CANTRAP_CVTI (cvtqssuid)
GAMMA_CANTRAP_CVTI (cvtqtsuid)

GAMMA_CANTRAP_CVTL (cvtlq)
GAMMA_CANTRAP_CVTL (cvtql)
GAMMA_CANTRAP_CVTL (cvtqlv)
GAMMA_CANTRAP_CVTL (cvtqlsv)

#define GAMMA_EMPTY_OPC14(NAME) GAMMA_EMPTY_FP_OPER(op_opc14,opc14_ ##NAME,NAME)

GAMMA_EMPTY_OPC14(sqrttc)
GAMMA_EMPTY_OPC14(sqrttm)
GAMMA_COMPLEX_FP(opc14,sqrtt)
GAMMA_EMPTY_OPC14(sqrttd)
GAMMA_EMPTY_OPC14(sqrttuc)
GAMMA_EMPTY_OPC14(sqrttum)
GAMMA_EMPTY_OPC14(sqrttu)
GAMMA_EMPTY_OPC14(sqrttud)
GAMMA_EMPTY_OPC14(sqrttsuc)
GAMMA_EMPTY_OPC14(sqrttsum)
GAMMA_EMPTY_OPC14(sqrttsu)
GAMMA_EMPTY_OPC14(sqrttsud)
GAMMA_EMPTY_OPC14(sqrttsuic)
GAMMA_EMPTY_OPC14(sqrttsuim)
GAMMA_EMPTY_OPC14(sqrttsui)
GAMMA_EMPTY_OPC14(sqrttsuid)

GAMMA_EMPTY_OPC14(sqrtsc)
GAMMA_EMPTY_OPC14(sqrtsm)
GAMMA_COMPLEX_FP(opc14,sqrts)
GAMMA_EMPTY_OPC14(sqrtsd)
GAMMA_EMPTY_OPC14(sqrtsuc)
GAMMA_EMPTY_OPC14(sqrtsum)
GAMMA_EMPTY_OPC14(sqrtsu)
GAMMA_EMPTY_OPC14(sqrtsud)
GAMMA_EMPTY_OPC14(sqrtssuc)
GAMMA_EMPTY_OPC14(sqrtssum)
GAMMA_EMPTY_OPC14(sqrtssu)
GAMMA_EMPTY_OPC14(sqrtssud)
GAMMA_EMPTY_OPC14(sqrtssuic)
GAMMA_EMPTY_OPC14(sqrtssuim)
GAMMA_EMPTY_OPC14(sqrtssui)
GAMMA_EMPTY_OPC14(sqrtssuid)


GAMMA_COMPLEX_FP(opc14,itofs)
GAMMA_COMPLEX_FP(opc14,itoff)
GAMMA_COMPLEX_FP(opc14,itoft)
