/*
 * histogram.h -- Periodic sampling histogram data structure
 */

typedef struct {
   int min;
   int max;
   int granularity;
   long *buckets;
} SamHistogram;

#define HistogramInit(H, MIN, MAX, GRANULARITY)                               \
   (H).min = MIN;                                                             \
   (H).max = MAX;                                                             \
   (H).granularity = GRANULARITY;                                             \
   (H).buckets = (long *)calloc(((MAX) - (MIN) + 1) / (GRANULARITY), sizeof(long));

#define HistogramAdd(H, VALUE) {                        \
   int _bucket = ((VALUE) - (H).min) / (H).granularity; \
   ASSERT((VALUE) >= (H).min);                          \
   ASSERT((VALUE) <= (H).max);                          \
   (H).buckets[_bucket]++;                              \
}

#define HistogramDump(H) {                                      \
   int _i;                                                      \
   for (_i = (H).min; _i <= (H).max; _i += (H).granularity) {   \
      if ((H).granularity > 1) {                                \
         CPUPrint("%d - %d: ", _i, _i + (H).granularity - 1);   \
      } else {                                                  \
         CPUPrint("%d: ", _i);                                  \
      }                                                         \
      CPUPrint("%ld\n", (H).buckets[_i]);                       \
   }                                                            \
}


