/*
 * print.c -- support for dumping pipeline state to the log
 */

#include "sam.h"
#include "cpu_interface.h"
#include "disassembler.h"

void
SamPrintCycle(SamState *sp)
{
   int cpuNum = sp->PE->myNum;
   if (samPrintLevel[cpuNum] > 0) {
      CPUPrint("<PIPE> %2d begin %lld\n", cpuNum, (uint64)CPUVec.CycleCount(cpuNum));
   }
}

void
SamPrintIwinEntry(SamState *sp, SamIwinEntry *ip)
{
}

void
SamPrintIwin(SamState *sp)
{
   List_Links *lp;
   LIST_FORALL(sp->reorderBuffer, lp) {
      SamIwinEntry *ip = ReorderListToIwin(lp);
      SamPrintIwinEntry(sp, ip);
   }
}

void
SamPrintFetch(SamState *sp, SamIwinEntry *ip)
{
   int cpuNum = sp->PE->myNum;
   if (samPrintLevel[cpuNum] > 0) {
      int inum = ip - sp->iwin;
      
      char buf[200];
      long piG1=0, piG2=0, piG3=0;
      disasm(buf, ip->pc, ip->instr.word, &piG1, &piG2, &piG3);
      
      CPUPrint("<PIPE> %2d entry: %2d", cpuNum, inum);
      if (ip->iwinFlags & IwinDepA) {
         SamIwinEntry *op = ip->ra.reg->owner;
         int onum = op - sp->iwin;
         CPUPrint(" %2d", onum);
      } else {
         CPUPrint("   ");
      }
      if (ip->iwinFlags & IwinDepB) {
         SamIwinEntry *op = ip->rb.reg->owner;
         int onum = op - sp->iwin;
         CPUPrint(" %2d", onum);
      } else {
         CPUPrint("   ");
      }
      CPUPrint(" %s\n", buf);
   }
}

void
SamPrintStage(SamState *sp, SamIwinEntry *ip, char *stage)
{
   int cpuNum = sp->PE->myNum;
   if (samPrintLevel[cpuNum] > 1) {
      int inum = ip - sp->iwin;
      CPUPrint("<PIPE> %2d stage: %2d %s\n", cpuNum, inum, stage);
   }
}
