/*
 * sam_params.c -- Tcl interface for Sam configuration parameters
 */

#include "sam_params.h"
#include "machine_params.h"
#include "print.h"

SamParams samParams;

/* SamPrintLevelTclCmd: called by "samPrintLevel" Tcl command */

static int
SamPrintLevelTclCmd(ClientData clientData, Tcl_Interp *interp, 
                    int argc, char *argv[]) 
{
   int level;
   int cpuNum;

   if (argc != 3) {
      Tcl_AppendResult(interp, "Usage: ", argv[0], " <cpuNum|all> level", NULL);
      return TCL_ERROR;
   }
   if (Tcl_GetInt(interp, argv[2], &level) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad print level ", argv[2], 
                       "passed to ", argv[0], "command", NULL);
      return TCL_ERROR;
   }
   if (!strcmp(argv[1], "all")) {
      for (cpuNum = 0; cpuNum < TOTAL_CPUS; cpuNum++) {
         samPrintLevel[cpuNum] = level;
      }
   } else if (Tcl_GetInt(interp, argv[1], &cpuNum) == TCL_OK) {
      samPrintLevel[cpuNum] = level;
   } else {
      Tcl_AppendResult(interp, "Bad CPU number ", argv[1], 
                       "passed to ", argv[0], "command", NULL);
      return TCL_ERROR;
   }
   return TCL_OK;
}

/* SamTclInit: called from AlphaTclInit() */

void
SamTclInit(void)
{
   ParamRegister("PARAM(CPU.Sam.InOrder)",
                 (char *)&samParams.inOrder, PARAM_BOOLEAN);
   ParamRegister("PARAM(CPU.Sam.InOrder.NumIssueSlots)",
                 (char *)&samParams.numIssueSlots, PARAM_INT);
   
   ParamRegister("PARAM(CPU.Sam.NumGPRs)",
                 (char *)&samParams.numGPRs, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.NumFPRs)",
                 (char *)&samParams.numFPRs, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.NumInstructionSlots)",
                 (char *)&samParams.numInstructionSlots, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.NumThreads)",
                 (char *)&samParams.numThreads, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.NumLdstBuffers)",
                 (char *)&samParams.numLdstBuffers, PARAM_INT);

   ParamRegister("PARAM(CPU.Sam.FetchWidth)",
                 (char *)&samParams.fetchWidth, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.IssueWidth)",
                 (char *)&samParams.issueWidth, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.GradWidth)",
                 (char *)&samParams.gradWidth, PARAM_INT);

   ParamRegister("PARAM(CPU.Sam.FixedUnit.Count)",
                 (char *)&samParams.numFixedUnits, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.FixedUnit.Latency)",
                 (char *)&samParams.fixedUnitLatency, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.FixedUnit.RepeatRate)",
                 (char *)&samParams.fixedUnitRepeatRate, PARAM_INT);
   
   ParamRegister("PARAM(CPU.Sam.UnifiedUnit.Count)",
                 (char *)&samParams.numUnifiedUnits, PARAM_INT);

   ParamRegister("PARAM(CPU.Sam.ALU.Count)",
                 (char *)&samParams.numALUs, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.ALU.Latency)",
                 (char *)&samParams.ALULatency, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.ALU.RepeatRate)",
                 (char *)&samParams.ALURepeatRate, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.FPU.Count)",
                 (char *)&samParams.numFPUs, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.FPU.Latency)",
                 (char *)&samParams.FPULatency, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.FPU.RepeatRate)",
                 (char *)&samParams.FPURepeatRate, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.LSU.Count)",
                 (char *)&samParams.numLSUs, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.LSU.Latency)",
                 (char *)&samParams.LSULatency, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.LSU.RepeatRate)",
                 (char *)&samParams.LSURepeatRate, PARAM_INT);

   ParamRegister("PARAM(CPU.Sam.LoadToUseLatency)",
                 (char *)&samParams.loadToUseLatency, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.RetireStoresEarly)",
                 (char *)&samParams.retireStoresEarly, PARAM_BOOLEAN);

   ParamRegister("PARAM(CPU.Sam.BP.TableSize)",
                 (char *)&samParams.branchPrediction.tableSize, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.BP.ReturnStackSize)",
                 (char *)&samParams.branchPrediction.returnStack, PARAM_INT);
   ParamRegister("PARAM(CPU.Sam.BP.MispredictLatency)",
                 (char *)&samParams.branchPrediction.mispredictLatency, PARAM_INT);

   Tcl_CreateCommand(TCLInterp, "samPrintLevel", SamPrintLevelTclCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
}
