/*
 * thread.h -- the Sam thread structure and macros
 */

#ifndef SAM_THREAD_H
#define SAM_THREAD_H

#include "list.h"
#include "register.h"
#include "trap.h"

typedef enum { Taken, NotTaken, NotPredicted } SamBranchResult;

typedef struct {
   List_Links links;   /* Active/free list management -- must come first */

   enum { ThFree, ThActive, ThSpec } status;
   SimTime mispredictRestartCycle;
   int stallFetch;
   bool cmov2Fetch;
   SamTrapState trap;

   /* Register state */

   SamRegState *gprName[33];
   SamRegState *fprName[33];
   VA pc;

   /* Branch prediction state */

   int branchSP;
   SamBranchResult prediction;
   SamBranchResult resolution;
   VA predictPC;
   VA oldPrediction;
   bool call;

} SamThreadState;

/* stallFetch: Reasons for stalling thread ifetch */

#define StallNone         0x0000
#define StallICache       0x0001
#define StallTrap         0x0002
#define StallITB          0x0004
#define StallREI          0x0008
#define StallBarrier      0x0010
#define StallBranch       0x0020
#define StallThread       0x0040
#define StallMispredict   0x0080
#define StallSC           0x0100

/* Convert between thread and list pointers */

#define ThreadToList(TH) ((List_Links*)(TH))
#define ListToThread(LP) ((SamThreadState*)(LP))

/* Macro to access the current fetch/grad threads */

#define SamGetGradThread(TH, SP)                        \
{                                                       \
   List_Links *_lp = List_First((SP)->activeThreads);   \
   (TH) = ListToThread(_lp);                            \
   ASSERT(_lp != (SP)->activeThreads);                  \
}

#define SamGetFetchThread(TH, SP)                       \
{                                                       \
   List_Links *_lp = List_Last((SP)->activeThreads);    \
   (TH) = ListToThread(_lp);                            \
   ASSERT(_lp != (SP)->activeThreads);                  \
}

/* Check if there is a free thread */

#define SamThreadIsAvailable(SP) !List_IsEmpty((SP)->freeThreads)

/* Claim a thread from the free list */

#define SamGetFreeThread(TH, SP)                        \
{                                                       \
   List_Links *_lp = List_First((SP)->freeThreads);     \
   if (_lp == (SP)->freeThreads) {                      \
      TH = NULL;                                        \
   } else {                                             \
      TH = ListToThread(_lp);                           \
      List_RemoveDebug(_lp);                            \
   }                                                    \
}

#endif /* SAM_THREAD_H */
