/*
 * trap.h --  the Sam trap structure and macros
 */

#ifndef SAM_TRAP_H
#define SAM_TRAP_H

#include <alpha/inst.h>
#include "ev5.h"
#include "ev5_ipr.h"

typedef struct {
   VA pc;
   int trap;
   AlphaTrapState *state;
} SamTrapState;

/* Pipeline-specific pseudo-traps */

#define TrapNone      -1
#define TrapReplay    -2
#define TrapSwitch    -3
/* For PAL trap codes, see ev5_ipr.h */


#define SamTrapOccurred(SP)                     \
   ((SP)->lastTrap.trap != TrapNone)
#define SamTrapPending(SP)                      \
   ((SP)->currentTrap.trap != TrapNone)
#define SamInterruptPending(SP)                 \
   ((SP)->interrupt.trap != TrapNone)

#define SamTrapClear(TRAP)                      \
   (TRAP).trap = TrapNone;                      \
   (TRAP).state = NULL;

#define SamTrapFree(TRAP)                       \
   freeTrapState((TRAP).state);                 \
   SamTrapClear(TRAP);

#define SamTrapCopy(SRC, DST)                           \
   if ((DST).state != NULL) freeTrapState((DST).state); \
   (DST).pc = (SRC).pc;                                 \
   (DST).trap = (SRC).trap;                             \
   (DST).state = (SRC).state;                           \
   SamTrapClear(SRC);

#endif /* SAM_TRAP_H */
