/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

#ifdef NT_FRAMEBUFFER

#ifndef _FRAMEBUF_H
#define _FRAMEBUF_H


/*
 * SimOS Memory Areas
 */
#define SIMOS_FB_REGADDR    0xffffffffb0000000
#define SIMOS_FB_MEMADDR    0xffffffffa8000000    

#define SIMOS_FB_REGS       ((struct SimOS_FB_Regs *)SIMOS_FB_REGADDR)

typedef unsigned long FBReg;

typedef struct SimOS_FB_Regs {
/*
 * Command fields (writing will initiate command)
 */
   FBReg Reset;          /* Reset the device */
   FBReg ChangeMode;     /* Change the video mode */
   FBReg ChangePalette;  /* Change the color palette */
   FBReg MapMemory;      /* Map the linear video memory buffer */
   FBReg UnMapMemory;    /* Unmap the linear video memory buffer */

/*
 * Data fields 
 */
   /* Set on IOCTL_VIDEO_SET_CURRENT_MODE */ 
   FBReg VideoMode;      /* Current video mode */
   FBReg XResolution;    /* X resolution */
   FBReg YResolution;    /* Y resolution */

   /* Set on IOCTL_VIDEO_SET_COLOR_REGISTER */
   FBReg Palette[256];   /* Color palette values */
} SimOS_FB_Regs;

int Framebuf_reg_access(int cpuNum, VA vA, int type, Reg *buff);

int Framebuf_mem_access(int cpuNum, VA vA, int type, Reg *buff);

#endif /* _FRAMEBUF_H */

#endif /* NT_FRAMEBUFFER */
