/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */

#ifdef NT_FRAMEBUFFER

#ifndef _KBD_MOUSE_H
#define _KBD_MOUSE_H

/*
 * Kbd and mouse commands
 */ 
#define KBM_READ_CONTROLLER_COMMAND_BYTE  0x20
#define KBM_WRITE_CONTROLLER_COMMAND_BYTE 0x60
#define KBM_DISABLE_MOUSE_DEVICE          0xA7
#define KBM_ENABLE_MOUSE_DEVICE           0xA8
#define KBM_AUXILIARY_DEVICE_TEST         0xA9
#define KBM_KEYBOARD_DEVICE_TEST          0xAB
#define KBM_DISABLE_KEYBOARD_DEVICE       0xAD
#define KBM_ENABLE_KEYBOARD_DEVICE        0xAE
#define KBM_WRITE_TO_AUXILIARY_DEVICE     0xD4

/*
 * Controller command bytes 
 */
#define CCB_ENABLE_KEYBOARD_INTERRUPT     0x01
#define CCB_ENABLE_MOUSE_INTERRUPT        0x02
#define CCB_DISABLE_KEYBOARD_DEVICE       0x10
#define CCB_DISABLE_MOUSE_DEVICE          0x20
#define CCB_KEYBOARD_TRANSLATE_MODE       0x40

/*
 * Status register bits
 */
#define KBM_OUTPUT_BUFFER_FULL            0x01
#define KBM_INPUT_BUFFER_FULL             0x02
#define KBM_MOUSE_OUTPUT_BUFFER_FULL      0x20

/*
 * Responses
 */
#define KBM_ACKNOWLEDGE                   0xFA
#define KBM_RESEND                        0xFE

/*
 * Define commands to the keyboard (through the data port).
 */
#define SET_KEYBOARD_INDICATORS           0xED
#define SELECT_SCAN_CODE_SET              0xF0
#define READ_KEYBOARD_ID                  0xF2
#define SET_KEYBOARD_TYPEMATIC            0xF3
#define SET_ALL_TYPEMATIC_MAKE_BREAK      0xFA
#define KEYBOARD_RESET                    0xFF

/*
 * Define the keyboard responses.
 */
#define KEYBOARD_COMPLETE_SUCCESS 0xAA
#define KEYBOARD_COMPLETE_FAILURE 0xFC
#define KEYBOARD_BREAK_CODE       0xF0
#define KEYBOARD_DEBUG_HOTKEY_ENH 0x37 // SysReq scan code for Enhanced Keyboard
#define KEYBOARD_DEBUG_HOTKEY_AT  0x54 // SysReq scan code for 84-key Keyboard

/*
 * Define commands to the mouse (through the data port).
 */
#define SET_MOUSE_RESOLUTION              0xE8
#define SET_MOUSE_SAMPLING_RATE           0xF3
#define MOUSE_RESET                       0xFF
#define ENABLE_MOUSE_TRANSMISSION         0xF4
#define SET_MOUSE_SCALING_1TO1            0xE6
#define READ_MOUSE_STATUS                 0xE9
#define GET_DEVICE_ID                     0xF2

/*
 * Define the mouse responses.
 */
#define MOUSE_COMPLETE      0xAA
#define MOUSE_ID_BYTE       0x00
#define WHEELMOUSE_ID_BYTE  0x03


/*
 * Our buffer sizes
 */
#define REPLY_BUFF_SIZE      8
#define KEYBOARD_BUFF_SIZE   16
#define MOUSE_BUFF_SIZE      48

int KbdMouse_Initialize(int machine);
int KbdMouse_KeyPress(int machine, unsigned char scancode);
int KbdMouse_MouseEvent(int machine, int x_delta, int y_delta, unsigned int button_state); 
int KbdMouse_Status_Read(int cpuNum, VA vA, int type, Reg *buff);
int KbdMouse_Command_Write(int cpuNum, VA vA, int type, Reg *buff);
int KbdMouse_Data_Read(int cpuNum, VA vA, int type, Reg *buff);
int KbdMouse_Data_Write(int cpuNum, VA vA, int type, Reg *buff);
int KbdMouse_PendingKbd(int cpuNum);
int KbdMouse_PendingMouse(int cpuNum);

#endif /* _KBD_MOUSE_H */

#endif /* NT_FRAMEBUFFER */
