/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */


/* 
 * This file contains the simos side of the mappings
 * and structures for the ethernet driver. It is mirrored
 * on the os side in io/dec/netif/if_simreg.h
 * Changes may need corresponding changes in the OS
 */


#ifndef _IF_SIMETHER_H_
#define _IF_SIMETHER_H_

typedef unsigned long DevRegister64;

#define DEV_ETHER_MAX_RCV_ENTRIES    64
#define DEV_ETHER_MAX_SND_ENTRIES    64
#define DEV_ETHER_MAX_SND_CHUNKS     128

#define DEV_ETHER_MAX_TRANSFER_SIZE  1800
#define DEV_ETHER_OS_OWNED           (DevRegister64)1
#define DEV_ETHER_CONTROLLER_OWNED   (DevRegister64)2

#define DEV_ETHER_REGISTERS  ((volatile DevEtherRegisters*)(0xfffffcc510000000))
#define DEV_ETHER_REGISTERS_32  ((volatile DevEtherRegisters*)(0xffffffffa5000000))


typedef struct DevEtherRegisters {
  DevRegister64 etheraddr[6];       /* controller tells OS its ethernet addr */
  DevRegister64 numRcvEntries;      /* read by OS, indicates how many receive
				   * ring buffer entries will be used. OS must
				   * allocate a receive buffer for each of
				   * these entries */
  DevRegister64 numSndEntries;      /* read by OS, indicates how many send ring
				   * buffer entries will be used so it knows
				   * when to wrap its index pointer */
  DevRegister64 numSndChunks;       /* same as numSndEntries */

  struct {
    DevRegister64 pAddr;
    DevRegister64 maxLen;
    DevRegister64 len;
    DevRegister64 flag;         
  } rcvEntries[ DEV_ETHER_MAX_RCV_ENTRIES ];

  struct {
    DevRegister64 firstChunk;
    DevRegister64 lastChunk;
    DevRegister64 flag; /* triggers send */
  } sndEntries[ DEV_ETHER_MAX_SND_ENTRIES ];

  struct { 
    DevRegister64 pAddr;
    DevRegister64 len;
  } sndChunks[ DEV_ETHER_MAX_SND_CHUNKS ]; 

  /* note: sndChunks is last because we might extend the number of
   * send chunks in the future and don't want to break OS compatibility when
   * we do it */
} DevEtherRegisters;

void Simether_Init(int restoreFromChkpt);
extern int Simether_access(int cpuNum, VA vA, int type, DevRegister64 *datap); 
extern void Simether_interrupt(int n, int on);
extern int Simether_Pending(int cpuNum);
extern void Simether_Poll (void) ;

#endif	/* _IF_SIMETHER_H_ */
