/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



/*
 * TurboLaser address layout and backdoor
 *
 * The UART stuff is needlessly broken, as a result of a PALcode bug (now fixed)
 * that screwed up the IPL logic. Should clean up.
 */


#include <stddef.h>
#include <stdlib.h>
#include <sys/time.h>

#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"

#include "devices.h"
#include "tlaser.h"
#include "tlaser_clock.h"
#include "tlaser_uart.h"
#include "tlaser_pcia.h"
#include "tlaser_mbox.h"
#include "simos_scsi.h"
#include "simether.h"
#include "framebuf.h"
#include "kbd_mouse.h"



struct TLaserState {
   int intrSumType;
   int ipiPending;
} tlaserState[SIM_MAXCPUS];


/* ************************************************************
 * On-board clock (wallclock time). Right now, we return 0, but we should
 * eventually return either today's date or an abitrary number, just like
 * for the mips simulator.
 * ***********************************************************/

#define UNIX_YEAR_OFFSET 52

int TLaser_mc146818__access(int cpuNum, VA vA, int type, Reg *buff)
{
   static int init=0;
   static struct tm *tm;

   VA off = vA  & 0xfffff;
   VA ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (!init) {
      struct timeval tv;
      char *str;
      gettimeofday(&tv,0);
      tm = gmtime(&tv.tv_sec);
      str = asctime(tm);
      CPUWarning("TOY clock set to %s\n",str);
      init=1;
   }

   if (type==BDOOR_LOAD_DOUBLE||type==BDOOR_LOAD_WORD) {
      uint val = 0;
      int set = 0;
      *(Reg *)buff = 0;

      switch(off) {

      case (RTC_SECOND << GBUS_BIT_SHIFT):
         set = 1; val = tm->tm_sec;
         break;
      case (RTC_MINUTE << GBUS_BIT_SHIFT):
         set = 2; val = tm->tm_min;
         break;
      case (RTC_HOUR << GBUS_BIT_SHIFT):
         set = 3; val = tm->tm_hour;
         break;
      case (RTC_DAY_OF_WEEK << GBUS_BIT_SHIFT):
         set = 4; val = tm->tm_mday;
         break;
      case (RTC_MONTH << GBUS_BIT_SHIFT):
         set = 5; val = tm->tm_mon+1;
         break;
      case (RTC_YEAR << GBUS_BIT_SHIFT):
         set = 6; val = tm->tm_year - UNIX_YEAR_OFFSET;
         break;

      case (RTC_CONTROL_REGISTERA << GBUS_BIT_SHIFT):
	 *(Reg *)buff = 0x20; /*RTA_DV_OSC_ENABLE */
	 return SUCCESS;

      case (RTC_CONTROL_REGISTERB << GBUS_BIT_SHIFT):
	 *(Reg*)buff = 0x6; /* #define RTB_EXPECT  (u_char)(RTB_DMBINARY | RTB_24HR) */
	 return SUCCESS;

      case (RTC_CONTROL_REGISTERC << GBUS_BIT_SHIFT):
         *buff = 0;
         TLaserClockAck(cpuNum);
         return 0;

      case (RTC_CONTROL_REGISTERD << GBUS_BIT_SHIFT):
	 *(Reg*)buff=0x80; /* RTD_VRT */
	 return SUCCESS;

      }

      if (set) {
         CPUWarning("CLOCK READ set=%d val=%d \n",set,val);
         *(Reg*)buff = val;
	 return SUCCESS;
      }
   }
	     
   if (type==BDOOR_LOAD_DOUBLE && off == (RTC_CONTROL_REGISTERC << GBUS_BIT_SHIFT)) {
      *buff = 0;
      TLaserClockAck(cpuNum);
      return 0;
   }

   CPUPrint("TLaser_mc146818 : OFF 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
              off, PE[cpuNum]->PC, ra, vA,bdoorType[type]); 

   switch(type) { 
   case BDOOR_LOAD_WORD:
   case BDOOR_LOAD_DOUBLE:
   case BDOOR_LOAD_HALF:
   case BDOOR_LOAD_BYTE:
      *(uint64*)buff = 0;
      break;
   default:
       break;
   }

   return SUCCESS;
}



#ifdef BOOT_NT

static int TLaserRegisters(int cpuNum, VA vA, int type, Reg *buff)
{
   VA off = vA  & 0xfffff;
   Reg ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (type==BDOOR_LOAD_WORD) {
      uint32 reg=0;
      switch(off) {
         
         /* 
          * Common registers 
          */

      case TLDEV_REG:
         *buff = DEVICE_TYPE_TIOP | 0x20; /* TIOP_LDEV */
         return 0;

      case TLBER_REG:
         *buff = 0x0; /* no error */
         return 0;

      case TLCNR_REG:
         *buff = 0; /* ok */
         return 0;

         /* 
          * I/O registers 
          */

      case TLILID1_REG:
         if (SimOS_SCSI_Pending(cpuNum)) {
#ifdef BOOT_NT
            *buff = 0x39;
#else
            *buff = 0x820;
#endif
#ifdef NT_FRAMEBUFFER
         } else if (KbdMouse_PendingMouse(cpuNum)) {
            *buff = 0x3c; 
         } else if (KbdMouse_PendingKbd(cpuNum)) {
            *buff = 0x3b;
#endif
         } else {
            *buff = 0x0;
         }
         return 0;

      case IDPNSE1_REG:
         /* tlsb_iop.c::tiopconfl1  iopreg->idpnse1; */
         *buff = 0x6;  /* ((IDPNSE_CBLOK | IDPNSE_PWROKC) ) */
         return 0;

         /*
          * CPU registers
          */
         
      case TLINTRSUM0_REG:
      case TLINTRSUM1_REG:
         if (cpuNum == 0 && sim_console_int_status_noSE(0)) {
	    /* first CPU only  */
	    reg |= (1<<0);
	 }
	 if (tlaserState[cpuNum].ipiPending)    reg |= (1<<5);
	 if (TLaserClockIntrPending(cpuNum))    reg |= (1<<6);
	 if (SimOS_SCSI_Pending(cpuNum)) {
	    reg |= (1<<2);   /* summary bit set */
	    reg |= (1<<12);  /* node4   bit */
	 }
#ifdef NT_FRAMEBUFFER
         if (KbdMouse_PendingKbd(cpuNum) || KbdMouse_PendingMouse(cpuNum)) {
            reg |= (1<<2);
            reg |= (1<<12);
         }
#endif
         *buff = reg;
         return 0;

      case TLEPAERR_REG:
         *buff = 0;
         return 0;

      }   
   }


   if (type==BDOOR_STORE_WORD) {
      switch (off) {

         /* 
          * I/O registers 
          */
         
      case TLMBPR_REG:

	 /* 
	  * Mailbox (store conditional)
	  */
	 TLaserMbox(M_FROM_CPU(cpuNum),vA,*buff);
	 return 0;

      case TLILID0_REG:
      case TLILID1_REG:
      case TLILID2_REG:
      case TLILID3_REG:
         return 0;

      case TLINTRSUM0_REG:
      case TLINTRSUM1_REG:
         if (*buff & (1 << 0)) {   /* UART interrupt */
            /* Not acknowledged from here */
         } 
         if (*buff & (1 << 2)) {   /* SCSI (IPL15) interrupt */
            ASSERT(*buff & (1 << 12));  /* Ensure it is on node 4 */
            /* Not acknowledged from here */
         }
         if (*buff & (1 << 5)) {   /* IPI interrupt */
            ASSERT(tlaserState[cpuNum].ipiPending>0);
            CPUPrint("IPI ACK cpu=%d pending=%d \n",cpuNum,tlaserState[cpuNum].ipiPending);
            tlaserState[cpuNum].ipiPending--;
            if (!tlaserState[cpuNum].ipiPending) { 
               EV5_ClearExtInterrupt(PE[cpuNum],INTLEVEL_IRQ2);
            }
         }
         if (*buff & (1 << 6)) {   /* Timer interrupt */
            TLaserClockAck(cpuNum);
         }
	 return 0; 

      case TLEPAERR_REG:
         return 0;

      }
   }

   CPUPrint("TLaserRegisters: OFF 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
            off, PE[cpuNum]->PC, ra, vA,bdoorType[type]);  
   if (type==BDOOR_LOAD_WORD || type==BDOOR_LOAD_DOUBLE) { 
      *buff = 0;
   }
   return 0;
}

#else /* BOOT_NT */

static int TLaserRegisters(int cpuNum, VA vA, int type, Reg *buff)
{
   VA off = vA  & 0xfffff;
   Reg ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);
#if 0
   CPUPrint("TLaserRegisters:  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
	    PE[cpuNum]->PC,
	    ra,
	    vA,bdoorType[type]);  
#endif
   if (type==BDOOR_LOAD_WORD) {
      uint32 reg=0;
      switch(off) {
      case TLDEV_REG:
	 if (0) { 
	    *buff = 0x5000; /* TLSB_MEM */
	 } else {
	    *buff = 0x00002020; /* TLSB_ITIOP_LDEV */
	 }
	 /* from tlsb_init */

	 return 0;    
         
      case TLCNR_REG: 
	 /* tlsb_conf: nxv->lsb_lcnr */
	 *buff = 0; /* ok */
	 return 0;
     
      case IDPNSE1_REG: /*0x2140*/
         /* tlsb_iop.c::tiopconfl1  iopreg->idpnse1; */
         *buff = 0x6;  /* ((IDPNSE_CBLOK | IDPNSE_PWROKC) ) */
         return 0;

      case TLILID1_REG: 
    
         /* 
          * return the interrupt vector.
          * Accessed for I/O nodes only
          *
          * For now, hack in 0x820 (which corresponds to the entry
          * used by cam_simos.c
          *
          * XXX should have OS notify device of interrupt vector
          * XXX FIXTHIS (requries OS mod) bugnion
          *
          * Added to this hack for networking - verghese
          * Maybe I'll come back and clean it up
          *
          */

         if (SimOS_SCSI_Pending(cpuNum)) {
            *buff = 0x820;
         } else if (Simether_Pending(cpuNum)) {
            *buff = 0x830;
         } else {
            *buff = 0x0;
         }
         return 0;
  
      case TLINTRSUM0_REG:
      case TLINTRSUM1_REG:
	 /*
	  * TLINTRSUMx interrupt source register read
	  */
	 if (cpuNum == 0 && sim_console_int_status_noSE(0)) {
	    /* first CPU only  */
	    reg |= (1<<0);
	 }
	 if (TLaserClockIntrPending(cpuNum))    reg |= (1<<6);
	 if (tlaserState[cpuNum].ipiPending)    reg |= (1<<5);
	 if (SimOS_SCSI_Pending(cpuNum)) {
	    reg |= (1<<2);   /* summary bit set */
	    reg |= (1<<12);  /* node4   bit */
	 }
         if (Simether_Pending(cpuNum)) {
	    reg |= (1<<2);   /* summary bit set */
	    reg |= (1<<12);  /* node4   bit */
	 }
	 if (!reg) {
            /*
             * XXX
             * XXX bugnion. known possible bug. System will hang. 
             * XXX under certain race conditions, the system will get into 
             * XXX the state where it issues sprious interrupts at infinitum
             * XXX 
             * XXX Mail me if you can reproduce it but can't fix it. I
             * XXX have not seen it in a long long time.
             */
	    CPUError("Tlaser.c: spurious interrupt cpu=%d pc=0x%lx \n",cpuNum,PE[cpuNum]->PC);
            reg = (1<<0);
	 }
         
	 *buff = reg;	 
	 return 0;
	 break;
        
      case IDPNSE0_REG: /*0x2a40*/
         CPUPrint("TLaserRegisters: EMPTY READ  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                  PE[cpuNum]->PC,
                  ra,
                  vA,bdoorType[type]);      
         *buff = 0;
         return 0;
      }
      
   }
   

   if (type==BDOOR_STORE_WORD) {
      switch (off) {
      case TLMBPR_REG:
	 /* 
	  * Mailbox (store conditional)
	  */
	 TLaserMbox(M_FROM_CPU(cpuNum),vA,*buff);
	 return 0;
      case ICCNSE_REG:
      case IDPVR_REG:
      case IDPNSE0_REG:
      case IDPNSE1_REG:
      case IDPNSE2_REG:
      case IDPNSE3_REG:
      case TLCPUMASK_REG:
         CPUPrint("TLaserRegisters: EMPTY WRITE  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                  PE[cpuNum]->PC,
                  ra,
                  vA,bdoorType[type]);  
         return 0;
      }
      
   }

   if (type==BDOOR_STORE_DOUBLE) {
      switch (off) {
      case TLINTRSUM0_REG:
	 ASSERT( (cpuNum & 1)==0);
	 tlaserState[cpuNum].intrSumType =  *(uint64*)buff;
	 return 0;
      case TLINTRSUM1_REG:
	 ASSERT( (cpuNum & 1)==1);
	 tlaserState[cpuNum].intrSumType =  *(uint64*)buff;
	 return 0; 
      case TLMBPR_REG:
         CPUPrint("TLaserRegisters: EMPTY WRITE  PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                  PE[cpuNum]->PC, ra, vA, bdoorType[type]);  
         return 0;
      }
   }
   if (type==BDOOR_LOAD_DOUBLE) {
      switch(off) { 
      case TLINTRSUM0_REG:
      case TLINTRSUM1_REG: 
         switch (tlaserState[cpuNum].intrSumType) {
         case TLASER_INTRSUM_UART:
            *buff = 1; /* at sys_int_20, nothing else there  */
            return 0;
         case TLASER_INTRSUM_INTIM: 
            *buff = 0;
            /*TLaserClockAck(cpuNum);*/
            return 0;

         case TLASER_INTRSUM_IPI:
            ASSERT(tlaserState[cpuNum].ipiPending>0);
            CPUPrint("IPI ACK cpu=%d pending=%d \n",cpuNum,tlaserState[cpuNum].ipiPending);
            tlaserState[cpuNum].ipiPending--;
            if (!tlaserState[cpuNum].ipiPending) { 
               EV5_ClearExtInterrupt(PE[cpuNum],INTLEVEL_IRQ2);
            }
            return 0;
         }
      }
  
   }
         
   CPUWarning("TLaserRegisters:  ERROR PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
              PE[cpuNum]->PC,
              ra,
              vA,bdoorType[type]);  
   if (type==BDOOR_LOAD_WORD || type==BDOOR_LOAD_DOUBLE) { 
      *buff = 0;
   }
   return 0;
}

#endif /* !BOOT_NT */



static int TLaser_IPI(int cpuNum, VA vA, int type, long *buff)
{
   int i;
   Reg ra;
   int machine = M_FROM_CPU(cpuNum);

   CPUVec.GetRegister(cpuNum, REG_RA, &ra);
#if 0
   CPUWarning("TLaserIPI:  CPU %d PC 0x%lx RA 0x%lx va=0x%lx type=%s val=%d\n",
	      cpuNum,PE[cpuNum]->PC,
	      ra,
	      vA,bdoorType[type],*buff);  
#endif   
   if ((vA==0xfffffcff8e000000+TLIPINTR_REG) && 
       (type==BDOOR_STORE_WORD || type==BDOOR_STORE_DOUBLE)) {
      int target = *buff;

      for (i=FIRST_CPU(machine);i<=LAST_CPU(machine);i++) {
	 if ( (1LL<<(i-FIRST_CPU(machine))) & target) {  
	    CPUPrint("IPI SND machine=%d cpu=%d pending=%d from=%d \n",
                     machine,i,tlaserState[i].ipiPending+1,cpuNum);
	    if (!tlaserState[i].ipiPending) {
	       EV5_SetExtInterrupt(PE[i],INTLEVEL_IRQ2);
	    }
	    tlaserState[i].ipiPending++;
	 }
      }
      return SUCCESS;
   }
   CPUError("TLaserIPI:  CPU %d PC 0x%lx RA 0x%lx va=0x%lx type=%s val=%d\n",
	      cpuNum,PE[cpuNum]->PC,
	      ra,
	      vA,bdoorType[type],*buff); 
   return 0;
}

#ifdef BOOT_NT

typedef union _WHATAMI{
  struct{
     unsigned char CpuNumber: 1;      // Cpu number for that module (0 or 1)
     unsigned char NodeId: 3;         // Node number for the cpu module.
     unsigned char TlsbBad: 1;
     unsigned char TlsbConwin: 1;
     unsigned char MnfgMode: 1;
     unsigned char Reserved: 1;
  } d;
   unsigned long all;
} WHATAMI, *PWHATAMI;

static int TLaserWHATAMI_access(int cpuNum, VA vA, int type, long *buff)
{
   if (vA==0xfffffcffc0000000 && type==BDOOR_LOAD_WORD) {
      PWHATAMI result = (PWHATAMI)buff;	
      int machine = M_FROM_CPU(cpuNum);
      result->all = 0;
      result->d.CpuNumber = (cpuNum-FIRST_CPU(machine)) & 0x1;
      result->d.NodeId = (cpuNum-FIRST_CPU(machine)) >> 1;
      return SUCCESS;
   }
   CPUError("TLaserWHATAMI_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n", 
            cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
   return 0;
}

/*
 * Defining a totally arbritrary serial number to return to NT
 */
#define SERIAL_NUMBER 0xFAFAFAFA  

static int TLaserSERNUM_access(int cpuNum, VA vA, int type, long *buff)
{
   static uint32 serialNumber = SERIAL_NUMBER;
   if (vA==0xfffffcffc7000000 && type==BDOOR_LOAD_WORD) {
      *buff = 0;
      *(uint32*)buff = serialNumber;
      CPUPrint(
                 "TLaserSERNUM_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
                 cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
      return SUCCESS;
   }
   else if (vA==0xfffffcffc7000000 && type==BDOOR_STORE_WORD) {
      serialNumber = *(uint32*)buff;
      CPUPrint(
                 "TLaserSERNUM_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
                 cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
      return SUCCESS;
   }
   CPUError("TLaserSERNUM_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
            cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
   return 0;
}

static int TLaserPCIA_SG_access(int cpuNum, VA vA, int type, long *buff)
{
   VA off = vA  & 0xffffff;
   Reg ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);
   if (type==BDOOR_STORE_WORD) {
#if 0
      if (*buff) {
         CPUWarning(
             "TLaserPCIA_SG_access: OFF 0x%lx CPU %d PC 0x%lx ra 0x%lx va=0x%lx type=%s val=%d\n",
              off,cpuNum,PE[cpuNum]->PC,ra,vA,bdoorType[type],*buff);
      }
#endif
      return SUCCESS;
   } else if (type==BDOOR_LOAD_WORD) {
      *buff = 0;
#if 0
      CPUWarning(
          "TLaserPCIA_SG_access: OFF 0x%lx CPU %d PC 0x%lx ra 0x%lx va=0x%lx type=%s val=%d\n",
          off,cpuNum,PE[cpuNum]->PC,ra,vA,bdoorType[type],*buff);
#endif
      return SUCCESS;
   }
   return 0;
}

static int TLaserFLASH_access(int cpuNum, VA vA, int type, long *buff)
{
   CPUPrint(
              "TLaserFLASH_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
              cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
   if (type==BDOOR_STORE_WORD) {
      return SUCCESS;
   } else if (type==BDOOR_LOAD_WORD) {
      *buff = 0;
      return SUCCESS;
   }
   return 0;
}

/* Just ignore PCI interrupt commands */

static int TLaserPCI_fake_access(int cpuNum, VA vA, int type, long *buff)
{
   CPUPrint(
              "TLaserPCI_fake_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
              cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
   if (type==BDOOR_STORE_WORD) {
      return SUCCESS;
   } else if (type==BDOOR_LOAD_WORD) {
      *buff = 0;
      return SUCCESS;
   }
   return 0;
}
#endif /* BOOT_NT */

#ifdef BOOT_LINUX

static int Timer_access(int cpuNum, VA vA, int type, long *buff)
{
    CPUWarning(
		"Timer_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
        cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
	if (type==BDOOR_STORE_WORD) {
		return SUCCESS;
        } else if (type==BDOOR_LOAD_WORD) {
		*buff = 0;
                return SUCCESS;
        }
	return 0;
}

static Unknown_access(int cpuNum, VA vA, int type, long *buff)
{
    CPUWarning(
		"Unknown_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
        cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
	if (type==BDOOR_STORE_WORD) {
		return SUCCESS;
        } else if (type==BDOOR_LOAD_WORD) {
		*buff = 0;
                return SUCCESS;
        }
	return 0;
}

static ciaInit_access(int cpuNum, VA vA, int type, long *buff)
{
    CPUWarning(
		"ciaInit_access: CPU %d PC 0x%lx va=0x%lx type=%s val=%d\n",
        cpuNum,PE[cpuNum]->PC, vA,bdoorType[type],*buff);
	if (type==BDOOR_STORE_WORD) {
		return SUCCESS;
        } else if (type==BDOOR_LOAD_WORD) {
		*buff = 0;
                return SUCCESS;
        }
	return 0;
}

#endif /* BOOT_LINUX */

/*
 * Note: Registry ranges are now in increasing order in the address space.
 *       This makes management easier, so please try to keep it this way.
 * - benw
 */
void InitTLaser(void)
{
   int n=0;

   /*
    * PCI fake registers
    */
#ifdef BOOT_NT
   RegistryAddRange(0xfffffcc380400000,0x100000,REG_FUNC,
                    (void*)TLaserPCI_fake_access,"tlaser pci_fake");

   /*
    * Port access ranges
    */
#ifdef NOT_USED
   RegistryAddRange(0xfffffcc501000000,0x1000000,REG_FUNC,
                    (void*)TLaserPCIA_SG_access,"tlaser eisa");
#endif
   RegistryAddRange(0xfffffcc600000000,0x1000000,REG_FUNC,
                    (void*)TLaserISP_access,"tlaser eisa");
#endif


   /*
    * PCI Bus ranges
    */

   RegistryAddRange(0xfffffcc700000000,0x1000000,REG_FUNC,
		    (void*)TLaserPCIA_access,"tlaser pcia");

   RegistryAddRange(0xfffffcc780000000, 0x1000000,REG_FUNC,
		    (void*)TLaserPCIA_access,"tlaser pcia");

#ifdef BOOT_NT
   RegistryAddRange(0xfffffcc781000000, 0x1000000,REG_FUNC,
		    (void*)TLaserPCIA_SG_access,"tlaser pcia_sg");
#endif


#ifdef BOOT_LINUX

    RegistryAddRange(0xfffffc8580000000,0x1000000,REG_FUNC,
                     (void*)Timer_access,"timer");

    RegistryAddRange(0xfffffc8740000000,0x1000000,REG_FUNC,
                     (void*)Unknown_access,"unknown");

    RegistryAddRange(0xfffffc8760000000,0x1000000,REG_FUNC,
                     (void*)ciaInit_access,"ciaInit");
    
#endif

   /*
    * Turbolaser Bus ranges 
    */


   for (n=0;n<10;n++) {
      RegistryAddRange(0xfffffcff88000000 + (n*0x400000),0x4000,REG_FUNC,
		       (void*)TLaserRegisters,"tlaser regs");
   }

   /*
    * Broadcast
    */

   RegistryAddRange(0xfffffcff8e000000 ,0x2000,REG_FUNC,
		    (void*)TLaser_IPI,"tlaser_ipi");


   /*
    * GBUS ranges
    */

#ifdef BOOT_NT
   RegistryAddRange(GBUS_BASE + FLASH_BASE, 0x1000000,
                    REG_FUNC, (void*)TLaserFLASH_access,"tlaser flash");
#endif

   RegistryAddRange(GBUS_BASE + UART_BASE, 0x1000,
		    REG_FUNC, (void*)ALPHA_8530_CN_access, "alpha_8530");
   
   RegistryAddRange(GBUS_BASE + WATCH_CSR_BASE, 0x2000,
                    REG_FUNC, (void*)TLaser_mc146818__access, "mc146818 clock");

#ifdef BOOT_NT
   RegistryAddRange(GBUS_BASE + WHATAMI_REG, 0x2000,
                    REG_FUNC, (void*)TLaserWHATAMI_access, "tlaser whatami");

   RegistryAddRange(GBUS_BASE + SERNUM_REG, 0x2000,
                    REG_FUNC, (void*)TLaserSERNUM_access, "tlaser sernum");
#endif


   /*
    * SimOS fake addressing range for SIMOS-SCSI PIOS
    */

   RegistryAddRange(SIMOS_SCSI_ADDR, SIMOS_SCSI_MAXCTLR*sizeof(SimOS_SCSI),REG_FUNC,
                    (void*)SimOS_SCSI_access,"simos scsi");
#ifdef BOOT_NT
   RegistryAddRange(SIMOS_SCSI_ADDR_32, sizeof(SimOS_SCSI),REG_FUNC,
                    (void*)SimOS_SCSI_access,"simos scsi");
#endif



   /*
    * SimOS fake addressing range for SIMETHER PIOS
    */

   RegistryAddRange((VA)DEV_ETHER_REGISTERS, sizeof(DevEtherRegisters),REG_FUNC,
                    (void*)Simether_access, "simos ethernet");
#ifdef BOOT_NT
   RegistryAddRange((VA)DEV_ETHER_REGISTERS_32, sizeof(DevEtherRegisters),REG_FUNC,
                    (void*)Simether_access, "simos ethernet");
#endif

   /*
    * NT Framebuffer registers and memory 
    */
#ifdef NT_FRAMEBUFFER
   RegistryAddRange(SIMOS_FB_REGADDR, 0x1000, REG_FUNC,
                    (void*)Framebuf_reg_access, "simos framebuf_regs");

   RegistryAddRange(SIMOS_FB_MEMADDR, 0x8000000, REG_FUNC,
                    (void*)Framebuf_mem_access, "simos framebuf_mem");
#endif

   InitTLaserUart();
}

