/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */




#include <stddef.h>
#include <stdlib.h>


#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "tlaser.h"
#include "devices.h"





/* from io/dec/mbox/mbox.h */

struct mbox {
	/*** hardware portion ***/
	/* NB: 
	 * the hardware portion of the mailbox must be on a 64 byte boundary 
	 */
	unsigned int cmd;
	unsigned int mask:8,
                     rsv1:8,
                     hose:8,
                     rsv2:8; 
	unsigned long rbadr;
	unsigned long wdata;
	unsigned int unpredictable0[2];
	unsigned int rdata;
	unsigned int unpredictable1;
	union {
#define mb_status status_un.stat
#define mb_don_bit status_un.status_64.don
#define mb_err_bit status_un.status_64.err
		struct status_64 {
			unsigned long don:1,
		                      err:1,
		                      stat0:62;
		} status_64;
		unsigned long stat;
	} status_un;
	unsigned int unpredictable2[2];
	unsigned int unpredictable3[2];
	/*** software portion of the mailbox ***/
	unsigned int bus_timeout;  /* number of usec's to wait for comp */
	unsigned int mbox_inuse;   /* debug */
	vm_offset_t mbox_reg;      /* vaddr of mail box reg for this mbox */
	vm_offset_t mbox_pa;       /* mbox phys addr, set by mbox_pool_init */
	void (*mbox_cmd)();        /* adapter specific cmd field rtn */
	int (*err_rtn)();          /* adapter specific err rtn
				    * if != 0, then bus_ctlr_ptr must be
				    * filled in
				    */
#ifdef KERNEL
	bus_ctlr_common_t bus_ctlr_ptr; /* used by err_rtn */
#else
        void *bus_ctlr_ptr;
#endif
	struct mbox *next;		   /* used to link mbox structures when in a list */
#define PAD 8
	/* pad out to next 64 byte boundary */
	unsigned char pad[PAD];
};

typedef struct mbox * mbox_t;

#define BRDG_CSR    0x40000000     /* b(ridge) bit <30> 1-cmd to remote adpt */
#define WHOAREYOU   BRDG_CSR       /* who's on the other end of this hose? */



int TLaserMbox(int machine, VA vA, Reg val)
{
#ifdef DATA_HANDLING
   bomb. Must use DMA interface;
#endif
   struct mbox *mbox = (struct mbox*) PHYS_TO_MEMADDR(machine,val);
   ASSERT (val >=0 && val < MEM_SIZE(machine));
   CPUWarning("TLaser MBOX access ioAddr=0x%lx memAddr=0x%lx \n",
	      vA,val);
   CPUWarning("         cmd=0x%x mask=0x%x hose=0x%x rbadr=0x%lx wdata=0x%lx rdata=0x%x\n",
	      mbox->cmd,mbox->mask,mbox->hose,mbox->rbadr,mbox->wdata,mbox->rdata);
   
   switch (mbox->cmd) {
   case WHOAREYOU: 
      mbox->rdata = 0xef00;  /* PCIA */
      break;
   default: CPUError("TLaserMbox\n");
   }

   mbox->mb_err_bit = 0;
   mbox->mb_don_bit = 1;
   return 0;
   
}


