/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "tlaser.h"
#include "tlaser_clock.h"
#include "tlaser_uart.h"
#include "tlaser_pcia.h"
#include "kbd_mouse.h"

#include <stddef.h>
#include <stdlib.h>

extern char *bdoorType[];
#define PCI_DEBUG
int TLaserPCIA_access(int cpuNum, VA vA, int type, Reg *buff)
{  
   VA base = vA & BITMASK(31); /* clear top bits */
/*   uint64 width = (base >> WIDTH_SHIFT) & 0x3; */
   uint64 pciAddr = (base>>IO_SHIFT);
   uint64 offset  = pciAddr & 0xff;
   VA ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (type==BDOOR_LOAD_WORD) {
      uint64 res = 0;
      if ((pciAddr & 0xfff00) == 0x2000 || (pciAddr & 0xfff00) == 0x8000 || (pciAddr & 0xfff00) == 0x10000 || (pciAddr & 0xfff00) == 0x800) {
         switch (offset) { 
         case PCI_VENDOR_ID: 
            res = 0x1077;  /* QLOGIC_ID */ /* too complicated */
            res = 0x101A;  /* NCR_ID    */ /* does not work on tlaser (handler_add fails) */
            res = 0x1011;  /* DEC_ID    */
            res = 0x1291 | ((0x1291) << 16);  /* simos: August 1st */
            break;
         case PCI_DEVICE_ID: 
            res = 0x1020;  /* QLOGIC: ISP1020 SCSI Adapter */ 
            res = 0x0001;  /* NCR: NCR_810 */
            res = 0x0008;  /* DEC: PZA SCSI */
            res = 0x1291;  /* simos:  August 1st  */
            break;
         case PCI_HEADER_TYPE: 
            res = 0;       /* type 0 header */ 
            break;
         case PCI0_INTERRUPT_LINE:
            res = 0x9 | (0x1 << 8);  /* Line and pin (latency and grant to 0) */
            break;
         default: 
            res = 0;
            break;
         }
#ifdef PCI_DEBUG
            CPUPrint("TLaser_PCI SCSI OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx %s\n",
                       offset,
                       PE[cpuNum]->PC,ra,vA,pciAddr,bdoorType[type]);
#endif  
      } else if ((pciAddr & 0xfff00) == 0x4000) {
         switch (offset) { 
         case PCI_VENDOR_ID: 
            res = 0x1291;  /* simos ID  */
            break;
         case PCI_DEVICE_ID: 
            res = 0x1292;  /* SIMOS ethernet */ 
            break;
         case PCI_HEADER_TYPE: 
            res = 0;       /* type 0 header */ 
            break;
         case PCI0_INTERRUPT_LINE:
            res = 0x9;  /* Line ? */
            break;
         case PCI0_INTERRUPT_PIN:
            res = 0x1;  /* pin */
            break;
         default: 
            res = 0;
            break;
         }
#ifdef PCI_DEBUG
            CPUPrint("TLaser_PCI EtherSCSI OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx %s\n",
                       offset,
                       PE[cpuNum]->PC,ra,vA,pciAddr,bdoorType[type]);
#endif
      } else {
         res = 0;
         CPUPrint("TLaser_PCI Unknown OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx %s\n",
                  offset, PE[cpuNum]->PC,ra,vA,pciAddr,bdoorType[type]);  
      }
      
      /*
       * PCI->Tlaser data conversion. See io/dec/pci/pcia.c 
       */
      *buff = res<<((pciAddr&3)<<3);

      return 0;
   }

#ifdef PCI_DEBUG
   CPUPrint("TLaser_PCI NotLoad OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx value=0x%lx %s\n",
	      offset, PE[cpuNum]->PC,ra,vA,pciAddr,*buff,bdoorType[type]);  
#endif
   return 0;
}

int TLaserISP_access(int cpuNum, VA vA, int type, Reg *buff)
{  
   VA base = vA & 0xfffff;
/*   uint64 width = (base >> WIDTH_SHIFT) & 0x3; */
   uint64 port = (base>>IO_SHIFT);
   VA ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (type==BDOOR_LOAD_WORD) {
      switch (port) { 
#ifdef NT_FRAMEBUFFER
      case ISA_KBD_OUTPUT_BUFFER_PORT:
        return KbdMouse_Data_Read(cpuNum, vA, type, buff);
      case ISA_KBD_STATUS_PORT: 
         return KbdMouse_Status_Read(cpuNum, vA, type, buff);
#endif
      case EISA_TIMER1:  /* Timer 1 register (2 reads - LSB, then MSB) */
         *buff = 0;
         return 0;
      }
   }
   
   if (type==BDOOR_STORE_WORD) {
      switch (port) {
#ifdef NT_FRAMEBUFFER
      case ISA_KBD_OUTPUT_BUFFER_PORT: 
         return KbdMouse_Data_Write(cpuNum, vA, type, buff);
      case ISA_KBD_STATUS_PORT: 
         return KbdMouse_Command_Write(cpuNum, vA, type, buff);
#endif
      case EISA_COMMAND_MODE1:  /* Command 1 register */
         if(*buff == 0) /* SELECT_COUNTER_0 | SB_COUNTER_LATCH */
            return 0;
         else {
            CPUPrint("TLaser_ISP: Command1 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                       *buff, 
                       PE[cpuNum]->PC,
                       ra,
                       vA,bdoorType[type]);
            return 0;
         }
      }
   }

   CPUPrint("TLaser_ISP: Port 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
              port, 
              PE[cpuNum]->PC,
              ra,
              vA,bdoorType[type]);
   
   if (type==BDOOR_LOAD_WORD) {
      *buff = 0;
   }
   return 0;
}
