/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



#ifndef TLASER_PCIA_H
#define TLASER_PCIA_H

#define IO_SHIFT 5
#define WIDTH_SHIFT ( IO_SHIFT -2)

/*
 * Common PCI offsets
 */
#define PCI_VENDOR_ID    	0x00   /* ro */
#define PCI_DEVICE_ID		0x02   /* ro */
#define PCI_COMMAND		0x04   /* device control */
#define PCI_STATUS		0x06
#define PCI_REVISION_ID 	0x08   /* ro */
#define PCI_PROG_IF_CODE	0x09   /* ro */
#define PCI_SUB_CLASS_CODE	0x0A   /* ro */
#define PCI_BASE_CLASS_CODE	0x0B   /* ro */
#define PCI_CACHE_LINE_SIZE	0x0C   /* ro+ */
#define PCI_LATENCY_TIMER	0x0D   /* ro+ */
#define PCI_HEADER_TYPE		0x0E   /* ro */
#define PCI_BIST		0x0F   /* Built in self test */
/*
 * Type 0 PCI offsets 
 */
#define PCI0_BASE_ADDR          0x10
#define PCI0_CIS                0x28
#define PCI0_SUB_VENDOR_ID      0x2C
#define PCI0_SUB_SYSTEM_ID      0x2E
#define PCI0_ROM_BASE_ADDR      0x30
#define PCI0_INTERRUPT_LINE     0x3C
#define PCI0_INTERRUPT_PIN      0x3D   /* ro */
#define PCI0_MINIMUM_GRANT      0x3E   /* ro */
#define PCI0_MAXIMUM_LATENCY    0x3F   /* ro */
/*
 * Device specific offsets
 */
#define PCI_DEVICE_SPECIFIC     0x40   /* 192 bytes */


/*
 * Common EISA offsets
 */
#define EISA_DMA1_BASE             0x00
#define EISA_INTERRUPT1_CONTROL0   0x20
#define EISA_INTERRUPT1_CONTROL1   0x21
#define EISA_TIMER1                0x40
#define EISA_REFRESH_REQUEST       0x41
#define EISA_SPEAKER_TONE          0x42
#define EISA_COMMAND_MODE1         0x43
#define EISA_TIMER2                0x48
#define EISA_CPU_SPEED_CONTROL     0x4A
#define EISA_COMMAND_MODE2         0x4B
#define EISA_NMI_STATUS            0x61
#define EISA_NMI_ENABLE            0x70
#define EISA_DMA_PAGE_LOW_PORT     0x80
#define EISA_INTERRUPT2_CONTROL0   0xA0 
#define EISA_INTERRUPT2_CONTROL1   0xA1
#define EISA_DMA2_BASE             0xC0
/* This is only the start of this list */
/* Extend it as required */

/*
 * PC Standard ISA offsets [ISA == EISA here]
 */
#define ISA_KBD_OUTPUT_BUFFER_PORT   0x60
#define ISA_KBD_STATUS_PORT          0x64

#define ISA_RTC_ADDRESS_PORT         0x70
#define ISA_RTC_DATA_PORT            0x71
#define ISA_RTC_ALT_ADDRESS_PORT     0x170
#define ISA_RTC_ALT_DATA_PORT        0x171

#define ISA_FLOPPY_PORT              0x3F0
#define ISA_PARALLEL_PORT            0x3BC
#define ISA_COM1_PORT                0x2F8
#define ISA_COM2_PORT                0x3F8
#define ISA_CMOS_PORT                0x800


extern int TLaserPCIA_access(int cpuNum, VA vA, int type, Reg *buff);
extern int TLaserISP_access(int cpuNum, VA vA, int type, Reg *buff);

#endif 
