/*
 * Copyright (C) 1996 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * cpu_state.h
 *
 * Complete state of a PE... this structure is to be used by all cpu 
 * simulators so that switching between them is simple.
 *
 * Author: $Author: nikos $
 * Date: $Date: 1998/04/29 04:54:41 $
 *****************************************************************/

#ifndef CPU_STATE_H
#define CPU_STATE_H

#include "simtypes.h"
#include "cpu_interface.h"

typedef struct SegReg {
  uint16 selector;
  Reg64  descriptor;
} SegReg;

typedef struct CPUState {
  CPUStatus cpuStatus;

  Reg    eip;
  Reg    eflags;
  Reg    eax;
  Reg    ebx;
  Reg    ecx;
  Reg    edx; 
  Reg    esi;
  Reg    edi;
  Reg    ebp;
  Reg    esp;
  SegReg  cs; 
  SegReg  ss;
  SegReg  ds;
  SegReg  es;
  SegReg  fs;
  SegReg  gs;
  Reg32   cr[5];
  Reg64   gdtr;
  Reg64   idtr;
  SegReg  ldtr;
  SegReg  tr;
  Reg64   tsc; 
  Reg32   dr[8];
  int  outOfSlaveLoop;
  uint memSize;               /* Size of main memory in bytes */
  char *memoryPtr;            /* Main memory array */
  void *x86simCpu;            /* X86sim CPU handle */
} CPUState;


#endif /* CPU_STATE_H */

