
#include <stdlib.h>
#include "simtypes.h"
#include "iodevs.h"
#include "iospace.h"
#include "vserial.h"
#include "vconsole.h"
#include "vscsi.h"
#include "sim_error.h"
#include "intr.h"
#include "eventcallback.h"


IODevices *iodevs; /* Current we only support one set of devices */


#define DEFAULT_COM1_ADDR   0x3F8
#define DEFAULT_COM1_IRQ    4
#define DEFAULT_COM2_ADDR   0x2F8
#define DEFAULT_COM2_IRQ    3
#define DEFAULT_COM3_ADDR   0x3E8
#define DEFAULT_COM3_IRQ    4
#define DEFAULT_COM4_ADDR   0x2E8
#define DEFAULT_COM4_IRQ    3

#define CONSOLE_ADDR        DEFAULT_COM1_ADDR   /* console is hardcoded on com 0 */

/*************************************************************
 *   InitDevices
 *************************************************************/

void 
InitDevices(void)
{
   iodevs = (IODevices *) malloc(sizeof(IODevices));
   ASSERT(iodevs);
   IOSpace_Init(iodevs);
   Intr_Init(iodevs);
   
   VSerial_Init(&(iodevs->com[0]), iodevs, DEFAULT_COM1_ADDR,
                DEFAULT_COM1_IRQ);
   VSerial_Init(&(iodevs->com[1]), iodevs, DEFAULT_COM2_ADDR, 
                DEFAULT_COM2_IRQ);
   VSerial_Init(&(iodevs->com[2]), iodevs, DEFAULT_COM3_ADDR, 
                DEFAULT_COM3_IRQ);
   VSerial_Init(&(iodevs->com[3]), iodevs, DEFAULT_COM4_ADDR, 
                DEFAULT_COM4_IRQ);
   
   VConsole_Init(&(iodevs->console), iodevs); 

   VSerial_Connect(&(iodevs->com[0]), (SerialDevice*)&(iodevs->console));
   
   VScsi_Init(iodevs);
}


IODevices *
GetIODevices(int cpuNum)
{
   ASSERT(cpuNum == 0);
   return iodevs;
}


static EventCallbackHdr pollHdr;
#define DEVICE_POLL_INTERVAL (10*1000)
static void 
PollDevices(int cpuNum, EventCallbackHdr *hdr, void *arg)
{


  VConsole_poll(&(iodevs->console), iodevs);
  /* register callback for next poll round */
  EventDoCallback(0, PollDevices, &pollHdr, NULL, DEVICE_POLL_INTERVAL);
}


void
InstallPoller(void)
{
  static int installed = 0;
  if (!installed) {
     installed = 1;
     EventDoCallback(0, PollDevices, &pollHdr, NULL, DEVICE_POLL_INTERVAL);
  }
}
