/************************************************************
 *
 *  iospace.c
 *
 ************************************************************/

#include <stdlib.h>

#include "simtypes.h"
#include "sim_error.h"
#include "iodevs.h"
#include "iospace.h"

#ifdef SIMOS
#include "cpu_interface.h"
#endif
#define NUM_IOS_BUCKETS    128

#define HASH_ADDR(_addr)   ((_addr) % NUM_IOS_BUCKETS);

#define IOS_INVALID        0
#define IOS_VALID          1
#define IOS_DELETED        2

typedef struct IOS_Bucket {
   char          flag;
   uint16        addr;
   IOSpace_Func  f;
   void          *clientData;
} IOS_Bucket;


void IOSpace_Init(IODevices *iod)
{
   int i;
   
   iod->IOS_Hash = (IOS_Bucket*)malloc(NUM_IOS_BUCKETS * sizeof(IOS_Bucket));

   for (i=0; i<NUM_IOS_BUCKETS; i++) {
      iod->IOS_Hash[i].flag = IOS_INVALID;
   }
}

void IOSpace_Register(IODevices *iod, uint16 addr, IOSpace_Func f, void *clientData)
{
   int i, end;

   i = HASH_ADDR(addr);
   end = (i - 1) % NUM_IOS_BUCKETS;

   while (1) {
      ASSERT(i != end);
      if (iod->IOS_Hash[i].flag != IOS_VALID) {
         break;
      }
      ASSERT(iod->IOS_Hash[i].addr != addr);
      i = (i + 1) % NUM_IOS_BUCKETS;
   }

   iod->IOS_Hash[i].flag = IOS_VALID;
   iod->IOS_Hash[i].addr = addr;
   iod->IOS_Hash[i].f = f;
   iod->IOS_Hash[i].clientData = clientData;
}

void IOSpace_Unregister(IODevices *iod, uint16 addr, IOSpace_Func f)
{
   int i, end;

   i = HASH_ADDR(addr);
   end = (i - 1) % NUM_IOS_BUCKETS;

   while (1) {
      ASSERT(i != end);
      ASSERT(iod->IOS_Hash[i].flag != IOS_INVALID);
      if ((iod->IOS_Hash[i].flag == IOS_VALID)
          && (iod->IOS_Hash[i].addr == addr)) {
         break;
      }
      i = (i + 1) % NUM_IOS_BUCKETS;
   }

   iod->IOS_Hash[i].flag = IOS_DELETED;
}

int IOSpace_InOut(IODevices *iod, uint16 addr, int size, int inout, void *byte)
{
   int i, end;

   i = HASH_ADDR(addr);
   end = (i - 1) % NUM_IOS_BUCKETS;

   while (1) {
      if ((i == end) || (iod->IOS_Hash[i].flag == IOS_INVALID)) {
         CPUWarning("IOSpace_InOut: unimplemted %s to ioaddr 0x%x @ PC 0x%x\n", 
                    (inout == PORT_IN) ? "IN" : "OUT", 
                    addr,CPUVec.CurrentPC(0));
         return EXC_BUS_ERROR;
      }

      if ((iod->IOS_Hash[i].flag == IOS_VALID)
          && (iod->IOS_Hash[i].addr == addr)) {
         if (iod->IOS_Hash[i].f) {
            return (*iod->IOS_Hash[i].f)(iod, addr, size, inout, byte, iod->IOS_Hash[i].clientData);
         } else {
            return EXC_NONE;
         }
      }

      i = (i + 1) % NUM_IOS_BUCKETS;
   }

   return EXC_BUS_ERROR;
}
