/************************************************************
 *
 *  vconsole.c
 *
 ************************************************************/

#include "simtypes.h"
#include "sim_error.h"
#include "iodevs.h"
#include "vconsole.h"

#include "console.h"
#include "cpu_interface.h"

static void vconsole_connect(SerialDevice *sd, SerialPort *sp);
static void vconsole_transmitByte(SerialDevice *sd, uint8 byte);
static uint8 vconsole_receiveByte(SerialDevice *sd);

static VConsole *vcPtrs[MAX_CONSOLES];

static void
console_interrupt(int n, int on)
{
   int status;

   ASSERT(n == 0);  
   if (vcPtrs[n] == NULL) return;

   status = sim_console_int_status(n);
   if (status & CONS_INT_RX) {
      if (vcPtrs[n]->inputChar < 0) { 
         int ch = vcPtrs[n]->inputChar = sim_console_in(n);
         if (ch >= 0) { 
            vcPtrs[n]->sdhdr.RxAvail = 1;
            VSerial_BytesWaiting(vcPtrs[n]->sp);
         }
      } 
   } 
   /* XXX What should we do about transmit interrupts? */

}

void VConsole_Init(VConsole *vc, IODevices* iod)
{
   vc->sdhdr.connect = vconsole_connect;
   vc->sdhdr.transmitByte = vconsole_transmitByte;
   vc->sdhdr.receiveByte = vconsole_receiveByte;
   vc->iod = iod;
   vc->sdhdr.RxAvail = 0;

   vc->sp = NULL;
   vc->simosConsole = 0;
   vc->inputChar = -1;
   vcPtrs[0] = vc;
   sim_console_init(1, console_interrupt);

   sim_console_int_set(0, CONS_INT_RX|CONS_INT_TX);

}

static void vconsole_connect(SerialDevice *sd, SerialPort *sp)
{
   VConsole *vc = (VConsole *)sd;

   ASSERT(vc);
   ASSERT(sp);

   vc->sp = sp; 
}

static void vconsole_transmitByte(SerialDevice *sd, uint8 byte)
{
   VConsole *vc = (VConsole *)sd;

   ASSERT(vc);

   sim_console_out(vc->simosConsole,(char)byte);
}

static uint8 vconsole_receiveByte(SerialDevice *sd)
{
   VConsole *vc = (VConsole *)sd;
   int byte, status;

   ASSERT(vc);
   ASSERT(vc->sdhdr.RxAvail);

   byte = vc->inputChar;
   if (byte < 0) { 
      byte = sim_console_in(vc->simosConsole);
   }
   vc->inputChar = -1;
   ASSERT(byte != -1);

   status = sim_console_int_status(vc->simosConsole);
   if (status & CONS_INT_RX) {
        vc->inputChar = sim_console_in(vc->simosConsole);
        if (vc->inputChar >= 0) { 
           vc->sdhdr.RxAvail = 1;
        } else {
           vc->sdhdr.RxAvail = 0;
        }
   } else { 
        vc->sdhdr.RxAvail = 0;
   }
   return byte;
}

void
VConsole_poll(VConsole *vc, IODevices *iod)
{
   sim_console_poll();
   console_interrupt(vc->simosConsole,0);
}
