/************************************************************
 *
 *  vserial.c
 *
 *   XXX need vEISA
 *
 ************************************************************/
#include "errno.h"
#include "simtypes.h"
#include "iospace.h"
#include "intr.h"
#include "vscsi.h"


#include "cpu_interface.h"
#include "hd.h"
#include "sim.h"
#include "sim_error.h"

#if 0
#include "/maspar/mp1/devine/linux/drivers/scsi/vmx86_scsi.h" /* XXX */
#include <scsi/scsi.h>
#include <linux/unistd.h>
#endif

#define DID_OK          0x00 /* NO error                                */
#define DID_NO_CONNECT  0x01 /* Couldn't connect before timeout period  */
#define DID_BUS_BUSY    0x02 /* BUS stayed busy through time out period */
#define DID_TIME_OUT    0x03 /* TIMED OUT for other reason              */
#define DID_BAD_TARGET  0x04 /* BAD target.                             */
#define DID_ABORT       0x05 /* Told to abort for some other reason     */
#define DID_PARITY      0x06 /* Parity error                            */
#define DID_ERROR       0x07 /* Internal error                          */
#define DID_RESET       0x08 /* Reset by somebody.                      */
#define DID_BAD_INTR    0x09 /* Got an interrupt we weren't expecting.  */ 

#define MAX_CDB 12
#define MAX_SENSE 14
#define MAX_STATUS 32

#define CMD_TST_UNIT_RDY        0x0
#define CMD_REQ_SENSE           0x3
#define CMD_FORMAT              0x4
#define CMD_ADD_DEFECTS         0x7
#define CMD_READ_6              0x8
#define CMD_WRITE_6             0xa
#define CMD_INQUIRY             0x12
#define CMD_MODE_SELECT         0x15
#define CMD_MODE_SENSE          0x1a
#define CMD_STARTUNIT           0x1b
#define CMD_SEND_DIAG           0x1d
#define CMD_PREVREM             0x1e
#define CMD_READCAPACITY        0x25
#define CMD_READ                0x28
#define CMD_WRITE               0x2a
#define CMD_SEEK                0x2b
#define CMD_ERASE               0x2c
#define CMD_RDEFECTS            0x37

/* This is used with scatter-gather */
struct vmx86_scsi_chain {
  uint32  dataptr;         /* Location of data */
  uint32  datalen;         /* Size of this part of chain */
};

struct ecb {                    /* Enhanced Control Block 6.1 */
  uint8    inuse;                  /* Command Word */
  uint8    id;
  uint8    sg;
  uint32   dataptr;                /* Data or Scatter List ptr */
  uint32   datalen;                /* Data or Scatter List len */
  uint8    cdblen;                 /* CDB Length */
  uint8    cdb[MAX_CDB];           /* CDB area */
  uint32   statusptr;
  uint8    statuslen;
/* Hardware defined portion ends here, rest is driver defined */
  void  *SCpnt;             /* Link to the SCSI Command Block */
  void (*done)(void *);    /* Completion Function */
};


static struct vmx86_scsi_chain sgbuf[512];

int VScsi_Init(IODevices *iod)
{
   int id;
   for (id = 0; id < sizeof(iod->scsi0_drive)/sizeof(iod->scsi0_drive[0]); id++) {
      iod->scsi0_drive[id].errno = 0;
      iod->scsi0_drive[id].intrPending = 0;
      iod->scsi0_drive[id].bytesTransferred = 0;
      iod->scsi0_drive[id].ecbptr = 0;
   }
   iod->scsi0_IRQ = 10;

   return 0;
}

int 
mon_scsi_getconfig(int cpuNum, int slot, int *retVal)
{
   IODevices *iod = GetIODevices(cpuNum);
   
   /* XXX for now I have 1 controller and it is on slot 0 */
   if (slot != 0) {
      return -ENOSYS;
   }

   *retVal = iod->scsi0_IRQ;
   
   return 0;
}


int 
mon_scsi_command(int cpuNum, VA ecb_addr, int *retVal)
{
   IODevices *iod = GetIODevices(cpuNum);
   struct ecb ecb;
   Result r;
   int done, bytes, errnoVal;   

   r= CPUVec.GetMemory(cpuNum, ecb_addr, sizeof(struct ecb), (char*)&ecb);
   ASSERT(r == SUCCESS);


#ifdef notdef
   printk("VMX86: scsi received command 0x%x id=0x%x\n",
          (int)ecb.cdb[0], (int)ecb.id);
#endif
   
   switch (ecb.cdb[0]) {
   case CMD_TST_UNIT_RDY:
      {
         PA pa = 0;

         sim_disk_touch(0,0,ecb.id);
         sim_disk_startio(0,0,ecb.id, ecb.cdb, &pa, 0);
         sim_disk_status(0,0,ecb.id,&done,&bytes, &errnoVal);
         ASSERT(done);
         sim_disk_iodone(0,0,ecb.id);
         if (errnoVal == 0) { 
            iod->scsi0_drive[ecb.id].errno = DID_OK << 16;
         } else {
            iod->scsi0_drive[ecb.id].errno = DID_BAD_TARGET << 16;
         }
         iod->scsi0_drive[ecb.id].intrPending = 1;
         iod->scsi0_drive[ecb.id].ecbptr = ecb_addr;
         break;
      }

   case CMD_INQUIRY:
   case CMD_READCAPACITY:
      {
         Result r;
         PA pages[2];

         ASSERT(!ecb.sg);
         
         iod->scsi0_drive[ecb.id].errno = 0;
         iod->scsi0_drive[ecb.id].intrPending = 1;
         iod->scsi0_drive[ecb.id].ecbptr = ecb_addr;

         sim_disk_touch(0,0,ecb.id);
         r = CPUVec.TranslateVirtualNoSE(cpuNum, ecb.dataptr, &pages[0]);
         ASSERT(r == SUCCESS);         
         pages[1] = 0;
         sim_disk_startio(0,0,ecb.id, ecb.cdb, pages, pages[0] & 4095);
         sim_disk_status(0,0,ecb.id,&done,&bytes, &errnoVal);
         ASSERT(done);
         sim_disk_iodone(0,0,ecb.id);
         ASSERT(errnoVal == 0);        
         break;
      }
   case CMD_READ_6:
   case CMD_WRITE_6:
      {
         uint32 sector_start;
         uint32 num_sectors;
         int i;
         VA memPtr;
         PA pages[100], tmp;
         
         sim_disk_touch(0,0,ecb.id);
         iod->scsi0_drive[ecb.id].errno = 0;
         iod->scsi0_drive[ecb.id].intrPending = 1;
         iod->scsi0_drive[ecb.id].ecbptr = ecb_addr;

         sector_start = ((ecb.cdb[1] & 0x1f) << 16) | (ecb.cdb[2] << 8) | ecb.cdb[3];
         num_sectors = ecb.cdb[4];
         ASSERT(!ecb.cdb[5]);
            
         if (ecb.sg) {
            uint32 numsgs;
            uint32 num_sects_in_sg;
            uint32 total_sects_read;

            r = CPUVec.GetMemory(cpuNum, ecb.dataptr,  ecb.datalen, (char*)sgbuf);
            ASSERT(r == SUCCESS);
            
            numsgs = ecb.datalen / sizeof(struct vmx86_scsi_chain);
            total_sects_read = 0;
            
            for (i=0; i<numsgs; i++) {
               num_sects_in_sg = sgbuf[i].datalen / 512;
               ASSERT((sgbuf[i].datalen % 512) == 0);
               memPtr =  sgbuf[i].dataptr;
               
               r = CPUVec.TranslateVirtualNoSE(cpuNum, memPtr, &pages[0]);
               ASSERT(r == SUCCESS);         
               pages[1] = 0;
               r = CPUVec.TranslateVirtualNoSE(cpuNum, 
                                  memPtr+sgbuf[i].datalen-1, &tmp);
               ASSERT(r == SUCCESS);         
               ASSERT(tmp/4096 == pages[0]/4096);
               ecb.cdb[4] = num_sects_in_sg;
               ecb.cdb[1] = (sector_start>>16)&0x1f;
               ecb.cdb[2] = (sector_start>>8)&0xff;
               ecb.cdb[3] = (sector_start)&0xff;
               sim_disk_startio(0,0,ecb.id, ecb.cdb, pages, pages[0] & 4095);
               sim_disk_status(0,0,ecb.id,&done,&bytes, &errnoVal);
               ASSERT(done); ASSERT(errnoVal == 0); 
               ASSERT(bytes == sgbuf[i].datalen);
               sim_disk_iodone(0,0,ecb.id);

               sector_start += num_sects_in_sg;
               total_sects_read += num_sects_in_sg;
            }

            ASSERT(total_sects_read == num_sectors);
            
         } else {
               int nPages = 0;
               memPtr = ecb.dataptr;
               for (i = 0; i < ecb.datalen; ) { 
                  ASSERT(nPages < 100);
                  r = CPUVec.TranslateVirtualNoSE(cpuNum, memPtr+i, &pages[nPages]);
                  ASSERT(r == SUCCESS);         
                  nPages++;
                  if (i == 0) {
                     i = 4096 - (pages[0] % 4096);
                  } else {
                     i += 4096;
                  }
               }
               pages[nPages] = 0;
               sim_disk_startio(0,0,ecb.id, ecb.cdb, pages, pages[0] % 4096);
               sim_disk_status(0,0,ecb.id,&done,&bytes, &errnoVal);
               ASSERT(done); ASSERT(errnoVal == 0); 
               ASSERT(bytes == ecb.datalen);
               sim_disk_iodone(0,0,ecb.id);
         }
         
         break;
      }
   
   default:
      {
         Sim_Warning("VMX86: scsi received command 0x%x id=0x%x\n",
                (int)ecb.cdb[0], (int)ecb.id);
         abort();
         break;
      }
   }

   if (iod->scsi0_drive[ecb.id].intrPending) {
      Intr_Raise(iod, iod->scsi0_IRQ);
   }
   
   
   return 0;
}

int 
mon_scsi_ackintr(int cpuNum, int drive, int *retVal)
{
   IODevices *iod = GetIODevices(cpuNum);
   int id;

   for (id=0; id<8; id++) {
      if (iod->scsi0_drive[id].intrPending) {
         iod->scsi0_drive[id].intrPending = 0;
         *retVal = (int)iod->scsi0_drive[id].ecbptr;
         break;
      }
   }

   ASSERT(id < 8);

   return iod->scsi0_drive[id].errno;
}
