/* $Id: counter.cc,v 1.4 1998/04/29 04:54:45 nikos Exp $
 *
 * RCS Log:
 * $Log: counter.cc,v $
 * Revision 1.4  1998/04/29 04:54:45  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:35  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:20  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:54  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:54  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: counter.cc,v 1.4 1998/04/29 04:54:45 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/counter.cc,v 1.4 1998/04/29 04:54:45 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:23  bugnion
// stanford update (x86)
//
// Revision 1.1  1997/10/20  00:07:41  sean
// Initial revision
//
 */

static char rcsid[] __attribute__((unused))
        = "$Id: counter.cc,v 1.4 1998/04/29 04:54:45 nikos Exp $";

#include "counter.h"
#include <string.h>

InstructionCounter::InstructionCounter(void)
{
  total_instrs = 0;
  total_counted = 0;
  instr_names = 0;
  instr_counts = 0;
  array_lengths = 0;
}

InstructionCounter::~InstructionCounter(void)
{
  while(total_instrs)
    delete[] instr_names[--total_instrs];

  delete[] instr_names;
  delete[] instr_counts;
}

void InstructionCounter::Reset(void)
{
  while(total_instrs)
    delete[] instr_names[--total_instrs];
}

int InstructionCounter::AddInstruction(const char *instr_name, int to_add)
{
  int low, mid, high;
  int i, rslt;
  char **new_names;
  int *new_counts;

  // first, find the name...
  low = 0;
  high = total_instrs - 1;

  while(low <= high) {
    mid = (low + high) >> 1;
    rslt = strcmp(instr_name, instr_names[mid]);

    if(rslt < 0) high = mid-1; else
    if(rslt > 0) low  = mid+1; else {
      // already in list?  then increment and leave

      instr_counts[mid] += to_add;
      total_counted += to_add;
      return(instr_counts[mid]);
    }
  }

  // resize if necessary
  if(total_instrs >= array_lengths) {
    array_lengths += 32;

    new_names = new char *[array_lengths];
    new_counts = new int[array_lengths];

    for(i = 0; i < total_instrs; i++) {
      new_names[i] = instr_names[i];
      new_counts[i] = instr_counts[i];
    }

    delete[] instr_names;
    delete[] instr_counts;
    instr_names = new_names;
    instr_counts = new_counts;
  }

  for(i = total_instrs; i > low; i--) {
    instr_names[i] = instr_names[i-1];
    instr_counts[i] = instr_counts[i-1];
  }

  instr_names[i] = strdup(instr_name);
  instr_counts[i] = to_add;

  total_instrs++;
  total_counted += to_add;

  return(to_add);
}

void InstructionCounter::PrintReport(FILE *output) const
{
  int i;
  int half;

  half = (total_instrs + 1) >> 1;

  for(i = 0; i < half; i++) {
    fprintf(output, "%-10s: %7d calls (%6.2f%%)    ",
	    instr_names[i], instr_counts[i],
	    (total_counted ? (instr_counts[i] * 100.0 / total_counted) : 0.0));

    if((i + half) < total_instrs)
      fprintf(output, "%-10s: %7d calls (%6.2f%%)\n",
	      instr_names[i+half], instr_counts[i+half],
	      (total_counted ? (instr_counts[i+half] * 100.0 / total_counted) :
	                       0.0)); else
      fputc('\n', output);
  }
}
