/* -*- C++ -*-
 * $Id: counter.h,v 1.4 1998/04/29 04:54:46 nikos Exp $
 * 
 * RCS Log:
 * $Log: counter.h,v $
 * Revision 1.4  1998/04/29 04:54:46  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:36  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:21  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:36:54  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:54  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: counter.h,v 1.4 1998/04/29 04:54:46 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/counter.h,v 1.4 1998/04/29 04:54:46 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:23  bugnion
 * stanford update (x86)
 *
 * Revision 1.1  1997/10/20  00:07:41  sean
 * Initial revision
 *
 */

#ifndef _COUNTER_H
#define _COUNTER_H

#include <stdio.h>

class InstructionCounter {
 public:
  InstructionCounter(void);
  ~InstructionCounter(void);

  void Reset(void);
  int AddInstruction(const char *instr_name, int to_add = 1);

  void PrintReport(FILE *output) const;

 private:
  int total_instrs;   /* number of different instructions we've seen */
  int total_counted;  /* number of instructions counted so far */
  char **instr_names;
  int *instr_counts;
  int array_lengths;  /* current size of the arrays */
};

#endif
