/* $Id: cpu.cc,v 1.4 1998/04/29 04:54:46 nikos Exp $
 *
 * RCS Log:
 * $Log: cpu.cc,v $
 * Revision 1.4  1998/04/29 04:54:46  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:36  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:21  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:55  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:55  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: cpu.cc,v 1.4 1998/04/29 04:54:46 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/cpu.cc,v 1.4 1998/04/29 04:54:46 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:23  bugnion
// stanford update (x86)
//
// Revision 1.41  1997/12/12  11:28:56  yoel
// far CALL was calling LoadDescriptor() with Exceptions suppressed in a few
// places it shouldn't have in a few places and calling LoadDescriptor() without
// them suppressed and then generating the Exceptions by hand in a few places.
// Both fixed.
//
// Revised a some how LoadDescriptor() generates exceptions, so that if you're
// trying to load a TSS of some sort then #TS is generated instead of #GP.  I
// revised the #ifdef'ed out RPL, CPL, DPL checks some too, but still couldn't
// make them happy and left them #ifdef'ed out.
//
 * Revision 1.40  1997/12/11 22:14:28  mendel
 * 1) fixed a problem with the setting of the busy bit in the task
 * structure. We were not clearing the busy during a task switch always.
 *
 * 2) Fixed bug in reading and writing CR0-CR4 registers using the
 * debug interface.
 *
 * 3) Fixed bug in accessing memory thru the debug interface that was
 * causing the processor to take exceptions.
 *
 * Revision 1.39  1997/12/09 14:51:31  sjt
 * FPU emulation should now work completely - FPU exceptions are triggered when
 * 	they should be, and memory operands appear to work fine.  I removed the
 * 	printfs that were cluttering everything too.
 *
 * Mendel's fixes to BT/BTC/BTR/BTS and ADC have been stolen and checked in as
 * 	well.
 *
 * Revision 1.38  1997/12/09 05:33:09  mendel
 * Bugs patched around (These need to be fixed):
 *    1) ifdef'ed out the protection checking in LoadDescriptor() that
 *       was preventing IRET from working. See comment in code I added in
 *       cpu.cc
 *    2) ifdefed out the ASSERT(0) in the TLB flush routine in x86_instr.cc
 *    3) ifdef'ed out the protection checking in IRET that wasn't working.
 *
 *
 * Bugs:
 *
 * 1) Check in some bug fixes in fpu.cc that Sean gave me.
 *
 * 2) Fixed bug in reading stack segment selector in TRAP_GATE . We were
 * ASSERT(0)ing when the ReadFromVirtualAddress() was successful! in cpu.cc
 * HandleExecption()
 *
 * 3) Changed (CSDescriptor.dpl() != SSSelector & 0x3) to
 *            (CSDescriptor.dpl() != (SSSelector & 0x3))
 * for obvious reasons in cpu.cc / HandleException().
 *
 * 4) Test for GDT in TRAP_GATE code was backward in cpu.cc /
 * HandleException()
 *
 * 5) Since the prev_eip field of CPU class is used to set the return PC
 * of execeptions. We need to set prev_eip on instruction page faults to
 * be the faulting instruction. Otherwise we think the exception happen
 * on the PC of IRET instruction. ouch!
 *
 * 6) The MemoryOperand() code in instruction.cc wasn't setting the
 * default segment of the ESP register to be the stack segment! Found
 * this because Linux runs for a while with the user's DS selector
 * loaded and pushes things on the stack.
 *
 * 7) The is_task() method of seg_desc was completely bogus.
 *
 *
 * Clean up:
 *
 * 1) Removed fprintf() on exceptions in cpu.cc since we are taking far too many
 * execptions for this to be useful.
 *
 * Revision 1.37  1997/12/08 02:08:53  yoel
 * Fixed some reversed tests for whether or not a segment was conforming and
 * nudged the code to push an error code so that it would work for 16 bit code as
 * well.
 *
 * Revision 1.36  1997/12/07 21:04:42  mendel
 * Added annotations on IRET instructions. Useful for debugging.
 *
 * Fixed a couple of bugs in exception processing code:
 * 1) Was incorrectly pushing error code on stack on exception
 * 2) Was incorrectly computing space needed for exception frame.
 *
 * Revision 1.35  1997/12/07 06:57:06  mendel
 * Changes fprintf is cpu.cc to never occur on interrupts.
 *
 * Changed cpu class to export a translation routine so the rest
 * of SimOS can translate linear to physical address. This is
 * currently needed for the hacked up SCSI disk ( Scott's
 * Linux kernel passes virtual addreses to the monitor call)
 *
 * Revision 1.34  1997/12/07 02:53:52  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.33  1997/12/03 00:44:46  yoel
 * More exceptions.  Fixed the check in LoadExceptionDescriptor and removed the
 * #if 0.  Need a check that the IDT entry has the proper type, but that turned
 * out to be trickier than I had thought.  Also, most of the exceptions now
 * generated by HandleException() now add 1 to their error code if the exception
 * was externally generated.  The error codes are still wrong if generated by
 * LoadDescriptor() as called by HandleException(), but I want to talk to Sean
 * before mucking with it.
 *
 * Revision 1.32  1997/12/02 22:52:12  mendel
 * 1) Hooked up interrupts. Moved interrupt code from simos_interface.cc to
 *    cpu.cc .
 *
 * Several bug fixes:
 *
 * 1) Added check to push the error code only on processor generated
 *    exceptions and not on INT or external ones.
 *
 * 2) IDT gate descriptors were broken. I added a gate_use_32bit() since
 *    IDT gate descriptors use a different size field that normal segment
 *    descriptors.
 *
 * 3) Fixed bug in computing spaceNeeded for exception stack puke. Beware
 *    of C operator precedence.
 *
 * 4) Cleaned up some ugly and some incorrect code in HandleException()
 *
 * 5) #if 0 out some bogus check LoadExceptionDescriptor(). I couldn't
 *    tell want it was trying to do but is was wrong.
 *
 * 6) Compute the gate_offset correctly. We need all the bits of the
 *    flags for this to work.
 *
 * 7) Changed POP and MOV to not call WriteBackDescriptor() on selector
 *    zero. This caused WriteBackDescriptor() to die.
 *
// Revision 1.31  1997/12/02  05:26:06  yoel
// Annotations on Exceptions should now work (pray).  You can put an annotation
// on vec (any type of exception), exc (processor generated exceptions), int
// (software generated exceptions), or irq (external exceptions).
// Also added the DPL >= CPL check on Handling software Interupts.
//
 * Revision 1.30  1997/12/02 04:49:10  yoel
 *   Support added for taking Software Interupts and preparation for supporting
 * external interupts.  cpu.{cc,h} updated so that current_exception is an
 * int8 instead of the enum to allow exceptions 19-255 and enum ExceptionSource
 * added.  Cosmetic changes to fpu.cc and INT n / INTO now actually do
 * something (since we now support Software Interupts).
 *
 * Revision 1.29  1997/11/30 20:16:16  mendel
 * Hooked up I/O device interface to the IN and OUT instructions for x86sim.
 *
// Revision 1.28  1997/11/27  00:09:07  yoel
// x86_instr.cc:  Modified not to push return information on far JMPs (only
// CALLs).
//
// cpu.cc: Cosmetic changes plus some fixing of ExecuteOneInstruction()
// generating INVALID_OPCODE exceptions if decoding caused and Exception.
//
 * Revision 1.27  1997/11/26 18:22:02  sjt
 * Added 4MB page support to TranslateLinearToPhysical.
 * Fixed logic error in code descriptor permission checks.
 *
 * Revision 1.26  1997/11/25 21:34:31  mendel
 * cpu.cc:  Don't keep taking execptions after the CPU has halted.
 *
 * rest: Added support for gdb accessing. Breakpoints and single stepping
 * and memory access now work.
 *
// Revision 1.25  1997/11/21  07:02:20  whimsy
// Ooops ;)  Change the if(ExceptionOPending()) to a while(...) in ExecuteOne
// b/c of nested Exceptions.
//
 * Revision 1.24  1997-11-20 22:59:37-08  whimsy
 * Exception Handling code is now all written.  It even compiles.  God only
 * knows if it works.  I doubt it.  But the complete first attempt is there.
 *
 * Revision 1.23  1997-10-30 16:59:22-08  whimsy
 * FInished up the Task Switching stuff.  Added SaveTSS(), LoadTSS() as
 * helpers.
 *
 * Revision 1.22  1997-10-29 14:46:03-08  sean
 * fixed typo in SwitchTasks() (had CPU_EFL instead of CPU_EFLAGS)
 *
 * Revision 1.21  1997/10/28 08:38:27  whimsy
 * Added the first half of SwitchTaksssks.
 *
 * Revision 1.20  1997-10-23 20:51:18-07  whimsy
 * Nudged LoadDescriptor() to generate #TS exceptions for Task type
 * descriptors instead of only #GPs.
 *
 * Revision 1.19  1997-10-17 17:38:28-07  sean
 * added support for instruction counter and simulation of FPU ops on the
 *   actual chip
 *
 * Revision 1.18  1997/09/27 18:41:02  sean
 * Added code for comparing simulator stack to lockstep stack, and glue code
 *   for which flags to ignore in lockstep comparison
 *
 * Revision 1.17  1997/09/24 08:10:41  sean
 * Tweaked call to VerifyRegisters to know what instruction caused the
 * mismatch.
 *
 * Revision 1.16  1997/09/24 02:40:55  sjt
 * support for single-stepping of REP'd instructions added
 *
 * Revision 1.15  1997/09/06 00:39:51  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.14  1997/03/18 19:15:49  sean
 * fixed stupid bug in ReadFromLinearAddress() and WriteToLinearAddress() to
 *   correct handling of read/writes that cross page boundaries
 *
 * Revision 1.13  1997/03/14 21:39:39  sean
 * modified DumpState() to print contents of the Time Stamp Counter - removed
 *   call to DumpState() in ExecuteOneInstruction() (it doesn't belong there
 *   at all)
 *
 * Revision 1.12  1997/03/07 05:44:14  sean
 * modified DumpState() to show which flags are on
 *
 * Revision 1.11  1997/03/06 02:52:35  sean
 * Added accessors for register file, descriptor list, and create the
 *   DumpState() method, which dumps registers, stack, and code stream to
 *   stderr
 *
 * Revision 1.10  1997/02/23 03:25:45  sean
 * added functionality for read/writes to I/O ports (just stubs so far), for
 *   reading/writing descriptors based on selectors, and a helper function
 *   for checking permissions on specified I/O ports.
 *
 * Revision 1.9  1997/02/11 05:07:19  sjt
 * changed ExecuteOneInstruction() to create/use MemoryStreams for
 * instruction data
 *
 * Revision 1.8  1997/02/10 07:29:26  sean
 * Modularized memory accessing code (again).  Created nice, well-behaved
 * TranslateVirtualToLinear() and TranslateLinearToPhysical() methods, paving the
 * way for the MemoryStream class.
 *
 * Revision 1.7  1997/02/04 22:37:25  sean
 * modified ExecuteOneInstruction to count and return clock cycles spent
 *
 * Revision 1.6  1997/02/04 10:38:36  sean
 * added functions to set just s,z,p flags, or just o,c flags
 *
 * Revision 1.5  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: cpu.cc,v 1.4 1998/04/29 04:54:46 nikos Exp $";

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "debug.h"
#include "cpu.h"
#include "instruction.h"
#include "data_stream.h"
#include "lockstep.h"
#include "counter.h"

extern "C" {
#include "annotations.h"
}

#ifdef SIMOS
extern "C" {
#include "simos_interface.h"
}
#endif

CPU::CPU(int32 _last_byte)
{
  _last_was_syscall = 0;
  last_memory_byte = _last_byte;
  memory = new int8[_last_byte + 1];
  mem_alloced = true;
  instr_counter = 0;
  fpu_state = 0;

  Reset();
}

CPU::CPU(void *_memory_base, int32 _last_byte)
{
  _last_was_syscall = 0;
  last_memory_byte = _last_byte;
  memory = (int8 *) _memory_base;
  mem_alloced = false;
  instr_counter = 0;
  fpu_state = 0;

  Reset();
}

CPU::~CPU(void)
{
  if(mem_alloced)
    delete[] memory;
}

void CPU::Reset(void)
{
  int i;

  // we start out not halted, and with no exceptions pending
  halted = false;
  exception_occurred = false;

  // by default, we'll run REP'd instructions all at once
  single_stepping = false;

  // most registers start out zeroed
  bzero(regfile,sizeof(regfile));

  // The following settings are from page 16-4
  CPU_EFLAGS = 0x00000002;
  CPU_EIP    = 0x0000FFF0;
  CPU_CR0    = 0x60000010;
  CPU_EDX    = 0x00000500;
  CPU_DR6    = 0xFFFF0FF0;
  CPU_DR7    = 0x00000400;

  CPU_CS     = 0xF000;

  seg_descs[SDESC_ES ].set_base_limit(0x00000000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_CS ].set_base_limit(0xFFFF0000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_SS ].set_base_limit(0x00000000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_DS ].set_base_limit(0x00000000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_FS ].set_base_limit(0x00000000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_GS ].set_base_limit(0x00000000,0xFFFF,PRESENT_RW_ACCESSED);
  seg_descs[SDESC_GDT].set_base_limit(0x00000000,0xFFFF,PRESENT_RW);
  seg_descs[SDESC_IDT].set_base_limit(0x00000000,0xFFFF,PRESENT_RW);
  seg_descs[SDESC_LDT].set_base_limit(0x00000000,0xFFFF,PRESENT_RW);

  // reset our instruction counter, if we've got one
  if(instr_counter) instr_counter->Reset();
}

//  Physical addresses are the actual places in the RAM chips to look
//   for data (i.e. this is after segmentation and paging)
//  Reads and writes to physical memory cannot result in errors that the
//   CPU can see.  Reads from non-existent addresses yield all 0's.  Writes
//   to non-existent or read-only addresses do nothing
void CPU::ReadFromPhysicalAddress(int32 physical, void *buffer, int len)
{
  ASSERT(len > 0);

  // first, clear out the buffer, in case parts of the read fail.
  bzero(buffer,len);

  if(physical <= last_memory_byte) {
    if((last_memory_byte - physical) < ((int32)len - 1)) {
      DEBUG(DEBUG_MEMORY,"Read from physical memory (@ 0x%08x, %d bytes) truncated to %d bytes.",physical,len,(last_memory_byte - (physical - 1)));
      len = last_memory_byte - (physical - 1);
    }

    memcpy(buffer,memory+physical,len);
  } else
    DEBUG(DEBUG_MEMORY,"Read from physical memory (@ 0x%08x, %d bytes) out of range.",physical,len);
}

void CPU::WriteToPhysicalAddress(int32 physical, const void *buffer, int len)
{
  ASSERT(len > 0);

  if(physical <= last_memory_byte) {
    if((last_memory_byte - physical) < ((int32)len - 1)) {
      DEBUG(DEBUG_MEMORY,"Write to physical memory (@ 0x%08x, %d bytes) truncated to %d bytes.",physical,len,(last_memory_byte - (physical - 1)));
      len = last_memory_byte - (physical - 1);
    }

    memcpy(memory+physical,buffer,len);
  } else
    DEBUG(DEBUG_MEMORY,"Write to physical memory (@ 0x%08x, %d bytes) out of range.",physical,len);
}

//  I/O Ports are pretty much like memory, except they live in a different
//   address space, and there's no paging, or segments, or anything weird
//   (unless you use memory-mapped I/O).
//  I/O is also not directly dependent on the Pentium architecture, so these
//   functions are essentially wrappers around simulators for related
//   hardware.
void CPU::ReadFromIOPort(int32 port, void *buffer, int len)
{
  DEBUG(DEBUG_MEMORY,"Read from I/O port (@ 0x%08x, %d bytes)",port,len);
#ifdef SIMOS
  X86DoPortIO(0, TRUE, port, len, (void *)buffer);
#endif
}

void CPU::WriteToIOPort(int32 port, const void *buffer, int len)
{
  DEBUG(DEBUG_MEMORY,"Write to I/O port (@ 0x%08x, %d bytes)",port,len);
#ifdef SIMOS
  X86DoPortIO(0, FALSE, port, len, (void *)buffer);
#endif
}

bool CPU::CheckPagePermissions(int32 entry, bool user, bool writing, 
			       int32 linear_addr, bool quiet)
{
  // First rule: A not-present page ALWAYS causes a page fault
  if(BITS(entry,0,1) == 0) {
    if(!quiet) {
      ExceptionOccurred(EXC_PAGE_FAULT,  // error code (below) defined on 14-23
			(user ? 4 : 0) + (writing ? 2 : 0) + 0);
      CPU_CR2 = linear_addr;
    }
    return(false);
  }

  // Next rule: A user cannot see a supervisor-only page (GPF)
  if(user && (BITS(entry,2,1) == 0)) {
    if(!quiet) {
      ExceptionOccurred(EXC_PAGE_FAULT,
			(user ? 4 : 0) + (writing ? 2 : 0) + 1);
      CPU_CR2 = linear_addr;
    }
    return(false);
  }

  // Last rule: A user (or supervisor when WP is set) cannot write to
  //  a read-only page (GPF)
  if(writing && (BITS(entry,1,1) == 0) &&
     (user || (BITS(CPU_CR0,16,1) != 0))) {
    if(!quiet) {
      ExceptionOccurred(EXC_PAGE_FAULT,
			(user ? 4 : 0) + (writing ? 2 : 0) + 1);
      CPU_CR2 = linear_addr;
    }
    return(false);
  }

  return(true);
}

// This function attempts to translate a 'linear' (pre-paging)
//  address to a 'physical' (post-paging) address.  'user' should be true
//  if the access is coming from a CPL=3 source, 'writing' should be true
//  if the access is a write (as opposed to a read), and 'quiet' should be
//  true if exceptions should be masked.
//
// The return value is 0 if the address could not be translated, and the
//  number of bytes (counting the given address) for which the translation is
//  valid.
int32 CPU::TranslateLinearToPhysical(int32 linear, int32& physical, bool user,
				     bool writing, bool quiet)
{
  int32 table_base, page_base;

  // Check to make sure paging is enabled before doing anything
  if(BITS(CPU_CR0,31,1) != 0) {
    // First step is to look up the page table in the page directory
    ReadFromPhysicalAddress((BITS(CPU_PDBR,12,20) << 12) |
			    (BITS(linear,22,10) << 2),&table_base,4);

    // Put our table base in a byte order we understand
    table_base = xtoh32(table_base);

    // Check that we can look at the page table
    if(!CheckPagePermissions(table_base,user,writing,linear,quiet))
      return(0);

    // Check to see if we're using 4MB pages (basically a 1-level page
    //  structure instead of the standard 2-level)
    if((BITS(CPU_CR4,4,1) != 0) && (BITS(table_base,7,1) != 0)) {
      // We have a 4MB page, and the PSE (page size extension) bit is on, so
      //  honor it...

      // All we have to do is form the physical address, and we're done
      physical = (BITS(table_base,22,10) << 22) | BITS(linear,0,22);
      return(4*1024*1024 - BITS(physical,0,22));
    } else {
      // Either it's a 4KB page, or it wants to be a 4MB page, but PSE is off
      //  so we consider it a 4KB page anyway - means we have to follow next
      //  level of page directory

      // Now, same idea, use page table to get actual address
      ReadFromPhysicalAddress((BITS(table_base,12,20) << 12) |
  			      (BITS(linear,12,10) << 2),&page_base,4);

      // Put our page base in a byte order we understand
      page_base = xtoh32(page_base);

      // Check that we can look at the actual page
      if(!CheckPagePermissions(page_base,user,writing,linear,quiet))
        return(0);

      // If everything goes ok, we can form the physical address
      physical = (BITS(page_base,12,20) << 12) | BITS(linear,0,12);
      return(PAGE_SIZE - BITS(physical,0,PAGE_OFFSET_BITS));
    }
  } else
    physical = linear;

  return(PAGE_SIZE - BITS(physical,0,PAGE_OFFSET_BITS));
}

//  Linear addresses are what you have after you deal with segmentation,
//   but before paging (if paging is off, linear = physical)
int CPU::ReadFromLinearAddress(int32 linear, void *buffer, int len,
			       bool user, int min_len)
{
  int to_read;
  int count = 0;
  int32 physical;

  if(min_len == 0) min_len = len;

  while(count < len) {
    to_read = MIN(len - count, PAGE_SIZE - BITS(linear,0,PAGE_OFFSET_BITS));
    
    if(!TranslateLinearToPhysical(linear+count,physical,user,false,
				  (count >= min_len))) 
      return(count);

    ReadFromPhysicalAddress(physical,buffer+count,to_read);
    count += to_read;
  }

  return(count);
}

int CPU::WriteToLinearAddress(int32 linear, const void *buffer, int len,
			      bool user, int min_len)
{
  int to_read;
  int count = 0;
  int32 physical;

  if(min_len == 0) min_len = len;

  while(count < len) {
    to_read = MIN(len - count, PAGE_SIZE - BITS(linear,0,PAGE_OFFSET_BITS));
    
    if(!TranslateLinearToPhysical(linear+count,physical,user,true,
				  (count >= min_len))) 
      return(count);

    WriteToPhysicalAddress(physical,buffer+count,to_read);
    count += to_read;
  }

  return(count);
}

// This function attempts to translate a virtual address (described by a
//  'segment' and an 'offset') into a 'linear' (pre-paging) address.
//  'writing' should be true if the access will be a write (as opposed to a
//  read or fetch), 'fetching' should be true if the access is an instruction
//  fetch (as opposed to a normal read or write).  'is_stack' should be true
//  if 'segment' refers to the stack segment, and 'quiet' should be true if
//  exceptions should be masked.
//
// The return value is 0 if the address could not be translated, and the
//  number of bytes (past the given address) for which the translation is
//  valid.
int32 CPU::TranslateVirtualToLinear(SegmentDescriptor& segment, int32 offset,
				    int32& linear, bool writing,
				    bool fetching, bool is_stack, bool quiet)
{
  // First check, the segment must be valid
  if(!segment.is_valid()) {
    if(!quiet)
      ExceptionOccurred((is_stack ? EXC_STACK_FAULT : EXC_PROTECTION), 0);
    return(0);
  }

  if(writing) {
    // A write must be to a writeable (data) segment
    // change: or we'll allow writes to 32-bit tss's for now too.
    if(!segment.is_writeable() &&
       (segment.type() != DESC_32BIT_TSS_AVAIL) &&
       (segment.type() != DESC_32BIT_TSS_BUSY)) {
      if(!quiet)
	ExceptionOccurred(EXC_PROTECTION, 0);
      return(0);
    }
  } else
  if(fetching) {
    // An instruction fetch must be from a code segment, or we must be in real
    //  mode
    if(!segment.is_code() && in_protected_mode()) {
      if(!quiet)
	ExceptionOccurred(EXC_PROTECTION, 0);
      return(0);
    }
  } else {
    // A read must be from a data or readable code segment
    if(!segment.is_readable()) {
      if(!quiet)
	ExceptionOccurred(EXC_PROTECTION, 0);
      return(0);
    }
  }

  // Finally, check to make sure the offset into the segment is valid
  if(!segment.offset_valid(offset)) {
    if(!quiet)
      ExceptionOccurred((is_stack ? EXC_STACK_FAULT : EXC_PROTECTION), 0);
    return(0);
  }

  linear = segment.base() + offset;
  return(segment.bytes_available(offset));
}

//  Virtual addresses are pre-segmentation (i.e. what actual programs are
//   using) - they're annoying because you need both the segment
//   descriptor and the offset
int CPU::ReadFromVirtualAddress(SegmentDescriptor& segment, int32 offset,
				void *buffer, int len, int cpl, int min_len,
				bool is_instr_fetch)
{
  int32 avail_len, linear;

  ASSERT(len > 0);
  if(min_len == 0) min_len = len;

  avail_len = TranslateVirtualToLinear(segment,offset,linear,
				       false, is_instr_fetch,
				       (&segment == (seg_descs+SDESC_SS)),
				       false);

  // if the translation failed, die
  if(avail_len == 0) return(0);
  
  // Adjust the number of bytes, if necessary
  len = MIN(len,avail_len);

  // If the above adjustment drops us below our minimum read length, complain
  if(len < min_len) {
    ExceptionOccurred(((&segment == (seg_descs+SDESC_SS)) ? EXC_STACK_FAULT :
		                                            EXC_PROTECTION),0);
    return(0);
  }

  // Do (or try) the read
  len = ReadFromLinearAddress(linear,buffer,len,(cpl == 3),min_len);

  // If we read anything, set the accessed bit
  if(len > 0) segment.set_accessed(true);

  return(len);
}

int CPU::WriteToVirtualAddress(SegmentDescriptor& segment, int32 offset, 
			       const void *buffer, int len, int cpl, 
			       int min_len)
{
  int32 avail_len, linear;

  ASSERT(len > 0);
  if(min_len == 0) min_len = len;

  avail_len = TranslateVirtualToLinear(segment,offset,linear,
				       true, false,
				       (&segment == (seg_descs+SDESC_SS)),
				       false);

  // if the translation failed, die
  if(avail_len == 0) return(0);
  
  // Adjust the number of bytes, if necessary
  len = MIN(len,avail_len);

  // If the above adjustment drops us below our minimum read length, complain
  if(len < min_len) {
    ExceptionOccurred(((&segment == (seg_descs+SDESC_SS)) ? EXC_STACK_FAULT :
		                                            EXC_PROTECTION),0);
    return(0);
  }

  // Do (or try) the write
  len = WriteToLinearAddress(segment.base() + offset,buffer,len,(cpl == 3),
			     min_len);

  // If we wrote anything, set the accessed bit
  if(len > 0) segment.set_accessed(true);

  return(len);
}

int CPU::GetCPL(void) const
{
  return(in_protected_mode() ?
	   (in_v86_mode() ?
             3 : 
	     BITS(CPU_CS,0,2)) : 
	   0);
}

// This function handles the evilness of loading a descriptor specified by
//  'selector' into the 'target' descriptor.  Only valid descriptors with
//  a type permitted by the 'type_mask' (which should be one of the DMASK_
//  macros defined in seg_desc.h) will be loaded, and if 'need_present' is
//  set, the descriptor must be marked present as well.
// If the descriptor is ok in all respects, it is copied into 'target', and
//  the function returns true.
// If any error occurs, the function returns false, and if 'quiet' is false,
//  an exception is raised.
bool CPU::LoadDescriptor(SegmentDescriptor& target, int16 selector,
			 int32 type_mask, bool need_present, 
			 bool allow_null, bool quiet)
{
  int8 desc_data[8];
  SegmentDescriptor temp_desc;
  Exception exc_caused;

  // real and vm86 modes are VERY simple - protected is ugly as hell
  if(in_protected_mode() && !in_v86_mode()) {

    // Okay, if the only types we're try to load are a subset of the TSS
    // types, we generate #TS() on problems, else #GP()
    if (type_mask & ~(DMASK_AVAIL_TSS | DMASK_BUSY_TSS))
      exc_caused = EXC_PROTECTION;
    else
      exc_caused = EXC_INVALID_TSS;
    
    // special case: if selector is in [0,3], AND null descriptors are allowed
    //  we mark the descriptor as invalid, and quit (if nulls aren't allowed,
    //  we gpf)
    if((selector >> 2) == 0) {
      if(allow_null) {
	target.mark_invalid();
	return(true);
      } else {
	if(!quiet) ExceptionOccurred(EXC_PROTECTION,0);
	return(false);
      }
    }

    // based on the TI bit in the selector, use the GDT or LDT
    if(BITS(selector,2,1) == 0) {
      // GDT
      // make sure GDT is valid (if it isn't, we've a BIG problem)
      if(!seg_descs[SDESC_GDT].is_valid()) {
	DEBUG(DEBUG_MEMORY,"Global Descriptor Table Register is invalid!");
	ASSERT(0);
      }

      // read our 8 bytes out of the table (ReadFromVirtualAddress() will take
      //  care of limit checks and such) - Note that we assume CPL=0 so that
      //  we don't have any chance of a permission problem on the read
      if(ReadFromVirtualAddress(seg_descs[SDESC_GDT],(selector & 0x0FFF8),
                                desc_data,8,0) < 8)
	return(false);
    } else {
      // LDT
      // make sure LDT is valid (if it isn't, we GPF)
      if(!seg_descs[SDESC_GDT].is_valid()) {
	if(!quiet) ExceptionOccurred(EXC_PROTECTION,selector);
	return(false);
      }

      // read our 8 bytes out of the table (ReadFromVirtualAddress() will take
      //  care of limit checks and such) - Note that we assume CPL=0 so that
      //  we don't have any chance of a permission problem on the read
      if(ReadFromVirtualAddress(seg_descs[SDESC_LDT],(selector & 0x0FFF8),
                                desc_data,8,0) < 8) {
	//  Okay, this is probably sick and wrong, but if we're here then
	// we've violated descriptor table limits and generated an
	// exception.  However, it won't be the right type necessarily, so
	// violate all sorts of things and change the current exception.
	// Note that #GP and #TS are the same severity level so we don't
	// have to worry about missing a double fault here.
	if (current_exception != EXC_DOUBLE_FAULT)
	  current_exception = exc_caused;
	return(false);
      }
    }

    // now that we've got the descriptor, we can check for validity, presense,
    //   type, and permissions
    temp_desc.extract_descriptor(desc_data);

    if(need_present && !temp_desc.is_present()) {
      if(!quiet) ExceptionOccurred(EXC_NO_SEGMENT,selector);
      return(false);
    }

    // check type against bitmap
    if(BITS(type_mask,temp_desc.type(),1) == 0) {
      if(!quiet) ExceptionOccurred(exc_caused,selector);
      return(false);
    }

#ifdef notdef
    //  Okay, the exact nature of the checks below is kinda uncertain and
    // what is current there doesn't work for some reason with cross
    // privilege IRET.  There are two basic concerns-- when LoadDescriptor
    // loads Segemnt Registers in which case CS, SS, and all the rest each
    // have slightly different rules and the case when we're loading a
    // Descriptor that isn't to a code or data segment (a gate or TSS).  I
    // haven't been able to quite come up with something that makes them all
    // happy, so I'm leaving this #ifdef'ed out for now. --Jeremy.
        
    // basic rule for permission levels: "Instructions may load a segment
    //  register only if the DPL of the segment is the same or a less
    //  priviledged level (greater priviledge number) than the less priviledged
    //  of the CPL and selector's RPL." (12-9) - I read this as requiring 
    //  (DPL >= CPL) && (DPL >= RPL)
    // As a completely random exception, for loading code segments, we need
    //  (DPL == CPL) or ((DPL < CPL) and segment is conforming)
    // Actually, I believe we need to check RPL on non-conforming code
    //  segments as well.  I've sent Sean mail about this.
    if(temp_desc.is_code()) {		//  Note is_code() implies !is_task()
      if((temp_desc.dpl() > GetCPL()) ||
	 ((temp_desc.dpl() < GetCPL()) && !temp_desc.is_conforming())) {
	if(!quiet)
	  ExceptionOccurred(EXC_PROTECTION, selector);
	return(false);
      }
    } else {
      if((temp_desc.dpl() < GetCPL()) || 
	 (temp_desc.dpl() < BITS(selector,0,2))) {
	if(!quiet) 
	  ExceptionOccurred(temp_desc.is_task() ? 
	                            EXC_INVALID_TSS : EXC_PROTECTION,selector);
	return(false);
      }
    }
#endif
    
    // At this point, we're happy
    // Load the real descriptor from the same data as the temp descriptor...
    target.extract_descriptor(desc_data);

    // ... and return
    return(true);
  } else {
    // just load the selector in as the base << 4
    target.set_base_limit(((int32) selector) << 4, 0x0FFFF, PRESENT_RW);
    return(true);
  }
}

// This function writes the 'source' segment descriptor back into the place
//  specified by 'selector' - return value is whether or not the write was
//  either unnecessary (in real or vm86 modes) or successful
bool CPU::WriteBackDescriptor(const SegmentDescriptor& source, int16 selector)
{
  int8 desc_data[8];

  // first, only protected mode gets to write back descriptors
  if(in_protected_mode() && !in_v86_mode()) {
    // special case: if selector is in [0,3], the descriptor is invalid, and
    //  we should not have been called!
    // CHANGED: we allow "writing back" of a NULL descriptor, but it's a NOP
    if((selector >> 2) == 0) {
      return(true);
    }

    // put the data into the descriptor memory image
    source.compress_descriptor(desc_data);

    // now, pick which table to write to, and do it
    // based on the TI bit in the selector, use the GDT or LDT
    if(BITS(selector,2,1) == 0) {
      // GDT
      // make sure GDT is valid (if it isn't, we've a BIG problem)
      if(!seg_descs[SDESC_GDT].is_valid()) {
	DEBUG(DEBUG_MEMORY,"Global Descriptor Table Register is invalid!");
	ASSERT(0);
      }

      // write the data (WriteToVirtualAddress() will take
      //  care of limit checks and such) - Note that we assume CPL=0 so that
      //  we don't have any chance of a permission problem on the write
      if(WriteToVirtualAddress(seg_descs[SDESC_GDT],(selector & 0x0FFF8),
                               desc_data,8,0) < 8)
	return(false);
    } else {
      // LDT
      // make sure LDT is valid (if it isn't, we GPF)
      if(!seg_descs[SDESC_GDT].is_valid()) {
	ExceptionOccurred(EXC_PROTECTION,selector);
	return(false);
      }

      // write the data (WriteToVirtualAddress() will take
      //  care of limit checks and such) - Note that we assume CPL=0 so that
      //  we don't have any chance of a permission problem on the write
      if(WriteToVirtualAddress(seg_descs[SDESC_LDT],(selector & 0x0FFF8),
                               desc_data,8,0) < 8)
	return(false);
    }
  }

  return(true);
}

// This function checks to see if the current process is allowed to access
//  the I/O ports starting at 'port_base' with a size of 'size'.  It returns
//  true if access is allowed, and false if access is not allowed.  The only
//  exception that can occur inside of this code is a page fault.
bool CPU::CheckIOPermissions(int32 port_base, int size)
{
  int32 offset;
  int16 io_bits;

  // if we're in real mode, or protected mode with cpl <= iopl, all ports are
  //  fair game...
  if(!in_protected_mode() || (!in_v86_mode() && (GetCPL() <= GetIOPL())))
    return(true);

  // first, only 32-bit TSS's have I/O permission bitmaps.
  if(seg_descs[SDESC_TR].type() != DESC_32BIT_TSS_BUSY) return(false);

  // figure the offset of the first bit we need (guaranteed to get all the
  //  rest, since 'size' is less than 9, and we're reading 2 bytes)
  offset = TSS_IO_MAP_BASE + (port_base >> 3);

  // now, make sure we have room in the TSS segment to load the bitmap
  if(seg_descs[SDESC_TR].bytes_available(offset) < 2) return(false);

  // this should only fail on a page fault...
  if(!ReadFromVirtualAddress(seg_descs[SDESC_TR],offset,&io_bits,2,GetCPL()))
    return(false);

  // now just check the proper bits
  return(BITS(io_bits,(port_base & 7),size) == 0);
}

void CPU::SetArithFlags(bool of, bool sf, bool zf, bool af, bool pf, bool cf)
{
  int32 flags;

  flags = CPU_EFLAGS;

  SET_BITS(flags,FLAG_OF,1,(of ? 1 : 0));
  SET_BITS(flags,FLAG_SF,1,(sf ? 1 : 0));
  SET_BITS(flags,FLAG_ZF,1,(zf ? 1 : 0));
  SET_BITS(flags,FLAG_AF,1,(af ? 1 : 0));
  SET_BITS(flags,FLAG_PF,1,(pf ? 1 : 0));
  SET_BITS(flags,FLAG_CF,1,(cf ? 1 : 0));

  CPU_EFLAGS = flags;
}

void CPU::Set_SZP_Flags(bool sf, bool zf, bool pf)
{
  int32 flags;

  flags = CPU_EFLAGS;

  SET_BITS(flags,FLAG_SF,1,(sf ? 1 : 0));
  SET_BITS(flags,FLAG_ZF,1,(zf ? 1 : 0));
  SET_BITS(flags,FLAG_PF,1,(pf ? 1 : 0));

  CPU_EFLAGS = flags;
}

void CPU::Set_OC_Flags(bool of, bool cf)
{
  int32 flags;

  flags = CPU_EFLAGS;

  SET_BITS(flags,FLAG_OF,1,(of ? 1 : 0));
  SET_BITS(flags,FLAG_CF,1,(cf ? 1 : 0));

  CPU_EFLAGS = flags;
}

void CPU::ExceptionOccurred(int8 exception_type, int32 error_code,
			    ExceptionSource exception_origin)
{
  int32 curValue;

  // Special case: an NMI knocks the CPU out of HLT mode
  if(exception_type == EXC_NMI) halted = false;

  // Print a message, in case anyone is watching
  DEBUG(DEBUG_EXCEPTION,"Exception #%d(%d)\n",exception_type,error_code);

  // There are three cases to deal with:
  // Case 1: There are no outstanding exceptions, so just record this one
  if(!exception_occurred) {
    exception_occurred = true;
    current_exception = exception_type;
    current_error_code = error_code;
    current_origin = exception_origin;
    return;
  }

  // Case 2: There is an exception other than a double fault, outstanding
  //  which could lead either to a double fault or else the second fault
  //  needs to be handled and the instruction restarted so the first fault
  //  is dealth with.
  if(current_exception != EXC_DOUBLE_FAULT) {
    if(((exception_class[current_exception] == EXC_CLASS_PAGE_FAULT) &&
	(exception_class[exception_type]    != EXC_CLASS_BENIGN)) ||
       ((exception_class[current_exception] != EXC_CLASS_BENIGN) &&
	(exception_class[exception_type]    == EXC_CLASS_CONTRIBUTORY))) {
      fprintf(stderr, "Second exception (%d)... Double Faulting!\n", exception_type);

      current_exception = EXC_DOUBLE_FAULT;
      current_error_code = 0;
      current_origin = EXC_PROCESSOR;
    } else {
      fprintf(stderr, "Second exception (%d)... Nesting!\n", exception_type);
      current_exception = exception_type;
      current_error_code = error_code;
      current_origin = exception_origin;
      CPU_EIP = prev_eip;              // Nested exception is always a restart.
    }
    return;
  }

  // Case 3: There is a double fault outstanding, and the specified (14-18)
  //  behavior is to freak out and enter shutdown (HLT) mode.
  fprintf(stderr,"Exception after double fault!  Shutting down...\n");
  halted = true;
}

void CPU::ClearException(void)
{
  exception_occurred = false;
}

void CPU::HandleException(void)
{
  SegmentDescriptor Descriptor, CSDescriptor, SSDescriptor;
  char vector[4];
  int8 spaceNeeded, desc_data[8];
  int16 CSSelector, SSSelector;
  int32 Offset, newESP, tmpEFL;
  bool EXT = (current_origin == EXC_EXTERNAL);

  DEBUG(DEBUG_EXCEPTION,"In HandleException (exception_occurred = %d)",exception_occurred);

  // We should never be called unless an exception actually occurred
  ASSERT(exception_occurred);

  DEBUG(DEBUG_EXCEPTION,"Exception handler called for #%d(%d)",
	current_exception,current_error_code);

  // Try and run annotations.  Cross fingers and pray.
  AnnExec(AnnFindInt("vec", current_exception));     // General Exception Annotations
  switch(current_origin) {
  case EXC_PROCESSOR:   // Internal Exception Annotations
    AnnExec(AnnFindInt("exc", current_exception));
    break;
  case EXC_INT:   // Software Exception Annotations
    AnnExec(AnnFindInt("int", current_exception));
    break;
  case EXC_EXTERNAL:   // IRQ / External Annotations
    AnnExec(AnnFindInt("irq", current_exception));
    break;
  }

  //  Fiddle with EIP.  On a restart we can safely set EIP here, none
  //  of the code below ever reads it and a nested exception is always
  //  a restart so it won't care.  Note also that a Double Fault sets
  //  EIP as if it were a restart (see 14-29), even though under no
  //  conditions can a Double Faulting Exception be restarted.  Odd.
  if ((current_origin == EXC_PROCESSOR) && 
      ((current_exception == EXC_DOUBLE_FAULT) ||
       (current_exception == EXC_DIV_BY_ZERO) ||
       (current_exception == EXC_BOUNDS) ||
       (current_exception == EXC_INVALID_OPCODE) ||
       (current_exception == EXC_NO_DEVICE) ||
       (current_exception == EXC_DOUBLE_FAULT) ||
       (current_exception == EXC_INVALID_TSS) ||
       (current_exception == EXC_NO_SEGMENT) ||
       (current_exception == EXC_STACK_FAULT) ||
       (current_exception == EXC_PROTECTION) ||
       (current_exception == EXC_PAGE_FAULT) ||
       (current_exception == EXC_FPU_ERROR) ||
       (current_exception == EXC_ALIGNMENT)))
    CPU_EIP = prev_eip;

  if (in_real_mode()) {
    WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_SP, 
			  &CPU_FLAGS, SIZE_16BIT, GetCPL());
    CPU_FLAGS &= 0x40300;
    WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_SP, 
			  &CPU_CS, SIZE_16BIT, GetCPL());
    WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_SP, 
			  &CPU_IP, SIZE_16BIT, GetCPL());
    ReadFromVirtualAddress(seg_descs[SDESC_IDT], current_exception<<2,
			   &CPU_IP, SIZE_16BIT, GetCPL());
    ReadFromVirtualAddress(seg_descs[SDESC_IDT], current_exception<<2 + 2,
			   &CPU_CS, SIZE_16BIT, GetCPL());
    // Load the new CS Descriptor, this may not be necessary, but can't hurt.
    LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true, false, true);
    return;
  }
  
  if(!LoadExceptionDescriptor(Descriptor)) {
   //  Exception occurred loading the descriptor, exit and restart the
   // handler to cope with it.
   return;
  }

  switch(Descriptor.type()) {
  case DESC_TASK_GATE:
    // Same as FAR CALL stuff, block-copied from x86_instr.cc
    CSSelector = Descriptor.gate_selector();
    if (CSSelector & (1<<2)) {                    // Must be in GDT
      ExceptionOccurred(EXC_INVALID_TSS, CSSelector); return;
    }

    if (LoadDescriptor(Descriptor, CSSelector, DMASK_AVAIL_TSS,
                                             false, false, false) == false) {
      ExceptionOccurred(EXC_INVALID_TSS, CSSelector);
      return;
    }
    ASSERT(SwitchTasks(CSSelector, Descriptor));
    if (CPU_EIP > seg_descs[SDESC_CS].limit()) {
      ExceptionOccurred(EXC_INVALID_TSS, 0); return;
    }
    return;
    break;
   
  case DESC_16BIT_INT_GATE:
  case DESC_32BIT_INT_GATE:
  case DESC_16BIT_TRAP_GATE:
  case DESC_32BIT_TRAP_GATE:
     // Grab the new Descriptor from the gate, punt if it generates an
     // exception.  I think we can clobber the old Descriptor at this
     // point.
     CSSelector = Descriptor.gate_selector();
     Offset = Descriptor.gate_offset();
     if (!LoadDescriptor(CSDescriptor,CSSelector,DMASK_CODE,true,false,true))
       return;

     // Very much like a CALL to a Call Gate, but with some changes...
     if (!CSDescriptor.is_conforming() && (CSDescriptor.dpl() < GetCPL())) {
       // Sigh, cross privilege code is messy.
       if(ReadFromVirtualAddress(seg_descs[SDESC_TR], 
                                 8 + 8*CSDescriptor.dpl(), &SSSelector, 
                                 sizeof(SSSelector), GetCPL()) != sizeof(SSSelector))
	 ASSERT(0);                             // Can't have TSS errors here

       //  Okay, hand code LoadDescriptor() here since we need to do
       // slightly odd things in order to load the new SS Descriptor.
       if (((SSSelector & ~0x7) == 0) || 
	   (CSDescriptor.dpl() != (SSSelector & 0x3))) {
	 ExceptionOccurred(EXC_INVALID_TSS, EXT);
	 return;
       }
       
       if (BITS(SSSelector, 2, 1) == 0) { // GDT
	 if(!seg_descs[SDESC_GDT].is_valid()) {
	   DEBUG(DEBUG_MEMORY,"Global Descriptor Table Register is invalid!");
	   ASSERT(0);
	 }
	 if(ReadFromVirtualAddress(seg_descs[SDESC_GDT],(SSSelector & 0x0FFF8),
				   desc_data,8,0) < 0)
	   return;
       } else { // LDT
	 if (!seg_descs[SDESC_LDT].is_valid()) 
	   { ExceptionOccurred(EXC_PROTECTION, SSSelector); return; }
	 if(!ReadFromVirtualAddress(seg_descs[SDESC_LDT],
				    (SSSelector & 0x0FFF8), desc_data, 8, 0))
	   return;
       }
       SSDescriptor.extract_descriptor(desc_data);
       if (!BITS(DMASK_WRITEABLE, SSDescriptor.type(), 1) || 
	   (SSDescriptor.dpl() != CSDescriptor.dpl())) 
	 { ExceptionOccurred(EXC_INVALID_TSS, SSSelector + EXT); return; }
       if (!SSDescriptor.is_present()) 
	 { ExceptionOccurred(EXC_STACK_FAULT + EXT); return; }
       
      if(!ReadFromVirtualAddress(seg_descs[SDESC_TR], 4 + 8*CSDescriptor.dpl(),
				 &newESP, sizeof(newESP), CSDescriptor.dpl()))
	return;

      // Space to push everything we need to save?
      if(Descriptor.gate_use_32bit()) {
	if (!SSDescriptor.offset_valid(newESP - 20))
	  { ExceptionOccurred(EXC_STACK_FAULT, 0); return; }
      } else {
	if (!SSDescriptor.offset_valid(newESP - 10))
	  { ExceptionOccurred(EXC_STACK_FAULT, 0); return; }
      }

      // Is the new eIP valid for the new CS?
      if (!CSDescriptor.offset_valid(Offset)) {
	ExceptionOccurred(EXC_PROTECTION, 0);
	return;
      }

      if (in_v86_mode()) {
	goto HandleV86;                     //  I feel like a BASIC programer.
      }

      //  Point of No Return.  About to muck with the Stack and
      // registers, no more exceptions better happen.

      //  Okay, now the SSDescriptor is loaded, time to push things
      // on the new stack.  Note that the procedure in the book
      // (25-175) has one first load the new registers and then push
      // the old values.  That strikes me as kind of a neat trick...
      // I'm going to reverse that (push the current values then load
      // the new ones) rather than use a ton of temporary variables.
      if (Descriptor.gate_use_32bit()) {
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT, &CPU_SS,
			      SIZE_32BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT, &CPU_ESP,
			      SIZE_32BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT, 
			      &CPU_EFLAGS, SIZE_32BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT, &CPU_CS,
			      SIZE_32BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT, &CPU_EIP,
			      SIZE_32BIT, CSDescriptor.dpl());
	if ((current_origin == EXC_PROCESSOR) && exception_error[current_exception])
	  WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_32BIT,
				&current_error_code, SIZE_32BIT, 
				CSDescriptor.dpl());
	CPU_EIP = Offset;
	CPU_ESP = newESP;
      } else {
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT, &CPU_SS,
			      SIZE_16BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT, &CPU_SP,
			      SIZE_16BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT, &CPU_FLAGS,
			      SIZE_16BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT, &CPU_CS,
			      SIZE_16BIT, CSDescriptor.dpl());
	WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT, &CPU_IP,
			SIZE_16BIT, CSDescriptor.dpl());
	if ((current_origin == EXC_PROCESSOR) && exception_error[current_exception])
	  WriteToVirtualAddress(SSDescriptor, newESP -= SIZE_16BIT,
				&current_error_code, SIZE_16BIT, 
				CSDescriptor.dpl());
	CPU_IP = 0xFFFF & Offset;
	CPU_SS = newESP;
      }

      //  Load the common values, Segment registers and their descriptors.
      SET_BITS(CSSelector, 0, 2, CSDescriptor.dpl());
      CPU_CS = CSSelector;
      CPU_SS = SSSelector;
      seg_descs[SDESC_CS] = CSDescriptor;
      seg_descs[SDESC_SS] = SSDescriptor;

      if ((Descriptor.type() == DESC_16BIT_INT_GATE) ||
	  (Descriptor.type() == DESC_32BIT_INT_GATE))
	CPU_EFLAGS &= ~(1<<9);				// Clear IF
      CPU_EFLAGS &= ~(0x41<<8);			        // Clear NT, TF
      exception_occurred = false;
      return;
     
HandleV86:  // This should be a function.  I'm so lame.  Will fix someday
      if (GetIOPL() < 3) {
	ExceptionOccurred(EXC_PROTECTION, 0);
	return;
      }	else {
	if (Descriptor.dpl() == 3) {
	  if (CSDescriptor.dpl() != 0) 
	    { ExceptionOccurred(EXC_PROTECTION, 0); return; }
	  
	  tmpEFL = CPU_EFLAGS;
	  CPU_EFLAGS &= ~(0x201<<8);                 // TF, VM = 0
	  if ((Descriptor.type() == DESC_16BIT_INT_GATE) ||
	      (Descriptor.type() == DESC_32BIT_INT_GATE))
	    CPU_EFLAGS &= ~(1<<9);	 	     // Clear IF
	  SSSelector = CPU_SS;
	  newESP = CPU_ESP;

	  //  Read CPL = 0 SS and eSP from TSS.  TSS _must_ be valid
	  // and the new SS Descriptor _must_ be there.
	  if (!ReadFromVirtualAddress(
                         seg_descs[SDESC_TR], 8, &CPU_SS, SIZE_16BIT, 0) ||
	      !ReadFromVirtualAddress(
                         seg_descs[SDESC_TR], 4, &CPU_ESP, SIZE_32BIT, 0) || 
	      !LoadDescriptor(seg_descs[SDESC_SS], 
                                    CPU_SS, DMASK_DATA, true, false, true)) {
	    ASSERT(0);
	  }

	  //  Point of No Return, we're about to push things on the
	  // stack.  Note that unlike the Protected Mode setup, we
	  // haven't made sure there's enough space on the new stack
	  // yet.  I have no idea what happens if we run out of space
	  // though...
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_GS, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_FS, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_DS, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_ES, SIZE_32BIT, 0);
	  CPU_GS = CPU_FS = CPU_DS = CPU_ES = 0;

	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP -= SIZE_32BIT, 
			       &SSSelector, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &newESP, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &tmpEFL, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_CS, SIZE_32BIT, 0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], 
			       CPU_ESP -= SIZE_32BIT, &CPU_EIP, SIZE_32BIT, 0);
	  CPU_CS = CSSelector;
	  CPU_EIP = Offset;
	  exception_occurred = false;
	  return; // V86 Mode Exception.

	} else {
	  ExceptionOccurred(EXC_PROTECTION, 0);
	  return;
	}
      }
     } else // Conforming or Same Privilege Handler.  
       if (CSDescriptor.is_conforming()||(CSDescriptor.dpl() == GetCPL())) {
	 // Space to push everything?  
	 if (Descriptor.gate_use_32bit()) 
	   spaceNeeded = 12 + (((current_exception < 19) && ((current_origin == EXC_PROCESSOR) && exception_error[current_exception])) ? 4 : 0); 
	 else spaceNeeded = 6 + (((current_exception < 19) && ((current_origin == EXC_PROCESSOR) && exception_error[current_exception])) ? 2 : 0);

        if(!seg_descs[SDESC_SS].offset_valid(
	          (Descriptor.gate_use_32bit() ?  CPU_ESP : CPU_SP) - spaceNeeded))
	  { ExceptionOccurred(EXC_STACK_FAULT, 0); return; }

	if (!CSDescriptor.offset_valid(Offset)) 
	  { ExceptionOccurred(EXC_PROTECTION, 0); return; }

	//  Point of No Return, We're about to change registers and
	// stacks.  Cannot generate any interupts past here.

	if(Descriptor.gate_use_32bit()) {
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_32BIT, &CPU_EFLAGS,SIZE_32BIT,GetCPL());
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_32BIT, &CPU_CS,SIZE_32BIT,GetCPL());
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_32BIT, &CPU_EIP,SIZE_32BIT,GetCPL());
	  CPU_EIP = Offset;
	} else {
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_16BIT, &CPU_FLAGS,SIZE_16BIT,GetCPL());
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_16BIT, &CPU_CS,SIZE_16BIT,GetCPL());
	  WriteToVirtualAddress(seg_descs[SDESC_SS],
		      CPU_ESP -= SIZE_16BIT, &CPU_IP,SIZE_16BIT,GetCPL());
	  CPU_IP = Offset;
        }
	SET_BITS(CSSelector, 0, 2, GetCPL());                 // RPL = CPL
	CPU_CS = CSSelector;
	seg_descs[SDESC_CS] = CSDescriptor;
	if ((current_origin == EXC_PROCESSOR) && 
            exception_error[current_exception]) {       // Push error code
	   if (Descriptor.gate_use_32bit())
	       WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP -= SIZE_32BIT,
				     &current_error_code, SIZE_32BIT, GetCPL());
	   else
	       WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_SP -= SIZE_16BIT,
				     &current_error_code, SIZE_16BIT, GetCPL());
	}
        
	if ((Descriptor.type() == DESC_16BIT_INT_GATE) ||
	    (Descriptor.type() == DESC_32BIT_INT_GATE))
	 CPU_EFLAGS &= ~(1<<FLAG_IF);				// Clear IF
         CPU_EFLAGS &=  ~((1 << FLAG_NT) | (1 << FLAG_TF)); 	// Clear NT, TF
 	 exception_occurred = false;
	 return;
     } else {
       ExceptionOccurred(EXC_PROTECTION, CSSelector + EXT);
       return;
     }

     break;
   
  default:
     //  Um, I don't know what the hell's in the IDT, but it ain't legal.
     fprintf(stderr, "Invalid Descriptor in IDT.  Type = %d\n",
							Descriptor.type());
     ASSERT(0);
  }
}

// This function attempts to execute a single instruction
int64 CPU::ExecuteOneInstruction(LockStep *singlestep)
{
  Instruction *instruction;
  int64 cycles_used = 0;
  MemoryStream instr_data(this);
  uint8 exc_code;

  if (BITS(CPU_EFLAGS,FLAG_IF,1) && (X86HasIntrPending(0, &exc_code))) {
     ExceptionOccurred(exc_code,0,EXC_EXTERNAL);
     while(exception_occurred && !is_halted())
        HandleException();
     return 0;
  }
  prev_eip = CPU_EIP;  /* Save old PC in case we need to start over */
  if(instr_data.ResetStream(seg_descs[SDESC_CS],CPU_EIP,
			    MAX_BYTES_PER_INSTRUCTION,true,false,
			    (GetCPL() == 3))) {
    // Decode the instruction
    instruction = Instruction::DecodeInstruction(instr_data,
				    (seg_descs[SDESC_CS].use_32bit() ?
				     SIZE_32BIT : SIZE_16BIT));

    if(instruction != 0) {
      // If we've got the instruction:
      //   1) Increment the EIP register, saving the old value
      prev_eip = CPU_EIP;
      CPU_EIP += instruction->Length();

      //   1b) If we're using 16-bit code segments, EIP can't be > 0xFFFF
      if(!seg_descs[SDESC_CS].use_32bit())
	CPU_EIP &= 0x0000FFFF;

      //   1.5) Make sure we single step a REP'd instruction if neccesary
      single_stepping = (singlestep != NULL);

      //   1.7) The instruction has to explicitly ask to ignore flags.
      flags_to_ignore = 0;

      //   2) Actually execute the instruction, and count clock cycles
      cycles_used = (instruction->GetOperation())(*instruction);
      AdjustTimeStampCounter(cycles_used);

      //   2b) If LockSteppingdo that here and make sure the registers are
      //  correct.  Note this has to occur before deleting the instruction
      //  object so that when we die we can examine the instruction with
      //  the debugger.
      if (singlestep != NULL) {
        //  If we just executed a system call, skip the SingleStep this
	// time b/c of a bug in Linux ptrace() which executes two
	// instructions at a time when one is a syscall.
        if (_last_was_syscall == 1) {
	  _last_was_syscall = 0;
        } else {
          switch(singlestep->SingleStep()) {
          case 1:
            break;
          case 0:
            fprintf(stderr, "SingleStep has exited\n");
            break;
          case -1:
            fprintf(stderr, "SingleStep failed!\n");
            exit(1);
          default:
            fprintf(stderr, "SingleStep returned an invalid value!");
            ASSERT(0);
            exit(1);
          }
          singlestep->VerifyRegisterFile(&regfile, instruction,
					 flags_to_ignore);
        }
      }

      //   3) Clean up instruction's memory
      delete instruction;

      //   4) Check for exceptions...  Done below
    } else
      // Could have returned NULL on either a page fault or a decode failure.
      if (!exception_occurred) ExceptionOccurred(EXC_INVALID_OPCODE);
  } else {
    // Somebody else has already generated an exception, so we'll handle it
    //  below...
  }
        
  //  Handle exceptions.  If a nested exception occurs, HandleException()
  // will return, but there will still be a pending exception, so we use a
  // while() to keep going until we die or handle everything.
  // HandleException() also patched up EIP in case of an instruction restart.
  while(exception_occurred && !is_halted())
    HandleException();

  return(cycles_used);
}

void CPU::DumpState(void)
{
  int8 buffer[15];
  int32 val;
  int32 i, len;
  int32 esp, eip;  char e;

  if(seg_descs[SDESC_SS].use_32bit()) {
    esp = CPU_ESP;
    e = 'E';
  } else {
    esp = CPU_SP;
    e = ' ';
  }

  fprintf(stderr,"CS: %04X   EIP: %08X   TSC: %08X:%08X        ",
	  CPU_CS,CPU_EIP,CPU_TSC_HI,CPU_TSC_LO);
  fprintf(stderr,"[SS:%cSP+24] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+24,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"DS: %04X   EFL: %08X (",CPU_DS,CPU_EFLAGS);
  fputc((BITS(CPU_EFLAGS,FLAG_OF,1) ? 'O' : 'o'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_DF,1) ? 'D' : 'd'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_IF,1) ? 'I' : 'i'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_TF,1) ? 'T' : 't'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_SF,1) ? 'S' : 's'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_ZF,1) ? 'Z' : 'z'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_AF,1) ? 'A' : 'a'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_PF,1) ? 'P' : 'p'),stderr);
  fputc((BITS(CPU_EFLAGS,FLAG_CF,1) ? 'C' : 'c'),stderr);

  fprintf(stderr,")                     [SS:%cSP+20] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+20,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"ES: %04X   EAX: %08X   ESP: %08X                 ",
	  CPU_ES,CPU_EAX,CPU_ESP);
  fprintf(stderr,"[SS:%cSP+16] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+16,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"FS: %04X   EBX: %08X   EBP: %08X                 ",
	  CPU_FS,CPU_EBX,CPU_EBP);
  fprintf(stderr,"[SS:%cSP+12] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+12,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"GS: %04X   ECX: %08X   ESI: %08X                 ",
	  CPU_GS,CPU_ECX,CPU_ESI);
  fprintf(stderr,"[SS:%cSP+ 8] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+8,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"SS: %04X   EDX: %08X   EDI: %08X                 ",
	  CPU_SS,CPU_EDX,CPU_EDI);

  fprintf(stderr,"[SS:%cSP+ 4] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp+4,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  fprintf(stderr,"                                                         ");
  fprintf(stderr,"[SS:%cSP+ 0] = ",e);
  if(ReadFromVirtualAddress(seg_descs[SDESC_SS],esp,&val,4,GetCPL(),0) > 3)
    fprintf(stderr,"%08X\n",xtoh32(val)); else
    fprintf(stderr,"????????\n");

  if(seg_descs[SDESC_CS].use_32bit()) {
    eip = CPU_EIP;
    e = 'E';
  } else {
    eip = CPU_IP;
    e = ' ';
  }

  fprintf(stderr,"[CS:%cIP] =",e);

  len = ReadFromVirtualAddress(seg_descs[SDESC_CS],eip,buffer,15,GetCPL(),0,true);
  for(i = 0; i < 15; i++)
    if(i < len)
      fprintf(stderr," %02X",buffer[i]); else
      fprintf(stderr," ??");

  fprintf(stderr,"\n\n");
}

// Compares the CPU's stack to the 'new_stack', reporting any differences.
//  Returns true if the stacks are identical, false if not
bool CPU::CompareStacks(int32 esp, int32 len, const int32 *new_stack)
{
  int32 val;
  int32 i, mismatches;

  if(CPU_ESP != esp) {
    DEBUG(DEBUG_LOADER, "ESP Mismatch! Sim ESP = %08X, Real = %08X", 
	  CPU_ESP, esp);
    return(false);
  }

  mismatches = 0;
  for(i = 0; i < len; i++) {
    if(!ReadFromVirtualAddress(seg_descs[SDESC_SS], esp+(i<<2), &val, 4, 0)) {
      DEBUG(DEBUG_LOADER, "Stack read error at 0x%08X!", esp+(i<<2));
      mismatches++;
    } else
    if(val != new_stack[i]) {
      DEBUG(DEBUG_LOADER, "Stack value mismatch @0x%08X: S:%X vs. R:%X",
	    esp+(i<<2), val, new_stack[i]);
      mismatches++;
    }
  }

  return(mismatches == 0);
}

//  Copy in a stack.  In theory this is one WriteToVirtual Call.
int CPU::InitStack(int32 esp, int32 len, const int32 *new_stack)
{
  CPU_ESP = esp;

  if(WriteToVirtualAddress(seg_descs[SDESC_SS], esp, new_stack, len<<2, 0))
    return 0;
  else
    return -1;
}

bool CPU::SaveTSS(int16 newSelector, SegmentDescriptor &newDescriptor)
{
  int32 newTSS[16];

  // Fill in the new data
  newTSS[0] = CPU_EIP;
  newTSS[1] = CPU_EFLAGS;
  newTSS[2] = CPU_EAX;
  newTSS[3] = CPU_ECX;
  newTSS[4] = CPU_EDX;
  newTSS[5] = CPU_EBX;
  newTSS[6] = CPU_ESP;
  newTSS[7] = CPU_EBP;
  newTSS[8] = CPU_ESI;
  newTSS[9] = CPU_EDI;
  newTSS[10] = CPU_ES;			// Zero extended magically, cool eh?
  newTSS[11] = CPU_CS;
  newTSS[12] = CPU_SS;
  newTSS[13] = CPU_DS;
  newTSS[14] = CPU_FS;
  newTSS[15] = CPU_GS;

  // Write the new data
  WriteToVirtualAddress(seg_descs[SDESC_TR], 0x20, 
					  newTSS, sizeof(newTSS), GetCPL());
  if (ExceptionPending()) return false;			// Uh oh.
  
  // Write out TSS Selector to the new task just in case it ever returns...
  WriteToVirtualAddress(newDescriptor, 0, 
					    &CPU_TR, sizeof(int16), GetCPL());
  if (ExceptionPending()) return false;			// Uh oh.

  return true;
}

bool CPU::LoadTSS(void)
{
  int32 newTSS[18];

  ReadFromVirtualAddress(seg_descs[SDESC_TR], 0x1C, newTSS, 
						     sizeof(newTSS), GetCPL());
  if (ExceptionPending()) return false;

  // Now load a ton of state.
  CPU_CR3 = newTSS[0];
  CPU_EIP = newTSS[1];
  CPU_EFLAGS = newTSS[2];
  CPU_EAX = newTSS[3];
  CPU_ECX = newTSS[4];
  CPU_EDX = newTSS[5];
  CPU_EBX = newTSS[6];
  CPU_ESP = newTSS[7];
  CPU_EBP = newTSS[8];
  CPU_ESI = newTSS[9];
  CPU_EDI = newTSS[10];

  //  Segment registers and their descriptors.  This can be the fun part,
  // what to do if a descriptor isn't there?
  CPU_ES = newTSS[11] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_ES], CPU_ES, DMASK_DATA, true, true, true))
    return false; 			// Uh oh.
  CPU_CS = newTSS[12] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true, true, true))
    return false; 			// Uh oh.
  CPU_SS = newTSS[13] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_SS], CPU_SS, DMASK_DATA, true, true, true))
    return false; 			// Uh oh.
  CPU_DS = newTSS[14] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_DS], CPU_DS, DMASK_DATA, true, true, true))
    return false; 			// Uh oh.
  CPU_FS = newTSS[15] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_FS], CPU_FS, DMASK_DATA, true, true, true))
    return false; 			// Uh oh.
  CPU_GS = newTSS[16] & 0xFFFF;
  if (!LoadDescriptor(seg_descs[SDESC_GS], CPU_GS, DMASK_DATA, true, true, true))
    return false; 			// Uh oh.
  CPU_LDT = newTSS[17] & 0xFFFF;
  if(!LoadDescriptor(seg_descs[SDESC_LDT], CPU_LDT, DMASK_LDT, true, true, true))
    return false; 			// Uh oh.

  return true;
}

bool CPU::SwitchTasks(int16 newSelector, SegmentDescriptor &newDescriptor)
{

  if (!SaveTSS(newSelector, newDescriptor)) return false;
  seg_descs[SDESC_TR].set_tss_busy(false);
  if(!WriteBackDescriptor(seg_descs[SDESC_TR],CPU_TR)) ASSERT(0);

  CPU_TR = newSelector;
  seg_descs[SDESC_TR] = newDescriptor;
  seg_descs[SDESC_TR].set_tss_busy();
  if(!WriteBackDescriptor(seg_descs[SDESC_TR],CPU_TR)) ASSERT(0);
  CPU_CR0 |= 1<<3;					// Set TS bit

  return(LoadTSS());
}

//  Like CPU::LoadDescriptor, but looks in the IDT and generates Exceptions
// as appropriate for the an ExceptionHandler rather than a generic case.
bool CPU::LoadExceptionDescriptor(SegmentDescriptor& target)
{
  int8 desc_data[8];
  bool EXT = (current_origin == EXC_EXTERNAL);
  
  //  ReadFromVirtual will handle any problems with the limits and such.
  if (ReadFromVirtualAddress(seg_descs[SDESC_IDT], current_exception<<3,
							desc_data,8,0) < 8) {
    ExceptionOccurred(EXC_PROTECTION, current_exception<<3 + 2 + EXT);
    return false;
  }
  target.extract_descriptor(desc_data);

  // Make sure the descriptor type is valid for something in an IDT.
  if(BITS(DMASK_IDT,target.type(),1) == 0) {
    ExceptionOccurred(EXC_PROTECTION, current_exception<<3 + 2 + EXT);
    return false;
  }

  // Only need DPL >= CPL on INT n / INTO (Software Interupts).
  if ((current_origin == EXC_INT) && (target.dpl() < GetCPL())) {
    ExceptionOccurred(EXC_PROTECTION, current_exception<<3 + 2 + EXT);
    return false;
  }

  if (!target.is_present()) {
    ExceptionOccurred(EXC_NO_SEGMENT, current_exception<<3 + 2 + EXT);
    return false;
  }

  return true;
}
