/* -*- C++ -*-
 *
 * $Id: cpu.h,v 1.4 1998/04/29 04:54:47 nikos Exp $
 * 
 * RCS Log:
 * $Log: cpu.h,v $
 * Revision 1.4  1998/04/29 04:54:47  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:37  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:22  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:36:55  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:55  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: cpu.h,v 1.4 1998/04/29 04:54:47 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/cpu.h,v 1.4 1998/04/29 04:54:47 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:24  bugnion
 * stanford update (x86)
 *
 * Revision 1.25  1997/12/09  14:51:42  sjt
 * FPU emulation should now work completely - FPU exceptions are triggered when
 * 	they should be, and memory operands appear to work fine.  I removed the
 * 	printfs that were cluttering everything too.
 *
 * Mendel's fixes to BT/BTC/BTR/BTS and ADC have been stolen and checked in as
 * 	well.
 *
 * Revision 1.24  1997/12/07 06:57:08  mendel
 * Changes fprintf is cpu.cc to never occur on interrupts.
 *
 * Changed cpu class to export a translation routine so the rest
 * of SimOS can translate linear to physical address. This is
 * currently needed for the hacked up SCSI disk ( Scott's
 * Linux kernel passes virtual addreses to the monitor call)
 *
 * Revision 1.23  1997/12/02 22:52:14  mendel
 * 1) Hooked up interrupts. Moved interrupt code from simos_interface.cc to
 *    cpu.cc .
 *
 * Several bug fixes:
 *
 * 1) Added check to push the error code only on processor generated
 *    exceptions and not on INT or external ones.
 *
 * 2) IDT gate descriptors were broken. I added a gate_use_32bit() since
 *    IDT gate descriptors use a different size field that normal segment
 *    descriptors.
 *
 * 3) Fixed bug in computing spaceNeeded for exception stack puke. Beware
 *    of C operator precedence.
 *
 * 4) Cleaned up some ugly and some incorrect code in HandleException()
 *
 * 5) #if 0 out some bogus check LoadExceptionDescriptor(). I couldn't
 *    tell want it was trying to do but is was wrong.
 *
 * 6) Compute the gate_offset correctly. We need all the bits of the
 *    flags for this to work.
 *
 * 7) Changed POP and MOV to not call WriteBackDescriptor() on selector
 *    zero. This caused WriteBackDescriptor() to die.
 *
 * Revision 1.22  1997/12/02  04:49:18  yoel
 *   Support added for taking Software Interupts and preparation for supporting
 * external interupts.  cpu.{cc,h} updated so that current_exception is an
 * int8 instead of the enum to allow exceptions 19-255 and enum ExceptionSource
 * added.  Cosmetic changes to fpu.cc and INT n / INTO now actually do
 * something (since we now support Software Interupts).
 *
 * Revision 1.21  1997/11/22 07:47:40  whimsy
 * SimOS Compatibility, #undef'ed the SimOS PAGE_SIZE so we could use our own.
 *
 * Revision 1.20  1997-11-13 21:32:08-08  whimsy
 * Added exception_error array.
 *
 * Revision 1.19  1997-11-09 21:57:51-08  whimsy
 * Fixed paramters to LoadExceptionDescriptor()
 *
 * Revision 1.18  1997-11-09 16:08:05-08  whimsy
 * Added LoadINteruptDescriptor() method.
 *
 * Revision 1.17  1997-10-30 17:00:01-08  whimsy
 * Added SaveTSS(), LoadTSS() as helpers for Task Switching.
 *
 * Revision 1.16  1997-10-29 14:45:34-08  sean
 * Added prototypes for simulating FPU ops, and for task switching
 *
 * Revision 1.15  1997/10/18 00:38:28  sean
 * added support for instruction counter and simulation of FPU ops on the
 *   actual chip
 *
 * Revision 1.14  1997/09/27 18:41:02  sean
 * Added code for comparing simulator stack to lockstep stack, and glue code
 *   for which flags to ignore in lockstep comparison
 *
 * Revision 1.13  1997/09/24 02:40:55  sjt
 * support for single-stepping of REP'd instructions added
 *
 * Revision 1.12  1997/09/06 00:39:51  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.11  1997/03/18 19:14:29  sean
 * changed prototype for SetFlagsAfterOp to allow silly SUB behavior of
 *   flipping CF and AF from what they'd intuitively be...
 *
 * Revision 1.10  1997/03/09 21:16:00  sean
 * added #include for fpu_instr.h - makes FPU instructions exist (sorta)
 *
 * Revision 1.9  1997/03/06 02:52:35  sean
 * Added accessors for register file, descriptor list, and create the
 *   DumpState() method, which dumps registers, stack, and code stream to
 *   stderr
 *
 * Revision 1.8  1997/02/23 03:25:45  sean
 * added functionality for read/writes to I/O ports (just stubs so far), for
 *   reading/writing descriptors based on selectors, and a helper function
 *   for checking permissions on specified I/O ports.
 *
 * Revision 1.7  1997/02/11 05:06:56  sjt
 * made class MemoryStream friend of class CPU
 *
 * Revision 1.6  1997/02/10 07:29:51  sean
 * Modularized memory accessing code (again).  Created nice, well-behaved
 * TranslateVirtualToLinear() and TranslateLinearToPhysical() methods, paving the
 * way for the MemoryStream class.
 *
 * Revision 1.5  1997/02/04 22:36:49  sean
 * created accessors, mutators for new TSC register, made ExecuteOneInstruction
 * return number of clock cycles used
 *
 * Revision 1.4  1997/02/04 10:39:29  sean
 * added functions to set just s,z,p flags, or just o,c flags
 *
 * Revision 1.3  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef _CPU_H
#define _CPU_H

#include "registers.h"
#include "seg_desc.h"

#if defined(SIMOS) && defined(PAGE_SIZE)
#undef PAGE_SIZE
#endif

#define MAX_BYTES_PER_INSTRUCTION       15

#define PAGE_OFFSET_BITS                12
#define PAGE_SIZE                       (1 << PAGE_OFFSET_BITS)

class Instruction;
class LockStep;
class InstructionCounter;

enum Exception { EXC_DIV_BY_ZERO    =  0,
		 EXC_DEBUG          =  1,
		 EXC_NMI            =  2,
		 EXC_BREAKPOINT     =  3,
		 EXC_OVERFLOW       =  4,
		 EXC_BOUNDS         =  5,
		 EXC_INVALID_OPCODE =  6,
		 EXC_NO_DEVICE      =  7,
		 EXC_DOUBLE_FAULT   =  8,
		 EXC_INVALID_TSS    = 10,
		 EXC_NO_SEGMENT     = 11,
		 EXC_STACK_FAULT    = 12,
		 EXC_PROTECTION     = 13,
		 EXC_PAGE_FAULT     = 14,
		 EXC_FPU_ERROR      = 16,
		 EXC_ALIGNMENT      = 17,
		 EXC_MACHINE_CHECK  = 18};

enum ExceptionClass { EXC_CLASS_BENIGN, 
		      EXC_CLASS_CONTRIBUTORY,
		      EXC_CLASS_PAGE_FAULT,
                      EXC_CLASS_EVIL_NASTY };

enum ExceptionSource { EXC_PROCESSOR,              // Generated by the processor
		       EXC_INT,                    // Generated by an INT n / INTO
		       EXC_EXTERNAL };             // Hardware Interupt

const ExceptionClass exception_class[] = {
  EXC_CLASS_CONTRIBUTORY,     // EXC_DIV_BY_ZERO    =  0,
  EXC_CLASS_BENIGN,           // EXC_DEBUG          =  1,
  EXC_CLASS_BENIGN,	      // EXC_NMI            =  2,
  EXC_CLASS_BENIGN,	      // EXC_BREAKPOINT     =  3,
  EXC_CLASS_BENIGN,           // EXC_OVERFLOW       =  4,
  EXC_CLASS_BENIGN,           // EXC_BOUNDS         =  5,
  EXC_CLASS_BENIGN,           // EXC_INVALID_OPCODE =  6,
  EXC_CLASS_BENIGN,           // EXC_NO_DEVICE      =  7,
  EXC_CLASS_EVIL_NASTY,       // EXC_DOUBLE_FAULT   =  8,
  EXC_CLASS_BENIGN,
  EXC_CLASS_CONTRIBUTORY,     // EXC_INVALID_TSS    = 10,
  EXC_CLASS_CONTRIBUTORY,     // EXC_NO_SEGMENT     = 11,
  EXC_CLASS_CONTRIBUTORY,     // EXC_STACK_FAULT    = 12,
  EXC_CLASS_CONTRIBUTORY,     // EXC_PROTECTION     = 13,
  EXC_CLASS_PAGE_FAULT,       // EXC_PAGE_FAULT     = 14,
  EXC_CLASS_BENIGN,
  EXC_CLASS_BENIGN,           // EXC_FPU_ERROR      = 16,
  EXC_CLASS_BENIGN,           // EXC_ALIGNMENT      = 17,
  EXC_CLASS_BENIGN            // EXC_MACHINE_CHECK  = 18
};

const bool exception_error[] = {
  false,           // EXC_DIV_BY_ZERO    =  0,
  false,           // EXC_DEBUG          =  1,
  false,           // EXC_NMI            =  2,
  false,           // EXC_BREAKPOINT     =  3,
  false,           // EXC_OVERFLOW       =  4,
  false,           // EXC_BOUNDS         =  5,
  false,           // EXC_INVALID_OPCODE =  6,
  false,           // EXC_NO_DEVICE      =  7,
  true,            // EXC_DOUBLE_FAULT   =  8,
  true,
  true,            // EXC_INVALID_TSS    = 10,
  true,            // EXC_NO_SEGMENT     = 11,
  true,            // EXC_STACK_FAULT    = 12,
  true,            // EXC_PROTECTION     = 13,
  true,            // EXC_PAGE_FAULT     = 14,
  false,
  false,           // EXC_FPU_ERROR      = 16,
  true,            // EXC_ALIGNMENT      = 17,
  false            // EXC_MACHINE_CHECK  = 18
};

// This is to make it easier to play with registers in general

// 64-bit registers
#define CPU_TSC         REG64(regfile,REG_TSC)

// 32-bit registers
#define CPU_EAX         REG32(regfile,REG_EAX)
#define CPU_ECX         REG32(regfile,REG_ECX)
#define CPU_EDX         REG32(regfile,REG_EDX)
#define CPU_EBX         REG32(regfile,REG_EBX)
#define CPU_ESP         REG32(regfile,REG_ESP)
#define CPU_EBP         REG32(regfile,REG_EBP)
#define CPU_ESI         REG32(regfile,REG_ESI)
#define CPU_EDI         REG32(regfile,REG_EDI)
#define CPU_EFLAGS      REG32(regfile,REG_EFLAGS)
#define CPU_EIP         REG32(regfile,REG_EIP)

#define CPU_CR(num)     REG32(regfile,REG_CRBASE + (num))
#define CPU_CR0         REG32(regfile,REG_CR0)
#define CPU_CR1         REG32(regfile,REG_CR1)
#define CPU_CR2         REG32(regfile,REG_CR2)
#define CPU_CR3         REG32(regfile,REG_CR3)
#define CPU_CR4         REG32(regfile,REG_CR4)

#define CPU_PDBR        REG32(regfile,REG_PDBR)

#define CPU_DR(num)     REG32(regfile,REG_DRBASE + (num))
#define CPU_DR0         REG32(regfile,REG_DR0)
#define CPU_DR1         REG32(regfile,REG_DR1)
#define CPU_DR2         REG32(regfile,REG_DR2)
#define CPU_DR3         REG32(regfile,REG_DR3)
#define CPU_DR4         REG32(regfile,REG_DR4)
#define CPU_DR5         REG32(regfile,REG_DR5)
#define CPU_DR6         REG32(regfile,REG_DR6)
#define CPU_DR7         REG32(regfile,REG_DR7)

#define CPU_TSC_LO      REG32(regfile,REG_TSC_LO)
#define CPU_TSC_HI      REG32(regfile,REG_TSC_HI)

// 16-bit registers
#define CPU_AX          REG16(regfile,REG_AX)
#define CPU_CX          REG16(regfile,REG_CX)
#define CPU_DX          REG16(regfile,REG_DX)
#define CPU_BX          REG16(regfile,REG_BX)
#define CPU_SP          REG16(regfile,REG_SP)
#define CPU_BP          REG16(regfile,REG_BP)
#define CPU_SI          REG16(regfile,REG_SI)
#define CPU_DI          REG16(regfile,REG_DI)
#define CPU_FLAGS       REG16(regfile,REG_FLAGS)
#define CPU_IP          REG16(regfile,REG_IP)

#define CPU_ES          REG16(regfile,REG_ES)
#define CPU_CS          REG16(regfile,REG_CS)
#define CPU_SS          REG16(regfile,REG_SS)
#define CPU_DS          REG16(regfile,REG_DS)
#define CPU_FS          REG16(regfile,REG_FS)
#define CPU_GS          REG16(regfile,REG_GS)
#define CPU_TR          REG16(regfile,REG_TR)
#define CPU_LDT         REG16(regfile,REG_LDT)

// 8-bit registers
#define CPU_AL          REG8(regfile,REG_AL)
#define CPU_CL          REG8(regfile,REG_CL)
#define CPU_DL          REG8(regfile,REG_DL)
#define CPU_BL          REG8(regfile,REG_BL)
#define CPU_AH          REG8(regfile,REG_AH)
#define CPU_CH          REG8(regfile,REG_CH)
#define CPU_DH          REG8(regfile,REG_DH)
#define CPU_BH          REG8(regfile,REG_BH)

// other helpful (but random) defines
#define TSS_IO_MAP_BASE 0x0066

class CPU {
 public:
  CPU(int32 _last_byte);
  CPU(void *_memory_base, int32 _last_byte = 0x0FFFFFFFF);
  ~CPU(void);

  // This function attempts to execute a single instruction, returns the
  //  number of clock cycles spent
  int64 ExecuteOneInstruction(LockStep *lockstep = 0);

  // This function resets all registers, etc. to initial states
  void Reset();

  // This is the magical instruction decode function
  //  Up to 'data_len' bytes of opcode data is provided in 'opcode_data',
  //  and the function is supposed to return a decoded instruction (although
  //  we need a way to say how many bytes of data we burned on the instruction
  //  also...)
//  Instruction *DecodeInstruction(int8 *opcode_data, int data_len,
//				 int default_datasize);

  // Functions for reading and writing to/from memory 
  //  All take some form of address, a buffer to read/write, and the number
  //   of bytes to read/write - the return value is the number of bytes
  //   actually read/written

  //  Physical addresses are the actual places in the RAM chips to look
  //   for data (i.e. this is after segmentation and paging)
  //  Reads and writes to physical memory cannot result in errors that the
  //   CPU can see.  Reads from non-existent addresses yield all 0's.  Writes
  //   to non-existent or read-only addresses do nothing
  void ReadFromPhysicalAddress(int32 physical, void *buffer, int len);
  void WriteToPhysicalAddress(int32 physical, const void *buffer, int len);

  //  I/O Ports are pretty much like memory, except they live in a different
  //   address space, and there's no paging, or segments, or anything weird
  //   (unless you use memory-mapped I/O).
  //  I/O is also not directly dependent on the Pentium architecture, so these
  //   functions are essentially wrappers around simulators for related
  //   hardware.
  void ReadFromIOPort(int32 port, void *buffer, int len);
  void WriteToIOPort(int32 port, const void *buffer, int len);

  //  Linear addresses are what you have after you deal with segmentation,
  //   but before paging (if paging is off, linear = physical)
  int ReadFromLinearAddress(int32 linear, void *buffer, int len, bool user,
			    int min_len = 0);
  int WriteToLinearAddress(int32 linear, const void *buffer, int len, 
			   bool user, int min_len = 0);

  //  Virtual addresses are pre-segmentation (i.e. what actual programs are
  //   using) - they're annoying because you need both the segment
  //   descriptor and the offset
  int ReadFromVirtualAddress(SegmentDescriptor& segment, int32 offset, 
			     void *buffer, int len, int cpl, int min_len = 0,
			     bool is_instr_fetch = false);
  int WriteToVirtualAddress(SegmentDescriptor& segment, int32 offset, 
			    const void *buffer, int len, int cpl, 
			    int min_len = 0);

  inline bool is_halted(void) const { return(halted); }

  inline bool in_real_mode(void) const { return(BITS(CPU_CR0,0,1) == 0); }
  inline bool in_protected_mode(void) const { return(BITS(CPU_CR0,0,1) == 1); }
  inline bool in_v86_mode(void) const {return(in_protected_mode() && 
                                             (BITS(CPU_EFLAGS,FLAG_VM,1) == 1)); }

  inline int64 ReadTimeStampCounter(void) const { return(CPU_TSC); }

  // Accessors for registers and segment descriptors to let people maul them
  inline RegisterFile *GetRegisters(void) { return(&regfile); }
  inline SegmentDescriptor *GetDescriptors(void) { return(seg_descs); }

  void DumpState(void);

  // Compares the CPU's stack to the 'new_stack', reporting any differences.
  //  Returns true if the stacks are identical, false if not
  bool CompareStacks(int32 esp, int32 len, const int32 *new_stack);

  //  Takes a stack (presumably valid) and overwrites the simulated stack.
  int InitStack(int32 esp, int32 len, const int32 *new_stack);

  // Gives the CPU an instruction counter to keep tally
  inline void SetInstructionCounter(InstructionCounter *_counter)
                                                { instr_counter = _counter; }

  void ExceptionOccurred(int8 exception_type, int32 error_code = 0,
			 ExceptionSource exception_origin = EXC_PROCESSOR);


  // This function attempts to translate a 'linear' (pre-paging)
  //  address to a 'physical' (post-paging) address.  'user' should be true
  //  if the access is coming from a CPL=3 source, 'writing' should be true
  //  if the access is a write (as opposed to a read), and 'quiet' should be
  //  true if exceptions should be masked.
  //
  // The return value is 0 if the address could not be translated, and the
  //  number of bytes (past the given address) for which the translation is
  //  valid.
  int32 TranslateLinearToPhysical(int32 linear, int32& physical, bool user, 
				  bool writing, bool quiet);

protected:
  // This function attempts to translate a virtual address (described by a
  //  'segment' and an 'offset') into a 'linear' (pre-paging) address.
  //  'writing' should be true if the access will be a write (as opposed to a
  //  read or fetch), 'fetching' should be true if the access is an instruction
  //  fetch (as opposed to a normal read or write).  'is_stack' should be true
  //  if 'segment' refers to the stack segment, and 'quiet' should be true if
  //  exceptions should be masked.
  //
  // The return value is 0 if the address could not be translated, and the
  //  number of bytes (past the given address) for which the translation is
  //  valid.
  int32 TranslateVirtualToLinear(SegmentDescriptor& segment, int32 offset,
				 int32& linear, bool writing,
				 bool fetching, bool is_stack, bool quiet);

  // This function handles the evilness of loading a descriptor specified by
  //  'selector' into the 'target' descriptor.  Only valid descriptors with
  //  a type permitted by the 'type_mask' (which should be one of the DMASK_
  //  macros defined in seg_desc.h) will be loaded, and if 'need_present' is
  //  set, the descriptor must be marked present as well.
  // If the descriptor is ok in all respects, it is copied into 'target', and
  //  the function returns true.
  // If any error occurs, the function returns false, and if 'quiet' is false,
  //  an exception is raised.
  bool LoadDescriptor(SegmentDescriptor& target, int16 selector,
		      int32 type_mask, bool need_present, 
		      bool allow_null, bool quiet);

  // This function writes the 'source' segment descriptor back into the place
  //  specified by 'selector' - return value is whether or not the write was
  //  either unnecessary (in real or vm86 modes) or successful
  bool WriteBackDescriptor(const SegmentDescriptor& source, int16 selector);

  inline void AdjustTimeStampCounter(int64 to_add) { CPU_TSC += to_add; }

  // This function checks to see if the current process is allowed to access
  //  the I/O ports starting at 'port_base' with a size of 'size'.  It returns
  //  true if access is allowed, and false if access is not allowed.  The only
  //  exception that can occur inside of this code is a page fault.
  bool CheckIOPermissions(int32 port_base, int size);

  bool CheckPagePermissions(int32 entry, bool user, bool writing, 
			    int32 linear_addr, bool quiet);

  // This function performs the steps necessary to switch to a new TSS
  bool SwitchTasks(int16 newSelector, SegmentDescriptor &newDescriptor);

  //  Writes the current state to the TSS
  bool SaveTSS(int16 newSelector, SegmentDescriptor &newDescriptor);
  
  //  Reads the new state from the stored TSS
  bool LoadTSS();

  void SetArithFlags(bool of, bool sf, bool zf, bool af, bool pf, bool cf);
  void Set_SZP_Flags(bool sf, bool zf, bool pf);
  void Set_OC_Flags(bool of, bool cf);

  void SetFlagsAfterOp(int32 a, int32 b, int32 x, int size, bool flip_carry);
  
  int GetCPL(void) const;
  inline int GetIOPL(void) { return(BITS(CPU_EFLAGS,FLAG_IOPL,2)); }

  bool LoadExceptionDescriptor(SegmentDescriptor& target);
  inline bool ExceptionPending(void) const { return(exception_occurred); }
  void ClearException(void);
  void HandleException(void);

  // used in fpu_instr.cc to simulate fpu calls
  void SimulateFPUonRealFPU(const Instruction &inst);
  bool CheckForFPUException(void);

  RegisterFile regfile;
  SegmentDescriptor seg_descs[SDESC_TOTAL];
  int32 last_memory_byte;
  int8 *memory;
  bool mem_alloced;

  // All of the operand functions need to be public to allow tables to be
  //  built...
  // Actually, protected should work if Instructions and CPUs are friends.
#include "x86_instr.h"

  // Add in the FPU instructions too...
#include "fpu_instr.h"

  bool halted;

  // For handling of lockstep simulation, some instructions have undefined
  //  effects on flags, and we want to ignore those when comparing to lockstep.
  int32 flags_to_ignore;

  // Single-stepping mode - when active, this only executes REP'd instructions
  //  once per call to ExecuteOneInstruction.  It runs slower this way,
  //  especially since it decodes the REP'd instruction n times instead of 1,
  //  but it is necessary for lockstep debugging, and when the TF is enabled.
  bool single_stepping;

  // Exception stuff
  int32 prev_eip;    // during an instruction, we need the previous eip, so's
                     //   we can back up to it on an exception or REP prefix
  bool exception_occurred;
  int8 current_exception;
  int32 current_error_code;
  ExceptionSource current_origin;

  //  Syscall tracking for b/c of Linux ptrace() bug.
  bool _last_was_syscall;

  // So we can count which instructions have been executed
  InstructionCounter *instr_counter;

  // For now, keep FPU state in a block of memory
  int8 *fpu_state;

  // Keep track of our friends
  friend class Instruction;
  friend class Operand;      // Do we need subclasses of Operand too?
  friend class RegisterOperand;
  friend class MemoryOperand;
  friend class MemoryStream;
};

#endif
