/* $Id: debug.cc,v 1.4 1998/04/29 04:54:48 nikos Exp $
 *
 * RCS Log:
 * $Log: debug.cc,v $
 * Revision 1.4  1998/04/29 04:54:48  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:38  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:24  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:57  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:57  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: debug.cc,v 1.4 1998/04/29 04:54:48 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/debug.cc,v 1.4 1998/04/29 04:54:48 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:24  bugnion
// stanford update (x86)
//
// Revision 1.4  1997/03/18  19:01:00  sean
// created DEBUG_ALL ('+'), which effectively turns on all debug flags - moved
//   DEBUG_FPU ('F'), DEBUG_LOADER ('L'), and DEBUG_SYSCALL ('S') into debug.h
//
 * Revision 1.3  1997/03/15 02:18:20  yoel
 * Added functions to turn on/off specific DEBUGGING flags.
 *
 * Revision 1.2  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: debug.cc,v 1.4 1998/04/29 04:54:48 nikos Exp $";

//  File debug.cc, Author Sean Treichler / sjt@leland.stanford.edu
//  Last Revised 28 January 1997
//
//  A slightly nicer version of a DEBUG message call...

#include <stdio.h>
#include <stdarg.h>
#include "debug.h"

bool DebugCodes[256] = {
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false, false, false, false, false,
  false, false, false, false, false, false 
};

static inline bool CheckDebugCode(char msg_code)
{
  return(DebugCodes[msg_code] || DebugCodes['+']);
}

void TurnOnDebugCode(char msg_code)
{
  DebugCodes[msg_code] = true;
}

void TurnOffDebugCode(char msg_code)
{
  DebugCodes[msg_code] = false;
}

void PrintDebugMessage(char msg_code, const char *fmt, ...)
{
  va_list args;

  if(CheckDebugCode(msg_code)) {
    fprintf(stderr,"DEBUG(%c): ",msg_code);

    va_start(args,fmt);
    vfprintf(stderr,fmt,args);
    va_end(args);

    fputc('\n',stderr);
  }
}
