/* $Id: debug.h,v 1.4 1998/04/29 04:54:49 nikos Exp $
 * 
 * RCS Log:
 * $Log: debug.h,v $
 * Revision 1.4  1998/04/29 04:54:49  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:39  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:25  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:36:58  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:57  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: debug.h,v 1.4 1998/04/29 04:54:49 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/debug.h,v 1.4 1998/04/29 04:54:49 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:25  bugnion
 * stanford update (x86)
 *
 * Revision 1.8  1997/03/18  19:01:00  sean
 * created DEBUG_ALL ('+'), which effectively turns on all debug flags - moved
 *   DEBUG_FPU ('F'), DEBUG_LOADER ('L'), and DEBUG_SYSCALL ('S') into debug.h
 *
 * Revision 1.7  1997/03/15 02:18:20  yoel
 * Added code to Turn On/Off specific debugging flags.
 *
 * Revision 1.6  1997/02/04 01:48:11  sjt
 * stupid cosmetic changes
 *
 * Revision 1.5  1997/02/04 01:36:12  sjt
 * Stupid fix for ASSERT macro (needed to #include <stdio.h>)
 * 
 * Revision 1.4  1997/02/04 01:19:21  sjt
 * Added operand debug flag
 * 
 */

#ifndef __DEBUG_H__
#define __DEBUG_H__

// Need stdio.h for stderr...
#include <stdio.h>

// We could include all of stdlib.h, or just prototype abort...
extern "C" {
void abort(void);
};

//  Various DEBUG constants
#define DEBUG_ALL       '+'
#define DEBUG_DECODE	'D'
#define DEBUG_EXCEPTION 'E'
#define DEBUG_FPU       'F'
#define DEBUG_LOADER    'L'
#define DEBUG_MEMORY    'M'
#define DEBUG_OPERAND   'O'
#define DEBUG_SYSCALL   'S'

#ifdef NO_DEBUG_MSGS
#define DEBUG        (void)
#else
#define DEBUG        PrintDebugMessage
#endif

extern void PrintDebugMessage(char msg_code, const char *fmt, ...);
extern void TurnOnDebugCode(char msg_code);
extern void TurnOffDebugCode(char msg_code);

//----------------------------------------------------------------------
// ASSERT - borrowed (stolen?) from Nachos source code
//
//      If condition is false,  print a message and dump core.
//      Useful for documenting assumptions in the code.
//
//      NOTE: needs to be a #define, to be able to print the location
//      where the error occurred.
//----------------------------------------------------------------------
#define ASSERT(condition)                                                 \
  if(!(condition)) {                                                      \
    fprintf(stderr, "Assertion failed: line %d, file \"%s\"\n",           \
                    __LINE__, __FILE__);                                  \
    fflush(stderr);                                                       \
    abort();                                                              \
  }

#endif
