/* $Id: decode.cc,v 1.4 1998/04/29 04:54:49 nikos Exp $
 *
 * RCS Log:
 * $Log: decode.cc,v $
 * Revision 1.4  1998/04/29 04:54:49  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:39  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:25  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:58  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:58  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: decode.cc,v 1.4 1998/04/29 04:54:49 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/decode.cc,v 1.4 1998/04/29 04:54:49 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:25  bugnion
// stanford update (x86)
//
// Revision 1.17  1997/03/18  19:02:57  sean
// fixed typo in call to DecodeFPUInstruction - passing second byte to
//   'firstByte' didn't make much sense to me... :)
//
 * Revision 1.16  1997/03/14 08:03:34  yoel
 * Added hooks to call the FPU decoder.
 *
 * Revision 1.15  1997/03/09 08:36:57  yoel
 * Hmm, cleaned up some fluff, unnecessary nDecoded's and stuff.
 *
 * Revision 1.14  1997/03/07 05:54:20  sean
 * added special case decoding for Jcc's and SETcc's, fixed Grp3(F7) decoding
 *
 * Revision 1.13  1997/03/06 18:54:48  sean
 * fixed Grp1 weirdness in operand tables (the opcode map in the book lies
 *   about 0x83) - changed DecodeInstruction() to take better care of which
 *   bytes are unread (should always be the ModR/M)
 *
 * Revision 1.12  1997/03/06 09:22:59  yoel
 * Added support for all of the GrpN instructions.  Note that if the
 * operand encoding depends on the ModRM byte, you cannot push it back into
 * the stream until you've checked the OPerand tables.
 *
 * Revision 1.11  1997/03/06 02:13:01  sean
 * fixed decoding maps, fixed DecodeInstruction bugs (too many prefix bytes,
 *   wrong instruction length reporting)
 *
 * Revision 1.10  1997/02/28 09:24:18  yoel
 * Added the code to do the Operand decoding.  I.e. the table lookups and
 * then calls to the magic decode functions.
 *
 * Revision 1.9  1997/02/19 06:51:59  yoel
 * Shouldn't have made many changes here, probably all cosmetic.
 *
 * Revision 1.8  1997/02/14 06:14:14  yoel
 * Made the decoder now InstructionStream aware and added the first bits of
 * Operand decoding.  Modulo debugging, opcode decoding and prefix decoding
 * should now be done.  Maybe plus one hook for prefix to track REP prefixes.
 *
 * Revision 1.7  1997/02/10 10:19:07  yoel
 * Lots of changes.  With the new static Opcode tables, I've added code to
 * handle all of the opcode decoding and prefix decoding, plus detecting
 * error conditions in invalid opcodes.  And made it compile ;)  All that
 * remains is operands and debugging.
 *
 * Revision 1.6  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: decode.cc,v 1.4 1998/04/29 04:54:49 nikos Exp $";

//  File decode.cc, Author Jeremy Sugerman (yoel@stanford.edu)
//
//  This is the file that is completely dedicated to instruction decoding.

#include <stdio.h>
#include <stdlib.h>
#include "data_stream.h"
#include "decode.h"
#include "cpu.h"
#include "instruction.h"
#include "prefix.h"
#include "debug.h"

#define GET_BYTE(instr_data, curByte, nDecoded) \
  if (!TryToFetch(instr_data, curByte, nDecoded)) return NULL;

#define PUT_BYTE(instr_data, curByte, nDecoded) \
if (instr_data.UnReadOneByte(curByte)) nDecoded--;   \
  else return NULL; 

Instruction *Instruction::DecodeInstruction(DataStream& instr_data, 
                                                          int default_datasize)
{
  Prefix iPrefix(default_datasize);
  InstEvalFunc iFunction;
  DecodeOperandDecl operandFn;
  int8 nDecoded = 0;
  int8 curByte;  // curByte is now used only in prefix decoding...
  int8 opcode[3] = { 0, 0, 0 };
  Operand *ops[3] = { NULL, NULL, NULL };
  int orig_stream_pos = instr_data.BytesRead();

  // Fetch the first byte
  GET_BYTE(instr_data, curByte, nDecoded);

  while(nDecoded <= 4) {
    switch (iPrefix.CheckPrefix(curByte)) {
    case -1:  //  Duplicated prefix group
      DEBUG(DEBUG_DECODE, "Prefix byte (%X) when there was already a prefix "
	                  "from its group.", curByte);
      return NULL;

    case 0:  // Not a prefix so there cannot be any more prefix bytes.
      // A continue is not enough to break out of the while loop, so goto...
      goto done_with_prefixes;

    default:
      // that byte was a prefix, so get the next one
      GET_BYTE(instr_data, curByte, nDecoded);      
      break;
    }
  }
done_with_prefixes:

  // Okay, curByte is now the first opcode byte.
  opcode[0] = curByte;

  //  Lookup the first byte.  For one-byte opcodes this is a one step
  // issue.  For two-bytes, escapes, and errors, this requires at least one
  // more lookup in the fall-back table and then further handling as
  // necessary.
 
  // Modifications made to how data from the instruction stream is read/unread
  //   Since in order to determine the operation, we must read every byte 
  //   before the ModR/M, and also the ModR/M if it matters, we always put the
  //   ModR/M byte back into the stream at this stage.
  // You always have enough information to determine the encoding once you've
  //   determined the operation, so all of the operand encoding stuff can look
  //   at the stored values in opcode[0..2]
  if ((iFunction = OneByteOpcodes[opcode[0]]) == NULL) 
    switch(BadFirstByte[opcode[0]]) {
    case NULL:
      DEBUG(DEBUG_DECODE, "Naughty people!  You're using opcodes that are "
	                  "Intel (tm) reserved.  Quit it.  Use 0x0F0B and "
			  "0x0FB9 for Illegal Opcodes.");
      return NULL;
    
    case TWO_BYTE:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      
      //  Just like the one-byte case.  Lookup in the Opcode table first
      // and if that fails figure out how to handle the special cases.
      if ((iFunction = TwoByteOpcodes[opcode[1]]) == NULL)
        switch(BadSecondByte[opcode[1]]) {
        case NULL:
	  DEBUG(DEBUG_DECODE, "Naughty people!  You're using opcodes that are "
	 		      "Intel (tm) reserved.  Quit it.  Use 0x0F0B and "
			      "0x0FB9 for Illegal Opcodes.");
	  return NULL;

        case GRP6:
          GET_BYTE(instr_data, opcode[2], nDecoded);
	  iFunction = Grp6Codes[GET_MOD_BITS(opcode[2])];
	  if (iFunction == NULL) {
	    DEBUG(DEBUG_DECODE, "Invalid Grp6 opcode %X", opcode[2]);
	    return NULL;
	  }
	  PUT_BYTE(instr_data, opcode[2], nDecoded); // put back ModR/M
	  break;
       
	case GRP7:
          GET_BYTE(instr_data, opcode[2], nDecoded);
	  iFunction = Grp7Codes[GET_MOD_BITS(opcode[2])];
	  if (iFunction == NULL) {
	    DEBUG(DEBUG_DECODE, "Invalid Grp7 opcode %X", opcode[2]);
	    return NULL;
	  }
	  PUT_BYTE(instr_data, opcode[2], nDecoded); // put back ModR/M
	  break;
       
	case GRP8:
          GET_BYTE(instr_data, opcode[2], nDecoded);
	  iFunction = Grp8Codes[GET_MOD_BITS(opcode[2])];
	  if (iFunction == NULL) {
	    DEBUG(DEBUG_DECODE, "Invalid Grp8 opcode %X", opcode[2]);
	    return NULL;
	  }
          PUT_BYTE(instr_data, opcode[2], nDecoded); // put back ModR/M
	  break;
       
	case GRP9:
          GET_BYTE(instr_data, opcode[2], nDecoded);
	  iFunction = Grp9Codes[GET_MOD_BITS(opcode[2])];
	  if (iFunction == NULL) {
	    DEBUG(DEBUG_DECODE, "Invalid Grp9 opcode %X", opcode[2]);
	    return NULL;
	  }
	  PUT_BYTE(instr_data, opcode[2], nDecoded); // put back ModR/M
	  break;
	
	case ILLEGAL_OPCODE:
          DEBUG(DEBUG_DECODE, "Kewl!  You used the Illegal Opcode "
	                      "instruction.  I don't know what to do with "
			      "it though.");
	  return NULL;
	
	default:
	  DEBUG(DEBUG_DECODE, "Amazing!  Second Opcode byte %X is unhandled.  "
	                      "This should not be possible.", 
			      opcode[1]);
	  return NULL;
	}
      break;
    
    case FPU:
      return DecodeFPUInstruction(instr_data, default_datasize, 
				  iPrefix, opcode[0]);
    
    case IS_PREFIX:
      //  Either we already had 4 valid prefixes or we read a byte that
      // wasn't a prefix and now have one that is.
      DEBUG(DEBUG_DECODE, "No more than 4 prefixes please, and no "
                          "interweaving them with valid opcode bytes.");
      return NULL;
    
    case GRP1:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      iFunction = Grp1Codes[GET_MOD_BITS(opcode[1])];
      if (iFunction == NULL) {
        DEBUG(DEBUG_DECODE, "Invalid Grp1 opcode %X", opcode[1]);
        return NULL;
      }
      PUT_BYTE(instr_data, opcode[1], nDecoded); // put back ModR/M
      break;
    
    case GRP2:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      iFunction = Grp2Codes[GET_MOD_BITS(opcode[1])];
      if (iFunction == NULL) {
        DEBUG(DEBUG_DECODE, "Invalid Grp2 opcode %X", opcode[1]);
        return NULL;
      }
      PUT_BYTE(instr_data, opcode[1], nDecoded); // put back ModR/M
      break;
    
    case GRP3:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      iFunction = Grp3Codes[GET_MOD_BITS(opcode[1])];
      if (iFunction == NULL) {
        DEBUG(DEBUG_DECODE, "Invalid Grp3 opcode %X", opcode[1]);
        return NULL;
      }
      PUT_BYTE(instr_data, opcode[1], nDecoded); // put back ModR/M
      break;
    
    case GRP4:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      iFunction = Grp4Codes[GET_MOD_BITS(opcode[1])];
      if (iFunction == NULL) {
        DEBUG(DEBUG_DECODE, "Invalid Grp4 opcode %X", opcode[1]);
        return NULL;
      }
      PUT_BYTE(instr_data, opcode[1], nDecoded); // put back ModR/M
      break;
    
    case GRP5:
      GET_BYTE(instr_data, opcode[1], nDecoded);
      iFunction = Grp5Codes[GET_MOD_BITS(opcode[1])];
      if (iFunction == NULL) {
        DEBUG(DEBUG_DECODE, "Invalid Grp5 opcode %X", opcode[1]);
        return NULL;
      }
      PUT_BYTE(instr_data, opcode[1], nDecoded); // put back ModR/M
      break;
    
    default:
      DEBUG(DEBUG_DECODE, "Amazing!  Opcode byte %X is unhandled.  This "
                          "should not be possible.", opcode[0]);
      return NULL;
      break;
    }
   
  //  Handle operand decoding - relying solely on values in opcode[0..2]
  if ((operandFn = OneByteOperand[opcode[0]]) == NULL) {
    switch(opcode[0]) {
    case 0x70:  case 0x71:  case 0x72:  case 0x73:
    case 0x74:  case 0x75:  case 0x76:  case 0x77:
    case 0x78:  case 0x79:  case 0x7A:  case 0x7B:
    case 0x7C:  case 0x7D:  case 0x7E:  case 0x7F: // near Jcc's
      if(!JB(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
              iPrefix.SegReg(), ops+1)) 
        return NULL;
      ops[0] = new ImmediateOperand(SIZE_8BIT, (opcode[0] - 0x70));
      break;

    case 0xCE:  // INTO is really an INT 4.
      ops[0] = new ImmediateOperand(SIZE_8BIT, 4);
      break;

    case 0xD0:  // Grp2 of form Eb, 1
      if (!EB(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
              iPrefix.SegReg(), ops)) 
        return NULL;
      ops[1] = new ImmediateOperand(SIZE_8BIT, 1);
      break;

    case 0xD1:  // Grp2 of form Ev, 1
      if (!EV(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
          iPrefix.SegReg(), ops))
        return NULL;
      ops[1] = new ImmediateOperand(SIZE_8BIT, 1);
      break;

    case 0xF6:  // Grp3 with Byte operands...
      if ((operandFn = Grp3OperandsF6[GET_MOD_BITS(opcode[1])]) != NULL) {
	if (!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
		       iPrefix.SegReg(), ops))
	return NULL;
      }
      break;

    case 0xF7:  // Grp3 with OpSize operands...
      if ((operandFn = Grp3OperandsF7[GET_MOD_BITS(opcode[1])]) != NULL) {
	if (!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
		       iPrefix.SegReg(), ops))
	return NULL;
      }
      break;
      
    case 0xFF:  // Grp5's.  These are almost all Ev's.
      if ((operandFn = Grp5Operands[GET_MOD_BITS(opcode[1])]) != NULL) {
	if (!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
	               iPrefix.SegReg(), ops))
          return NULL;
      }
      break;

    case 0x0F:  // Two Byte opcode...
      if ((operandFn = TwoByteOperand[opcode[1]]) != NULL) {
	if(!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
		      iPrefix.SegReg(), ops))
	  return NULL;
      } else {
	switch(opcode[1]) {
	case 0x00:   // Group 6
	  if (!EW(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
	          iPrefix.SegReg(), ops)) 
	    return NULL;
	  break;

	case 0x01:   // Group 7
	  if ((operandFn = Grp7Operands[GET_MOD_BITS(opcode[2])]) != NULL) {
	    if (!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
	                   iPrefix.SegReg(), ops))
	      return NULL;
       	  }
	  break;
	  
	case 0x80:  case 0x81:  case 0x82:  case 0x83:
	case 0x84:  case 0x85:  case 0x86:  case 0x87:
	case 0x88:  case 0x89:  case 0x8A:  case 0x8B:
	case 0x8C:  case 0x8D:  case 0x8E:  case 0x8F: // long Jcc's
	  if(!JV(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
		 iPrefix.SegReg(), ops+1)) 
	    return NULL;
	  ops[0] = new ImmediateOperand(SIZE_8BIT, (opcode[1] - 0x80));
	  break;

	case 0x90:  case 0x91:  case 0x92:  case 0x93:
	case 0x94:  case 0x95:  case 0x96:  case 0x97:
	case 0x98:  case 0x99:  case 0x9A:  case 0x9B:
	case 0x9C:  case 0x9D:  case 0x9E:  case 0x9F: // SETcc's
	  if(!EB(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(), 
		 iPrefix.SegReg(), ops+1)) 
	    return NULL;
	  ops[0] = new ImmediateOperand(SIZE_8BIT, (opcode[1] - 0x90));
	  break;

	case 0xC7:   // Group 9 (all Group 8's are Ev, Ib)
	  if (GET_MOD_BITS(opcode[2]) == 0x1) {
	    if (!MQ(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
	            iPrefix.SegReg(), ops))
	      return NULL;
	  }
	  break;
	}
	break;
      }
    }
  } else {
    if(!operandFn(instr_data, iPrefix.OpSize(), iPrefix.AddrSize(),
		                                        iPrefix.SegReg(), ops))
      return NULL;
  }

  // Can't use 'nDecoded' to figure out how many bytes we read, since it'd
  //  be too hard to do that in all the operand decode fns - instead, we ask
  //  the data stream for before and after positions in the stream - also
  //  have to use MaxBytesRead here on the off chance that we read the Mod R/M
  //  byte to figure out the opcode, but then put it back and never pull it
  //  back out to do operands...
  return (new Instruction(iFunction, 
			  (instr_data.MaxBytesRead() - orig_stream_pos),
			  iPrefix, ops[0], ops[1], ops[2]));
}

// I think the name gives this away...
bool Instruction::TryToFetch(DataStream& instr_data, 
                                            int8& nextByte, int8& nDecoded)
{
  if (!instr_data.ReadOneByte(nextByte)) {
    DEBUG(DEBUG_DECODE, "Premature end of instruction (i.e. end of data "
	                "without decoding a complete instruction.");
    return false;
  }
  
  DEBUG(DEBUG_DECODE, "Fetched opcode byte %02X (byte #%2d) from stream.\n",
	nextByte,nDecoded);
  nDecoded++;
  return true;
}
