// -*- C++ -*-
/* $Id: decode.h,v 1.4 1998/04/29 04:54:50 nikos Exp $
 * 
 * RCS Log:
 * $Log: decode.h,v $
 * Revision 1.4  1998/04/29 04:54:50  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:40  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:25  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:36:59  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:58  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: decode.h,v 1.4 1998/04/29 04:54:50 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/decode.h,v 1.4 1998/04/29 04:54:50 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:25  bugnion
 * stanford update (x86)
 *
 * Revision 1.27  1997/12/07  02:54:03  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.26  1997/11/30 23:51:48  mendel
 * Fixed problem with decoding operations of the BTS instruction.
 *
 * Added calls to interrupt controller.
 *
 * Revision 1.25  1997/11/17  08:27:29  whimsy
 * Adjusted operand decode map for LSS, LFS, LGS to Gv_Mp, the table in the
 * book is wrong.
 *
 * Revision 1.24  1997-09-27 11:31:05-07  sean
 * Fixed IMUL decoding to always return three operands (even if first two are
 *   the same)
 *
 * Revision 1.23  1997/09/24 08:16:43  sean
 * Typo in Grp3 table (was NOT instead of NEG for #3)
 *
 * Revision 1.22  1997/03/14 21:42:04  sean
 * fixed operand entries for XCHG EAX,E?? (0x91 - 0x97) instructions
 *
 * Revision 1.21  1997/03/13 18:54:33  yoel
 * THere was an EB_Ob that should have been an EB_Ib for C0.
 *
 * Revision 1.20  1997/03/07 05:52:18  sean
 * fixed Jcc, SETcc operand encodings (NULL -> special case), fixed
 *   MOV Exx, Iv (0xB8 - 0xBF)
 *
 * Revision 1.19  1997/03/06 18:54:48  sean
 * fixed Grp1 weirdness in operand tables (the opcode map in the book lies
 *   about 0x83) - changed DecodeInstruction() to take better care of which
 *   bytes are unread (should always be the ModR/M)
 *
 * Revision 1.18  1997/03/06 09:22:59  yoel
 * Filled in and expanded the tables for GrpNOperand, removed the unneccesary
 * ones.
 *
 * Revision 1.17  1997/03/06 02:48:13  sean
 * fixed decoding maps, fixed DecodeInstruction bugs (too many prefix bytes,
 *   wrong instruction length reporting)
 *
 * Revision 1.16  1997/03/02 00:46:09  yoel
 * There was a typo in GET_MOD_BITS().  I was only shifting 2 bits instead of
 * 3 and the bottom _3_ bits are the R/M field.
 *
 * Revision 1.15  1997/02/28 23:53:03  sean
 * changed AI_IB's in table to AL_IB's (which is what they should be)
 *
 * Revision 1.14  1997/02/28 09:24:18  yoel
 * Cleaned up a few table errors (EAX_LV).
 *
 * Revision 1.13  1997/02/28 03:34:58  sean
 * changed RET operation to RETnear and RETfar
 *
 * Revision 1.12  1997/02/19 06:51:59  yoel
 * Whee, beat the heck out of the proto-types for the OperandDecoding arrays.
 * They still confuse me, but at least they compile.
 *
 * Revision 1.11  1997/02/14 06:14:14  yoel
 * Added stubs for the GrpN operand decoding tables and the tables for the
 * OneByte and TwoByte Operand decodings.
 *
 * Revision 1.10  1997/02/10 10:19:07  yoel
 * Cleaned up typos to make it compile.  Removed an extra * from the decl
 * of TwoByteOpcodes and added G's to CMPXCHG and CMPXCHG8B.  Added #define
 * for ILLEGAL_OPCODE.
 *
 * Revision 1.9  1997/02/10 09:11:00  yoel
 * Curtesy of perl, I have now generated the tables for all of the opcodes
 * we decode.  Also removed the odd function pointer casting stuff since
 * stupid g++ can't cope with it :(  Now I need to put proto-types in
 * instruction.h and actually write the code that uses the tables ;)
 *
 * Revision 1.8  1997/02/04 02:36:16  sjt
 * Played around with messed up opcode tables...  Determined method pointers are
 * not simple numbers.  They tend to look like:
 *    (gdb) p Instruction::OpCodes[0]
 *    $2 = {__delta = 0, __index = -1, __pfn_or_delta2 = {__pfn = &ADD (
 *        Instruction &), __delta2 = 1}}
 * Solution is to use function numbers (defined by CPU, and deref'd by CPU), or
 *   use NULLs, and a second table for special cases.
 *
 */

//  decode.h		-*- C++ -*-
//
//  Contains the array which correlates opcodes to their corresponding
// function pointers.

#ifndef __DECODE_H__
#define __DECODE_H__

#include "instruction.h"

// Need cpu.h for operand function pointers
#include "cpu.h"

// Rip out bits 5,4,3 of the ModR/M byte for table lookups
#define GET_MOD_BITS(opcode)	((opcode>>3) & 0x7)

#define ILLEGAL_OPCODE	(0x1)
#define TWO_BYTE	(0x2)
#define FPU		(0x3)
#define IS_PREFIX	(0x4)
#define GRP1		(0x5)
#define GRP2		(0x6)
#define GRP3		(0x7)
#define GRP4		(0x8)
#define GRP5		(0x9)
#define GRP6		(0xA)
#define GRP7		(0xB)
#define GRP8		(0xC)
#define GRP9		(0xD)
#define MAX_ODD_OPCODE	GRP9

//  ========================================================================
//  NOTE:  Everything below is the initialization of static const class
// member of the class Instruction.  These are basically jump tables
// stolen directly from the Opcode Map in the Pentium book and are used
// for fast decoding of opcodes.

const InstEvalFunc Instruction::OneByteOpcodes[256] = {
// 0x00
  CPU::ADD,	CPU::ADD,	CPU::ADD,	CPU::ADD,
  CPU::ADD,	CPU::ADD,	CPU::PUSH,	CPU::POP,
  CPU::OR,	CPU::OR,	CPU::OR,	CPU::OR,
  CPU::OR,	CPU::OR,	CPU::PUSH,	NULL,
// 0x10
  CPU::ADC,	CPU::ADC,	CPU::ADC,	CPU::ADC,
  CPU::ADC,	CPU::ADC,	CPU::PUSH,	CPU::POP,
  CPU::SBB,	CPU::SBB,	CPU::SBB,	CPU::SBB,
  CPU::SBB,	CPU::SBB,	CPU::PUSH,	CPU::POP,
// 0x20
  CPU::AND,	CPU::AND,	CPU::AND,	CPU::AND,
  CPU::AND,	CPU::AND,	NULL,		CPU::DAA,
  CPU::SUB,	CPU::SUB,	CPU::SUB,	CPU::SUB,
  CPU::SUB,	CPU::SUB,	NULL,		CPU::DAS,
// 0x30
  CPU::XOR,	CPU::XOR,	CPU::XOR,	CPU::XOR,
  CPU::XOR,	CPU::XOR,	NULL,		CPU::AAA,
  CPU::CMP,	CPU::CMP,	CPU::CMP,	CPU::CMP,
  CPU::CMP,	CPU::CMP,	NULL,		CPU::AAS,
// 0x40
  CPU::INC,	CPU::INC,	CPU::INC,	CPU::INC,
  CPU::INC,	CPU::INC,	CPU::INC,	CPU::INC,	
  CPU::DEC,	CPU::DEC,	CPU::DEC,	CPU::DEC,
  CPU::DEC,	CPU::DEC,	CPU::DEC,	CPU::DEC,	
// 0x50
  CPU::PUSH,	CPU::PUSH,	CPU::PUSH,	CPU::PUSH,
  CPU::PUSH,	CPU::PUSH,	CPU::PUSH,	CPU::PUSH,
  CPU::POP,	CPU::POP,	CPU::POP,	CPU::POP,
  CPU::POP,	CPU::POP,	CPU::POP,	CPU::POP,
// 0x60
  CPU::PUSHA,	CPU::POPA,	CPU::BOUND,	CPU::ARPL,
  NULL,	 	NULL,	 	NULL,	 	NULL,	
  CPU::PUSH,	CPU::IMUL,	CPU::PUSH,	CPU::IMUL,
  CPU::INSB,	CPU::INSW,	CPU::OUTSB,	CPU::OUTSW,
// 0x70
  CPU::Jcc,	CPU::Jcc,	CPU::Jcc,	CPU::Jcc,
  CPU::Jcc,	CPU::Jcc,	CPU::Jcc,	CPU::Jcc,
  CPU::Jcc,	CPU::Jcc,	CPU::Jcc,	CPU::Jcc,
  CPU::Jcc,	CPU::Jcc,	CPU::Jcc,	CPU::Jcc,
// 0x80
  NULL,		NULL,		NULL,		NULL,	
  CPU::TEST,	CPU::TEST,	CPU::XCHG,	CPU::XCHG,
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,
  CPU::MOV,	CPU::LEA,	CPU::MOV,	CPU::POP,
// 0x90
  CPU::NOP,	CPU::XCHG,	CPU::XCHG,	CPU::XCHG,
  CPU::XCHG,	CPU::XCHG,	CPU::XCHG,	CPU::XCHG,
  CPU::CBW,	CPU::CWD,	CPU::CALL,	CPU::WAIT,
  CPU::PUSHF,	CPU::POPF,	CPU::SAHF,	CPU::LAHF,
// 0xA0  
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,
  CPU::MOVSB,	CPU::MOVSW,	CPU::CMPSB,	CPU::CMPSW,
  CPU::TEST,	CPU::TEST,	CPU::STOSB,	CPU::STOSW,
  CPU::LODSB,	CPU::LODSW,	CPU::SCASB,	CPU::SCASW,
// 0xB0
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,
  CPU::MOV,	CPU::MOV,	CPU::MOV,	CPU::MOV,	
// 0xC0
  NULL,		NULL,		CPU::RETnear,	CPU::RETnear,
  CPU::LES,	CPU::LDS,	CPU::MOV,	CPU::MOV,
  CPU::ENTER,	CPU::LEAVE,	CPU::RETfar,	CPU::RETfar,
  CPU::INT,	CPU::INT,	CPU::INTO,	CPU::IRET,
// 0xD0
  NULL,		NULL,		NULL,		NULL,	
  CPU::AAM,	CPU::AAD,	NULL,		CPU::XLAT,
  NULL,		NULL,		NULL,		NULL,	
  NULL,		NULL,		NULL,		NULL,	
// 0xE0
  CPU::LOOPNE,	CPU::LOOPE,	CPU::LOOP,	CPU::JCXZ,
  CPU::IN,	CPU::IN,	CPU::OUT,	CPU::OUT,
  CPU::CALL,	CPU::JMP,	CPU::JMP,	CPU::JMP,
  CPU::IN,	CPU::IN,	CPU::OUT,	CPU::OUT,
// 0xF0
  NULL,	 	NULL,	 	NULL,	 	NULL,	
  CPU::HLT,	CPU::CMC,	NULL,		NULL,	
  CPU::CLC,	CPU::STC,	CPU::CLI,	CPU::STI,
  CPU::CLD,	CPU::STD,	NULL,		NULL	
};

const int Instruction::BadFirstByte[256] = {
  0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,
  /*0F*/ TWO_BYTE,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,
  /*26*/ IS_PREFIX,		0,		0,		0,
  0,		0,		0,		0,
  /*2E*/ IS_PREFIX,		0,		0,		0,
  0,		0,		0,		0,
  /*36*/ IS_PREFIX,		0,		0,		0,
  0,		0,		0,		0,
  /*3E*/ IS_PREFIX,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,
  /*64*/ IS_PREFIX,		/*65*/ IS_PREFIX,
  /*66*/ IS_PREFIX,		/*67*/ IS_PREFIX,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		/*80*/ GRP1,
  /*81*/ GRP1,	/*82*/ GRP1,	/*83*/ GRP1,	0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		/*C0*/ GRP2,
  /*C1*/ GRP2,	0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  /*D0*/ GRP2,	/*D1*/ GRP2,	/*D2*/ GRP2,	/*D3*/ GRP2,	0,
  0,		0,		0,		/*D8*/ FPU,
  /*D9*/ FPU,	/*DA*/ FPU,	/*DB*/ FPU,	/*DC*/ FPU,
  /*DD*/ FPU,	/*DE*/ FPU,	/*DF*/ FPU,	0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,
  /*F0*/ IS_PREFIX,		0,		/*F2*/ IS_PREFIX,
  /*F3*/ IS_PREFIX,		0,		0,		/*F6*/ GRP3,
  /*F7*/ GRP3,	0,		0,		0,		0,
  0,		0,		/*FE*/ GRP4,	/*FF*/ GRP5,
};

const InstEvalFunc Instruction::TwoByteOpcodes[256] = {
  NULL,         NULL,		/*02*/ CPU::LAR,
  /*03*/ CPU::LSL,	NULL,		NULL,
  /*06*/ CPU::CLTS,		NULL,
  /*08*/ CPU::INVD,
  /*09*/ CPU::WBINVD,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  /*20*/ CPU::MOV,	/*21*/ CPU::MOV,
  /*22*/ CPU::MOV,	/*23*/ CPU::MOV,	NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		/*30*/ CPU::WRMSR,
  /*31*/ CPU::RDTSC,		/*32*/ CPU::RDMSR,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		/*80*/ CPU::Jcc,
  /*81*/ CPU::Jcc,	/*82*/ CPU::Jcc,
  /*83*/ CPU::Jcc,	/*84*/ CPU::Jcc,
  /*85*/ CPU::Jcc,	/*86*/ CPU::Jcc,
  /*87*/ CPU::Jcc,	/*88*/ CPU::Jcc,
  /*89*/ CPU::Jcc,	/*8A*/ CPU::Jcc,
  /*8B*/ CPU::Jcc,	/*8C*/ CPU::Jcc,
  /*8D*/ CPU::Jcc,	/*8E*/ CPU::Jcc,
  /*8F*/ CPU::Jcc,	/*90*/ CPU::SETcc,
  /*91*/ CPU::SETcc,		/*92*/ CPU::SETcc,
  /*93*/ CPU::SETcc,		/*94*/ CPU::SETcc,
  /*95*/ CPU::SETcc,
  /*96*/ CPU::SETcc,		/*97*/ CPU::SETcc,
  /*98*/ CPU::SETcc,		/*99*/ CPU::SETcc,
  /*9A*/ CPU::SETcc,		/*9B*/ CPU::SETcc,
  /*9C*/ CPU::SETcc,		/*9D*/ CPU::SETcc,
  /*9E*/ CPU::SETcc,		/*9F*/ CPU::SETcc,
  /*A0*/ CPU::PUSH,		/*A1*/ CPU::POP,
  /*A2*/ CPU::CPUID,		/*A3*/ CPU::BT,
  /*A4*/ CPU::SHLD,		/*A5*/ CPU::SHLD,
  NULL,		NULL,		/*A8*/ CPU::PUSH,
  /*A9*/ CPU::POP,	/*AA*/ CPU::RSM,
  /*AB*/ CPU::BTS,	/*AC*/ CPU::SHRD,
  /*AD*/ CPU::SHRD,		NULL,
  /*AF*/ CPU::IMUL,
  /*B0*/ CPU::CMPXCHG,
  /*B1*/ CPU::CMPXCHG,		/*B2*/ CPU::LSS,
  /*B3*/ CPU::BTR,	/*B4*/ CPU::LFS,
  /*B5*/ CPU::LGS,	/*B6*/ CPU::MOVZX,
  /*B7*/ CPU::MOVZX,		NULL,		NULL,
  NULL,		/*BB*/ CPU::BTC,
  /*BC*/ CPU::BSF,	/*BD*/ CPU::BSR,
  /*BE*/ CPU::MOVSX,		/*BF*/ CPU::MOVSX,
  /*C0*/ CPU::XADD,		/*C1*/ CPU::XADD,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		/*C8*/ CPU::BSWAP,
  /*C9*/ CPU::BSWAP,		/*CA*/ CPU::BSWAP,
  /*CB*/ CPU::BSWAP,		/*CC*/ CPU::BSWAP,
  /*CD*/ CPU::BSWAP,		/*CE*/ CPU::BSWAP,
  /*CF*/ CPU::BSWAP,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,
};

const int Instruction::BadSecondByte[256] = {
  /*00*/ GRP6,	/*01*/ GRP7,	0,		0,		0,
  0,		0,		0,		0,		0,
  0,		/*0B*/ ILLEGAL_OPCODE,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		/*B9*/ ILLEGAL_OPCODE,		/*BA*/ GRP8,	0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		/*C7*/ GRP9,	0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,		0,		0,
  0,		0,		0,
};

const InstEvalFunc Instruction::Grp1Codes[8] = {
  CPU::ADD,	CPU::OR,	CPU::ADC,	CPU::SBB,
  CPU::AND,	CPU::SUB,	CPU::XOR,	CPU::CMP
};

const InstEvalFunc Instruction::Grp2Codes[8] = {
  CPU::ROL,	CPU::ROR,	CPU::RCL,	CPU::RCR,
  CPU::SHL,	CPU::SHR,	NULL,		CPU::SAR
};

const InstEvalFunc Instruction::Grp3Codes[8] = {
  CPU::TEST,	NULL,		CPU::NOT,	CPU::NEG,
  CPU::MUL,	CPU::IMUL,	CPU::DIV,	CPU::IDIV
};

const InstEvalFunc Instruction::Grp4Codes[8] = {
  CPU::INC,	CPU::DEC,	NULL,		NULL,	
  NULL,	 	NULL,		NULL,		NULL,	
};

const InstEvalFunc Instruction::Grp5Codes[8] = {
  CPU::INC,	CPU::DEC,	CPU::CALL,	CPU::CALL,
  CPU::JMP,	CPU::JMP,	CPU::PUSH,	NULL	
};

const InstEvalFunc Instruction::Grp6Codes[8] = {
  CPU::SLDT,	CPU::STR,	CPU::LLDT,	CPU::LTR,
  CPU::VERR,	CPU::VERW,	NULL,		NULL
};

const InstEvalFunc Instruction::Grp7Codes[8] = {
  CPU::SGDT,	CPU::SIDT,	CPU::LGDT,	CPU::LIDT,
  CPU::SMSW,	NULL,		CPU::LMSW,	CPU::INVLPG
};

const InstEvalFunc Instruction::Grp8Codes[8] = {
  NULL,		NULL,		NULL,		NULL,
  CPU::BT,	CPU::BTS,	CPU::BTR,	CPU::BTC
};

const InstEvalFunc Instruction::Grp9Codes[8] = {
  NULL,		CPU::CMPXCHG8B,	NULL,		NULL,
  NULL,		NULL,		NULL,		NULL
};

const DecodeOperandDecl Instruction::OneByteOperand[256] = {
  /*00*/ Instruction::EB_GB,	/*01*/ Instruction::EV_GV,
  /*02*/ Instruction::GB_EB,	/*03*/ Instruction::GV_EV,
  /*04*/ Instruction::AL_IB,	/*05*/ Instruction::EAX_IV,
  /*06*/ Instruction::ES,	/*07*/ Instruction::ES,
  /*08*/ Instruction::EB_GB,	/*09*/ Instruction::EV_GV,
  /*0A*/ Instruction::GB_EB,	/*0B*/ Instruction::GV_EV,
  /*0C*/ Instruction::AL_IB,	/*0D*/ Instruction::EAX_IV,
  /*0E*/ Instruction::CS,	NULL,
  /*10*/ Instruction::EB_GB,	/*11*/ Instruction::EV_GV,
  /*12*/ Instruction::GB_EB,	/*13*/ Instruction::GV_EV,
  /*14*/ Instruction::AL_IB,	/*15*/ Instruction::EAX_IV,
  /*16*/ Instruction::SS,	/*17*/ Instruction::SS,
  /*18*/ Instruction::EB_GB,	/*19*/ Instruction::EV_GV,
  /*1A*/ Instruction::GB_EB,	/*1B*/ Instruction::GV_EV,
  /*1C*/ Instruction::AL_IB,	/*1D*/ Instruction::EAX_IV,
  /*1E*/ Instruction::DS,	/*1F*/ Instruction::DS,
  /*20*/ Instruction::EB_GB,	/*21*/ Instruction::EV_GV,
  /*22*/ Instruction::GB_EB,	/*23*/ Instruction::GV_EV,
  /*24*/ Instruction::AL_IB,	/*25*/ Instruction::EAX_IV,
  NULL,	NULL,
  /*28*/ Instruction::EB_GB,	/*29*/ Instruction::EV_GV,
  /*2A*/ Instruction::GB_EB,	/*2B*/ Instruction::GV_EV,
  /*2C*/ Instruction::AL_IB,	/*2D*/ Instruction::EAX_IV,
  NULL,	NULL,
  /*30*/ Instruction::EB_GB,	/*31*/ Instruction::EV_GV,
  /*32*/ Instruction::GB_EB,	/*33*/ Instruction::GV_EV,
  /*34*/ Instruction::AL_IB,	/*35*/ Instruction::EAX_IV,
  NULL,	NULL,
  /*38*/ Instruction::EB_GB,	/*39*/ Instruction::EV_GV,
  /*3A*/ Instruction::GB_EB,	/*3B*/ Instruction::GV_EV,
  /*3C*/ Instruction::AL_IB,	/*3D*/ Instruction::EAX_IV,
  NULL,	NULL,
  /*40*/ Instruction::EAX,	/*41*/ Instruction::ECX,
  /*42*/ Instruction::EDX,	/*43*/ Instruction::EBX,
  /*44*/ Instruction::ESP,	/*45*/ Instruction::EBP,
  /*46*/ Instruction::ESI,	/*47*/ Instruction::EDI,
  /*48*/ Instruction::EAX,	/*49*/ Instruction::ECX,
  /*4A*/ Instruction::EDX,	/*4B*/ Instruction::EBX,
  /*4C*/ Instruction::ESP,	/*4D*/ Instruction::EBP,
  /*4E*/ Instruction::ESI,	/*4F*/ Instruction::EDI,
  /*50*/ Instruction::EAX,	/*51*/ Instruction::ECX,
  /*52*/ Instruction::EDX,	/*53*/ Instruction::EBX,
  /*54*/ Instruction::ESP,	/*55*/ Instruction::EBP,
  /*56*/ Instruction::ESI,	/*57*/ Instruction::EDI,
  /*58*/ Instruction::EAX,	/*59*/ Instruction::ECX,
  /*5A*/ Instruction::EDX,	/*5B*/ Instruction::EBX,
  /*5C*/ Instruction::ESP,	/*5D*/ Instruction::EBP,
  /*5E*/ Instruction::ESI,	/*5F*/ Instruction::EDI,
  NULL,	NULL,
  /*62*/ Instruction::GV_MA,	/*63*/ Instruction::EW_GW,	NULL,
  NULL,		NULL,		NULL,		/*68*/ Instruction::IV,
  /*69*/ Instruction::GV_EV_IV,	/*6A*/ Instruction::IB,
  /*6B*/ Instruction::GV_EV_IB,	/*6C*/ Instruction::YB_DX,
  /*6D*/ Instruction::YV_DX,	/*6E*/ Instruction::DX_XB,
  /*6F*/ Instruction::DX_XV,	NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		/*80*/ Instruction::EB_IB,
  /*81*/ Instruction::EV_IV,	/*82*/ Instruction::EB_IB,
  /*83*/ Instruction::EV_IB,	/*84*/ Instruction::EB_GB,
  /*85*/ Instruction::EV_GV,	/*86*/ Instruction::EB_GB,
  /*87*/ Instruction::EV_GV,	/*88*/ Instruction::EB_GB,
  /*89*/ Instruction::EV_GV,	/*8A*/ Instruction::GB_EB,
  /*8B*/ Instruction::GV_EV,	/*8C*/ Instruction::EW_SW,
  /*8D*/ Instruction::GV_M,	/*8E*/ Instruction::SW_EW,
  /*8F*/ Instruction::EV,	NULL,
  /*91*/ Instruction::EAX_ECX,	/*92*/ Instruction::EAX_EDX,
  /*93*/ Instruction::EAX_EBX,	/*94*/ Instruction::EAX_ESP,
  /*95*/ Instruction::EAX_EBP,	/*96*/ Instruction::EAX_ESI,
  /*97*/ Instruction::EAX_EDI,	NULL,
  NULL,	/*9A*/ Instruction::AP,
  NULL,	/*9C*/ Instruction::FV,
  /*9D*/ Instruction::FV,	NULL,
  NULL,	/*A0*/ Instruction::AL_OB,
  /*A1*/ Instruction::EAX_OV,	/*A2*/ Instruction::OB_AL,
  /*A3*/ Instruction::OV_EAX,	/*A4*/ Instruction::XB_YB,
  /*A5*/ Instruction::XV_YV,	/*A6*/ Instruction::XB_YB,
  /*A7*/ Instruction::XV_YV,	/*A8*/ Instruction::AL_IB,
  /*A9*/ Instruction::EAX_IV,	/*AA*/ Instruction::YB_AL,
  /*AB*/ Instruction::YV_EAX,	/*AC*/ Instruction::AL_XB,
  /*AD*/ Instruction::EAX_XV,	/*AE*/ Instruction::AL_YB,
  /*AF*/ Instruction::EAX_YV,	/*B0*/ Instruction::AL_IB,
  /*B1*/ Instruction::CL_IB,	/*B2*/ Instruction::DL_IB,
  /*B3*/ Instruction::BL_IB,	/*B4*/ Instruction::AH_IB,
  /*B5*/ Instruction::CH_IB,	/*B6*/ Instruction::DH_IB,
  /*B7*/ Instruction::BH_IB,	/*B8*/ Instruction::EAX_IV,
  /*B9*/ Instruction::ECX_IV,	/*BA*/ Instruction::EDX_IV,
  /*BB*/ Instruction::EBX_IV,	/*BC*/ Instruction::ESP_IV,
  /*BD*/ Instruction::EBP_IV,	/*BE*/ Instruction::ESI_IV,
  /*BF*/ Instruction::EDI_IV,	/*C0*/ Instruction::EB_IB,
  /*C1*/ Instruction::EV_IB,	/*C2*/ Instruction::IW,
  NULL,	/*C4*/ Instruction::GV_MP,
  /*C5*/ Instruction::GV_MP,	/*C6*/ Instruction::EB_IB,
  /*C7*/ Instruction::EV_IV,	/*C8*/ Instruction::IW_IB,
  NULL,	/*CA*/ Instruction::IW,
  NULL,	NULL,		/*CD*/ Instruction::IB,
  NULL,		NULL,		/*D0*/ Instruction::EB,
  /*D1*/ Instruction::EV,	/*D2*/ Instruction::EB_CL,
  /*D3*/ Instruction::EV_CL,	NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,
  /*E0*/ Instruction::JB,	/*E1*/ Instruction::JB,
  /*E2*/ Instruction::JB,	/*E3*/ Instruction::JB,
  /*E4*/ Instruction::AL_IB,	/*E5*/ Instruction::EAX_IB,
  /*E6*/ Instruction::IB_AL,	/*E7*/ Instruction::IB_EAX,
  /*E8*/ Instruction::JV,	/*E9*/ Instruction::JV,
  /*EA*/ Instruction::AP,	/*EB*/ Instruction::JB,
  /*EC*/ Instruction::AL_DX,	/*ED*/ Instruction::EAX_DX,
  /*EE*/ Instruction::DX_AL,	/*EF*/ Instruction::DX_EAX,	NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  /*FE*/ Instruction::EB,       NULL,
};

const DecodeOperandDecl Instruction::TwoByteOperand[256] = {
  NULL,         NULL,		/*02*/ Instruction::GV_EW,
  /*03*/ Instruction::GV_EW,	NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  /*20*/ Instruction::RD_CD,	/*21*/ Instruction::RD_DD,
  /*22*/ Instruction::CD_RD,	/*23*/ Instruction::DD_RD,	NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,         NULL,
  /*A0*/ Instruction::FS,	/*A1*/ Instruction::FS,		NULL,
  /*A3*/ Instruction::EV_GV,	/*A4*/ Instruction::EV_GV_IB,
  /*A5*/ Instruction::EV_GV_CL,	NULL,		NULL,
  /*A8*/ Instruction::GS,	/*A9*/ Instruction::GS,		NULL,
  /*AB*/ Instruction::EV_GV,	/*AC*/ Instruction::EV_GV_IB,
  /*AD*/ Instruction::EV_GV_CL,	NULL,		/*AF*/ Instruction::GV_GV_EV,
  /*B0*/ Instruction::EB_GB,	/*B1*/ Instruction::EV_GV,
  /*B2*/ Instruction::GV_MP,	/*B3*/ Instruction::EV_GV,
  /*B4*/ Instruction::GV_MP,	/*B5*/ Instruction::GV_MP,
  /*B6*/ Instruction::GV_EB,	/*B7*/ Instruction::GV_EW,	NULL,
  NULL,		/*BA*/ Instruction::EV_IB, /*BB*/ Instruction::EV_GV,
  /*BC*/ Instruction::GV_EV,	/*BD*/ Instruction::GV_EV,
  /*BE*/ Instruction::GV_EB,	/*BF*/ Instruction::GV_EW,
  /*C0*/ Instruction::EB_GB,	/*C1*/ Instruction::EV_GV,	NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  /*C8*/ Instruction::EAX,	/*C9*/ Instruction::ECX,
  /*CA*/ Instruction::EDX,	/*CB*/ Instruction::EBX,
  /*CC*/ Instruction::ESP,	/*CD*/ Instruction::EBP,
  /*CE*/ Instruction::ESI,	/*CF*/ Instruction::EDI,	NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,		NULL,		NULL,		NULL,
  NULL,		NULL,
};

const DecodeOperandDecl Instruction::Grp3OperandsF6[8] = {
  Instruction::EB_IB, 	NULL,	Instruction::EB,	Instruction::EB,
  Instruction::AX_AL_EB,	Instruction::AX_AL_EB,	
  Instruction::EB_AL,           Instruction::EB_AL
};

const DecodeOperandDecl Instruction::Grp3OperandsF7[8] = {
  Instruction::EV_IV, 	NULL,	Instruction::EV,	Instruction::EV,
  Instruction::EDXEAX_EAX_EV,	Instruction::EDXEAX_EAX_EV,	
  Instruction::EV_EAX,          Instruction::EV_EAX
};

const DecodeOperandDecl Instruction::Grp5Operands[8] = {
  Instruction::EV, 	Instruction::EV,	Instruction::EV,
  Instruction::EP,	Instruction::EV,	Instruction::EP,
  Instruction::EV,	NULL
};

const DecodeOperandDecl Instruction::Grp7Operands[8] = {
  Instruction::MS,	Instruction::MS,	Instruction::MS,
  Instruction::MS,	Instruction::EW,	NULL,
  Instruction::EW,	NULL
};

#endif
