/* $Id: fpu.cc,v 1.4 1998/04/29 04:54:50 nikos Exp $
 *
 * RCS Log:
 * $Log: fpu.cc,v $
 * Revision 1.4  1998/04/29 04:54:50  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:40  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:26  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:36:59  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:19:59  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: fpu.cc,v 1.4 1998/04/29 04:54:50 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/fpu.cc,v 1.4 1998/04/29 04:54:50 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:26  bugnion
// stanford update (x86)
//
// Revision 1.6  1997/12/14  02:34:54  mendel
// Fixed two bugs in fpu code:
//
// 1) The FUCOMPP instruction was in the wrong place in the decode table.
//
// 2) The fake FPU code didn't handle instructions with prefixes.
//
// I patched around the problem 2. We need a real fix.
//
 * Revision 1.5  1997/12/09 05:33:18  mendel
 * Bugs patched around (These need to be fixed):
 *    1) ifdef'ed out the protection checking in LoadDescriptor() that
 *       was preventing IRET from working. See comment in code I added in
 *       cpu.cc
 *    2) ifdefed out the ASSERT(0) in the TLB flush routine in x86_instr.cc
 *    3) ifdef'ed out the protection checking in IRET that wasn't working.
 *
 *
 * Bugs:
 *
 * 1) Check in some bug fixes in fpu.cc that Sean gave me.
 *
 * 2) Fixed bug in reading stack segment selector in TRAP_GATE . We were
 * ASSERT(0)ing when the ReadFromVirtualAddress() was successful! in cpu.cc
 * HandleExecption()
 *
 * 3) Changed (CSDescriptor.dpl() != SSSelector & 0x3) to
 *            (CSDescriptor.dpl() != (SSSelector & 0x3))
 * for obvious reasons in cpu.cc / HandleException().
 *
 * 4) Test for GDT in TRAP_GATE code was backward in cpu.cc /
 * HandleException()
 *
 * 5) Since the prev_eip field of CPU class is used to set the return PC
 * of execeptions. We need to set prev_eip on instruction page faults to
 * be the faulting instruction. Otherwise we think the exception happen
 * on the PC of IRET instruction. ouch!
 *
 * 6) The MemoryOperand() code in instruction.cc wasn't setting the
 * default segment of the ESP register to be the stack segment! Found
 * this because Linux runs for a while with the user's DS selector
 * loaded and pushes things on the stack.
 *
 * 7) The is_task() method of seg_desc was completely bogus.
 *
 *
 * Clean up:
 *
 * 1) Removed fprintf() on exceptions in cpu.cc since we are taking far too many
 * execptions for this to be useful.
 *
 * Revision 1.4  1997/12/02 04:49:21  yoel
 *   Support added for taking Software Interupts and preparation for supporting
 * external interupts.  cpu.{cc,h} updated so that current_exception is an
 * int8 instead of the enum to allow exceptions 19-255 and enum ExceptionSource
 * added.  Cosmetic changes to fpu.cc and INT n / INTO now actually do
 * something (since we now support Software Interupts).
 *
 * Revision 1.3  1997/11/22 07:45:04  whimsy
 * Fixed bug in GET_OP_BITS.  It should have looked at the the bottom 6 bits
 * rather than only 3.
 *
 * Revision 1.2  1997-03-18 11:07:34-08  sean
 * added check to make sure 'firstByte' was in range - fixed type in array
 *   accesses to make sure 0xD8-0xDF mapped to 0-7
 *
 * Revision 1.1  1997/03/14 08:03:34  yoel
 * Initial revision
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: fpu.cc,v 1.4 1998/04/29 04:54:50 nikos Exp $";

//   File fpu.cc, Author Jeremy Sugerman (yoel@cs.stanford.edu)
//
// This file is devoted to operand and opcode decoding of x86 bytecodes for
// floating point calls.

#include "instruction.h"
#include "data_stream.h"
#include "fpu.h"
#include "debug.h"

// Isn't the x86 architecture great?  Depending on whether the modRM is below
//   0xC0 or above, we either need bits 5 through 3 or bits 5 through 0.
#define GET_BELOW_C0_OP_BITS(modRM) (((modRM) >> 3) & 0x7)
#define GET_ABOVE_C0_OP_BITS(modRM) ((modRM) & 0x3f)

Instruction *Instruction::DecodeFPUInstruction(DataStream& instr_data, 
                                               int default_datasize, 
				   	       Prefix& prefix, int8 firstByte) 
{
  int8 modRM;
  Operand *ops[3] = { NULL, NULL, NULL };
  InstEvalFunc iFunction;
  DecodeOperandDecl operandFn;
  int startPos = instr_data.MaxBytesRead();

  // We should only be called with a firstByte between 0xD8 and 0xDF
  ASSERT((firstByte >= 0xD8) && (firstByte <= 0xDF));
  
  // All FPU instructions have a modR/M Byte and need it to decode.
  if (!instr_data.ReadOneByte(modRM)) return NULL;

  //  Look up the Eval Functions and the Operand Decoder
  if (modRM < 0xC0) {
    if ((iFunction = FPU_Low[firstByte - 0xD8][GET_BELOW_C0_OP_BITS(modRM)]) == NULL)
      return NULL;
    operandFn = FPU_LowOperand[firstByte - 0xD8][GET_BELOW_C0_OP_BITS(modRM)]; 
  } else {
    if ((iFunction = FPU_High[firstByte - 0xD8][GET_ABOVE_C0_OP_BITS(modRM)]) == NULL)
      return NULL;
    operandFn = FPU_HighOperand[firstByte - 0xD8][GET_ABOVE_C0_OP_BITS(modRM)]; 
  }

  // Decode Operands if necessary
  if (operandFn != NULL) {
    // In order to share functions with the non-FPU, we must put back modRM
    if (!instr_data.UnReadOneByte(modRM)) return NULL;

    if (!operandFn(instr_data, prefix.OpSize(), 
                                   prefix.AddrSize(), prefix.SegReg(), ops))
       return NULL;
  }

  //  If we've made it this far, we have a complete instruction.  The
  // legnth is the number of new bytes we read for operands plus the one
  // for the Dx.
  return (new Instruction(iFunction, instr_data.MaxBytesRead(),
                          prefix, ops[0], ops[1], ops[2]));

}
