/* -*- C++ -*-
 * $Id: fpu.h,v 1.4 1998/04/29 04:54:51 nikos Exp $
 * 
 * RCS Log:
 * $Log: fpu.h,v $
 * Revision 1.4  1998/04/29 04:54:51  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:41  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:26  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:00  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:19:59  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: fpu.h,v 1.4 1998/04/29 04:54:51 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/fpu.h,v 1.4 1998/04/29 04:54:51 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:26  bugnion
 * stanford update (x86)
 *
 * Revision 1.3  1997/12/14  02:34:56  mendel
 * Fixed two bugs in fpu code:
 *
 * 1) The FUCOMPP instruction was in the wrong place in the decode table.
 *
 * 2) The fake FPU code didn't handle instructions with prefixes.
 *
 * I patched around the problem 2. We need a real fix.
 *
 * Revision 1.2  1997/12/07  02:54:09  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.1  1997/03/14 08:03:34  yoel
 * Initial revision
 *
 */

//  File fpu.h, Author Jeremy Sugerman (yoel@cs.stanford.edu)
//
// Tables for decoding FPU Opcodes and Operands.

#ifndef __FPU_H__
#define __FPU_H__

#include "instruction.h"
#include "cpu.h"

#ifndef NULL
#define NULL  0x0
#endif

//  The ones where the ModR/M doesn't has opcode bytes:
const InstEvalFunc Instruction::FPU_Low[8][8] = {
// 0xD8
  {  CPU::FADD, CPU::FMUL, CPU::FCOM, CPU::FCOMP, CPU::FSUB, CPU::FSUBR,
     CPU::FDIV, CPU::FDIVR },
// 0xD9
  {  CPU::FLD, NULL, CPU::FST, CPU::FSTP, CPU::FLDENV, CPU::FLDCW,
     CPU::FSTENV, CPU::FSTCW },
// 0xDA
  {  CPU::FIADD, CPU::FIMUL, CPU::FICOM, CPU::FICOMP, CPU::FISUB,
     CPU::FISUBR, CPU::FIDIV, CPU::FIDIVR },
// 0xDB
  {  CPU::FILD, NULL, CPU::FIST, CPU::FISTP, NULL, CPU::FLD, NULL, CPU::FSTP },
//0xDC
  {  CPU::FADD, CPU::FMUL, CPU::FCOM, CPU::FCOMP, CPU::FSUB, CPU::FSUBR,
     CPU::FDIV, CPU::FDIVR },
// 0xDD
  {  CPU::FLD, NULL, CPU::FST, CPU::FSTP, CPU::FRSTOR, NULL, CPU::FSAVE,
     CPU::FSTSW },
// 0xDE
  {  CPU::FIADD, CPU::FIMUL, CPU::FICOM, CPU::FICOMP, CPU::FISUB,
     CPU::FISUBR, CPU::FIDIV, CPU::FIDIVR },
// 0xDF
  {  CPU::FILD, NULL, CPU::FIST, CPU::FISTP, CPU::FBLD, CPU::FILD,
     CPU::FBSTP, CPU::FISTP }
};

// The ones where the opcode is set by the nnn bits in ModR/M:
const InstEvalFunc Instruction::FPU_High[8][64] = {
// 0xD8
  { CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD,
    CPU::FADD, CPU::FADD, CPU::FMUL, CPU::FMUL, CPU::FMUL, CPU::FMUL,
    CPU::FMUL, CPU::FMUL, CPU::FMUL, CPU::FMUL, CPU::FCOM, CPU::FCOM, 
    CPU::FCOM, CPU::FCOM, CPU::FCOM, CPU::FCOM, CPU::FCOM, CPU::FCOM,
    CPU::FCOMP, CPU::FCOMP, CPU::FCOMP, CPU::FCOMP, CPU::FCOMP, CPU::FCOMP,
    CPU::FCOMP, CPU::FCOMP, CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FSUB, 
    CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FSUBR,  CPU::FSUBR,
    CPU::FSUBR, CPU::FSUBR, CPU::FSUBR, CPU::FSUBR, CPU::FSUBR,
    CPU::FSUBR, CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIV,
    CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR,
    CPU::FDIVR, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR 
  },
// 0xD9
  { CPU::FLD, CPU::FLD, CPU::FLD, CPU::FLD, CPU::FLD, CPU::FLD, CPU::FLD,
    CPU::FLD, CPU::FXCH, CPU::FXCH, CPU::FXCH, CPU::FXCH, CPU::FXCH,
    CPU::FXCH, CPU::FXCH, CPU::FXCH, CPU::FNOP, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    CPU::FCHS, CPU::FABS, NULL, NULL, CPU::FTST, CPU::FXAM, NULL, NULL,
    CPU::FLD1, CPU::FLDL2T, CPU::FLDL2E, CPU::FLDPI, CPU::FLDLG2,
    CPU::FLDLN2, CPU::FLDZ, NULL, CPU::F2XM1, CPU::FYL2X, CPU::FPTAN,
    CPU::FPATAN, CPU::FXTRACT, CPU::FPREM1, CPU::FDECSTP, CPU::FINCSTP,
    CPU::FPREM, CPU::FYL2XP1, CPU::FSQRT, CPU::FSINCOS, CPU::FRNDINT,
    CPU::FSCALE, CPU::FSIN, CPU::FCOS
   },
// 0xDA
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, CPU::FUCOMPP, NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL
  },
// 0xDB
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, CPU::FCLEX,
    CPU::FINIT, 
    CPU::NOP /* Really FSETPM, but only supported on the 80287, but apparently
		tolerated on other FPUs - we'll just NOP */, 
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL 
  },
// 0xDC
  { CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD, CPU::FADD,
    CPU::FADD, CPU::FADD, CPU::FMUL, CPU::FMUL, CPU::FMUL, CPU::FMUL,
    CPU::FMUL, CPU::FMUL, CPU::FMUL, CPU::FMUL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    CPU::FSUBR, CPU::FSUBR, CPU::FSUBR, CPU::FSUBR, CPU::FSUBR, CPU::FSUBR,
    CPU::FSUBR, CPU::FSUBR, CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FSUB,
    CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FSUB, CPU::FDIVR, CPU::FDIVR,
    CPU::FDIVR, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR, CPU::FDIVR,
    CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIV, CPU::FDIV,
    CPU::FDIV, CPU::FDIV
  },
// 0xDD
  { CPU::FFREE, CPU::FFREE, CPU::FFREE, CPU::FFREE, CPU::FFREE, CPU::FFREE,
    CPU::FFREE, CPU::FFREE, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    CPU::FST, CPU::FST, CPU::FST, CPU::FST, CPU::FST, CPU::FST, CPU::FST,
    CPU::FST, CPU::FSTP, CPU::FSTP, CPU::FSTP, CPU::FSTP, CPU::FSTP,
    CPU::FSTP, CPU::FSTP, CPU::FSTP, CPU::FUCOM, CPU::FUCOM, CPU::FUCOM,
    CPU::FUCOM, CPU::FUCOM, CPU::FUCOM, CPU::FUCOM, CPU::FUCOM,
    CPU::FUCOMP, CPU::FUCOMP, CPU::FUCOMP, CPU::FUCOMP, CPU::FUCOMP,
    CPU::FUCOMP, CPU::FUCOMP, CPU::FUCOMP, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
  },
// 0xDE
  { CPU::FADDP, CPU::FADDP, CPU::FADDP, CPU::FADDP, CPU::FADDP, CPU::FADDP,
    CPU::FADDP, CPU::FADDP, CPU::FMULP, CPU::FMULP, CPU::FMULP, CPU::FMULP,
    CPU::FMULP, CPU::FMULP, CPU::FMULP, CPU::FMULP, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, CPU::FCOMPP, NULL, NULL, NULL, NULL,
    NULL, NULL, CPU::FSUBRP, CPU::FSUBRP, CPU::FSUBRP, CPU::FSUBRP,
    CPU::FSUBRP, CPU::FSUBRP, CPU::FSUBRP, CPU::FSUBRP, CPU::FSUBP,
    CPU::FSUBP, CPU::FSUBP, CPU::FSUBP, CPU::FSUBP, CPU::FSUBP, CPU::FSUBP,
    CPU::FSUBP, CPU::FDIVRP, CPU::FDIVRP, CPU::FDIVRP, CPU::FDIVRP,
    CPU::FDIVRP, CPU::FDIVRP, CPU::FDIVRP, CPU::FDIVRP, CPU::FDIVP,
    CPU::FDIVP, CPU::FDIVP, CPU::FDIVP, CPU::FDIVP, CPU::FDIVP, CPU::FDIVP,
    CPU::FDIVP
  },
// 0xDF
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, CPU::FSTSW, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL
  }
};

//  Operand Decoding if ModRM < C0h
const DecodeOperandDecl Instruction::FPU_LowOperand[8][8] = {
  // 0xD8
  { Instruction::MD, Instruction::MD, Instruction::MD, Instruction::MD,
    Instruction::MD, Instruction::MD, Instruction::MD, Instruction::MD
  },

  // 0xD9
  { Instruction::MD, NULL, Instruction::MD, Instruction::MD,
    Instruction::MENV, Instruction::MW, Instruction::MENV, Instruction::MW
  },

  // 0xDA
  { Instruction::MD, Instruction::MD, Instruction::MD, Instruction::MD,
    Instruction::MD, Instruction::MD, Instruction::MD, Instruction::MD
  },

  // 0xDB
  { Instruction::MD, NULL, Instruction::MD, Instruction::MD, NULL,
    Instruction::M80, NULL, Instruction::M80
  },

  // 0xDC
  { Instruction::MQ, Instruction::MQ, Instruction::MQ, Instruction::MQ,
    Instruction::MQ, Instruction::MQ, Instruction::MQ, Instruction::MQ
  },

  // 0xDD
  { Instruction::MQ, NULL, Instruction::MQ, Instruction::MQ,
    Instruction::MSTATE, NULL, Instruction::MSTATE, Instruction::MW
  },

  // 0xDE
  { Instruction::MW, Instruction::MW, Instruction::MW, Instruction::MW,
    Instruction::MW, Instruction::MW, Instruction::MW, Instruction::MW 
  },

  // 0xDF
  { Instruction::MW, NULL, Instruction::MW, Instruction::MW,
    Instruction::M80, Instruction::MQ, Instruction::M80, Instruction::MQ
  }
};

// Operand Decoding if ModRm > BFh
const DecodeOperandDecl Instruction::FPU_HighOperand[8][64] = {
  // 0xD8
  { Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N
  },

  // 0xD9
  { Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    Instruction::ST_N, Instruction::ST_N, Instruction::ST_N, Instruction::ST_N,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  },

  // 0xDA
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL
  },

  // 0xDB
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL
  },

  // 0xDC
  { Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST
  },

  // 0xDD
  { Instruction::N, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, Instruction::N, Instruction::N, Instruction::N,
    Instruction::N, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL
  },

  // 0xDE
  { Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST, Instruction::N_ST, Instruction::N_ST,
    Instruction::N_ST 
  },

  // 0xDF
  { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, Instruction::AX, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL
  }
};

#endif

