/* $Id: fpu_instr.cc,v 1.4 1998/04/29 04:54:51 nikos Exp $
 *
 * RCS Log:
 * $Log: fpu_instr.cc,v $
 * Revision 1.4  1998/04/29 04:54:51  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:42  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:27  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:02  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:00  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: fpu_instr.cc,v 1.4 1998/04/29 04:54:51 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/fpu_instr.cc,v 1.4 1998/04/29 04:54:51 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:26  bugnion
// stanford update (x86)
//
// Revision 1.8  1997/12/14  02:35:00  mendel
// Fixed two bugs in fpu code:
//
// 1) The FUCOMPP instruction was in the wrong place in the decode table.
//
// 2) The fake FPU code didn't handle instructions with prefixes.
//
// I patched around the problem 2. We need a real fix.
//
 * Revision 1.7  1997/12/09 14:51:47  sjt
 * FPU emulation should now work completely - FPU exceptions are triggered when
 * 	they should be, and memory operands appear to work fine.  I removed the
 * 	printfs that were cluttering everything too.
 *
 * Mendel's fixes to BT/BTC/BTR/BTS and ADC have been stolen and checked in as
 * 	well.
 *
 * Revision 1.6  1997/12/07 02:54:12  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.5  1997/10/29 22:46:50  sean
 * Added simulation (instead of just ignoring) of FPU ops.
 *
 * Revision 1.4  1997/09/27 18:31:45  sean
 * added END_DEFNs to make code cleaner - hacked FSTCW to always store 0x37F,
 *   to allow lockstep to compare ok
 *
 * Revision 1.3  1997/09/24 08:15:47  sean
 * removed left-braces for new, string-giving, operation header (see
 *  x86_instr.h)
 *
 * Revision 1.2  1997/03/18 19:09:05  sean
 * changed ASSERT-failing behavior of all FPU instructions into NOPs because
 *   FPU instructions have a nasty habit of popping up in weird places (like
 *   libc startup code)
 *
 * Revision 1.1  1997/03/09 21:19:26  sean
 * Initial revision
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: fpu_instr.cc,v 1.4 1998/04/29 04:54:51 nikos Exp $";

#include "cpu.h"
#include "instruction.h"
#include "debug.h"
#include "simlib.h"
#include "counter.h"

//  For now, FPU instructions are NOPs (we don't assert fail because there
//   are FPU ops in libc startup code, and we'd like to be able to simulate
//   something at least.
#define NO_FPU_YET                                                         \
  DEBUG(DEBUG_FPU,"FPU instructions are not implemented yet - ignored.");  \
  return(0);

//  This lets us fake FPU ops.
#define FAKE_FPU_OP                     \
  if(CheckForFPUException()) return(0); \
  SimulateFPUonRealFPU(inst); return(1)

//  This lets us fake FPU ops, but doesn't check for exceptions - certain
//   control instructions have this behavior
#define FAKE_FPU_OP_NO_CHECK                     \
  SimulateFPUonRealFPU(inst); return(1)

// more cheating with macros: quick way to get at operands
#define OP(num)                  inst.GetOperand(num)
#define GET_OP(num)              (inst.GetOperand(num)->FetchValue(*this))
#define SET_OP(num,val)          inst.GetOperand(num)->WriteValue(*this,val)
#define OP_SIZE(num)             inst.GetOperand(num)->OperandSize()

// GET_OP was obvious; this one is the same, except zero extended
#define GET_OP_Z(num)            (inst.GetOperand(num)->FetchValue(*this,0))

// These three allow an extra displacement too
#define GET_OP_DISP(num,disp)    (inst.GetOperand(num)->FetchValue(*this,1,disp))
#define GET_OP_Z_DISP(num,disp)  (inst.GetOperand(num)->FetchValue(*this,0,disp))
#define SET_OP_DISP(num,val,disp) inst.GetOperand(num)->WriteValue(*this,val,disp)

bool CPU::CheckForFPUException(void)
{
  // first, if we haven't had any fpu ops yet, no exceptions.
  if(!fpu_state) return(false);

  // if the ES bit (bit 7) of the status word is set, we have an unmasked
  //  exception, and we need to fire an exception
  if(BITS(((int32 *)fpu_state)[1], 7, 1) == 1) {
    // Technically, we should check the NE bit of CR0 to see if we raise
    //  exception 16, or some evil external interrupt that essentially halts
    //  the CPU.

    //    fprintf(stderr, "FPU Exception at 0x%08lx!\r\n", prev_eip);
    ExceptionOccurred(EXC_FPU_ERROR);
    return(true);
  }

  return(false);
}

void CPU::SimulateFPUonRealFPU(const Instruction &inst)
{
  static int8 instr_buffer[15];
  int8 mem_op_buffer[108];      // this should be as large as the largest
                                //  memory operand we can have...
  int i, inst_len, mem_op;
  SegmentDescriptor *desc;      // to hold EA of memory operand
  int32 offset;

  if(!fpu_state) {
    fpu_state = new int8[108];
    __asm__("movl  %0, %%ebx;"
	    "finit;"
	    "fsave  (%%ebx);" : /* no output */ : "m" (fpu_state) : "ebx"
	    );
  }

  inst_len = inst.Length();
  //  fprintf(stderr, "Handling FPU op at 0x%08lx (%d bytes)\r\n", 
  //	  prev_eip, inst_len);

  if(!ReadFromVirtualAddress(seg_descs[SDESC_CS], prev_eip,
			     instr_buffer, inst_len, GetCPL(),
			     0, true)) {
    // fprintf(stderr, "Could not read instruction!\r\n");
    return;
  }

  // see if we've got a memory operand...
  for(mem_op = 0; mem_op < OPERANDS_PER_INSTRUCTION; mem_op++)
    if(OP(mem_op) && (OP(mem_op)->Type() == OPTYPE_MEMORY)) break;

  if(mem_op < OPERANDS_PER_INSTRUCTION) {
    // fprintf(stderr, "Oh, goodie!  We've got a memory operand! (%d)\r\n", mem_op);
    desc = OP(mem_op)->Fetch_EA_Segment(*this);
    offset = OP(mem_op)->Fetch_EA_Offset(*this);
    // fprintf(stderr, "EA:  Segment: %p, Offset: %08lx\r\n", desc, offset);

    if(!OP(mem_op)->FetchBytes(*this, mem_op_buffer, OP_SIZE(mem_op))) {
      // fprintf(stderr, "Attempt to fetch contents of memory operand failed.\r\n");
      return;
    }

    // now we have to find the modRM byte and maul it...
    // this code assumes that all FPU ops are 2-byte (0xD? modRM), and that
    //   there are never immediate operands afterwards
    i = ((instr_buffer[0] == 0x9B) ? 1 : 0);  // skip over WAIT prefix

    // rip out the modRM byte (except for the op bits) and put in absolute
    //   addressing modRM byte (0x05)
    instr_buffer[i+1] = 0x05 | (instr_buffer[i+1] & 0x38);

    // now drop in the address of 'mem_op_buffer'
    *(int32 *)(instr_buffer+i+2) = (int32)mem_op_buffer;

    // finally, fix the instruction length
    inst_len = i + 6;

    // do the rewritten instruction
    ExecuteFPUOp(instr_buffer, inst_len, fpu_state, CPU_EAX);

    // now write the memory operand back out - NOTE that this will bomb on
    //   read-only segments, even if the real FPU instruction is read-only
    if(!OP(mem_op)->WriteBytes(*this, mem_op_buffer, OP_SIZE(mem_op))) {
      // fprintf(stderr, "Attempt to write contents of memory operand failed.\r\n");
      return;
    }

    // fprintf(stderr, "Done with FPU/memory op.\r\n");
  } else {
    // fprintf(stderr, "No memory operands - continuing with execution.\r\n");

    if ((instr_buffer[0] == 0x66) && 
        (instr_buffer[1] == 0xdf)) {
       // XXX hack - fix this 
       fprintf(stderr, "Found 0x66dfe0 op at 0x%x\r\n", prev_eip);
       instr_buffer[0] = 0xdf;
       instr_buffer[1] = 0xe0;
    }
       
    ExecuteFPUOp(instr_buffer, inst.Length(), fpu_state, CPU_EAX);
    // fprintf(stderr, "Done with FPU op.\r\n");
  }
}

DEFN_INST(F2XM1)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FABS)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FADD)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FADDP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FIADD)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FBLD)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FBSTP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FCHS)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FCLEX)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FCOM)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FCOMP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FCOMPP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FCOS)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FDECSTP)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FDIV)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FDIVP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FIDIV)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FDIVR)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FDIVRP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FIDIVR)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FFREE)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FICOM)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FICOMP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FILD)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FINCSTP)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FINIT)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FIST)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FISTP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLD)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLD1)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDL2T)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDL2E)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDPI)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDLG2)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDLN2)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDZ)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FLDCW)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FLDENV)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FMUL)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FMULP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FIMUL)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FNOP)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FPATAN)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FPREM)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FPREM1)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FPTAN)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FRNDINT)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FRSTOR)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FSAVE)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FSCALE)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSIN)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSINCOS)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSQRT)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FST)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSTP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSTCW)
{
  // Setting it to 0x37F seems to be a good idea right now...
  //  DEBUG(DEBUG_FPU, "FSTCW just storing 0x37F for now...");

  //  SET_OP(0, 0x037F);
  // return(0);

  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FSTENV)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FSTSW)
{
  FAKE_FPU_OP_NO_CHECK;
}
END_DEFN

DEFN_INST(FSUB)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSUBP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FISUB)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSUBR)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FSUBRP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FISUBR)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FTST)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FUCOM)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FUCOMP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FUCOMPP)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FXAM)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FXCH)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FXTRACT)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FYL2X)
{
  FAKE_FPU_OP;
}
END_DEFN

DEFN_INST(FYL2XP1)
{
  FAKE_FPU_OP;
}
END_DEFN
