/* -*- C++ -*-
 *
 * $Id: fpu_instr.h,v 1.4 1998/04/29 04:54:52 nikos Exp $
 * 
 * RCS Log:
 * $Log: fpu_instr.h,v $
 * Revision 1.4  1998/04/29 04:54:52  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:44  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:28  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:02  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:00  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: fpu_instr.h,v 1.4 1998/04/29 04:54:52 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/fpu_instr.h,v 1.4 1998/04/29 04:54:52 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:26  bugnion
 * stanford update (x86)
 *
 * Revision 1.1  1997/03/09  21:18:05  sean
 * Initial revision
 *
 */

#ifndef _FPU_INSTR_H
#define _FPU_INSTR_H

// This file is essentially a sequel to 'x86_instr.h' and defines the FPU
//  instructions supported (more or less) by the simulator
//
// Like 'x86_instr.h', this file is embedded inside of the 'class CPU' 
//  declaration, so declaring classes and such inside this file may not
//  do what you think it does...

// We need x86_instr.h for the method formats:
#include "x86_instr.h"

// Defined in x86_instr.h
//define DECL_INST(mnemonic)      int64 mnemonic(const Instruction& inst);
//define DEFN_INST(mnemonic)      int64 CPU::mnemonic(const Instruction& inst)

// Instruction synonyms go here...
#define FNCLEX    FCLEX
#define FNINIT    FINIT
#define FNSAVE    FSAVE
#define FNSTCW    FSTCW
#define FNSTENV   FSTENV
#define FNSTSW    FSTSW
#define FWAIT     WAIT

// Instructions follow below
DECL_INST(F2XM1)
DECL_INST(FABS)
DECL_INST(FADD)
DECL_INST(FADDP)
DECL_INST(FIADD)
DECL_INST(FBLD)
DECL_INST(FBSTP)
DECL_INST(FCHS)
DECL_INST(FCLEX)
DECL_INST(FCOM)
DECL_INST(FCOMP)
DECL_INST(FCOMPP)
DECL_INST(FCOS)
DECL_INST(FDECSTP)
DECL_INST(FDIV)
DECL_INST(FDIVP)
DECL_INST(FIDIV)
DECL_INST(FDIVR)
DECL_INST(FDIVRP)
DECL_INST(FIDIVR)
DECL_INST(FFREE)
DECL_INST(FICOM)
DECL_INST(FICOMP)
DECL_INST(FILD)
DECL_INST(FINCSTP)
DECL_INST(FINIT)
DECL_INST(FIST)
DECL_INST(FISTP)
DECL_INST(FLD)
DECL_INST(FLD1)
DECL_INST(FLDL2T)
DECL_INST(FLDL2E)
DECL_INST(FLDPI)
DECL_INST(FLDLG2)
DECL_INST(FLDLN2)
DECL_INST(FLDZ)
DECL_INST(FLDCW)
DECL_INST(FLDENV)
DECL_INST(FMUL)
DECL_INST(FMULP)
DECL_INST(FIMUL)
DECL_INST(FNOP)
DECL_INST(FPATAN)
DECL_INST(FPREM)
DECL_INST(FPREM1)
DECL_INST(FPTAN)
DECL_INST(FRNDINT)
DECL_INST(FRSTOR)
DECL_INST(FSAVE)
DECL_INST(FSCALE)
DECL_INST(FSIN)
DECL_INST(FSINCOS)
DECL_INST(FSQRT)
DECL_INST(FST)
DECL_INST(FSTP)
DECL_INST(FSTCW)
DECL_INST(FSTENV)
DECL_INST(FSTSW)
DECL_INST(FSUB)
DECL_INST(FSUBP)
DECL_INST(FISUB)
DECL_INST(FSUBR)
DECL_INST(FSUBRP)
DECL_INST(FISUBR)
DECL_INST(FTST)
DECL_INST(FUCOM)
DECL_INST(FUCOMP)
DECL_INST(FUCOMPP)
DECL_INST(FXAM)
DECL_INST(FXCH)
DECL_INST(FXTRACT)
DECL_INST(FYL2X)
DECL_INST(FYL2XP1)

// Helper functions come below...

#endif
