/* $Id: instruction.cc,v 1.4 1998/04/29 04:54:52 nikos Exp $
 *
 * RCS Log:
 * $Log: instruction.cc,v $
 * Revision 1.4  1998/04/29 04:54:52  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:44  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:29  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:03  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:01  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: instruction.cc,v 1.4 1998/04/29 04:54:52 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/instruction.cc,v 1.4 1998/04/29 04:54:52 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:27  bugnion
// stanford update (x86)
//
// Revision 1.14  1997/12/09  05:33:21  mendel
// Bugs patched around (These need to be fixed):
//    1) ifdef'ed out the protection checking in LoadDescriptor() that
//       was preventing IRET from working. See comment in code I added in
//       cpu.cc
//    2) ifdefed out the ASSERT(0) in the TLB flush routine in x86_instr.cc
//    3) ifdef'ed out the protection checking in IRET that wasn't working.
//
//
// Bugs:
//
// 1) Check in some bug fixes in fpu.cc that Sean gave me.
//
// 2) Fixed bug in reading stack segment selector in TRAP_GATE . We were
// ASSERT(0)ing when the ReadFromVirtualAddress() was successful! in cpu.cc
// HandleExecption()
//
// 3) Changed (CSDescriptor.dpl() != SSSelector & 0x3) to
//            (CSDescriptor.dpl() != (SSSelector & 0x3))
// for obvious reasons in cpu.cc / HandleException().
//
// 4) Test for GDT in TRAP_GATE code was backward in cpu.cc /
// HandleException()
//
// 5) Since the prev_eip field of CPU class is used to set the return PC
// of execeptions. We need to set prev_eip on instruction page faults to
// be the faulting instruction. Otherwise we think the exception happen
// on the PC of IRET instruction. ouch!
//
// 6) The MemoryOperand() code in instruction.cc wasn't setting the
// default segment of the ESP register to be the stack segment! Found
// this because Linux runs for a while with the user's DS selector
// loaded and pushes things on the stack.
//
// 7) The is_task() method of seg_desc was completely bogus.
//
//
// Clean up:
//
// 1) Removed fprintf() on exceptions in cpu.cc since we are taking far too many
// execptions for this to be useful.
//
 * Revision 1.13  1997/09/27 18:38:27  sean
 * Tweaked RegisterOperand to support the EDXEAX register, which writes (or
 *   reads) top half to/from (E)DX and bottom to/from (E)AX - used by IMUL
 *
 * Revision 1.12  1997/09/24 08:14:04  sean
 * Instructions (and Operands, too) now know how to print themselves - very
 *  useful for debugging.
 *
 * Revision 1.11  1997/03/09 21:12:55  sean
 * used failure to compile as an excuse to generalize the resizing function
 *   of the MemoryOperand copy constructor - make Operand::OperandSize inline
 *
 * Revision 1.10  1997/03/09 08:36:57  yoel
 * Copy constructor for MemoryOp's now sets an opSize as well.
 *
 * Revision 1.9  1997/03/07 05:51:22  sean
 * Fixed broken sign-extension in RegisterOperand::FetchValue()
 *
 * Revision 1.8  1997/03/06 18:47:54  sean
 * created new MemoryOperand constructor that makes a copy of an existing
 *   MemoryOperand, adding in an immediate offset
 *
 * Revision 1.7  1997/02/28 09:24:18  yoel
 * Removed _repeat which is now obsolete.  Changed the constructor to
 * call the Prefix(const Prefix&) constructor for iPrefix rather than
 * use =.
 *
 * Revision 1.6  1997/02/14 05:00:18  sean
 * made operands know what kind they are, made memory operands take an optional
 * displacement on fetch/write
 *
 * Revision 1.5  1997/02/04 22:36:01  sean
 * changed [hx]to[xh][sl] macros to new names (..16,..32)
 *
 * Revision 1.4  1997/02/04 09:37:46  sean
 * make 16-bit addressing modes work (or at least made them less broken) - killed
 *   overflow into upper bits of offset
 *
 * Revision 1.3  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: instruction.cc,v 1.4 1998/04/29 04:54:52 nikos Exp $";

#include "instruction.h"
#include "cpu.h"
#include "debug.h"
#include "prefix.h"

#include <string.h>

// Yes, this assumes there are 3 operands/instruction, but it's a pain to
//  do it any other way...
Instruction::Instruction(InstEvalFunc _operation, int _inst_len, Prefix
			 _prefix, Operand *op1, Operand *op2,
			 Operand *op3) : iPrefix(_prefix)
{
  int i;

  operation = _operation;
  inst_len = _inst_len;

  for(i = 0; i < OPERANDS_PER_INSTRUCTION; i++) operand[i] = 0;
  operand[0] = op1;
  operand[1] = op2;
  operand[2] = op3;
}

// This'll destroy any operands it's holding also...
Instruction::~Instruction(void)
{
  int i;

  for(i = 0; i < OPERANDS_PER_INSTRUCTION; i++) delete operand[i];
}

// for debugging...
void Instruction::PrintSelf(FILE *output) const
{
  switch(iPrefix.RepOrLock()) {
  case RepRepZ:
    fputs("repe ", output);
    break;

  case RepRepNZ:
    fputs("repne ", output);
    break;

  case RepLock:
    fputs("lock ", output);
    break;

  case RepNone:
  default:
    break;
  }

  // this may qualify as the world's ugliest hack, but I wasn't about to
  //  write the name of each instruction by hand...
  CPU *cpu; // This is necessary to avoid an internal compiler error 191 :)
  fputs((char *) ((cpu ->* (operation))  // Trust me, this works...
		  (* (const Instruction *) 0)), 
	output);
  //  fputs("instr?", output);

  if(operand[0] != NULL) {
    fputs(" ", output);
    operand[0]->PrintSelf(output);

    if(operand[1] != NULL) {
      fputs(", ", output);
      operand[1]->PrintSelf(output);

      if(operand[2] != NULL) {
	fputs(", ", output);
	operand[2]->PrintSelf(output);
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////

Operand::Operand(int operand_size, OperandType _op_type)
{
  size = operand_size;
  op_type = _op_type;
}

Operand::~Operand(void) {}

///////////////////////////////////////////////////////////////////////////

RegisterOperand::RegisterOperand(int register_size, int register_num) :
  Operand(register_size,OPTYPE_REGISTER)
{
  reg_num = register_num;
}

RegisterOperand::~RegisterOperand(void) {}

SegmentDescriptor *RegisterOperand::Fetch_EA_Segment(CPU &cpu) const
{
  ASSERT(0);
  return(0);
}

int32 RegisterOperand::Fetch_EA_Offset(CPU &cpu, int32 disp2) const
{
  ASSERT(0);
  return(0);
}

int32 RegisterOperand::FetchValue(CPU& cpu, bool sign_extend, 
				  int32 disp2) const
{
  s_int32 value;

  ASSERT(disp2 == 0);

  switch(size) {
  case SIZE_8BIT:
    value = (s_int8) REG8(cpu.regfile,reg_num);
    return(sign_extend ? value : CHOP8(value));
    break;
  case SIZE_16BIT:
    value = (s_int16) REG16(cpu.regfile,reg_num);
    return(sign_extend ? value : CHOP16(value));
    break;
  case SIZE_32BIT:
    if(reg_num != REG_EDXEAX)
      return(REG32(cpu.regfile,reg_num)); else
      return((REG16(cpu.regfile, REG_DX) << 16) | REG16(cpu.regfile, REG_AX));
    break;
  default:
    ASSERT(0);
    return(0);
    break;
  }
}

bool RegisterOperand::WriteValue(CPU& cpu, int32 new_value, int32 disp2) const
{
  ASSERT(disp2 == 0);

  switch(size) {
  case SIZE_8BIT: 
    REG8(cpu.regfile,reg_num) = new_value;
    return(true);
    break;
  case SIZE_16BIT:
    REG16(cpu.regfile,reg_num) = new_value;
    return(true);
    break;
  case SIZE_32BIT:
    if(reg_num == REG_EDXEAX) {
      REG16(cpu.regfile, REG_DX) = (new_value >> 16);
      REG16(cpu.regfile, REG_AX) = (new_value & 0x0FFFF);
    } else
      REG32(cpu.regfile,reg_num) = new_value;
    return(true);
    break;
  default:
    ASSERT(0);
    return(false);
    break;
  }
}

int RegisterOperand::FetchBytes(CPU& cpu, void *buffer, int len, 
				int32 disp2) const
{
  ASSERT(0);
  return(-1);
}

int RegisterOperand::WriteBytes(CPU& cpu, const void *buffer, int len, 
				int32 disp2) const
{
  ASSERT(0);
  return(-1);
}

// these probably belong in a registers.cc...
static const char *RegNames32[] = { 
  "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi",
  "cr0", "cr1", "cr2", "cr3", "cr4", "cr5", "dr0", "dr1",
  "dr2", "dr3", "dr4", "dr5", "dr6", "dr7", "!22", "!23",
  "!24", "!25", "!26", "!27", "tsl", "tsh", "eip" };

static const char *RegNames16[] = { 
  "ax/!!0", "cx/!!1", "dx/!!2", "bx/!!3", 
  "sp/!!4", "bp/!!5", "si/!!6", "di/!!7",
  "!!8", "!!9", "!!10", "!!11", "!!12", "!!13", "!!14", "!!15",
  "!!16", "!!17", "!!18", "!!19", "!!20", "!!21", "es/tr", "cs/!!23",
  "ss/ldt", "ds/!!25", "fs/!!26", "gs/!!27", "!!28", "!!29", "ip/!!30" };

static const char *RegNames8[] = { 
  "al", "cl", "dl", "bl", "ah", "ch", "dh", "bh" };

static void PrintRegisterName(FILE *output, int reg_size, int reg_num)
{
  const char *pos;

  switch(reg_size) {
  case SIZE_8BIT: 
    fputs(RegNames8[reg_num], output);
    break;
  case SIZE_16BIT:
    if((pos = strchr(RegNames16[reg_num & 127], '/')) != NULL) {
      if(reg_num & 128)
	fputs(pos+1, output); else
	fprintf(output, "%.*s", (pos - RegNames16[reg_num & 127]),
		RegNames16[reg_num & 127]);
    } else
      fputs(RegNames16[reg_num & 127], output);
    break;
  case SIZE_32BIT:
    fputs(RegNames32[reg_num], output);
    break;
  default:
    ASSERT(0);
    break;
  }
}

// for debugging...
void RegisterOperand::PrintSelf(FILE *output) const
{
  PrintRegisterName(output, OperandSize(), reg_num);
}

///////////////////////////////////////////////////////////////////////////

ImmediateOperand::ImmediateOperand(int operand_size, s_int32 operand_value) :
  Operand(operand_size,OPTYPE_IMMEDIATE)
{
  value = (int32) operand_value;
}

ImmediateOperand::~ImmediateOperand(void) {}

SegmentDescriptor *ImmediateOperand::Fetch_EA_Segment(CPU &cpu) const
{
  ASSERT(0);
  return(0);
}

int32 ImmediateOperand::Fetch_EA_Offset(CPU &cpu, int32 disp2) const
{
  ASSERT(0);
  return(0);
}

int32 ImmediateOperand::FetchValue(CPU& cpu, bool sign_extend, 
				   int32 disp2) const
{
  ASSERT(disp2 == 0);

  if(sign_extend || (size == SIZE_32BIT)) return(value);

  return((size == SIZE_16BIT) ? CHOP16(value) : CHOP8(value));
}

bool ImmediateOperand::WriteValue(CPU& cpu, int32 new_value, int32 disp2) const
{
  ASSERT(0);
  return(false);
}

int ImmediateOperand::FetchBytes(CPU& cpu, void *buffer, int len, 
				 int32 disp2) const
{
  ASSERT(0);
  return(-1);
}

int ImmediateOperand::WriteBytes(CPU& cpu, const void *buffer, int len, 
				 int32 disp2) const
{
  ASSERT(0);
  return(-1);
}

// for debugging...
void ImmediateOperand::PrintSelf(FILE *output) const
{
  fprintf(output, "$%x{%d}", value, OperandSize());
}

///////////////////////////////////////////////////////////////////////////

MemoryOperand::MemoryOperand(int operand_size, int _address_size,
			     int segment_override,
			     int _base_reg, int _index_reg, int _index_scale,
			     s_int32 _displacement) :
  Operand(operand_size,OPTYPE_MEMORY)
{
  address_size = _address_size;

  if(segment_override == REG_NONE) {
    // REG_EBP == REG_BP, so just check, ignoring address_size
    if((_base_reg == REG_EBP) || (_index_reg == REG_EBP) || 
       (_base_reg == REG_ESP))
      segment_reg = REG_SS; else
      segment_reg = REG_DS;
  } else
    segment_reg = segment_override;

  base_reg = _base_reg;
  index_reg = _index_reg;
  index_scale = _index_scale;
  displacement = _displacement;
}

// This constructor lets one take a memory operand, and create another one
//  that points '_offset' bytes past the original (but using same segment,
//  base and index registers)
MemoryOperand::MemoryOperand(const MemoryOperand& copy_from, 
			     s_int32 _offset, int _op_size) 
  : Operand((_op_size ? _op_size : copy_from.OperandSize()), OPTYPE_MEMORY) 
{
  address_size = copy_from.address_size;
  segment_reg  = copy_from.segment_reg;
  base_reg     = copy_from.base_reg;
  index_reg    = copy_from.index_reg;
  index_scale  = copy_from.index_scale;
  displacement = copy_from.displacement + _offset;
}

MemoryOperand::~MemoryOperand(void) {}

SegmentDescriptor *MemoryOperand::Fetch_EA_Segment(CPU &cpu) const
{
  ASSERT((segment_reg >= REG_SEGBASE) && (segment_reg < (REG_SEGBASE + SDESC_TOTAL)));

  return(cpu.seg_descs + (segment_reg - REG_SEGBASE));
}

int32 MemoryOperand::Fetch_EA_Offset(CPU &cpu, int32 disp2) const
{
  int32 offset;

  // Start with the displacement
  offset = displacement;

  // Add on the base register, if present
  if(base_reg != REG_NONE)
    offset += ((address_size == SIZE_32BIT) ? REG32(cpu.regfile,base_reg) :
		                              REG16(cpu.regfile,base_reg));

  // Add on the index register, if present, and shift it properly
  if(index_reg != REG_NONE)
    offset += (((address_size == SIZE_32BIT) ? REG32(cpu.regfile,index_reg) :
		                               REG16(cpu.regfile,index_reg))
	       << index_scale);

  // If we had a secondary displacement specified (like for XLAT), use it
  offset += disp2;

  // The last step before returning is to, if we are in a 16-bit addressing 
  //  mode, chop off the top 16 bits.
  return((address_size == SIZE_32BIT) ? offset : CHOP16(offset));
}

int32 MemoryOperand::FetchValue(CPU& cpu, bool sign_extend, int32 disp2) const
{
  s_int8 value8;
  s_int16 value16;
  s_int32 value32;

  switch(size) {
  case SIZE_8BIT:
    FetchBytes(cpu,&value8,1,disp2);
    value32 = value8;             // Free sign-extension (I hope)
    return((int32) (sign_extend ? value32 : CHOP8(value32)));
    break;
  case SIZE_16BIT:
    FetchBytes(cpu,&value16,2,disp2);
    value16 = xtoh16(value16);      // Make sure the bytes are in host order
    value32 = value16;             // Free sign-extension (I hope)
    return((int32) (sign_extend ? value32 : CHOP16(value32)));
    break;
  case SIZE_32BIT:
    FetchBytes(cpu,&value32,4,disp2);
    value32 = xtoh32(value32);      // Make sure the bytes are in host order
    return((int32) value32);
    break;
  default:
    ASSERT(0);
    return(0);
    break;
  }
}

bool MemoryOperand::WriteValue(CPU& cpu, int32 new_value, int32 disp2) const
{
  int8 value8;
  int16 value16;

  switch(size) {
  case SIZE_8BIT:
    value8 = new_value;
    return(WriteBytes(cpu,&value8,1,disp2) == 1);
    break;
  case SIZE_16BIT:
    value16 = htox16(new_value);    // Make sure we use the right byte order...
    return(WriteBytes(cpu,&value16,2,disp2) == 2);
    break;
  case SIZE_32BIT:
    new_value = htox32(new_value);  // Make sure we use the right byte order...
    return(WriteBytes(cpu,&new_value,4,disp2) == 4);
    break;
  default:
    ASSERT(0);
    return(false);
    break;
  }
}

int MemoryOperand::FetchBytes(CPU& cpu, void *buffer, int len, 
			      int32 disp2) const
{
  return(cpu.ReadFromVirtualAddress(*(Fetch_EA_Segment(cpu)),
				    Fetch_EA_Offset(cpu,disp2),
				    buffer, len, cpu.GetCPL()));
}

int MemoryOperand::WriteBytes(CPU& cpu, const void *buffer, int len, 
			      int32 disp2) const
{
  return(cpu.WriteToVirtualAddress(*(Fetch_EA_Segment(cpu)),
				   Fetch_EA_Offset(cpu,disp2),
				   buffer, len, cpu.GetCPL()));
}

// for debugging...
void MemoryOperand::PrintSelf(FILE *output) const
{
  int printed = 0; // number of address elements printed

  // first, do segment
  PrintRegisterName(output, SIZE_16BIT, segment_reg);

  fputs(":[", output);

  // next is base register
  if(base_reg != -1) {
    PrintRegisterName(output, address_size, base_reg);
    printed++;
  }

  if(index_reg != -1) {
    if(printed) fputs(" + ", output);

    if(index_scale != SCALE_BY_1) fputs("(", output);
    PrintRegisterName(output, address_size, index_reg);

    // print scaling
    if(index_scale != SCALE_BY_1) 
      fprintf(output, " * %d)", (1 << index_scale));

    printed++;
  }

  if(displacement || !printed) {
    if(printed) fputs(" + ", output);

    fprintf(output, "%x", displacement);
  }

  fputs("]", output);
}
