/* -*- C++ -*-
 *
 * $Id: instruction.h,v 1.4 1998/04/29 04:54:53 nikos Exp $
 * 
 * RCS Log:
 * $Log: instruction.h,v $
 * Revision 1.4  1998/04/29 04:54:53  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:45  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:30  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:03  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:02  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: instruction.h,v 1.4 1998/04/29 04:54:53 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/instruction.h,v 1.4 1998/04/29 04:54:53 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:27  bugnion
 * stanford update (x86)
 *
 * Revision 1.30  1997/09/27  18:38:27  sean
 * Tweaked RegisterOperand to support the EDXEAX register, which writes (or
 *   reads) top half to/from (E)DX and bottom to/from (E)AX - used by IMUL
 *
 * Revision 1.29  1997/09/24 08:14:04  sean
 * Instructions (and Operands, too) now know how to print themselves - very
 *  useful for debugging.
 *
 * Revision 1.28  1997/03/14 08:03:34  yoel
 * Added the static functions and tables to do FPU decoding.
 *
 * Revision 1.27  1997/03/09 21:12:55  sean
 * used failure to compile as an excuse to generalize the resizing function
 *   of the MemoryOperand copy constructor - make Operand::OperandSize inline
 *
 * Revision 1.26  1997/03/09 08:36:57  yoel
 * CopyConstructor for MemOp now lets you change the operand size.
 *
 * Revision 1.25  1997/03/07 05:50:41  sean
 * added SIZE_48BIT and SIZE_64BIT #define's
 *
 * Revision 1.24  1997/03/06 18:47:54  sean
 * created new MemoryOperand constructor that makes a copy of an existing
 *   MemoryOperand, adding in an immediate offset
 *
 * Revision 1.23  1997/03/06 09:23:58  yoel
 * Removed the unneccessary GrpNOperand tables and added the two tables for
 * Grp3 (0xF6 has Byte operands and 0xF7 has either word or double word)
 *
 * Revision 1.22  1997/03/02 00:46:09  yoel
 * Added the static tables from modrm.h to class Instruction.
 *
 * Revision 1.21  1997/02/28 09:24:18  yoel
 * Fixed the accessors for GetRepeat, GetAddrSize, GetOpSize to use the new
 * Prefix class and finally work normally!  Changed the OperandDecode typedef
 * to use the expanded prototype.
 *
 * Revision 1.20  1997/02/27 08:48:03  yoel
 * Made the prototype return a bool now so I can indicate failure.  Also
 * removed a few functions that shouldn't exist where I read an I as an l.
 *
 * Revision 1.19  1997/02/20 05:52:17  yoel
 * Bleah, change prototype for DecodeOperandDecl to have *ops[]
 * removed DecodeOperand typedef completely.
 *
 * Revision 1.18  1997/02/19 06:51:59  yoel
 * Added typedef for DecodeOperandDecl and added a few items to the prototype
 *
 * Revision 1.17  1997/02/19 00:54:06  sean
 * fixed DecodeOperand typedef in order to allow compilation
 *
 * Revision 1.16  1997/02/14 06:09:33  yoel
 * Added prototypes for static Operand decoding tables.
 *
 * Revision 1.15  1997/02/14 05:51:53  yoel
 * Added a typedef for the DecodeOperand functions.
 *
 * Revision 1.14  1997/02/14 05:03:26  yoel
 * Changed DieIfNoSpace to FetchOrDie() to facilitate the DataStream.
 *
 * Revision 1.13  1997/02/14 05:00:18  sean
 * made operands know what kind they are, made memory operands take an optional
 * displacement on fetch/write
 *
 * Revision 1.12  1997/02/11 05:07:51  sjt
 * changed Instruction::DecodeInstruction() to take a DataStream for opcode
 * data instead of a buffer/length pair
 *
 * Revision 1.11  1997/02/10 10:19:07  yoel
 * FIxed the prototypes for the static opcode tables (I'd forgotten the
 * word static and had extra Instruction:: in them).
 *
 * Revision 1.10  1997/02/10 09:18:34  yoel
 * Added proto-types for all the static opcode tables I added to decode.h
 *
 * Revision 1.9  1997/02/06 09:51:39  sean
 * added GetAddressSize() and GetOperandSize() to Instruction class
 *
 * Revision 1.8  1997/02/04 22:30:31  sean
 * InstEvalFunc's (CPU's operation methods) now return int64's
 *
 * Revision 1.7  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef _INSTRUCTION_H
#define _INSTRUCTION_H

#include <stdio.h>

#include "prefix.h"
#include "registers.h"
#include "ints.h"

#define OPERANDS_PER_INSTRUCTION    3

#define SIZE_8BIT      1
#define SIZE_16BIT     2
#define SIZE_32BIT     4
#define SIZE_48BIT     6
#define SIZE_64BIT     8
#define SIZE_80BIT     10

// Not only do these nifty constants correspond to how many bits left a
//  number needs to be shifted, they correspond to the scaling coding in
//  the SIB byte.
#define SCALE_BY_1     0
#define SCALE_BY_2     1
#define SCALE_BY_4     2
#define SCALE_BY_8     3

// We need values for repeat codes (same as prefix codes in prefix.h)
#define REP_NONE        0
#define REP_REP         PRE_REP
#define REP_REPE        PRE_REPE
#define REP_REPNE       PRE_REPNE
#define REP_REPNZ       PRE_REPNZ
#define REP_REPZ        PRE_REPZ

// Something of a hack - since DX:AX (or EDX:EAX) is often a register target,
//   we create an imaginary register, REG_EDXEAX, which RegisterOperand handles
//   internally
#define REG_EDXEAX     999

// Condition codes
#define COND_O         0
#define COND_NO        1
#define COND_B         2
#define COND_NAE       COND_B
#define COND_NB        3
#define COND_AE        COND_NB
#define COND_E         4
#define COND_Z         COND_E
#define COND_NE        5
#define COND_NZ        COND_NE
#define COND_BE        6
#define COND_NA        COND_BE
#define COND_NBE       7
#define COND_A         COND_NBE
#define COND_S         8
#define COND_NS        9
#define COND_P        10
#define COND_PE        COND_P
#define COND_NP       11
#define COND_PO        COND_NP
#define COND_L        12
#define COND_NGE       COND_L
#define COND_NL       13
#define COND_GE        COND_NL
#define COND_LE       14
#define COND_NG        COND_LE
#define COND_NLE      15
#define COND_G         COND_NLE

class CPU;
class SegmentDescriptor;
class Instruction;
class Operand;
class DataStream;

typedef int64 (CPU::*InstEvalFunc)(const Instruction& inst);
typedef bool (*DecodeOperandDecl)(DataStream &instr_data, int opSize, int
				  addrSize, int segReg, Operand *ops[]);

class Instruction {
 public:
  // Yes, this assumes there are 3 operands/instruction, but it's a pain to
  //  do it any other way...
  Instruction(InstEvalFunc _operation, int _inst_len, Prefix _prefix, 
	      Operand *op1 = 0, Operand *op2 = 0, Operand *op3 = 0); 

  // This'll destroy any operands it's holding also...
  ~Instruction(void);

  // The magical instruction decoding function now lives here (where it 
  //  belongs)
  static Instruction *DecodeInstruction(DataStream& instr_data,
                                        int default_datasize);

  // FPU is living in its own world just for kicks.
  static Instruction *DecodeFPUInstruction(DataStream& instr_data,
                                           int default_datasize, 
				           Prefix& prefix, int8 firstByte);

  //  No Sean, I will not let you ignore data hiding completely.  These
  // are accessors, please use the instead of making all of my class
  // members public.
  int Length(void) const { return inst_len; }
  InstEvalFunc GetOperation(void) const { return operation; }
  Operand *GetOperand(int whichOp) const { return operand[whichOp]; }

  // The prefix is actually the one with all of the knowledge here...
  RepPrefix GetRepeat(void) const { return iPrefix.Rep(); }
  int GetAddressSize(void) const { return iPrefix.AddrSize(); }
  int GetOperandSize(void) const { return iPrefix.OpSize(); }

  // for debugging...
  void PrintSelf(FILE *output) const;

protected:
  InstEvalFunc operation;
  int          repeat;
  Prefix       iPrefix;
  int          inst_len;   // length of the instruction in bytes, COUNTING
                           //  prefixes!
  Operand      *operand[OPERANDS_PER_INSTRUCTION];

  //  Statics:
  static bool TryToFetch(DataStream& instr_data, 
                                              int8& nextByte, int8& nDecoded);

  //  These are all initialized in decode.h.  Note that this implies
  // decode.h can be included in exactly one .cc file.  If this becomes
  // problematic then they will move.
  static const InstEvalFunc OneByteOpcodes[];
  static const int BadFirstByte[];
  static const InstEvalFunc TwoByteOpcodes[];
  static const int BadSecondByte[];
  static const InstEvalFunc Grp1Codes[];
  static const InstEvalFunc Grp2Codes[];
  static const InstEvalFunc Grp3Codes[];
  static const InstEvalFunc Grp4Codes[];
  static const InstEvalFunc Grp5Codes[];
  static const InstEvalFunc Grp6Codes[];
  static const InstEvalFunc Grp7Codes[];
  static const InstEvalFunc Grp8Codes[];
  static const InstEvalFunc Grp9Codes[];
  static const DecodeOperandDecl OneByteOperand[];
  static const DecodeOperandDecl TwoByteOperand[];
  static const DecodeOperandDecl Grp3OperandsF6[];
  static const DecodeOperandDecl Grp3OperandsF7[];
  static const DecodeOperandDecl Grp5Operands[];
  static const DecodeOperandDecl Grp7Operands[];
  static const InstEvalFunc Instruction::FPU_Low[][];
  static const InstEvalFunc Instruction::FPU_High[][];
  static const DecodeOperandDecl Instruction::FPU_LowOperand[][];
  static const DecodeOperandDecl Instruction::FPU_HighOperand[][];

  //  Operand decoders are in x86_operand.h  and they're all static
#include "x86_operand.h"

  //  More decoding tables (for the ModR/M junk)
  static const int ModRM_CRegs[];
  static const int ModRM_DRegs[];
  static const int ModRM_SRegs[];

};

////////////////////////////////////////////////////////////////////////

enum OperandType { OPTYPE_INVALID, OPTYPE_IMMEDIATE, OPTYPE_REGISTER,
		   OPTYPE_MEMORY };

class Operand {
 public:
  Operand(int operand_size, OperandType _op_type);
  virtual ~Operand(void);

  inline OperandType Type(void) const { return(op_type); }

  inline int OperandSize(void) const { return(size); }

  virtual SegmentDescriptor *Fetch_EA_Segment(CPU &cpu) const = 0;
  virtual int32 Fetch_EA_Offset(CPU &cpu, int32 disp2 = 0) const = 0;

  virtual int32 FetchValue(CPU& cpu, bool sign_extend = 1, 
			   int32 disp2 = 0) const = 0;
  virtual bool WriteValue(CPU& cpu, int32 new_value, 
			  int32 disp2 = 0) const = 0;

  virtual int FetchBytes(CPU& cpu, void *buffer, int len, 
			 int32 disp2 = 0) const = 0;
  virtual int WriteBytes(CPU& cpu, const void *buffer, int len, 
			 int32 disp2 = 0) const = 0;

  // for debugging...
  virtual void PrintSelf(FILE *output) const = 0;

 protected:
  int size;
  OperandType op_type;
};

class RegisterOperand : public Operand {
 public:
  RegisterOperand(int register_size, int register_num);
  virtual ~RegisterOperand(void);

  inline int RegNum(void) const { return(reg_num); }

  virtual SegmentDescriptor *Fetch_EA_Segment(CPU &cpu) const;
  virtual int32 Fetch_EA_Offset(CPU &cpu, int32 disp2 = 0) const;

  virtual int32 FetchValue(CPU& cpu, bool sign_extend = 1, 
			   int32 disp2 = 0) const;
  virtual bool WriteValue(CPU& cpu, int32 new_value, 
			  int32 disp2 = 0) const;

  virtual int FetchBytes(CPU& cpu, void *buffer, int len, 
			 int32 disp2 = 0) const;
  virtual int WriteBytes(CPU& cpu, const void *buffer, int len, 
			 int32 disp2 = 0) const;

  // for debugging...
  virtual void PrintSelf(FILE *output) const;
 
 private:
  int reg_num;
};

class ImmediateOperand : public Operand {
 public:
  ImmediateOperand(int operand_size, s_int32 operand_value);
  virtual ~ImmediateOperand(void);

  virtual SegmentDescriptor *Fetch_EA_Segment(CPU &cpu) const;
  virtual int32 Fetch_EA_Offset(CPU &cpu, int32 disp2 = 0) const;

  virtual int32 FetchValue(CPU& cpu, bool sign_extend = 1, 
			   int32 disp2 = 0) const;
  virtual bool WriteValue(CPU& cpu, int32 new_value, 
			  int32 disp2 = 0) const;

  virtual int FetchBytes(CPU& cpu, void *buffer, int len, 
			 int32 disp2 = 0) const;
  virtual int WriteBytes(CPU& cpu, const void *buffer, int len, 
			 int32 disp2 = 0) const;

  // for debugging...
  virtual void PrintSelf(FILE *output) const;

 private:
  int32 value;
};

class MemoryOperand : public Operand {
 public:
  MemoryOperand(int operand_size, int _address_size,
		int segment_override = REG_NONE, int _base_reg = REG_NONE, 
		int _index_reg = REG_NONE, int _index_scale = SCALE_BY_1,
		s_int32 _displacement = 0);

  // This constructor lets one take a memory operand, and create another one
  //  that points '_offset' bytes past the original (but using same segment,
  //  base and index registers) - if you pass in a non-zero value for 
  //  '_op_size', that size will be used over the size of the 'copy_from'
  //  operand
  MemoryOperand(const MemoryOperand& copy_from, s_int32 _offset = 0, 
		int _op_size = 0);

  virtual ~MemoryOperand(void);

  virtual SegmentDescriptor *Fetch_EA_Segment(CPU &cpu) const;
  virtual int32 Fetch_EA_Offset(CPU &cpu, int32 disp2 = 0) const;

  virtual int32 FetchValue(CPU& cpu, bool sign_extend = 1, 
			   int32 disp2 = 0) const;
  virtual bool WriteValue(CPU& cpu, int32 new_value, 
			  int32 disp2 = 0) const;

  virtual int FetchBytes(CPU& cpu, void *buffer, int len, 
			 int32 disp2 = 0) const;
  virtual int WriteBytes(CPU& cpu, const void *buffer, int len, 
			 int32 disp2 = 0) const;

  // for debugging...
  virtual void PrintSelf(FILE *output) const;

protected:
  int address_size;
  int segment_reg, base_reg, index_reg, index_scale;
  s_int32 displacement;
};

#endif
