/* $Id: ints.h,v 1.4 1998/04/29 04:54:53 nikos Exp $
 * 
 * RCS Log:
 * $Log: ints.h,v $
 * Revision 1.4  1998/04/29 04:54:53  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:45  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:30  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:04  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:02  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: ints.h,v 1.4 1998/04/29 04:54:53 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/ints.h,v 1.4 1998/04/29 04:54:53 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:27  bugnion
 * stanford update (x86)
 *
 * Revision 1.6  1997/11/27  00:01:24  mendel
 * Patch to handle int8
 *
 * Revision 1.5  1997/11/22  07:43:32  whimsy
 * SimOS Compatibility, nade int32, etc #defines for types already in SimOS.
 *
 * Revision 1.4  1997-02-04 14:29:41-08  sean
 * created 64-bit integer type, and related macros
 *
 * Revision 1.3  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef _INTS_H
#define _INTS_H

#ifdef SIMOS
#include "simtypes.h"
#define int8 uint8
#define int16 uint16
#define int32 uint32
#define int64 uint64
#else
// These are the integer types we'll use for regularity...
typedef unsigned char int8;
typedef unsigned short int16;
typedef unsigned int int32;
typedef unsigned long long int64;
#endif


// We'll cast to these when we need signed values...
typedef signed char s_int8;
typedef signed short s_int16;
typedef signed int s_int32;
typedef signed long long s_int64;

// Since we'll usually need to use 32-bit integers to play with 64-bit, these
//  macros will help
#define TOP32(i64)      ((int32) ((i64) >> 32))
#define BOTTOM32(i64)   ((int32) (i64))
#define BUILD64(hi,lo)  ((((int64) (hi)) << 32) | ((int32) (lo)))

// This macro yields an integer with a string of 'size' 1's followed by a
//  string of 'shift' 0's.
#define MASK(size,shift)     (((1 << (size)) - 1) << (shift))

// This macro pulls out a bitfield that is 'size' bits wide out of 'field'
//  with bit number 'start' as the LSB of the extracted field
#define BITS(field,start,size)  (((field) >> (start)) & MASK(size,0))

// Like BITS, except it sets the specified bitfield to 'val'
#define SET_BITS(field,start,size,val) (field) = (((field) & ~MASK(size,start)) | ((val) << (start)))

#define CLEAR_BITS(field,start,size)   (field) &= ~MASK(size,start)

#define TOGGLE_BITS(field,start,size)  (field) ^= MASK(size,start)

// This function flips the byte order of a 16-bit integer
#define FLIP16(x)    (((x) << 8) | ((x) >> 8))

// This function flips the byte order of a 32-bit integer
#define FLIP32(x)    (((x) << 24) | (((x) & 0x0000ff00) << 8) | \
                      (((x) & 0x00ff0000) >> 8) | ((x) >> 24))

// This'll flip the bytes of a 64-bit integer
#define FLIP64(x)    ((FLIP32(x & 0x0ffffffff) << 32) | \
                      (FLIP32(x >> 32) & 0x0ffffffff))

// These functions define functions to translate from (h)ost byte order 
//  to/from (x)86 order...
#ifdef LITTLE_ENDIAN
// x86 is little endian, so these are nops...
#define htox16(h)       (h)
#define xtoh16(x)       (x)
#define htox32(h)       (h)
#define xtoh32(x)       (x)
#define htox64(h)       (h)
#define xtoh64(x)       (x)
#else
// big endian number have to be flipped
#define htox16(h)       FLIP16(h)
#define xtoh16(x)       FLIP16(x)
#define htox32(h)       FLIP32(h)
#define xtoh32(x)       FLIP32(x)
#define htox64(h)       FLIP64(h)
#define xtoh64(x)       FLIP64(x)
#endif

// These three functions chop larger integers down to smaller integers.
//  Basically, most integers will be sign-extended automatically, and
//  by chopping, we get the equivalent of zero-extension
#define CHOP32(x)      ((x) & 0x0ffffffff)
#define CHOP16(x)      ((x) & 0x00000ffff)
#define CHOP8(x)       ((x) & 0x0000000ff)

// Friendly max and min functions, using cool gcc operators
#define MIN(a,b)  ((a) <? (b))
#define MAX(a,b)  ((a) >? (b))

#endif
