/* -*- C++ -*-
 * $Id: lockstep.h,v 1.4 1998/04/29 04:54:55 nikos Exp $
 * 
 * RCS Log:
 * $Log: lockstep.h,v $
 * Revision 1.4  1998/04/29 04:54:55  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:47  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:32  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:06  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:04  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: lockstep.h,v 1.4 1998/04/29 04:54:55 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/lockstep.h,v 1.4 1998/04/29 04:54:55 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:28  bugnion
 * stanford update (x86)
 *
 * Revision 1.5  1997/09/27  18:41:51  sean
 * Added ability to ignore some flags in lockstep comparison.
 * Also added call to read data from lockstep'd process's memory space.
 *
 * Revision 1.4  1997/09/24 08:10:41  sean
 * Tweaked call to VerifyRegisters to know what instruction caused the
 * mismatch.
 *
 * Revision 1.3  1997/09/15 19:49:27  sean
 * tweaked register verification code to allow correction of mismatches
 *  rather than just aborting.  Copying from sim-land to real-land
 *  does not work yet, and it'd be nice if a mismatch killed the repetition
 *  of a command...
 *
 * Revision 1.2  1997/09/06 00:39:51  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.1  1997/04/13 23:37:31  yoel
 * Initial revision
 *
 */

//   LockStep.h, Author Jeremy Sugerman (yoel@Stanford.EDU)
//  class to run a binary one instruction at a time and facilitate
// comparison with a simulated binary.

#ifndef __LOCKSTEP_H__
#define __LOCKSTEP_H__

#include "registers.h"

class CPU;
class Instruction;

class LockStep {

public:
   static LockStep *Init(const char *cmd, char *argv[]);

   LockStep(int _pid);
   int SingleStep(void);
   int TraceRegisters(void);
   int MakeRegisterFile(RegisterFile *sim_registers);
   int VerifyRegisterFile(RegisterFile *sim_registers,
			  const Instruction *last_instr = 0 
			           /* the one that broke */,
			  int32 flag_ignore_mask = 0 
			           /* flags to not compare */);
   int ReadFromLockstepMemory(int32 address, void *buffer, int len);

   int CopyStack(CPU& cpu);

protected:
   void MismatchAction(int32 *sim_value, int real_regnum, int32 real_value);

   int GetRegisters(int32 registers[]);
   int pid;
   static const char *register_names[];
   static const int sim_order[];
   static const int reg_col1[];
   static const int reg_col2[];
   static const int reg_col3[];
};

#endif
