// -*- C++ -*-
/*
    $Id: modrm.h,v 1.4 1998/04/29 04:54:55 nikos Exp $

    $Log: modrm.h,v $
    Revision 1.4  1998/04/29 04:54:55  nikos
    Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
    only instructions with references (pc + ea) that miss in the L1 cache.
    MBs are always dumped.

    Revision 1.4  1998/03/25 19:16:47  nikos
    atf_tlb_fixed

 * Revision 1.3  1998/03/13  17:12:32  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
    Revision 1.3  1998/03/13 16:37:07  nikos
    .

 * Revision 1.2  1998/02/16  23:20:04  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: modrm.h,v 1.4 1998/04/29 04:54:55 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/modrm.h,v 1.4 1998/04/29 04:54:55 nikos Exp $, etc).
 *
    Revision 1.2  1998/02/11 14:24:55  nikos
    .

 * Revision 1.1  1997/12/16  01:28:28  bugnion
 * stanford update (x86)
 *
 * Revision 1.6  1997/03/07  05:49:12  sean
 * fixed SIB decoding, and freakish ModRM/SIB creations of disp32's...
 *
    Revision 1.5  1997/03/06 19:57:06  yoel
    Added SIB stuff.

    Revision 1.4  1997/03/06 18:15:51  sean
    fixed a couple typos - switched R/M 4 and 5 back to proper order in DO_E

    Revision 1.3  1997/03/06 10:05:36  yoel
    Added casts sicne all the displacements and such are signed, but the
    Data Strea wants unsigned variables.

    Revision 1.2  1997/03/06 07:30:16  yoel
    All of E except for SIB done.  Checking so Sean can get it.

    Revision 1.1  1997/03/02 00:45:23  yoel
    Initial revision

 */

//  File modrm.h,  Author Jeremy Sugerman (yoel@stanford.edu)
//
//  Contains all sorts of happy tables and #define's for dealing with the
// ModR/M Byte of an instruction.
#ifndef __MODRM_H__
#define __MODRM_H__

#include "registers.h"
#include "instruction.h"
#include "debug.h"

//  The tables below take either the R/M bits or REG bits of a ModR/M byte
// and return the number of the proper register.  Note that it might be
// possible to tweak registers.h such that one of these table is not
// needed, but the special case would probably be confusing.

//  Single Byte Registers & General Registers are actually #defined in
// exactly the right order courtesy of Sean, so they don't need arrays.

//  Control Registers
const int Instruction::ModRM_CRegs[8] = {
  REG_CR0, -1, REG_CR2, REG_CR3, REG_CR4, -1, -1, -1
};

//  Debug Registers, note DR4 and DR5 can NOT be used, but Sean catches that.
const int Instruction::ModRM_DRegs[8] = {
  REG_DR0, REG_DR1, REG_DR2, REG_DR3, REG_DR4, REG_DR5, REG_DR6, REG_DR7
};

//  Segment Registers
const int Instruction::ModRM_SRegs[8] = {
  REG_ES, REG_CS, REG_SS, REG_DS, REG_FS, REG_GS, -1, -1
};

#define ModRM_MOD(iByte) ((iByte>>6))
#define ModRM_REG(iByte) ((iByte>>3) & 0x7)
#define ModRM_RM(iByte)  ((iByte) & 0x7)

#define GET_MODRM(instr_data, modRM)    \
  if (!instr_data.ReadOneByte(modRM)) return false;

#define DO_CD(opNum, modRM) \
  ops[opNum] = new RegisterOperand(SIZE_32BIT, ModRM_CRegs[ModRM_REG(modRM)]);

#define DO_DD(opNum, modRM) \
  ops[opNum] = new RegisterOperand(SIZE_32BIT, ModRM_DRegs[ModRM_REG(modRM)]);

#define DO_GB(opNum, modRM) \
  ops[opNum]=new RegisterOperand(SIZE_8BIT, ModRM_REG(modRM));

#define DO_GV(opNum, modRM) \
  ops[opNum] = new RegisterOperand(opSize, ModRM_REG(modRM));

#define DO_GW(opNum, modRM) \
  ops[opNum] = new RegisterOperand(SIZE_16BIT, ModRM_REG(modRM));

#define DO_RD(opNum, modRM) \
  ops[opNum] = new RegisterOperand(SIZE_32BIT, ModRM_RM(modRM));

#define DO_SW(opNum, modRM) \
  ops[opNum] = new RegisterOperand(SIZE_16BIT, ModRM_SRegs[ModRM_REG(modRM)]);

#define SIB_SS(SIB)      ((SIB>>6))
#define SIB_INDEX(SIB)   ((SIB>>3) & 0x7)
#define SIB_BASE(SIB)    ((SIB) & 0x7)

static inline bool DO_SIB(int opNum, int8 modRM, int8 SIB, int opSize, 
			  int addrSize, int segReg, s_int32 disp32, 
			  Operand *ops[]) 
{
  int base;
  int index;

  // the caller will read the SIB for us, since it preceeds the displacement
  //  if (!instr_data.ReadOneByte(SIB)) return false;

  // if the MOD is 00, then what would be a [EBP] for the base is REG_NONE
  base = ((ModRM_MOD(modRM) == 0) && (SIB_BASE(SIB) == 0x5)) ? REG_NONE :
                                                               SIB_BASE(SIB); 

  index = (SIB_INDEX(SIB) == 0x4) ? REG_NONE : SIB_INDEX(SIB);

  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, base, index,
				 SIB_SS(SIB), disp32);
  return true;
}

// For EB, pass in opSize = SIZE_8BIT, for EV pass in opSize = opSize.
static inline bool DO_E(int opNum, DataStream& instr_data, int8 modRM, 
                        int opSize, int addrSize, int segReg, Operand
			*ops[], bool onlyMem = false)
{
  int8 SIB;
  s_int8 disp8;
  s_int16 disp16;
  s_int32 disp32;

  if (addrSize == SIZE_16BIT) { 
    if (ModRM_MOD(modRM) == 0x3) {
      if (onlyMem) {
	DEBUG(DEBUG_DECODE, "Puny mortal, Memory Only instructions cannot "
	                    "refer to registers.");
	return false;
      }
      ops[opNum] = new RegisterOperand(opSize, ModRM_RM(modRM));
    } else {
      switch(ModRM_MOD(modRM)) {
      case 0:
	disp32 = 0;
	break;
      case 1:
	if(!instr_data.ReadOneByte((int8&) disp8)) return false;
	disp32 = disp8;
	break;
      case 2:
	if(!instr_data.ReadTwoBytes((int16&) disp16)) return false;
	disp32 = disp16;
	break;
      }
      
      switch(ModRM_RM(modRM)) {
      case 0:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BX,
				       REG_SI, SCALE_BY_1, disp32);
	break;
      case 1:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BX,
				       REG_DI, SCALE_BY_1, disp32);
	break;
      case 2:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BP,
				       REG_SI, SCALE_BY_1, disp32);
	break;
      case 3:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BP,
				       REG_DI, SCALE_BY_1, disp32);
	break;
      case 4:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_SI,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 5:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_DI,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 6:
	if (ModRM_MOD(modRM) == 0x0) {
	  if (!instr_data.ReadTwoBytes((int16&) disp16)) return false;
	  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_NONE,
					 REG_NONE, SCALE_BY_1, disp16);
	} else { 
	  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BP,
					 REG_NONE, SCALE_BY_1, disp32);
	}
	break;
      case 7:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_BX,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      }
    }
  } else {
    if (ModRM_MOD(modRM) == 0x3) {
      if (onlyMem) {
        DEBUG(DEBUG_DECODE, "Puny mortal, Memory Only instructions cannot "
                            "refer to registers.");
        return false;
      }
      ops[opNum] = new RegisterOperand(opSize, ModRM_RM(modRM));
    } else {
      // need to read in the SIB now, if it exists...
      if(ModRM_RM(modRM) == 4) {
	if(!instr_data.ReadOneByte(SIB)) return(false);
      } else
	SIB = 0;  // this eliminates optimizer "use before init" warnings

      switch(ModRM_MOD(modRM)) {
      case 0x0:
	disp32 = 0;
	break;
      case 0x1:
	if (!instr_data.ReadOneByte((int8&) disp8)) return false;
	disp32 = disp8;
	break;
      case 0x2:
	if (!instr_data.ReadFourBytes((int32&) disp32)) return false;
	break;
      }

      switch (ModRM_RM(modRM)) {
      case 0:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_EAX,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 1:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_ECX,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 2:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_EDX,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 3:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_EBX,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 4:
	// Stupid exceptions: if MOD=00 and BASE=101, read disp32
	if((ModRM_MOD(modRM) == 0x0) && (SIB_BASE(SIB) == 5))
	  if (!instr_data.ReadFourBytes((int32&) disp32)) return false;

	if (!DO_SIB(opNum, modRM, SIB, opSize, addrSize, segReg, disp32, ops)) 
	  return false;
	break;
      case 5:
	if (ModRM_MOD(modRM) == 0x0) {
	  if (!instr_data.ReadFourBytes((int32&) disp32)) return false;
	  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_NONE,
					 REG_NONE,  SCALE_BY_1, disp32);
	} else {
	  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_EBP,
					 REG_NONE, SCALE_BY_1, disp32);
	}
	break;
      case 6:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_ESI,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      case 7:
	ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_EDI,
				       REG_NONE, SCALE_BY_1, disp32);
	break;
      }
    }
  }
  return true;
}

#endif
