/* $Id: prefix.cc,v 1.4 1998/04/29 04:54:56 nikos Exp $
 *
 * RCS Log:
 * $Log: prefix.cc,v $
 * Revision 1.4  1998/04/29 04:54:56  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:48  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:33  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:07  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:05  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: prefix.cc,v 1.4 1998/04/29 04:54:56 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/prefix.cc,v 1.4 1998/04/29 04:54:56 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:29  bugnion
// stanford update (x86)
//
// Revision 1.5  1997/02/28  09:24:18  yoel
// Totally rewrote prefixes to keep better track of each property--
// Repeat, Segment REgister, Operand Size, and Addr Size.
//
 * Revision 1.4  1997/02/14 04:17:10  yoel
 * Added Prefix counting code.
 *
 * Revision 1.3  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: prefix.cc,v 1.4 1998/04/29 04:54:56 nikos Exp $";

//  Prefix.cc, implement the tracking and decoding of Instruction prefixes.
//
//  Author: Jeremy Sugerman (yoel@stanford.edu)
//
#include "prefix.h"
#include "instruction.h"
#include "registers.h"
#include "debug.h"

Prefix::Prefix(const Prefix& inPrefix)
{
  pRep = inPrefix.RepOrLock();
  pSeg = inPrefix.SegReg();
  opSize = inPrefix.OpSize();
  addrSize = inPrefix.AddrSize();
  nPrefixes = inPrefix.NumPrefixes();
}

Prefix::Prefix(int defaultSize)
{
  pRep = RepNone;
  pSeg = REG_NONE;
  pOpSize = false;
  pAddrSize = false;
  opSize = addrSize = defaultSize;
  nPrefixes = 0;
}

int Prefix::CheckPrefix(int8 inByte)
{
  switch(inByte) {
  case PRE_REPZ:
    if (pRep != RepNone) return -1;
    pRep = RepRepZ;
    nPrefixes++;
    return 1;
  case PRE_REPNZ:
    if (pRep != RepNone) return -1;
    pRep = RepRepNZ;
    nPrefixes++;
    return 1;
  case PRE_LOCK:
    if (pRep != RepNone) return -1;
    pRep = RepLock;
    nPrefixes++;
    return 1;
  case PRE_SEGCS:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_CS;
    nPrefixes++;
    return 1;
  case PRE_SEGSS:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_SS;
    nPrefixes++;
    return 1;
  case PRE_SEGDS:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_DS;
    nPrefixes++;
    return 1;
  case PRE_SEGES:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_ES;
    nPrefixes++;
    return 1;
  case PRE_SEGFS:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_FS;
    nPrefixes++;
    return 1;
  case PRE_SEGGS:
    if (pSeg != REG_NONE) return -1;
    pSeg = REG_GS;
    nPrefixes++;
    return 1;
  case PRE_OPSIZE:
    if (pOpSize) return -1;
    pOpSize = true;
    opSize = (opSize == SIZE_16BIT) ? SIZE_32BIT : SIZE_16BIT;
    nPrefixes++;
    return 1;
  case PRE_ADDRSIZE:
    if (pAddrSize) return -1;
    pAddrSize = true;
    addrSize = (addrSize == SIZE_16BIT) ? SIZE_32BIT : SIZE_16BIT;
    nPrefixes++;
    return 1;
  }

  return 0;
}
