//  -*- C++ -*-
/* $Id: prefix.h,v 1.4 1998/04/29 04:54:56 nikos Exp $
 * 
 * RCS Log:
 * $Log: prefix.h,v $
 * Revision 1.4  1998/04/29 04:54:56  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:48  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:33  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:07  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:05  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: prefix.h,v 1.4 1998/04/29 04:54:56 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/prefix.h,v 1.4 1998/04/29 04:54:56 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:29  bugnion
 * stanford update (x86)
 *
 * Revision 1.6  1997/02/28  09:24:18  yoel
 * Totally rewrote prefixes, added all sorts of accessors since the Prefix
 * now properly keeps track of an instruction's operand size, addr size,
 * segment register, and repeat code (if any) and is willing to tell other
 * objects about it....
 *
 * Revision 1.5  1997/02/14 04:17:10  yoel
 * Added Prefix Counting code and NumPrefixes() accessor.
 *
 * Revision 1.4  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef __PREFIX_H__
#define __PREFIX_H__

#include "ints.h"

//  We need to be able to recognize all the prefixes when we decode an
// instruction so #def all of them.
//  Group 1:
#define PRE_REP		0xF3
#define PRE_REPZ	0xF3
#define PRE_REPNZ	0xF2
#define PRE_LOCK	0xF0
//  Group 2:
#define PRE_SEGCS	0x2E
#define PRE_SEGSS	0x36
#define PRE_SEGDS	0x3E
#define PRE_SEGES	0x26
#define PRE_SEGFS	0x64
#define PRE_SEGGS	0x65
//  Group 3:
#define PRE_OPSIZE	0x66
//  Group 4:
#define PRE_ADDRSIZE	0x67

enum RepPrefix { RepNone, RepRepZ, RepRepNZ, RepLock };

class Prefix {

public:

  Prefix(int defaultSize);
  Prefix(const Prefix& inPrefix);

  int8 NumPrefixes(void) const { return nPrefixes; }
  
  int OpSize(void) const { return opSize; }
  int AddrSize(void) const { return addrSize; }
  int SegReg(void) const { return pSeg; }
  RepPrefix Rep(void) const { return (pRep == RepLock) ? RepNone : pRep; }
  RepPrefix RepOrLock(void) const { return pRep; }
  
  //  Returns -1 on illegal prefix, 0 on not a prefix, 1 on valid prefix
  int CheckPrefix(int8 inByte);

protected:
  int8 nPrefixes;
  RepPrefix pRep;
  int pSeg;
  bool pOpSize, pAddrSize;
  int opSize, addrSize;

};

#endif
