/* -*- C++ -*-
 *
 * $Id: registers.h,v 1.4 1998/04/29 04:54:57 nikos Exp $
 * 
 * RCS Log:
 * $Log: registers.h,v $
 * Revision 1.4  1998/04/29 04:54:57  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:49  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:34  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:08  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:06  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: registers.h,v 1.4 1998/04/29 04:54:57 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/registers.h,v 1.4 1998/04/29 04:54:57 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:29  bugnion
 * stanford update (x86)
 *
 * Revision 1.6  1997/03/06  18:52:51  sean
 * fixed #define's for 16-bit registers so that they always match their 32-bit
 *   counterparts (we had REG_EIP = 30, but REG_IP = 28, and that's bad)
 *
 * Revision 1.5  1997/02/23 03:24:33  sean
 * added TR (task register) to register file, segment descriptors
 *
 * Revision 1.4  1997/02/06 09:48:32  sean
 * created EAX_TOP pseudo-register
 *
 * Revision 1.3  1997/02/04 22:32:03  sean
 * created macros for 64-bit registers, create TSC (time stamp counter)
 *
 * Revision 1.2  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef _REGISTERS_H
#define _REGISTERS_H

#include "ints.h"

// We use these two macros to decide how to address hi and lo halves of a
//  64-bit register
#ifdef LITTLE_ENDIAN
#define LO_HALF(reg64)   ((reg64) << 1)
#define HI_HALF(reg64)   (((reg64) << 1) + 1)
#else
#define LO_HALF(reg64)   (((reg64) << 1) + 1)
#define HI_HALF(reg64)   ((reg64) << 1)
#endif

#define REG_NONE      -1

// 64-bit registers
#define REG_TSC       14

// 32-bit registers
#define REG_EAX        0
#define REG_ECX        1
#define REG_EDX        2
#define REG_EBX        3
#define REG_ESP        4
#define REG_EBP        5
#define REG_ESI        6
#define REG_EDI        7
#define REG_EFLAGS     8

#define REG_CRBASE     9
#define REG_CR0        9
#define REG_CR1       10
#define REG_CR2       11
#define REG_CR3       12
#define REG_PDBR      REG_CR3
#define REG_CR4       13

#define REG_DRBASE    14
#define REG_DR0       14
#define REG_DR1       15
#define REG_DR2       16
#define REG_DR3       17
#define REG_DR4       18
#define REG_DR5       19
#define REG_DR6       20
#define REG_DR7       21

#define REG_SEGBASE   22
#define REG_ES_TR     22
#define REG_CS_XXX    23
#define REG_SS_LDT    24
#define REG_DS_XXX    25
#define REG_FS_XXX    26
#define REG_GS_XXX    27

// These are 28 and 29 (REG_TSC = 14)
#define REG_TSC_LO    LO_HALF(REG_TSC)
#define REG_TSC_HI    HI_HALF(REG_TSC)

#define REG_EIP       30

#define REG_TOTAL     31

// 16-bit registers  (bottom 16 bits of 32-bit registers have same number)
#define REG_AX         REG_EAX
#define REG_CX         REG_ECX
#define REG_DX         REG_EDX
#define REG_BX         REG_EBX
#define REG_SP         REG_ESP
#define REG_BP         REG_EBP
#define REG_SI         REG_ESI
#define REG_DI         REG_EDI
#define REG_FLAGS      REG_EFLAGS

// this isn't really a register, but it's nice to be able to get at
#define REG_EAX_TOP  128

#define REG_ES        22
#define REG_CS        23
#define REG_SS        24
#define REG_DS        25
#define REG_FS        26
#define REG_GS        27
#define REG_TR       150
#define REG_LDT      152
#define REG_IP         REG_EIP

// 8-bit registers
#define REG_AL         0
#define REG_CL         1
#define REG_DL         2
#define REG_BL         3
#define REG_AH         4
#define REG_CH         5
#define REG_DH         6
#define REG_BH         7

// Segment descriptor numbers
#define SDESC_ES       0
#define SDESC_CS       1
#define SDESC_SS       2
#define SDESC_DS       3
#define SDESC_FS       4
#define SDESC_GS       5
#define SDESC_GDT      6
#define SDESC_IDT      7
#define SDESC_LDT      8
#define SDESC_TR       9
#define SDESC_TOTAL   10

// flag numbers (in EFLAGS)
#define FLAG_CF        0
#define FLAG_PF        2
#define FLAG_AF        4
#define FLAG_ZF        6
#define FLAG_SF        7
#define FLAG_TF        8
#define FLAG_IF        9
#define FLAG_DF       10
#define FLAG_OF       11
#define FLAG_IOPL     12
#define FLAG_NT       14
#define FLAG_RF       16
#define FLAG_VM       17
#define FLAG_AC       18
#define FLAG_VIF      19
#define FLAG_VIP      20
#define FLAG_ID       21

#define FLAG_DEFINED  0x003F7FD5
#define FLAG_UNDEFVAL 0x00000002

typedef int32 RegisterFile[REG_TOTAL];

#define REG32(regfile,regnum)    (*(((int32 *)(regfile))+(regnum)))

#define REG64(regfile,regnum)    (*(((int64 *)(regfile))+(regnum)))

#ifdef LITTLE_ENDIAN
#define REG16(regfile,regnum)    (*(((int16 *)(regfile))+(((regnum) & 0x7f) << 1)+((regnum) >> 7)))
#define REG8(regfile,regnum)     (*(((int8 *)(regfile))+(((regnum) & 3) << 2)+((regnum) >> 2)))
#else
// Big Endian
#define REG16(regfile,regnum)    (*(((int16 *)(regfile))+(((regnum) & 0x7f) << 1)+1-((regnum) >> 7)))
#define REG8(regfile,regnum)     (*(((int8 *)(regfile))+(((regnum) & 3) << 2)+3-((regnum) >> 2)))
#endif

#endif
