/* $Id: seg_desc.cc,v 1.4 1998/04/29 04:54:57 nikos Exp $
 *
 * RCS Log:
 * $Log: seg_desc.cc,v $
 * Revision 1.4  1998/04/29 04:54:57  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:49  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:34  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:09  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:07  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: seg_desc.cc,v 1.4 1998/04/29 04:54:57 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/seg_desc.cc,v 1.4 1998/04/29 04:54:57 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:29  bugnion
// stanford update (x86)
//
// Revision 1.7  1997/12/03  00:45:40  mendel
// Some fixes to the IDT descriptor access routines.
//
// Revision 1.6  1997/12/02  22:52:15  mendel
// 1) Hooked up interrupts. Moved interrupt code from simos_interface.cc to
//    cpu.cc .
//
// Several bug fixes:
//
// 1) Added check to push the error code only on processor generated
//    exceptions and not on INT or external ones.
//
// 2) IDT gate descriptors were broken. I added a gate_use_32bit() since
//    IDT gate descriptors use a different size field that normal segment
//    descriptors.
//
// 3) Fixed bug in computing spaceNeeded for exception stack puke. Beware
//    of C operator precedence.
//
// 4) Cleaned up some ugly and some incorrect code in HandleException()
//
// 5) #if 0 out some bogus check LoadExceptionDescriptor(). I couldn't
//    tell want it was trying to do but is was wrong.
//
// 6) Compute the gate_offset correctly. We need all the bits of the
//    flags for this to work.
//
// 7) Changed POP and MOV to not call WriteBackDescriptor() on selector
//    zero. This caused WriteBackDescriptor() to die.
//
// Revision 1.5  1997/03/06  02:11:06  sean
// fixed segment limits to be 32-bit integers (really need only 20), and
//   fixed is_code() and is_data() methods (they were just plain borken)
//
 * Revision 1.4  1997/03/02 03:05:40  sean
 * minor restructuring of methods, member data - modifications to handle
 *   non code/data segment descriptor types
 *
 * Revision 1.3  1997/02/23 03:30:24  sean
 * segment descriptors now know what system segments are, and can give
 *   (some) useful information about them
 *
 * Revision 1.2  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: seg_desc.cc,v 1.4 1998/04/29 04:54:57 nikos Exp $";

#include "seg_desc.h"

// SegmentDescriptor stuff

SegmentDescriptor& SegmentDescriptor::operator=(const SegmentDescriptor& copy_from)
{
  desc_valid = copy_from.desc_valid;
  desc_base  = copy_from.desc_base;
  desc_limit = copy_from.desc_limit;
  flags      = copy_from.flags;

  return(*this);
}

void SegmentDescriptor::compress_descriptor(void *data) const
{
  ((int32 *)data)[0] = htox32((BITS(desc_base,0,16) << 16) | 
			      BITS(desc_limit,0,16));
  ((int32 *)data)[1] = htox32((BITS(desc_base,24,8) << 24) | (flags << 8) |
			      (BITS(desc_limit,16,4) << 16) | 
			      BITS(desc_base,16,8));
}

void SegmentDescriptor::extract_descriptor(const void *data)
{
  int32 data0, data1;

  data0 = xtoh32(((int32 *)data)[0]);
  data1 = xtoh32(((int32 *)data)[1]);

  desc_base = (BITS(data1,24,8) << 24) |
              (BITS(data1, 0,8) << 16) |
              BITS(data0,16,16);

  desc_limit = (BITS(data1,16,4) << 16) | 
               BITS(data0,0,16);

  flags = BITS(data1,8,16) & 0x0F0FF;

  desc_valid = true;
}

void SegmentDescriptor::set_base_limit(int32 _base, int32 _limit, int _flags)
{
  desc_base = _base;
  desc_limit = _limit;
  flags = _flags;
  desc_valid = true;
}

bool SegmentDescriptor::offset_valid(int32 offset)
{
  if(!use_32bit() && (offset > 0x0FFFF)) return(false);

  if(big_granularity()) offset >>= 12;

  return(expand_down() ? (offset > desc_limit) : (offset <= desc_limit));
}

int SegmentDescriptor::bytes_available(int32 offset)
{
  int32 max;

  if(!offset_valid(offset)) return(0);

  if(expand_down()) {
    max = 0xFFFFFFFF;
  } else
    if(big_granularity())
      max = (((int32) desc_limit) << 12) | 0x00000FFF; else
      max = desc_limit;

  if(!use_32bit()) max &= 0x0000FFFF;

  return((max - offset) + 1);
}
