/* -*- C++ -*-
 *
 * $Id: seg_desc.h,v 1.4 1998/04/29 04:54:58 nikos Exp $
 * 
 * RCS Log:
 * $Log: seg_desc.h,v $
 * Revision 1.4  1998/04/29 04:54:58  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:49  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:34  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:09  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:08  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: seg_desc.h,v 1.4 1998/04/29 04:54:58 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/seg_desc.h,v 1.4 1998/04/29 04:54:58 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:30  bugnion
 * stanford update (x86)
 *
 * Revision 1.14  1997/12/09  05:33:26  mendel
 * Bugs patched around (These need to be fixed):
 *    1) ifdef'ed out the protection checking in LoadDescriptor() that
 *       was preventing IRET from working. See comment in code I added in
 *       cpu.cc
 *    2) ifdefed out the ASSERT(0) in the TLB flush routine in x86_instr.cc
 *    3) ifdef'ed out the protection checking in IRET that wasn't working.
 *
 *
 * Bugs:
 *
 * 1) Check in some bug fixes in fpu.cc that Sean gave me.
 *
 * 2) Fixed bug in reading stack segment selector in TRAP_GATE . We were
 * ASSERT(0)ing when the ReadFromVirtualAddress() was successful! in cpu.cc
 * HandleExecption()
 *
 * 3) Changed (CSDescriptor.dpl() != SSSelector & 0x3) to
 *            (CSDescriptor.dpl() != (SSSelector & 0x3))
 * for obvious reasons in cpu.cc / HandleException().
 *
 * 4) Test for GDT in TRAP_GATE code was backward in cpu.cc /
 * HandleException()
 *
 * 5) Since the prev_eip field of CPU class is used to set the return PC
 * of execeptions. We need to set prev_eip on instruction page faults to
 * be the faulting instruction. Otherwise we think the exception happen
 * on the PC of IRET instruction. ouch!
 *
 * 6) The MemoryOperand() code in instruction.cc wasn't setting the
 * default segment of the ESP register to be the stack segment! Found
 * this because Linux runs for a while with the user's DS selector
 * loaded and pushes things on the stack.
 *
 * 7) The is_task() method of seg_desc was completely bogus.
 *
 *
 * Clean up:
 *
 * 1) Removed fprintf() on exceptions in cpu.cc since we are taking far too many
 * execptions for this to be useful.
 *
 * Revision 1.13  1997/12/03 01:15:30  mendel
 * Mistake from last checkin. This is pretty ugly stuff.
 *
 * Revision 1.12  1997/12/03  00:45:41  mendel
 * Some fixes to the IDT descriptor access routines.
 *
 * Revision 1.11  1997/12/02  22:52:16  mendel
 * 1) Hooked up interrupts. Moved interrupt code from simos_interface.cc to
 *    cpu.cc .
 *
 * Several bug fixes:
 *
 * 1) Added check to push the error code only on processor generated
 *    exceptions and not on INT or external ones.
 *
 * 2) IDT gate descriptors were broken. I added a gate_use_32bit() since
 *    IDT gate descriptors use a different size field that normal segment
 *    descriptors.
 *
 * 3) Fixed bug in computing spaceNeeded for exception stack puke. Beware
 *    of C operator precedence.
 *
 * 4) Cleaned up some ugly and some incorrect code in HandleException()
 *
 * 5) #if 0 out some bogus check LoadExceptionDescriptor(). I couldn't
 *    tell want it was trying to do but is was wrong.
 *
 * 6) Compute the gate_offset correctly. We need all the bits of the
 *    flags for this to work.
 *
 * 7) Changed POP and MOV to not call WriteBackDescriptor() on selector
 *    zero. This caused WriteBackDescriptor() to die.
 *
 * Revision 1.10  1997/11/17  08:17:42  whimsy
 * Stupid file doesn't know about the SIZE_* #def's,m so I hardcoded the sizes.
 *
 * Revision 1.9  1997-11-13 21:49:04-08  whimsy
 * Added op_size() method, veryt useful for Interupt code.
 *
 * Revision 1.8  1997-11-06 17:15:21-08  yoel
 * Added DMASK_IDT for handling interupts.
 *
 * Revision 1.7  1997-10-23 20:50:45-07  whimsy
 * Added is_task() method to check if a descriptor refers to a TSS or Task Gate
 * and so needs to #TS instead of #GP on permission errors.
 *
 * Revision 1.6  1997-10-19 19:33:55-07  whimsy
 * Added methods to treat a segment descriptor as a Gate of some sort (i.e. to
 * extract the offset and new selectors from Call / Task / Interupt Gates)
 *
 * Revision 1.5  1997-03-05 18:11:06-08  sean
 * fixed segment limits to be 32-bit integers (really need only 20), and
 *   fixed is_code() and is_data() methods (they were just plain borken)
 *
 * Revision 1.4  1997/03/02 03:05:40  sean
 * minor restructuring of methods, member data - modifications to handle
 *   non code/data segment descriptor types
 *
 * Revision 1.3  1997/02/23 03:30:24  sean
 * segment descriptors now know what system segments are, and can give
 *   (some) useful information about them
 *
 * Revision 1.2  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

#ifndef _SEG_DESC_H
#define _SEG_DESC_H

#include "ints.h"

// Segment Descriptor Types
enum DescriptorType { DESC_RESERVED0        =  0,
		      DESC_16BIT_TSS_AVAIL  =  1,
		      DESC_LDT              =  2,
		      DESC_16BIT_TSS_BUSY   =  3,
		      DESC_16BIT_CALL_GATE  =  4,
		      DESC_TASK_GATE        =  5,
		      DESC_16BIT_INT_GATE   =  6,
		      DESC_16BIT_TRAP_GATE  =  7,
		      DESC_RESERVED8        =  8,
		      DESC_32BIT_TSS_AVAIL  =  9,
		      DESC_RESERVED10       = 10,
		      DESC_32BIT_TSS_BUSY   = 11,
		      DESC_32BIT_CALL_GATE  = 12,
		      DESC_RESERVED13       = 13,
		      DESC_32BIT_INT_GATE   = 14,
		      DESC_32BIT_TRAP_GATE  = 15,
		      DESC_DATA_RO          = 16,
		      DESC_DATA_RO_ACC      = 17,
		      DESC_DATA_RW          = 18,
		      DESC_DATA_RW_ACC      = 19,
		      DESC_DATA_RO_DOWN     = 20,
		      DESC_DATA_RO_DOWN_ACC = 21,
		      DESC_DATA_RW_DOWN     = 22,
		      DESC_DATA_RW_DOWN_ACC = 23,
		      DESC_CODE_EO          = 24,
		      DESC_CODE_EO_ACC      = 25,
		      DESC_CODE_ER          = 26,
		      DESC_CODE_ER_ACC      = 27,
		      DESC_CODE_EO_CONF     = 28,
		      DESC_CODE_EO_CONF_ACC = 29,
		      DESC_CODE_ER_CONF     = 30,
		      DESC_CODE_ER_CONF_ACC = 31
};

// Masks for valid descriptor types (used by CPU::LoadDescriptor)
#define DMASK_16BIT_TSS_AVAIL       (1 << DESC_16BIT_TSS_AVAIL)
#define DMASK_LDT                   (1 << DESC_LDT)
#define DMASK_16BIT_TSS_BUSY        (1 << DESC_16BIT_TSS_BUSY)
#define DMASK_16BIT_CALL_GATE       (1 << DESC_16BIT_CALL_GATE)
#define DMASK_TASK_GATE             (1 << DESC_TASK_GATE)
#define DMASK_16BIT_INT_GATE        (1 << DESC_16BIT_INT_GATE)
#define DMASK_16BIT_TRAP_GATE       (1 << DESC_16BIT_TRAP_GATE)
#define DMASK_32BIT_TSS_AVAIL       (1 << DESC_32BIT_TSS_AVAIL)
#define DMASK_32BIT_TSS_BUSY        (1 << DESC_32BIT_TSS_BUSY)
#define DMASK_32BIT_CALL_GATE       (1 << DESC_32BIT_CALL_GATE)
#define DMASK_32BIT_INT_GATE        (1 << DESC_32BIT_INT_GATE)
#define DMASK_32BIT_TRAP_GATE       (1 << DESC_32BIT_TRAP_GATE)
#define DMASK_DATA_RO               (1 << DESC_DATA_RO)
#define DMASK_DATA_RO_ACC           (1 << DESC_DATA_RO_ACC)
#define DMASK_DATA_RW               (1 << DESC_DATA_RW)
#define DMASK_DATA_RW_ACC           (1 << DESC_DATA_RW_ACC)
#define DMASK_DATA_RO_DOWN          (1 << DESC_DATA_RO_DOWN)
#define DMASK_DATA_RO_DOWN_ACC      (1 << DESC_DATA_RO_DOWN_ACC)
#define DMASK_DATA_RW_DOWN          (1 << DESC_DATA_RW_DOWN)
#define DMASK_DATA_RW_DOWN_ACC      (1 << DESC_DATA_RW_DOWN_ACC)
#define DMASK_CODE_EO               (1 << DESC_CODE_EO)
#define DMASK_CODE_EO_ACC           (1 << DESC_CODE_EO_ACC)
#define DMASK_CODE_ER               (1 << DESC_CODE_ER)
#define DMASK_CODE_ER_ACC           (1 << DESC_CODE_ER_ACC)
#define DMASK_CODE_EO_CONF          (1 << DESC_CODE_EO_CONF)
#define DMASK_CODE_EO_CONF_ACC      (1 << DESC_CODE_EO_CONF_ACC)
#define DMASK_CODE_ER_CONF          (1 << DESC_CODE_ER_CONF)
#define DMASK_CODE_ER_CONF_ACC      (1 << DESC_CODE_ER_CONF_ACC)

// Now a few helpful mergers
#define DMASK_EXECUTABLE            (DMASK_CODE_EO          | \
				     DMASK_CODE_EO_ACC      | \
				     DMASK_CODE_ER          | \
				     DMASK_CODE_ER_ACC      | \
				     DMASK_CODE_EO_CONF     | \
				     DMASK_CODE_EO_CONF_ACC | \
				     DMASK_CODE_ER_CONF     | \
				     DMASK_CODE_ER_CONF_ACC)
#define DMASK_READABLE              (DMASK_DATA_RO          | \
				     DMASK_DATA_RO_ACC      | \
				     DMASK_DATA_RW          | \
				     DMASK_DATA_RW_ACC      | \
				     DMASK_DATA_RO_DOWN     | \
				     DMASK_DATA_RO_DOWN_ACC | \
				     DMASK_DATA_RW_DOWN     | \
				     DMASK_DATA_RW_DOWN_ACC | \
				     DMASK_CODE_ER          | \
				     DMASK_CODE_ER_ACC      | \
				     DMASK_CODE_ER_CONF     | \
				     DMASK_CODE_ER_CONF_ACC)
#define DMASK_WRITEABLE             (DMASK_DATA_RW          | \
				     DMASK_DATA_RW_ACC      | \
				     DMASK_DATA_RW_DOWN     | \
				     DMASK_DATA_RW_DOWN_ACC)
#define DMASK_DATA                  (DMASK_DATA_RO          | \
				     DMASK_DATA_RO_ACC      | \
				     DMASK_DATA_RW          | \
				     DMASK_DATA_RW_ACC      | \
				     DMASK_DATA_RO_DOWN     | \
				     DMASK_DATA_RO_DOWN_ACC | \
				     DMASK_DATA_RW_DOWN     | \
				     DMASK_DATA_RW_DOWN_ACC)
#define DMASK_CODE                  DMASK_EXECUTABLE
#define DMASK_HAS_AR_BYTE           (DMASK_CODE             | \
				     DMASK_DATA             | \
				     DMASK_16BIT_TSS_AVAIL  | \
				     DMASK_LDT              | \
				     DMASK_16BIT_TSS_BUSY   | \
				     DMASK_16BIT_CALL_GATE  | \
				     DMASK_TASK_GATE        | \
				     DMASK_32BIT_TSS_AVAIL  | \
				     DMASK_32BIT_TSS_BUSY   | \
				     DMASK_32BIT_CALL_GATE)
#define DMASK_HAS_LIMIT             (DMASK_CODE             | \
				     DMASK_DATA             | \
				     DMASK_16BIT_TSS_AVAIL  | \
				     DMASK_LDT              | \
				     DMASK_16BIT_TSS_BUSY   | \
				     DMASK_32BIT_TSS_AVAIL  | \
				     DMASK_32BIT_TSS_BUSY)
#define DMASK_AVAIL_TSS             (DMASK_16BIT_TSS_AVAIL  | \
				     DMASK_32BIT_TSS_AVAIL)
#define DMASK_BUSY_TSS              (DMASK_16BIT_TSS_BUSY  | \
				     DMASK_32BIT_TSS_BUSY)
#define DMASK_CAN_VERIFY            (DMASK_CODE             | \
				     DMASK_DATA)
#define DMASK_CAN_CALL              (DMASK_CODE             | \
				     DMASK_16BIT_TSS_AVAIL  | \
				     DMASK_16BIT_CALL_GATE  | \
				     DMASK_TASK_GATE        | \
				     DMASK_32BIT_TSS_AVAIL  | \
				     DMASK_32BIT_CALL_GATE)

// If this type of descriptor needs to generate a #TS instead of #GP on errors.
#define DMASK_TASK                 (DMASK_TASK_GATE        | \
                                    DMASK_16BIT_TSS_AVAIL  | \
				    DMASK_32BIT_TSS_AVAIL  | \
				    DMASK_16BIT_TSS_BUSY   | \
				    DMASK_32BIT_TSS_BUSY)

// Legal Descriptor in the IDT
#define DMASK_IDT		   (DMASK_TASK_GATE        | \
                                    DMASK_16BIT_INT_GATE   | \
                                    DMASK_16BIT_TRAP_GATE  | \
                                    DMASK_32BIT_INT_GATE   | \
                                    DMASK_32BIT_TRAP_GATE)

// Common segment descriptor flag settings
#define PRESENT_RW_ACCESSED     (0x0080 | DESC_DATA_RW_ACC)
#define PRESENT_RW              (0x0080 | DESC_DATA_RW)

// This encapsulates a segment descriptor.  As far as the CPU is concerned,
//  this is a read-only structure, with the single exception of the accessed
//  bit.
class SegmentDescriptor {
 public:
  SegmentDescriptor& operator=(const SegmentDescriptor& copy_from);

  void compress_descriptor(void *data) const;
  void extract_descriptor(const void *data);
  void set_base_limit(int32 _base, int32 _limit, int _flags = PRESENT_RW);
 
  inline int32 base(void) const           { return(desc_base); }
  inline int32 limit(void) const          { return(desc_limit); }
  inline int32 ar_byte(void) const        { return(((int32) flags) << 8); }

  inline bool is_valid(void) const        { return(desc_valid); }
  inline void mark_invalid(void)          { desc_valid = false; }

  bool offset_valid(int32 offset);
  int bytes_available(int32 offset);

  inline bool big_granularity(void) const { return(BITS(flags,15,1)); }
  inline bool use_32bit(void) const       { return(BITS(flags,14,1)); }
  inline int8 desc_size(void) const  { return (use_32bit() ?  4 : 2); }
  inline bool is_present(void) const      { return(BITS(flags, 7,1)); }
  inline char dpl(void) const             { return(BITS(flags, 5,2)); }

  // stuff to tell the type of the descriptor - these first four work on any
  //  type of descriptor
  inline DescriptorType type(void) const  { return((DescriptorType) BITS(flags, 0,5)); }

  inline bool application_seg(void) const { return(BITS(flags, 4,1)); }
  inline bool is_code(void) const         { return(BITS(flags, 3,2) == 3); }
  inline bool is_data(void) const         { return(BITS(flags, 3,2) == 2); }
  inline bool is_task(void) const         { return (!application_seg() && 
                                                    (((type()&0x5) == 0x3) || 
                                                     (type() == DESC_TASK_GATE))); }
  // these next four only work properly on code and data segments (basically,
  //  it's a pain to check to make sure it's not a system segment when we're
  //  never going to call these when it might be a system segment)
  inline bool expand_down(void) const     { return(BITS(flags, 2,1) && 
						   is_data()); }
  inline bool is_readable(void) const     { return(BITS(flags, 1,1) || 
						   is_data()); }
  inline bool is_writeable(void) const    { return(BITS(flags, 1,1) && 
						   is_data()); }
  inline bool is_conforming(void) const   { return(BITS(flags, 2,1) && 
						   is_code()); }

  // assuming the descriptor refers to a code or data segment, this says if
  //  it's been accessed
  inline bool accessed(void) const        { return(BITS(flags, 0,1)); }
  inline void set_accessed(bool a = true) { SET_BITS(flags,0,1,a); }
  
  // assuming the descriptor refers to a TSS, this says if it's busy
  inline bool tss_busy(void) const        { return(BITS(flags, 1,1)); }
  inline void set_tss_busy(bool b = true) { SET_BITS(flags,1,1,b); }

  //  Assuming the descriptor is a Gate of some sort we need to able to
  // extract somethings from it that aren't part of the normal internal
  // representation.
  inline int16 gate_selector(void) const { return (desc_base & 0xFFFF); }
  inline int32 gate_offset(void) const
   { return (desc_limit | ((flags & 0xF000)<<8) | (desc_base & 0xFF000000)); }
  inline int8 gate_nparam(void) const  { return BITS(desc_base, 16, 5); }
  inline bool gate_use_32bit(void) const  { return(BITS(flags,3,1)); }
 private:
  bool  desc_valid; // Is this segment descriptor valid?
  int32 desc_base;  // The linear base address of the segment
  int32 desc_limit; // The first invalid (or valid, for expand-down segs)
                    //  byte, relative to the base, NOT considering granularity
  int16 flags;      // The rest of the stuff in the descriptor can have
                    //  all sorts of meanings, so we stuff it all here, and
                    //  use inlined accessors to get at them - In most cases,
                    //  this is the AR (access rights) "byte"
};

#endif
