/* $Id: simlib.cc,v 1.4 1998/04/29 04:54:58 nikos Exp $
 *
 * RCS Log:
 * $Log: simlib.cc,v $
 * Revision 1.4  1998/04/29 04:54:58  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:50  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:35  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:10  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:08  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: simlib.cc,v 1.4 1998/04/29 04:54:58 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/simlib.cc,v 1.4 1998/04/29 04:54:58 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:30  bugnion
// stanford update (x86)
//
// Revision 1.10  1997/12/09  14:51:52  sjt
// FPU emulation should now work completely - FPU exceptions are triggered when
// 	they should be, and memory operands appear to work fine.  I removed the
// 	printfs that were cluttering everything too.
//
// Mendel's fixes to BT/BTC/BTR/BTS and ADC have been stolen and checked in as
// 	well.
//
 * Revision 1.9  1997/12/07 02:54:16  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.8  1997/11/22 07:41:42  whimsy
 * Added code to skip FPU instructions for SimOS incorporation.
 *
 * Revision 1.7  1997-10-29 14:47:45-08  sean
 * Changed SimulateLinuxSystemCall() to return whether or not CPU should be
 *   halted.
 * Added ExecuteBuffer() function to simulate FPU ops on the real FPU.
 *
 * Revision 1.6  1997/09/24 08:13:07  sean
 * brk and exit (sorta) system calls are wrappered, allowing better simulation
 *  of native code
 *
 * Revision 1.5  1997/03/18 19:10:03  sean
 * armored Linux-only stuff with #ifdef NATIVE_X86_HOST
 * moved the scary CopyAndSwitchCode() and CopyAndSwitchStacks() code in, so
 *   they could be with the rest of the evil assembler code
 *
 * Revision 1.4  1997/03/15 03:15:57  yoel
 * Removed the rdtsc stuff since no x86 before Pentium supports that.
 *
 * Revision 1.3  1997/03/15 02:18:20  yoel
 * Added stdlib.h for prototypes of atoi() and strtol(), start_sim now takes
 * a char * of debug flags.
 *
 * Revision 1.2  1997/03/14 21:43:44  sean
 * improved simulator to either single step (just hit enter), execute a
 *   certain number of instructions at a time (type in the number), execute
 *   until (E)IP is equal to a break point (enter the breakpoint in hex
 *   starting with 0x), or execute forever (enter -1).
 *
 * Revision 1.1  1997/03/07 05:56:01  sean
 * Initial revision
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: simlib.cc,v 1.4 1998/04/29 04:54:58 nikos Exp $";

#include <stdio.h>
#include <stdlib.h>

#ifdef NATIVE_X86_HOST
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/syscall.h>

#define PAGE_ALIGN(ptr)  (((ptr) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))
#endif

#include "simlib.h"
#include "registers.h"
#include "cpu.h"
#include "debug.h"

/* The link-in simulator only exists if we are a native Linux host -
     if not, we just comment out the entire functions */
#ifdef NATIVE_X86_HOST
// we need to keep track of the simulated program's 'brk' value so that
//  we can emulate the brk system call (otherwise the simulator's brk
//  value gets used.
int32 simulated_brk = 0;
int32 simulated_min_brk = 0;  // smallest brk is allowed to be (end of text)

#define STRINGIFY(s)         "(" ## #s ## ")"
// If you want the value of a #define made into a string, use STRING_VAL
#define STRING_VAL(x)        STRINGIFY(x)
#define R_POS(size,rnum,ofs) STRINGIFY((rnum << 2) + ofs) ## "(%ebx)"
#define R32_POS(rnum)        R_POS(REGFILESIZE,rnum,0)
#define R16_POS(rnum)        R_POS(REGFILESIZE,((rnum) & 0x7f),((rnum) >> 6))

// This is the size of the stack the the simulator needs to live in
#define SIM_STACK_SIZE       65536
#define SIM_STACK_SIZE_STR   STRING_VAL(SIM_STACK_SIZE)

void *start_sim(const char *debug_codes)
{
  RegisterFile *regptr;    /* this is at -4(%ebp) */
  int32        gdt[2];     /* this is at -12(%ebp) */
  int32        idt[2];     /* this is at -20(%ebp) */
  SegmentDescriptor *seg_descs;    /* at -24(%ebp) */

  int8         *new_stack; /* this is the stack for the simulator */
  CPU          *cpu;       /* the CPU we'll use to simulate */
  int          *quit_flag; /* the flag that is used to mark when we want to
			      quit the simulator (we set it to 0, we're done
			      when it's non-zero) */
  int          count;      /* number of cycles left to run */
  char         line[40];   /* input */
  int32        target;     /* target */

  // Enable any debug codes...
  while (*debug_codes != '\0') TurnOnDebugCode(*(debug_codes++));

  // this is somewhat strange - we are actually passing NULL as the base of
  //  the memory buffer we want our cpu to use, since we are using our "real"
  //  data segment as our simulated physical memory space.
  cpu = new CPU(0,0x0FFFFFFFF);
  quit_flag = new int;
  new_stack = new int8[SIM_STACK_SIZE];

  *quit_flag = 0;

  // we need to clobber the segment descriptors to get them to point at all
  //  of our memory space...
  seg_descs = cpu->GetDescriptors();

  seg_descs[SDESC_ES ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_CS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_CODE_ER_ACC));
  seg_descs[SDESC_SS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_DS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_FS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));
  seg_descs[SDESC_GS ].set_base_limit(0x00000000,0x0BFFFF,
				      (0x0C0E0 | DESC_DATA_RW_ACC));

  // now we can just ask the cpu for its register file, which we'll play with
  regptr = cpu->GetRegisters();

  // save old %ebx, and make %ebx point to the regfile
  __asm__("pushl %ebx; movl -4(%ebp), %ebx");

  // grab general registers
  __asm__("movl %eax, " R32_POS(REG_EAX));
  __asm__("movl %ecx, " R32_POS(REG_ECX));
  __asm__("movl %edx, " R32_POS(REG_EDX));
  __asm__("popl %eax");   // %eax now has the old value of %ebx
  __asm__("movl %eax, " R32_POS(REG_EBX));
  __asm__("movl %esp, " R32_POS(REG_ESP));
  __asm__("movl %ebp, " R32_POS(REG_EBP));
  __asm__("movl %esi, " R32_POS(REG_ESI));
  __asm__("movl %edi, " R32_POS(REG_EDI));

  // get the flags register
  __asm__("pushfl; popl %eax");
  __asm__("movl %eax, " R32_POS(REG_EFLAGS));

  // grab segment registers
  __asm__("movw %es,  " R16_POS(REG_ES));
  __asm__("movw %cs,  " R16_POS(REG_CS));
  __asm__("movw %ss,  " R16_POS(REG_SS));
  __asm__("movw %ds,  " R16_POS(REG_DS));
  __asm__("movw %fs,  " R16_POS(REG_FS));
  __asm__("movw %gs,  " R16_POS(REG_GS));
  __asm__("str        " R16_POS(REG_TR));
  __asm__("sldt       " R16_POS(REG_LDT));

  // slurp the time stamp counter
#ifdef 0
  // not included because: a) it isn't important, b) only works on pentiums
  __asm__(".byte 0x0f,0x31");   // rdtsc
#endif
  __asm__("xorl %eax, %eax; movl %eax, %edx");  // want 0 TSC anyway
  __asm__("movl %edx, " R32_POS(REG_TSC_HI));
  __asm__("movl %eax, " R32_POS(REG_TSC_LO));

  // can't read control, debug registers because we are at PL 3

  // make sure CR0 says we're in protected mode
  REG32(*regptr,REG_CR0) |= 1;

  // try to get GDT, IDT values
  __asm__("sgdt -10(%ebp)");   gdt[0] >>= 16;
  __asm__("sidt -18(%ebp)");   idt[0] >>= 16;

  // finally, grab the value we want to stick into EIP (the simulator
  //  basically thinks we did a jump from this line to the StartLabel,
  //  so we have to make sure the %esp is the same at these two points.
  __asm__("leal .StartLabel, %eax; movl %eax, "R32_POS(REG_EIP));

  // Now, before starting the simulator, copy and switch stacks
  __asm__("movl -28(%ebp), %ebx");  // ebx has base of new stack
  __asm__("addl $" SIM_STACK_SIZE_STR ", %ebx");  // now its the top

  __asm__("movl  -4(%ebp), %eax; movl %eax,  -4(%ebx)");   // copy 'regptr'
  __asm__("movl -28(%ebp), %eax; movl %eax, -28(%ebx)");   // copy 'new_stack'
  __asm__("movl -32(%ebp), %eax; movl %eax, -32(%ebx)");   // copy 'cpu'
  __asm__("movl -36(%ebp), %eax; movl %eax, -36(%ebx)");   // copy 'quit_flag'

  __asm__("movl %ebp, %eax; subl %esp, %eax");  // EAX = (EBP - ESP)
  __asm__("movl %ebx, %ebp");                   // set new EBP
  __asm__("subl %eax, %ebx; movl %ebx, %esp");  // set new ESP

  count = 0;
  target = 0;

  while(!*quit_flag && !cpu->is_halted()) {
    if(REG32(*regptr,REG_EIP) == target) count = 0;

    if(!count) {
      cpu->DumpState();
      fprintf(stderr,"> ");
      fgets(line,39,stdin);

      if(line[0] == 'q') {
	printf("Simulator aborted.\n");
	exit(0);
      }

      if((line[0] != '0') || (line[1] != 'x')) {
	count = atoi(line);
	if(!count) count = 1;
	target = 0;
      } else {
	count = -1;
	target = strtol(line+2, (char **)NULL, 16);
      }
    }

    if(count == -2) cpu->DumpState();
    if(count > 0) count--;

    cpu->ExecuteOneInstruction();
  }

  if(cpu->is_halted()) {
    fprintf(stderr,"CPU Halted - Simulator stopped!\n");
    exit(0);
  }

  // The following code is used to restore the proper machine state

  // we need to: 1) copy value from temp stack to real stack, 2) copy in
  //   new register values, 3) jump to the proper EIP (by RETing)

  // grab the pointer to the register file
  __asm__("movl -4(%ebp), %ebx");
  __asm__("movl " R32_POS(REG_EBP) ", %esi");  // ESI gets "real" EBP

  // copy values (new_stack, cpu, quit_flag)
  __asm__("movl -28(%ebp), %eax; movl %eax,  -4(%esi)");
  __asm__("movl -32(%ebp), %eax; movl %eax,  -8(%esi)");
  __asm__("movl -36(%ebp), %eax; movl %eax, -12(%esi)");

  // restore registers (being careful of ebx, eip, eflags) - ignoring segregs
  // move SP first, so that pushes and pops work ok
  __asm__("movl " R32_POS(REG_ESP) ", %esp");

  __asm__("movl " R32_POS(REG_EAX) ", %eax");
  __asm__("movl " R32_POS(REG_ECX) ", %ecx");
  __asm__("movl " R32_POS(REG_EDX) ", %edx");
  __asm__("movl %esi, %ebp");         // from above, we have EBP in ESI
  __asm__("movl " R32_POS(REG_ESI) ", %esi");
  __asm__("movl " R32_POS(REG_EDI) ", %edi");

  // now for the ones we'll have to pop off the stack
  __asm__("pushl " R32_POS(REG_EIP));
  __asm__("pushl " R32_POS(REG_EFLAGS));
  __asm__("pushl " R32_POS(REG_EBX));

  // pop the remaining couple registers and jump to the new eip
  __asm__("popl %ebx; popfl; ret");

  // This code is the entry point for the simulator (starts by returning)

  __asm__(".StartLabel:");

  __asm__("xorl %eax, %eax");
  //  __asm__("movl $0x1, %eax");
  //  __asm__("pushl %ebp; popl %esp; popl %ebp; ret");
  return((void *) quit_flag);
}

void stop_sim(void *sim)
{
  /* these will get the values from the simulator stack when it stops */
  int8         *new_stack; /* this is the stack for the simulator */
  CPU          *cpu;       /* the CPU we'll use to simulate */
  int          *quit_flag; /* the flag that is used to mark when we want to
			      quit the simulator (we set it to 0, we're done
			      when it's non-zero) */

  *(int *)sim = 1;  // this kills the simulator

  // now we clean up the space used by the simulator
  delete[] new_stack;
  delete cpu;
  delete quit_flag;
}
#endif /* ifdef NATIVE_X86_HOST */

#ifdef NATIVE_X86_HOST
// a bunch of system call emulation functions
static int32 Emulate_brk(int32 new_brk)
{
  int32 old_last_page, new_last_page;

  // if value is below min_brk, we just return current value
  if(new_brk < simulated_min_brk) return(simulated_brk);

  old_last_page = PAGE_ALIGN(simulated_brk);
  new_last_page = PAGE_ALIGN(new_brk);

  if(old_last_page < new_last_page) { // need more memory
    if(mmap(old_last_page, (new_last_page - old_last_page),
	    PROT_READ | PROT_WRITE | PROT_EXEC /* just in case */,
	    MAP_ANON | MAP_FIXED | MAP_PRIVATE, -1, 0) < 0)
      return(simulated_brk);
  } else
  if(old_last_page > new_last_page) { // can give memory back
    if(munmap(new_last_page, (old_last_page - new_last_page)) < 0)
      return(simulated_brk);
  }

  // otherwise, whatever they asked for (pending an mmap) goes...
  simulated_brk = new_brk;
  return(new_brk);
}
#endif

/* SimulatateLinuxSystemCall exists if we are on a non-native host, but it'll
    ASSERT-fail */
int SimulateLinuxSystemCall(void *registers_ptr)
{
#ifdef NATIVE_X86_HOST
  int32 result;
  RegisterFile *regs = (RegisterFile *) registers_ptr;

  DEBUG(DEBUG_SYSCALL,"Starting system call.  (Syscall #%d)",
	REG32(*(RegisterFile *)registers_ptr,REG_EAX));

  // catch some system calls, and emulate them...
  // in emulation, first argument is in EBX (I think)
  switch(REG32(*(RegisterFile *)registers_ptr,REG_EAX)) {
  case SYS_brk:
    result = Emulate_brk(REG32(regs, REG_EBX));

    DEBUG(DEBUG_SYSCALL, "brk(%d) = %08X", REG32(regs, REG_EBX), result);
    REG32(regs, REG_EAX) = result;
    break;

  case SYS_exit:
    fprintf(stderr, "Simulated program called exit(%d)!\n", 
	    REG32(regs, REG_EBX));
    return(1);
    break;

  default:
    __asm__("pushl %ebx; pushfl");  // easier to save these on the stack
    __asm__("movl 8(%ebp), %ebx");

    __asm__("xchgl " R32_POS(REG_EAX) ", %eax");
    __asm__("xchgl " R32_POS(REG_ECX) ", %ecx");
    __asm__("xchgl " R32_POS(REG_EDX) ", %edx");
    __asm__("xchgl " R32_POS(REG_ESP) ", %esp");
    __asm__("xchgl " R32_POS(REG_EBP) ", %ebp");
    __asm__("xchgl " R32_POS(REG_ESI) ", %esi");
    __asm__("xchgl " R32_POS(REG_EDI) ", %edi");

    __asm__("pushl " R32_POS(REG_EFLAGS));
    __asm__("pushl " R32_POS(REG_EBX));

    // we need to hide the base of our register array somewhere in the real
    //  memory space, but we can't stick it anywhere that might get 
    //  overwritten, so we'll try hiding it below the stack pointer...
    __asm__("addl $4, %esp; xchg -4(%esp), %ebx");
    __asm__("popfl");

    __asm__("int $0x80");

    __asm__("pushfl");

    // should now be able to get back the base of the register array
    __asm__("xchg -4(%esp), %ebx; subl $4, %esp");

    __asm__("popl " R32_POS(REG_EBX));
    __asm__("popl " R32_POS(REG_EFLAGS));

    __asm__("xchgl " R32_POS(REG_EAX) ", %eax");
    __asm__("xchgl " R32_POS(REG_ECX) ", %ecx");
    __asm__("xchgl " R32_POS(REG_EDX) ", %edx");
    __asm__("xchgl " R32_POS(REG_ESP) ", %esp");
    __asm__("xchgl " R32_POS(REG_EBP) ", %ebp");
    __asm__("xchgl " R32_POS(REG_ESI) ", %esi");
    __asm__("xchgl " R32_POS(REG_EDI) ", %edi");

    __asm__("popfl; pop %ebx");
    break;
  }

  DEBUG(DEBUG_SYSCALL,"Done with system call.  (Result = %08X)",
	REG32(*(RegisterFile *)registers_ptr,REG_EAX));
#else
  DEBUG(DEBUG_SYSCALL,"libx86sim.a compiled for non-native host!  Cannot "
	              "simulate system calls!");
  ASSERT(0);
#endif

  return(0);
}

#ifdef NATIVE_X86_HOST
/* These functions allow on-the-fly relocation of the code and stack segments.
   CopyAndSwitchStacks() appears to be relatively stable, and needed.
   CopyAndSwitchCode() has some serious problems with pointers to the code
     segment inside of the code segment (binaries do not have relocation info).
     Luckily, if you can convince 'ld' to link the binary somewhere else to
     start with, you never need to call CopyAndSwitchCode().  (I've mostly
     included it as an example of the sick things you can do in assembler. :)
*/
void CopyAndSwitchCode(int32 old_code_base, int32 new_code_base, int32 size,
		       int32 stack_top)
{
  __asm__("pushl %ebx");

  // old_code_base is 8(%ebp), new_code_base is 12(%ebp), size is 16(%ebp)
  //  stack_top is 20(%ebp)
  __asm__("movl 8(%ebp), %esi;"     // ESI gets source (old code)
	  "movl %esi, %edx;"        // save bottom of code in EDX
	  "movl 12(%ebp), %edi;"    // EDI gets dest (new code)
	  "movl 16(%ebp), %ecx;"    // ECX gets number of bytes to copy
	  "movl %edx, %ebx;"
	  "addl %ecx, %ebx;"        // EBX gets top of code
	  "cld;"                    // we want ESI and EDI to step up
	  "rep; movsb;"             // copy the stack (slowly :)
	  "movl %edi, %eax;"
	  "subl %esi, %eax;"        // EAX holds difference between new and old
	  "lea .new_codeptr, %esi;" // need a ip value for new space
	  "addl %eax, %esi;"        // relocate new eip into new code
	  "jmp *%esi;"              // set the new ip value
	  ".new_codeptr:"

	  // fix references to code in the stack (relatively safe)
	  "movl %esp, %edi;"        // EDI has bottom of stack to start scan
	  "movl 20(%ebp), %ecx;"
	  "subl %edi, %ecx;"        // ECX has # of bytes used in stack
	  "shrl $2, %ecx;"          // ECX has # of logns used in stack

	  ".sfix_start:"
	  "movl (%edi), %esi;"      // get word from stack
	  "cmpl %ebx, %esi;"        // compare value against top of code
	  "jae  .no_sfix_val;"
	  "cmpl %edx, %esi;"        // compare against bottom of code
	  "jb   .no_sfix_val;"
	  "addl %eax, %esi;"        // fix value
	  "movl %esi, (%edi);"      // put value back
	  ".no_sfix_val:"
	  "addl $4, %edi;"          // increment edi to point at next value
	  "loop .sfix_start;"       // repeat as needed
#if 0
	  // fix references to code in the code (this is a really unsafe
	  //  thing to be doing if you don't know the layout of your code,
	  //  as you could possibly change something that looks like a pointer
	  //  into your code that isn't - we'll make it a little safer by
	  //  assuming that any pointer into the code would be 4-byte aligned)
	  // actually, it turns out that we can't just look at 4-byte aligned
	  //  values, since immediates (used in compares for example) aren't
	  //  going to be aligned always (often?)
	  "movl 12(%ebp), %edi;"    // EDI has bottom of (new) code to scan
	  "movl 16(%ebp), %ecx;"    // ECX has # of bytes used in code
	  "subl $3, %ecx;"          // last 3 compares wouldn't be long enough

	  ".cfix_start:"
	  "movl (%edi), %esi;"      // get word from stack
	  "cmpl %ebx, %esi;"        // compare value against top of code
	  "ja   .no_cfix_val;"
	  "cmpl %edx, %esi;"        // compare against bottom of code
	  "jb   .no_cfix_val;"
	  "addl %eax, %esi;"        // fix value
	  "movl %esi, (%edi);"      // put value back
	  ".no_cfix_val:"
	  "addl $1, %edi;"          // increment edi to point at next value
	  "loop .cfix_start;"       // repeat as needed
#endif
	  );

  __asm__("popl %ebx");
}

void CopyAndSwitchStacks(int32 old_stack_top, int32 new_stack_top)
{
  __asm__("pushl %ebx");

  // old_stack_top is 8(%ebp), new_stack_top is 12(%ebp)
  __asm__("movl 8(%ebp), %esi;"     // ESI gets source (old stack)
	  "movl %esi, %ebx;"        // we'll need to remember this value later
	  "movl 12(%ebp), %edi;"    // EDI gets dest (new stack)
	  "movl %esi, %ecx;"
	  "subl %esp, %ecx;"        // ECX is amount of stack used (in bytes)
	  "shrl $2, %ecx;"          // ECX is amount of stack used (in longs)
	  "movl %ecx, %edx;"        // Save copy of ECX for 2nd pass
	  "subl $4, %esi;"          // ESI points to top word on old stack
	  "subl $4, %edi;"          // EDI points to top word on new stack
	  "std;"                    // we want ESI and EDI to step down
	  "rep; movsl;"             // copy the stack
	  "movl %edi, %eax;"
	  "subl %esi, %eax;"        // EAX gets different between stack locs
	  "addl %eax, %ebp;"        // fix EBP (pointing into new stack now)
	  "movl %esp, %ecx;"        // save old ESP for comparisons below
	  "addl %eax, %esp;"        // fix ESP (pointing into new stack now)

	  "addl %eax, environ;"     // while we're at it, fix libc's env ptr

	  "xchgl %edx, %ecx;"       // put size of stack in ECX again
	  ".fix_start:"
	  "addl $4, %edi;"          // increment edi to point at next value
	  "movl (%edi), %esi;"      // get word from stack
	  "cmpl %ebx, %esi;"        // compare value against top of stack
	  "jae  .no_fix_val;"
	  "cmpl %edx, %esi;"        // compare against bottom of stack
	  "jb   .no_fix_val;"
	  "addl %eax, %esi;"        // fix value
	  "movl %esi, (%edi);"      // put value back
	  ".no_fix_val:"
	  "loop .fix_start;"        // repeat as needed
	  );

  __asm__("popl %ebx");
}

int ExecuteBuffer(const void *buffer, int len,
		  void *registers_ptr,
		  void *fpu_state_ptr)
{
  unsigned char *fixed_buffer;  /* same as 'buffer', but a RET at the end */
  int8 saved_fpu_state[108];    /* used to store simulator's fpu state */
  void *return_addr;

  if (1) {
    fprintf(stderr, "Skipping FPU instruction at 0x%x\r\n", R32_POS(REG_EIP));
    return 0;
  }

  fixed_buffer = new unsigned char[len + 6];
  memcpy(fixed_buffer, buffer, len);
  fixed_buffer[len+0] = 0xFF;
  fixed_buffer[len+1] = 0x25;
  *(void **)(fixed_buffer+len+2) = &return_addr;

  __asm__("movl $exec_ret, -116(%ebp);"  // get our return address

	  "fsave  -112(%ebp);"       // save simulator's FPU state
	  "movl   20(%ebp), %ebx;"   // load address of simulated FPU state
	  "frstor (%ebx);"           // load in simulated FPU state

          "pushl %ebx; push %ecx; pushfl;" // easier to save these on the stack
          "movl 16(%ebp), %ebx;"
	  "movl -4(%ebp), %ecx;"

          "xchgl " R32_POS(REG_EAX) ", %eax;"
          "xchgl " R32_POS(REG_EDX) ", %edx;"
          "xchgl " R32_POS(REG_ESP) ", %esp;"
          "xchgl " R32_POS(REG_EBP) ", %ebp;"
          "xchgl " R32_POS(REG_ESI) ", %esi;"
          "xchgl " R32_POS(REG_EDI) ", %edi;"

          "pushl " R32_POS(REG_EFLAGS) ";"
          "pushl " R32_POS(REG_EBX) ";"
          "pushl " R32_POS(REG_ECX) ";"

  // we need to hide the base of our register array somewhere in the real
  //  memory space, but we can't stick it anywhere that might get 
  //  overwritten, so we'll try hiding it below the stack pointer...
          "addl $8, %esp; xchg -4(%esp), %ebx; xchg -8(%esp), %ecx;"
          "popfl;"

          "jmp    *-12(%esp);"        // execute buffer in simulated context

"exec_ret: pushfl;"

  // should now be able to get back the base of the register array
          "xchg -4(%esp), %ebx; xchg -8(%esp), %ecx; subl $8, %esp;"

          "popl " R32_POS(REG_ECX) ";"
          "popl " R32_POS(REG_EBX) ";"
          "popl " R32_POS(REG_EFLAGS) ";"

          "xchgl " R32_POS(REG_EAX) ", %eax;"
          "xchgl " R32_POS(REG_EDX) ", %edx;"
          "xchgl " R32_POS(REG_ESP) ", %esp;"
          "xchgl " R32_POS(REG_EBP) ", %ebp;"
          "xchgl " R32_POS(REG_ESI) ", %esi;"
          "xchgl " R32_POS(REG_EDI) ", %edi;"

          "popfl; pop %ecx; pop %ebx;"
	
	  "fsave  (%ebx);"           // save simulated FPU state back
	  "frstor -112(%ebp);"       // reload simulator's FPU state
	  );

  delete[] fixed_buffer;

  return(0);
}

// here we have a slightly more sedate version of ExecuteBuffer.  This one
//  relies on the caller having "fixed" any memory operands to be absolute
//  address operands, and just executes the instruction in the context of the
//  simulator (although it does take care to swap in the simulated program's
//  state).
int ExecuteFPUOp(const void *buffer, int len,
		 void *fpu_state_ptr,
		 int32& eax_to_use)
{
  int8 *fixed_buffer;  
                                /* same as 'buffer', but a RET (actually a 
				   jump so that we don't need a return address
                                   on the stack at the end */
  int8 saved_fpu_state[108];    /* used to store simulator's fpu state */
  void *return_addr;

  fixed_buffer = new int8[len + 6];
  memcpy(fixed_buffer, buffer, len);
  fixed_buffer[len+0] = 0xFF;
  fixed_buffer[len+1] = 0x25;
  *(void **)(fixed_buffer+len+2) = &return_addr;

  __asm__("movl $execfpu_ret, %1;"  // get our return address

	  "fsave  %2;"               // save simulator's FPU state
	  "movl   %3, %%ebx;"   // load address of simulated FPU state
	  "frstor (%%ebx);"           // load in simulated FPU state

	  "xchg   %%eax, %0;" // swap EAXs

          "jmp    *%5;"        // execute buffer in simulated context

"execfpu_ret:"
	  "xchg   %%eax, %0;" // swap EAXs
	
	  "fnsave  (%%ebx);"           // save simulated FPU state back
	  "frstor %2;"       // reload simulator's FPU state
	  : "=g"(eax_to_use)
	  : "g"(return_addr), "m"(saved_fpu_state), "g"(fpu_state_ptr),
	    "0"(eax_to_use), "m"(fixed_buffer)
	  : "ebx"
	  );

  delete[] fixed_buffer;

  return(0);
}
#endif
