/* $Id: ui.cc,v 1.4 1998/04/29 04:55:00 nikos Exp $
 *
 * RCS Log:
 * $Log: ui.cc,v $
 * Revision 1.4  1998/04/29 04:55:00  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:52  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:36  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:12  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:10  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: ui.cc,v 1.4 1998/04/29 04:55:00 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/ui.cc,v 1.4 1998/04/29 04:55:00 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:31  bugnion
// stanford update (x86)
//
// Revision 1.5  1997/10/18  00:39:10  sean
// fixed simulator loop to correctly stop on a halted CPU
//
 * Revision 1.4  1997/09/27 18:32:46  sean
 * fixed decode of mArg and mRepeatCount to allow values over 0x7ffffff
 *   (needed to use strtoul() instead of strtol()) - dumping data ('x')
 *   now works
 *
 * Revision 1.3  1997/09/15 05:23:02  yoel
 * Now removes leading spaces from the command string.
 *
 * Revision 1.2  1997-09-14 19:57:02-07  yoel
 * Redid a bit, added go and continue, fixed the parsing and allowed return to
 * repeat the last command.
 *
 * Revision 1.1  1997-09-14 16:19:26-07  yoel
 * Initial revision
 *
 * User interface.  Handles the interactive parts of the simulator loop.
 * This code started out in x86sim.cc.  Okay, here's how I'm going to
 * define the new command syntax (I freely admit I stole the format from
 * adb and gdb):
 *
 *		<letter><count> <arg>
 * 
 * Where letters cuurently allowed are:
 *	q		Quit
 *	b		Breakpoint
 *	r		Dump Registers
 *	s		Step
 *	x		Dump address <arg> as a hex long
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "ui.h"
#include "cpu.h"
#include "lockstep.h"

static char rcsid[] __attribute__((unused))
        = "$Id: ui.cc,v 1.4 1998/04/29 04:55:00 nikos Exp $";

UI::UI(CPU *cpu, LockStep *singlestep)
{
   mCPU = cpu;
   mSinglestep = singlestep;
   mTotalCycles = 0;
   mBreakPoint = 0x0;
   mCmd = '?';
   mRepeatCount = 1;
   mArg = 0x0;
}

int64 UI::SimLoop(void)
{
  RegisterFile *regptr;
  int i, j;			// Instead of clobbering mRepeatCount
  int32 val;                    // used when dumping...

  SimDump();

  while(!mCPU->is_halted()) {
    fputs("> ", stdout);
    SimParse();

    switch(mCmd) {
    case '?':
       fputs(X86SIM_HELP, stdout);
       break;
    case 'q':
       return (mTotalCycles);
       break;
    case 'b':
       if (mArg == 0x0) { 
          fputs("Must supply a (non-zero) argument.\n", stdout) ; break; 
       }
       mBreakPoint = mArg;
       break;
    case 'c':
    case 'g':
       while(!mCPU->is_halted()) {
	  regptr = mCPU->GetRegisters();
	  if(REG32(*regptr,REG_EIP) == mBreakPoint) { 
	    mBreakPoint = 0; mRepeatCount = 0; break; 
	  }
	  mCPU->ExecuteOneInstruction(mSinglestep);
	  mTotalCycles++;
       }
       SimDump();
       break;
    case 'r':
       SimDump();
       break;
    case 's':
       for (i = mRepeatCount ; (i > 0) && !mCPU->is_halted() ; i--) {
	  regptr = mCPU->GetRegisters();
	  if(REG32(*regptr,REG_EIP) == mBreakPoint) { 
	    mBreakPoint = 0; mRepeatCount = 0; break; 
	  }
	  mCPU->ExecuteOneInstruction(mSinglestep);
	  mTotalCycles++;
       }
       SimDump();
       break;
    case 'x':
       // 4-byte align data
       if(mArg & 3) {
	 mArg &= ~3UL;
	 mRepeatCount++;
       }

       for (i = mRepeatCount ; i > 0 ; i--, mArg += 4) {
	 printf(" 0x%08X:", mArg);

	 if(mCPU->ReadFromLinearAddress(mArg, &val, sizeof(val), false)) {
	   printf("  S:%08X (", val);

	   for(j = 0; j < 4; j++) {
	     fputc((isprint(val & 0x0FF) ? (val & 0x0FF) : '.'), stdout);
	     val >>= 8;
	   }

	   fputc(')', stdout);
	 } else
	   printf("  S:???????? (????)");

	 if(mSinglestep != NULL) {
	   if(mSinglestep->ReadFromLockstepMemory(mArg, &val, sizeof(val))) {
	     printf("  R:%08X (", val);

	     for(j = 0; j < 4; j++) {
	       fputc((isprint(val & 0x0FF) ? (val & 0x0FF) : '.'), stdout);
	       val >>= 8;
	     }

	     fputc(')', stdout);
	   } else
	     printf("  R:???????? (????)");
	 }

	 printf("\n");
       }
       break;
    default:
      fputs("Unknown command. Type ? for help.\n", stdout);
      break;
    }
  }

  return(mTotalCycles);
}

void UI::SimDump(void)
{
   if (mSinglestep != NULL) mSinglestep->TraceRegisters();
   mCPU->DumpState();
   return;
}

void UI::SimParse(void)
{
   char line[40];
   char *str;
   char *end[1];

   fgets(line,39,stdin);
   str = line;

   //  Really should kill leading whitespace.
   while((*str != '\0') && isspace(*str)) str++;

   if (*str == '\0') return;			// Preserve last command
   
   mCmd = *(str++);
   if ((*str != '\0') && (!isspace(*str))) {
     mRepeatCount = strtoul(str, end, 0);// Is strtol(str, &str, 0) legal?
     str = *end;
   } else
     mRepeatCount = 1;

   if (*str != '\0') 
      mArg = strtoul(str, (char **) NULL, 0);
   else
      mArg = 0x0;

   return;
}
