/* -*- C++ -*-
 * $Id: ui.h,v 1.4 1998/04/29 04:55:00 nikos Exp $
 * 
 * RCS Log:
 * $Log: ui.h,v $
 * Revision 1.4  1998/04/29 04:55:00  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:52  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:37  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:12  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:11  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: ui.h,v 1.4 1998/04/29 04:55:00 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/ui.h,v 1.4 1998/04/29 04:55:00 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:31  bugnion
 * stanford update (x86)
 *
 * Revision 1.2  1997/09/15  02:57:02  yoel
 * Renamed some of the functions, removed UI::SimPrompt b/c it was stupid.
 * Nudged the help files for the new commands.
 *
 * Revision 1.1  1997-09-14 16:19:26-07  yoel
 * Initial revision
 *
 * Implements the interactive portion of the simulator.
 */

#ifndef __UI_H__
#define __UI_H__

#include "ints.h"

#define X86SIM_HELP \
"\nThe format of all the commands is as follows:\n"                           \
"\t<letter><repeat-count> <argument>\n"                                       \
"Currently valid letters are as follows:\n"                                   \
"\t? -- Help.  This Message.\n"                                               \
"\tq -- Quit.  No Arguments.\n"                                               \
"\tb -- Breakpoint.  Set breakpoint as address <arg>.  No repeat-count.\n"    \
"\tc -- Continue.  Go forever from this point.  Same as g.\n"                 \
"\tg -- Go forever.  Same as c.\n"                                            \
"\tr -- Dump Registers.  Dumps LockStep & Simulated Registers.\n"             \
"\ts -- Step an instruction.  Execute <repeat-count> instruction. No arg.\n"  \
"\tx -- Display.  Dumps <repeat-count> 32-Bit quanties starting at <arg>\n"   \
"Note that not all commands make sense with a repeat-count or argument and\n" \
"in those cases it is ignored.\n"

class CPU;
class LockStep;

class UI {

public:
   UI(CPU *cpu, LockStep *singlestep);

   int64 SimLoop(void);
   void SimDump(void);
   void SimParse(void);

protected:
   CPU *mCPU;
   LockStep *mSinglestep;
   int64 mTotalCycles;
   int32 mBreakPoint;
   char mCmd;
   int32 mRepeatCount;
   int32 mArg;
};

#endif
