/* $Id: x86_instr.cc,v 1.4 1998/04/29 04:55:01 nikos Exp $
 *
 * RCS Log:
 * $Log: x86_instr.cc,v $
 * Revision 1.4  1998/04/29 04:55:01  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:53  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:38  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:12  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:12  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: x86_instr.cc,v 1.4 1998/04/29 04:55:01 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86_instr.cc,v 1.4 1998/04/29 04:55:01 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:32  bugnion
// stanford update (x86)
//
// Revision 1.38  1997/12/12  11:29:20  yoel
// far CALL was calling LoadDescriptor() with Exceptions suppressed in a few
// places it shouldn't have in a few places and calling LoadDescriptor() without
// them suppressed and then generating the Exceptions by hand in a few places.
// Both fixed.
//
// Revised a some how LoadDescriptor() generates exceptions, so that if you're
// trying to load a TSS of some sort then #TS is generated instead of #GP.  I
// revised the #ifdef'ed out RPL, CPL, DPL checks some too, but still couldn't
// make them happy and left them #ifdef'ed out.
//
 * Revision 1.37  1997/12/11 22:14:59  mendel
 * 1) fixed a problem with the setting of the busy bit in the task
 * structure. We were not clearing the busy during a task switch always.
 *
 * 2) Fixed bug in reading and writing CR0-CR4 registers using the
 * debug interface.
 *
 * 3) Fixed bug in accessing memory thru the debug interface that was
 * causing the processor to take exceptions.
 *
 * Revision 1.36  1997/12/10 00:55:31  yoel
 * Fixed broken privilege level checking of Segment Registers after an IRET
 * between privilege levels.  Added code to correct LoadDescriptor() priv
 * checking when it gets turned back on.
 *
 * Revision 1.35  1997/12/09 14:52:09  sjt
 * FPU emulation should now work completely - FPU exceptions are triggered when
 * 	they should be, and memory operands appear to work fine.  I removed the
 * 	printfs that were cluttering everything too.
 *
 * Mendel's fixes to BT/BTC/BTR/BTS and ADC have been stolen and checked in as
 * 	well.
 *
 * Revision 1.34  1997/12/09 05:33:41  mendel
 * Bugs patched around (These need to be fixed):
 *    1) ifdef'ed out the protection checking in LoadDescriptor() that
 *       was preventing IRET from working. See comment in code I added in
 *       cpu.cc
 *    2) ifdefed out the ASSERT(0) in the TLB flush routine in x86_instr.cc
 *    3) ifdef'ed out the protection checking in IRET that wasn't working.
 *
 *
 * Bugs:
 *
 * 1) Check in some bug fixes in fpu.cc that Sean gave me.
 *
 * 2) Fixed bug in reading stack segment selector in TRAP_GATE . We were
 * ASSERT(0)ing when the ReadFromVirtualAddress() was successful! in cpu.cc
 * HandleExecption()
 *
 * 3) Changed (CSDescriptor.dpl() != SSSelector & 0x3) to
 *            (CSDescriptor.dpl() != (SSSelector & 0x3))
 * for obvious reasons in cpu.cc / HandleException().
 *
 * 4) Test for GDT in TRAP_GATE code was backward in cpu.cc /
 * HandleException()
 *
 * 5) Since the prev_eip field of CPU class is used to set the return PC
 * of execeptions. We need to set prev_eip on instruction page faults to
 * be the faulting instruction. Otherwise we think the exception happen
 * on the PC of IRET instruction. ouch!
 *
 * 6) The MemoryOperand() code in instruction.cc wasn't setting the
 * default segment of the ESP register to be the stack segment! Found
 * this because Linux runs for a while with the user's DS selector
 * loaded and pushes things on the stack.
 *
 * 7) The is_task() method of seg_desc was completely bogus.
 *
 *
 * Clean up:
 *
 * 1) Removed fprintf() on exceptions in cpu.cc since we are taking far too many
 * execptions for this to be useful.
 *
 * Revision 1.33  1997/12/07 21:05:11  mendel
 * Added annotations on IRET instructions. Useful for debugging.
 *
 * Fixed a couple of bugs in exception processing code:
 * 1) Was incorrectly pushing error code on stack on exception
 * 2) Was incorrectly computing space needed for exception frame.
 *
 * Revision 1.32  1997/12/07 08:23:42  yoel
 * Implemented IRET. B/c Sean and I short-circuited CVS there's already a partial
 * IRET implementation checked in, but this one is complete.  Also fixed what I
 * think was a bug where we would generate two #TS exceptions on certain illegal
 * far calls through a Task Gate (not that that has ever been tested).  Still
 * need a RETfar, but it looks a lot like IRET, only easier.
 *
 * Revision 1.31  1997/12/07 02:54:32  sjt
 * Bunches of changes:
 *   1) Rewritten emulation of FPU ops is half-done.  Emulation is now done in
 *        the context of the simulator (although the FPU context of the simulated
 *        program is used), which dodges virtual->phyiscal mappings of addresses.
 *      Assembler has been written to be happy under fairly aggressive (i.e. it
 *        ignores the "volatile" keyword) optimization by g++.
 *      FPU operations using memory operands are unsupported, and ignored.
 *
 *   2) Minor changes in CPU::MUL and CPU::IMUL to support 64-bit products (you
 *        have to explicitly cast the int32's to int64's before the multiply, or
 *        you get 0's for the top 32 bits (even though they're available for free
 *        - seems g++ code gen could be written better).
 *
 *   3) Changes all over to get task switching to work.  Need to fix bunches of
 *        logic errors, uses of limit() instead of offset_valid(), make TSS's
 *        writeable under certain circumstances.
 *
 * Revision 1.30  1997/12/02 22:52:21  mendel
 * 1) Hooked up interrupts. Moved interrupt code from simos_interface.cc to
 *    cpu.cc .
 *
 * Several bug fixes:
 *
 * 1) Added check to push the error code only on processor generated
 *    exceptions and not on INT or external ones.
 *
 * 2) IDT gate descriptors were broken. I added a gate_use_32bit() since
 *    IDT gate descriptors use a different size field that normal segment
 *    descriptors.
 *
 * 3) Fixed bug in computing spaceNeeded for exception stack puke. Beware
 *    of C operator precedence.
 *
 * 4) Cleaned up some ugly and some incorrect code in HandleException()
 *
 * 5) #if 0 out some bogus check LoadExceptionDescriptor(). I couldn't
 *    tell want it was trying to do but is was wrong.
 *
 * 6) Compute the gate_offset correctly. We need all the bits of the
 *    flags for this to work.
 *
 * 7) Changed POP and MOV to not call WriteBackDescriptor() on selector
 *    zero. This caused WriteBackDescriptor() to die.
 *
// Revision 1.29  1997/12/02  04:49:32  yoel
//   Support added for taking Software Interupts and preparation for supporting
// external interupts.  cpu.{cc,h} updated so that current_exception is an
// int8 instead of the enum to allow exceptions 19-255 and enum ExceptionSource
// added.  Cosmetic changes to fpu.cc and INT n / INTO now actually do
// something (since we now support Software Interupts).
//
 * Revision 1.28  1997/12/01 06:02:16  mendel
 * Fixed bug in intr checking code ( was using FLAG_IP incorrectly)
 *
 * Added code to callout when we hit a vmlinux monitor call.
 * Currently just print it out.
 *
// Revision 1.27  1997/11/27  00:09:24  yoel
// x86_instr.cc:  Modified not to push return information on far JMPs (only
// CALLs).
//
// cpu.cc: Cosmetic changes plus some fixing of ExecuteOneInstruction()
// generating INVALID_OPCODE exceptions if decoding caused and Exception.
//
 * Revision 1.26  1997/11/26 18:19:43  sjt
 * Fixed setting of ZF for non 32-bit quantities.
 * Rigged interrupted, repeated instructions (i.e. SCAS, CMPS) to not worry about
 * 	flag settings.
 * Tweaked CALLorJMP to not push return address for normal far JMP (don't know
 * 	about rest of far JMPs)
 *
 * Revision 1.25  1997/11/25 21:40:09  mendel
 * Added gdb support to x86sim. Currently all INT 3 exceptions go to
 * simdebug.
 *
// Revision 1.24  1997/10/31  00:59:04  whimsy
// Added an ASSERT around SwitchTasks() since I don't know what to do on failure.
//
 * Revision 1.23  1997-10-29 14:48:52-08  sean
 * all the range checks for IDIV were backwards...  not any more (I don't
 *   think :)
 *
 * Revision 1.22  1997/10/28 08:38:04  whimsy
 * Added the call to SwitchTasks() in far CALKCALL.
 *
 * Revision 1.21  1997-10-23 20:50:03-07  whimsy
 * Fixed the segment loading stuff in far calls.  Exceptions no longer
 * suppresed on the LoadDEscriptor() call and whacked LoadDescriptor() tio
 * generate the proper exceptions.
 *
 * Revision 1.20  1997-10-19 19:33:01-07  whimsy
 * Added the Far CALL stuff and made it compile.  Can't exactly test it
 * yet and the Ta actual work of switching the tasks isn't done yet.
 *
 * Revision 1.19  1997-10-17 17:39:45-07  sean
 * added instruction counting code, and rigged it so some system calls
 *   (e.g. exit) can cause the CPU to halt, effectively terminating simulation
 *
 * Revision 1.18  1997/09/27 18:37:09  sean
 * Added calls to END_DEFN macro to make code cleaner.
 * Code to say which flags are left undefined by an instruction (killing
 *   false negatives when comparing to lockstep)
 * MOVS was moving in the wrong direction
 * SCAS compare had the +1 subtraction error
 *
 * Revision 1.17  1997/09/24 08:17:20  sean
 * 1) header of operation methods tweaked to return string name of operation
 *    to sufficiently whacked callers (see instruction.cc)
 * 2) SCAS fixed to be proper subtraction
 * 3) shift functions fixed to set AF (Pentium book says undefined, by my
 *    CPU apparently likes to always set AF for shifts, so I did it to make
 *    debugging easier.)
 *
 * Revision 1.16  1997/09/24 02:40:55  sjt
 * support for single-stepping of REP'd instructions added
 *
 * Revision 1.15  1997/09/15 19:50:57  sean
 * Fixed subtraction code (CMP, SUB, SBB) for correct carry flag
 *  implementation (the 1 that's added to the inverted second operand
 *  must be allowed to cause a carry - i.e. when CMP'ing with 0)
 *
 * Revision 1.14  1997/09/06 00:39:51  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.13  1997/03/18 19:24:34  sjt
 * flag related fixes:
 *   SetFlagsAfterOp() changed to allow flipping of CF and AF for SUB-type
 *     instructions
 *   Shift instructions now set SF, ZF, and PF as they're supposed to
 *
 * Revision 1.12  1997/03/14 21:48:50  sjt
 * fixed bugs:
 *   absolute far CAL/JMPs were using relative EIP
 *   problem with range-checking on 32-bit DIV
 *   operands were backwards for LODS
 *   16-bit RETs need to ignore top 16 bits of ESP
 *
 * Revision 1.11  1997/03/07 05:46:57  sean
 * fixed SCAS, repeat prefixes, added wrapper for INT 0x80 (Linux System
 *   Calls)
 *
 * Revision 1.10  1997/03/06 18:59:07  sean
 * fixed a few issues concerning operand sizes w.r.t. 8-bit operands
 *
 * Revision 1.9  1997/03/02 03:06:39  sean
 * split RET into RETnear and RETfar, RETnear is implemented now, so the
 *   list of un-implemented instructions are far control transfers and
 *   the six freakish ones
 *
 * Revision 1.8  1997/02/23 03:18:16  sean
 * added a ton of new instructions...  done with 110 out of 122 - remaining
 *   are the six control transfer functions (CALL, INT, INTO, IRET, JMP, RET),
 *   and six functions I don't know how to do yet (INVLPG, LMSW, RDMSR, RSM,
 *   SMSW, WRMSR)...
 *
 * Revision 1.7  1997/02/06 09:49:40  sean
 * added a few more instructions (CBW, CWD, CPUID, ARPL, maybe some more)
 *
 * Revision 1.6  1997/02/04 22:38:11  sean
 * made operation functions return number of clock cycles spent (sorta - all
 * return 1 right now), and SAL is a synonym for SHL
 *
 * Revision 1.5  1997/02/04 10:40:41  sean
 * added a bunch (like 24) of instructions - all unimplemented instructions
 * now assert-fail
 *
 * Revision 1.4  1997/02/04 02:42:55  sjt
 * cosmetic changes
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: x86_instr.cc,v 1.4 1998/04/29 04:55:01 nikos Exp $";

#include "cpu.h"
#include "instruction.h"
#include "debug.h"
#include "simlib.h"
#include "counter.h"

#ifdef SIMOS
extern "C" { 
#include "simos_interface.h"
}
#endif

// more cheating with macros: quick way to get at operands
#define OP(num)                  inst.GetOperand(num)
#define GET_OP(num)              (inst.GetOperand(num)->FetchValue(*this))
#define SET_OP(num,val)          inst.GetOperand(num)->WriteValue(*this,val)
#define OP_SIZE(num)             inst.GetOperand(num)->OperandSize()

// GET_OP was obvious; this one is the same, except zero extended
#define GET_OP_Z(num)            (inst.GetOperand(num)->FetchValue(*this,0))

// These three allow an extra displacement too
#define GET_OP_DISP(num,disp)    (inst.GetOperand(num)->FetchValue(*this,1,disp))
#define GET_OP_Z_DISP(num,disp)  (inst.GetOperand(num)->FetchValue(*this,0,disp))
#define SET_OP_DISP(num,val,disp) inst.GetOperand(num)->WriteValue(*this,val,disp)

// These macros let you know if a number is the min/max for its (signed) size
#define IS_MAX_VAL(num,size)  \
	((((size) == SIZE_8BIT)  && ((num) == 0x0000007F)) ||\
	 (((size) == SIZE_16BIT) && ((num) == 0x00007FFF)) ||\
	 (((size) == SIZE_32BIT) && ((num) == 0x7FFFFFFF)))
#define IS_MIN_VAL(num,size)  \
	((((size) == SIZE_8BIT)  && (CHOP8(num)  == 0x00000080)) ||\
	 (((size) == SIZE_16BIT) && (CHOP16(num) == 0x00008000)) ||\
	 (((size) == SIZE_32BIT) && ((num)       == 0x80000000))

// Eventually, we'll have a much fancier (and more accurate) way to count clock
//  cycles used by an instruction
#define COUNT_CLOCKS(cmd)    (1)

// Macros to say which flags we want to ignore in a lockstep comparison.
#define IGNORE_A       flags_to_ignore = 0 | (1 << FLAG_AF)
#define IGNORE_O       flags_to_ignore = 0 | (1 << FLAG_OF)
#define IGNORE_OA      flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_AF)
#define IGNORE_OAC     flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_AF) |\
	                                     (1 << FLAG_CF)
#define IGNORE_OSAPC   flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_SF) |\
	                                     (1 << FLAG_AF) | (1 << FLAG_PF) |\
	                                     (1 << FLAG_CF)
#define IGNORE_OSZAP   flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_SF) |\
	                                     (1 << FLAG_ZF) | (1 << FLAG_AF) |\
	                                     (1 << FLAG_PF)
#define IGNORE_OSZAPC  flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_SF) |\
	                                     (1 << FLAG_ZF) | (1 << FLAG_AF) |\
	                                     (1 << FLAG_PF) | (1 << FLAG_CF)
#define IGNORE_OSZP    flags_to_ignore = 0 | (1 << FLAG_OF) | (1 << FLAG_SF) |\
	                                     (1 << FLAG_ZF) | (1 << FLAG_PF)
#define IGNORE_SZAP    flags_to_ignore = 0 | (1 << FLAG_SF) | (1 << FLAG_ZF) |\
	                                     (1 << FLAG_AF) | (1 << FLAG_PF)

// returns true if the bottom byte of 'value' has even parity
static inline bool even_parity(int32 value)
{
  // go from 8 bits to 4 bits (bits past the bottom 8 don't matter at all)
  value ^= (value >> 4);

  // go from 4 bits down to 2
  value ^= (value >> 2);

  // reduce to a single bit
  value ^= (value >> 1);

  // even parity should have a 0 in bottom bit
  return((value & 1) == 0);
}

// returns true if the given 'flags' meet the specified 'condition'
static bool check_condition(int32 flags, int condition)
{
  int bit;

  switch(condition) {
  case COND_O:
  case COND_NO:
    bit = BITS(flags,FLAG_OF,1);
    break;
  case COND_B:
  case COND_NB:
    bit = BITS(flags,FLAG_CF,1);
    break;
  case COND_E:
  case COND_NE:
    bit = BITS(flags,FLAG_ZF,1);
    break;
  case COND_BE:
  case COND_NBE:
    bit = BITS(flags,FLAG_CF,1) | BITS(flags,FLAG_ZF,1);
    break;
  case COND_S:
  case COND_NS:
    bit = BITS(flags,FLAG_SF,1);
    break;
  case COND_P:
  case COND_NP:
    bit = BITS(flags,FLAG_PF,1);
    break;
  case COND_L:
  case COND_NL:
    bit = BITS(flags,FLAG_SF,1) ^ BITS(flags,FLAG_OF,1);
    break;
  case COND_LE:
  case COND_NLE:
    bit = (BITS(flags,FLAG_SF,1) ^ BITS(flags,FLAG_OF,1)) | 
             BITS(flags,FLAG_ZF,1);
    break;
  default:
    // Assert fail?
    return(false);
    break;
  }

  // Toggle the bit if we had a Ncond
  bit ^= (condition & 1);

  // if the bit is 1, we've met the condition
  return(bit == 1);
}

// true if bits representing operands a, b and result x show evidence of 
//  a carry out
static inline bool is_carry(int a, int b, int x)
{
  return((a + b - x) > 0);
}

// true if bits representing operands a, b and result x show evidence of 
//  an overflow
static inline bool is_overflow(int a, int b, int x)
{
  return((a == b) && (a != x));
}

void CPU::SetFlagsAfterOp(int32 a, int32 b, int32 x, int size, bool flip_carry)
{
  int a3, b3, x3;           // values of bit 3
  int a_top, b_top, x_top;  // values of top bits

  a3 = ((a >> 3) & 1);
  b3 = ((b >> 3) & 1);
  x3 = ((x >> 3) & 1);

  switch(size) {
  case SIZE_8BIT:
    a_top = ((a >> 7) & 1);
    b_top = ((b >> 7) & 1);
    x_top = ((x >> 7) & 1);
    x &= 0x000000FF;
    break;
  case SIZE_16BIT:
    a_top = ((a >> 15) & 1);
    b_top = ((b >> 15) & 1);
    x_top = ((x >> 15) & 1);
    x &= 0x0000FFFF;
    break;
  case SIZE_32BIT:
    a_top = ((a >> 31) & 1);
    b_top = ((b >> 31) & 1);
    x_top = ((x >> 31) & 1);
    break;
  default:
    // Assert fail?
    a_top = b_top = x_top = 0;
    break;
  }

  SetArithFlags(is_overflow(a_top,b_top,x_top),(x_top == 1),
		(x == 0),
		(flip_carry ? !is_carry(a3,b3,x3) : is_carry(a3,b3,x3)),
		even_parity(x & 0x000000FF),
		(flip_carry ? !is_carry(a_top,b_top,x_top) : 
		              is_carry(a_top,b_top,x_top)));
}

#define SET_SZP(val,width)                                                \
              Set_SZP_Flags(((((val) >> (((width) << 3) - 1)) & 1) == 1), \
                            ((((int32)(val)) << (32 - ((width) << 3))) == 0), \
			    even_parity(val))

DEFN_INST(AAA) 
{
  IGNORE_OSZP;

  if(((CPU_AL & 0x0F) > 9) || (BITS(CPU_EFLAGS,FLAG_AF,1) != 0)) {
    CPU_AH += ((CPU_AL > 0x00F9) ? 2 : 1);
    CPU_AL += 6;
    SET_BITS(CPU_EFLAGS,FLAG_AF,1,1);
    SET_BITS(CPU_EFLAGS,FLAG_CF,1,1);
  } else {
    CLEAR_BITS(CPU_EFLAGS,FLAG_AF,1);
    CLEAR_BITS(CPU_EFLAGS,FLAG_CF,1);
  }

  CPU_AL &= 0x0F;

  return(COUNT_CLOCKS(AAA));
}
END_DEFN

DEFN_INST(AAD)
{
  int8 tmp_al;

  IGNORE_OAC;

  // No need to check for an exception on the fetch, since the operand should
  //  be an immediate
  CPU_AL = tmp_al = ((CPU_AH * GET_OP(0)) + CPU_AL) & 0x00FF;
  CPU_AH = 0;

  Set_SZP_Flags(((tmp_al >> 7) == 1),(tmp_al == 0),even_parity(tmp_al));

  return(COUNT_CLOCKS(AAD));
}
END_DEFN

DEFN_INST(AAM)
{
  int8 tmp_al, divisor;

  IGNORE_OAC;

  tmp_al = CPU_AL;
  divisor = GET_OP(0);

  CPU_AH = tmp_al / divisor;
  CPU_AL = tmp_al % divisor;

  // Technically, we need to set SF, and ZF based on the new AX value, but
  //  we can cheat, since we know that the new AX will only be negative (or
  //  zero) when the original AL was negative (or zero)
  Set_SZP_Flags(((tmp_al >> 7) == 1),(tmp_al == 0),even_parity(CPU_AL));

  return(COUNT_CLOCKS(AAM));
}
END_DEFN

DEFN_INST(AAS)
{
  IGNORE_OSZP;

  if(((CPU_AL & 0x0F) > 9) || (BITS(CPU_EFLAGS,FLAG_AF,1) != 0)) {
    CPU_AH -= ((CPU_AL < 0x06) ? 2 : 1);
    CPU_AL -= 6;
    SET_BITS(CPU_EFLAGS,FLAG_AF,1,1);
    SET_BITS(CPU_EFLAGS,FLAG_CF,1,1);
  } else {
    CLEAR_BITS(CPU_EFLAGS,FLAG_AF,1);
    CLEAR_BITS(CPU_EFLAGS,FLAG_CF,1);
  }

  CPU_AL &= 0x0F;

  return(COUNT_CLOCKS(AAS));
}
END_DEFN

DEFN_INST(ADC)
{
  int32 a, b, x;
  bool cf;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // Do sum and write value, and (if no exceptions occur) set flags
  // We also need to sneak an increment in here, if the carry flag is set
  cf = (BITS(CPU_FLAGS,FLAG_CF,1) != 0);

  // This is really obnoxious - basically, we can't change the top bit of
  //  either of the two operands before the add
  if(cf && !IS_MAX_VAL(a,OP_SIZE(0))) {
    a++;
    cf = false;
  }

  if(cf && !IS_MAX_VAL(b,OP_SIZE(1))) {
    b++;
    cf = false;
  }

  x = a + b;

  // If we couldn't add to either, it is safe to add after, since the a+b is
  //  _guaranteed_ to be -2 (which we can increase to -1 without changing
  //  any flag outcomes)
  if(cf) x++;

  if(SET_OP(0,x)) SetFlagsAfterOp(a,b,x,OP_SIZE(0),false);

  return(COUNT_CLOCKS(ADC));
}
END_DEFN

DEFN_INST(ADD)
{
  int32 a, b, x;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // Do sum and write value, and (if no exceptions occur) set flags
  x = a + b;
  if(SET_OP(0,x)) SetFlagsAfterOp(a,b,x,OP_SIZE(0),false);

  return(COUNT_CLOCKS(ADD));
}
END_DEFN

DEFN_INST(AND)
{
  int32 a, b, x;

  IGNORE_A;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // Do operation and write value, and (if no exceptions occur) set flags
  x = a & b;
  if(SET_OP(0,x))
    SetArithFlags(0  /* OF = 0 */,
		  ((x >> 31) == 1),
		  (x == 0),
		  0  /* AF = 0 (value undefined in spec) */,
		  even_parity(x),
		  0  /* CF = 0 */);

  return(COUNT_CLOCKS(AND));
}
END_DEFN

DEFN_INST(ARPL)
{
  int16 dest, src;

  if(in_protected_mode() && !in_v86_mode()) {
    // read in the destination descriptor, checking for exceptions
    dest = GET_OP(0);
    if(ExceptionPending()) return(0);

    // read in the srouce descriptor (from a register - so no exc check)
    src =  GET_OP(1);

    if(BITS(dest,0,2) < BITS(src,0,2)) {
      SET_BITS(dest,0,2,BITS(src,0,2));

      // write back the descriptor, and set ZF
      if(SET_OP(0,dest))
	SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1);
    } else
      CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);
  } else
    ExceptionOccurred(EXC_INVALID_OPCODE);

  return(COUNT_CLOCKS(ARPL));
}
END_DEFN

DEFN_INST(BOUND)
{
  int8 bounds[8];
  int op_size;
  int32 val;

  // Don't have to check for exceptions, since operand 0 is a register
  val = GET_OP_Z(0);     // want to zero-extend here to make 16-bit comps
  op_size = OP_SIZE(0);

  // Try to load bounds data
  if(!OP(1)->FetchBytes(*this,bounds,op_size << 1)) return(0);

  // Handle 32- and 16-bit cases separately
  if(op_size == SIZE_32BIT) {
    if((val < ((int32 *)bounds)[0]) || (val > ((int32 *)bounds)[1]))
      ExceptionOccurred(EXC_BOUNDS);
  } else {
    if((val < ((int16 *)bounds)[0]) || (val > ((int16 *)bounds)[1]))
      ExceptionOccurred(EXC_BOUNDS);
  }

  return(COUNT_CLOCKS(BOUND));
}
END_DEFN

DEFN_INST(BSF)
{
  int32 val, shifts;

  IGNORE_OSAPC;

  // Load operand 1 (zero extend into high bits), and stop on exception
  val = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  if(val != 0) {
    shifts = 0;
    while((val & 1) == 0) {
      shifts++;
      val >>= 1;
    }

    // Write the number of shifts necessary into operand 0, and clear ZF
    if(SET_OP(0,shifts))
      CLEAR_BITS(CPU_FLAGS,FLAG_ZF,1);
  } else
    SET_BITS(CPU_FLAGS,FLAG_ZF,1,1);

  return(COUNT_CLOCKS(BSF));
}
END_DEFN

DEFN_INST(BSR)
{
  int32 val, shifts, mask;

  IGNORE_OSAPC;

  // Load operand 1 (zero extend into high bits), and stop on exception
  val = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  if(val != 0) {
    shifts = (OP_SIZE(1) * 8) - 1;
    mask = 1 << shifts;
    while((val & mask) == 0) {
      shifts--;
      mask >>= 1;
    }

    // Write the number of shifts necessary into operand 0, and clear ZF
    if(SET_OP(0,shifts))
      CLEAR_BITS(CPU_FLAGS,FLAG_ZF,1);
  } else
    SET_BITS(CPU_FLAGS,FLAG_ZF,1,1);

  return(COUNT_CLOCKS(BSR));
}
END_DEFN

DEFN_INST(BSWAP)
{
  // Since we should only be getting general-purpose registers as our 
  //  operand, there is no need to check for exceptions (and if someone is
  //  stupid enough to call BSWAP on a 16-bit register, well, we'll just clear
  //  the register for them.  :)
  SET_OP(0,FLIP32(GET_OP(0)));

  return(COUNT_CLOCKS(BSWAP));
}
END_DEFN

DEFN_INST(BT)
{
  int32 val, mask;

  IGNORE_OSZAP;

  if (OP(0)->Type() == OPTYPE_MEMORY) {
     int8 byte, bit;
     s_int32 bitNum = GET_OP(1);
     OP(0)->FetchBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     bit = (byte & (1 << (bitNum % 8)));
     SET_BITS(CPU_FLAGS,FLAG_CF,1,(bit ? 1 : 0));
  } else { 
     ASSERT(OP(0)->Type() ==  OPTYPE_REGISTER);
     // Fetch operand 0 (the value to test), watching for exceptions
     val = GET_OP(0);
     
     // Don't need to check for exceptions on this since it is reg or imm
     mask = 1 << (GET_OP(1) & 31);

     SET_BITS(CPU_FLAGS,FLAG_CF,1,(((val & mask) != 0) ? 1 : 0));
  }

  return(COUNT_CLOCKS(BT));
}
END_DEFN

DEFN_INST(BTC)
{
  int32 val, mask;
  int new_cf;

  IGNORE_OSZAP;

  if (OP(0)->Type() == OPTYPE_MEMORY) {
     int8 byte, bit;
     s_int32 bitNum = GET_OP(1);
     OP(0)->FetchBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     bit = (byte & (1 << (bitNum % 8)));
     byte ^= (1 << (bitNum % 8));
     OP(0)->WriteBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     SET_BITS(CPU_FLAGS,FLAG_CF,1,(bit ? 1 : 0));
  } else { 
     ASSERT(OP(0)->Type() ==  OPTYPE_REGISTER);

     // Fetch operand 0 (the value to test)
     val = GET_OP(0);
     if(ExceptionPending()) return(0);

     // Don't need to check for exceptions on this since it is reg or imm
     mask = 1 << (GET_OP(1) & 31);

     new_cf = (((val & mask) != 0) ? 1 : 0);

     // Try to complement the bit and write the new value, and CF
     if(SET_OP(0,val ^ mask))
        SET_BITS(CPU_FLAGS,FLAG_CF,1,new_cf);
  }
  return(COUNT_CLOCKS(BTC));
}
END_DEFN

DEFN_INST(BTR)
{
  int32 val, mask;
  int new_cf;

  IGNORE_OSZAP;

  if (OP(0)->Type() == OPTYPE_MEMORY) {
     int8 byte, bit;
     s_int32 bitNum = GET_OP(1);
     OP(0)->FetchBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     bit = (byte & (1 << (bitNum % 8)));
     byte &= ~(1 << (bitNum % 8));
     OP(0)->WriteBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     SET_BITS(CPU_FLAGS,FLAG_CF,1,(bit ? 1 : 0));
  } else { 
     ASSERT(OP(0)->Type() ==  OPTYPE_REGISTER);

     // Fetch operand 0 (the value to test), watching for exceptions
     val = GET_OP(0);
     if(ExceptionPending()) return(0);

     // Don't need to check for exceptions on this since it is reg or imm
     mask = 1 << (GET_OP(1) & 31);
     
     new_cf = (((val & mask) != 0) ? 1 : 0);

     // Try to reset the bit and write the new value, and CF
     if(SET_OP(0,val & ~mask))
        SET_BITS(CPU_FLAGS,FLAG_CF,1,new_cf);
  }
  return(COUNT_CLOCKS(BTR));
}
END_DEFN

DEFN_INST(BTS)
{
  int32 val, mask;
  int new_cf;

  IGNORE_OSZAP;

  if (OP(0)->Type() == OPTYPE_MEMORY) {
     int8 byte, bit;
     s_int32 bitNum = GET_OP(1);
     OP(0)->FetchBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     bit = (byte & (1 << (bitNum % 8)));
     byte |= (1 << (bitNum % 8));
     OP(0)->WriteBytes(*this,&byte,1, bitNum/8);
     if(ExceptionPending()) return(0);
     SET_BITS(CPU_FLAGS,FLAG_CF,1,(bit ? 1 : 0));
  } else { 
     ASSERT(OP(0)->Type() ==  OPTYPE_REGISTER);

     // Fetch operand 0 (the value to test), watching for exceptions
     val = GET_OP(0);
     if(ExceptionPending()) return(0);

     // Don't need to check for exceptions on this since it is reg or imm
     mask = 1 << (GET_OP(1) & 31);

     new_cf = (((val & mask) != 0) ? 1 : 0);

     // Try to set the bit and write the new value, and CF
     if(SET_OP(0,val | mask))
        SET_BITS(CPU_FLAGS,FLAG_CF,1,new_cf);
  }
  return(COUNT_CLOCKS(BTS));
}
END_DEFN

int64 CPU::CALLorJMP(const Instruction& inst, bool push_return)
{
  int16 new_cs;
  int32 new_eip, old_eip;
  int8 to_push[8];
  SegmentDescriptor cs_desc;
  int op_size;
  bool immediate; // this is true if the operands are immediates, which means
                  //  they aren't indirects...
  //bool push_return;  // for far jumps, we set this if we need to push the
                       //  return address on the stack

  // figure out if we have an immediate (i.e. relative) operand
  immediate = (OP(0)->Type() == OPTYPE_IMMEDIATE);

  // first, load the ip (sign-extend if it's relative - only on near jumps)
  new_eip = ((immediate && (OP(1) == 0)) ? GET_OP(0) : GET_OP_Z(0));
  if(ExceptionPending()) return(0);

  // again, we need to look at the instruction's operand size, because an 8-bit
  //  relative offset gets misleading
  op_size = inst.GetOperandSize();

  // we have a near jump if there's no second operand...
  if(OP(1) == 0) {
    // if we had an immediate offset, this is a relative CALL...
    if(immediate)
      if(op_size == SIZE_32BIT)
	new_eip += CPU_EIP; else
	new_eip = CHOP16(new_eip + CPU_IP);

    // make sure new (E)IP will fit...
    if(!seg_descs[SDESC_CS].offset_valid(new_eip)) {
      ExceptionOccurred(EXC_PROTECTION,0);
      return(0);
    }

    if(push_return) {
      // try to push old (E)IP...
      old_eip = htox32(CPU_EIP);  // works for 16-bit (IP) too...
      if(seg_descs[SDESC_SS].use_32bit()) {
	if(WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP - op_size,
				 &old_eip,op_size,GetCPL()))
	  CPU_ESP -= op_size; else
  	  return(0);
      } else {
	if(WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_SP - op_size,
				 &old_eip,op_size,GetCPL()))
	  CPU_SP -= op_size; else
	  return(0);
      }
    }

    // now we can just set the new (E)IP, and be done with it
    if(op_size == SIZE_32BIT)
      CPU_EIP = new_eip; else
      CPU_IP  = new_eip;
  } else {
    // Protected Mode, Far CALL
    int32 Offset, Parameters[32], newESP, nParam;
    int16 Selector, SSSelector;
    SegmentDescriptor Descriptor, CSDescriptor, SSDescriptor;

    Offset = GET_OP(0);
    if (ExceptionPending()) return(0);
    Selector = GET_OP(1);
    if (ExceptionPending()) return(0);

    if (Selector == 0) { ExceptionOccurred(EXC_PROTECTION, 0); return(0); }

    //  LoadDescriptor() generates all the right exceptions (at least it does
    // now ;) ) so we just call it and return if something went wrong.
    if (LoadDescriptor(Descriptor, Selector, DMASK_CAN_CALL, 
		       true, false, false) == false) {
      return(0);
    }

    switch(Descriptor.type()) {

    // NonConforming.  Execute-Only, Execute-Read, Accessed / Not.
    case DESC_CODE_EO:
    case DESC_CODE_EO_ACC:
    case DESC_CODE_ER:
    case DESC_CODE_ER_ACC:
      SET_BITS(Selector, 0, 2, GetCPL());			// RPL = CPL
      goto NormalFarCall;
      return(1);						// Not reached.
      break;

    //  Conforming.  Can be Execute-Only or Execute/Read, Accessed / Not.
    case DESC_CODE_EO_CONF:
    case DESC_CODE_EO_CONF_ACC:
    case DESC_CODE_ER_CONF:
    case DESC_CODE_ER_CONF_ACC:

NormalFarCall:
      // only push old CS/IP if we're asked to...
      if(push_return) {
	if (OP_SIZE(0) == SIZE_16BIT) {
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - SIZE_16BIT, 
				&CPU_CS, SIZE_16BIT,  GetCPL());
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - 2*SIZE_16BIT, 
				&CPU_IP, SIZE_16BIT,  GetCPL());
	  if (ExceptionPending()) return(0);
	  CPU_ESP -= 2*SIZE_16BIT;
	} else {
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - SIZE_32BIT, 
				&CPU_CS, SIZE_32BIT,  GetCPL());
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP - 2*SIZE_32BIT, 
				&CPU_EIP, SIZE_32BIT,  GetCPL());
	  if (ExceptionPending()) return(0);
	  CPU_ESP -= 2*SIZE_32BIT;
	}
      }

      // Set new register contents, I hope this is valid...
      CPU_CS = Selector;
      seg_descs[SDESC_CS] = Descriptor;
      CPU_EIP = Offset;
      if (! Descriptor.use_32bit())
      CPU_EIP &= 0xFFFF;			// Only 16 bit registers.
      return(1);
      break;
  
    //  Call Gates.
    case DESC_16BIT_CALL_GATE:
    case DESC_32BIT_CALL_GATE:
      Selector = Descriptor.gate_selector();
      if (Selector == 0) { ExceptionOccurred(EXC_PROTECTION, 0); return(0); }
      SET_BITS(Selector, 0, 2, GetCPL());		// RPL = CPL
  
      //  If NP, will exception later so that's fine, NULL doesn't seem valid.
      // In any case, suppress exceptions so we can #GP(Selector).  WRONG!
      // LoadDEscriptor already would generate exactly the same Exceptions.
      // Not metion that quiet here is false, so we would have been double
      // GPF'ing which would be uncool.
      if (LoadDescriptor(CSDescriptor, Selector, DMASK_CODE,
			 false, false, false) == false) {
	return(0);
      }
 
      //  NonConforming Code Segment:
      if (CSDescriptor.is_code() && !CSDescriptor.is_conforming() &&
	  (CSDescriptor.dpl() < GetCPL())) {

	// Get the SS for the new priority from the TSS.
	ReadFromVirtualAddress(seg_descs[SDESC_TR], 8 + 8*CSDescriptor.dpl(),
			       &SSSelector, sizeof(SSSelector), GetCPL());
	if (ExceptionPending()) {
	  // Intel book, 13-4 "It is an unrecoverable error to receive a
	  // page fault or general protection exception after the processor
	  // has started to read the TSS"  Seems harsh to me, but hey.
	  ASSERT(0);
	}
       
	if (SSSelector == 0) { 
	  ExceptionOccurred(EXC_INVALID_TSS, SSSelector); return(0);
	}
	if (LoadDescriptor(SSDescriptor, SSSelector, 
			   DMASK_WRITEABLE, false, false, true) == false) {
	  ExceptionOccurred(EXC_INVALID_TSS, SSSelector); return(0);
	}
	if (! SSDescriptor.is_present()) {
	  ExceptionOccurred(EXC_STACK_FAULT, SSSelector); return(0);
	}

	ReadFromVirtualAddress(seg_descs[SDESC_TR], 4 + 8*CSDescriptor.dpl(),
			       &newESP, sizeof(newESP), GetCPL());
	if (ExceptionPending()) {
	  ASSERT(0);
	}

	if (push_return) {  // Only have return addr and params on CALL.
	  //  Push Current SS and SP/eSP on the (to-be) new Stack
	  WriteToVirtualAddress(SSDescriptor, newESP -= OP_SIZE(0), &CPU_SS, 
				OP_SIZE(0), GetCPL()); 
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(SSDescriptor, newESP -= OP_SIZE(0), 
				OP_SIZE(0) == SIZE_16BIT ? (void *) &CPU_SP : 
				                           (void *) &CPU_ESP, 
				OP_SIZE(0), GetCPL()); 
	  if (ExceptionPending()) return(0);

	  // Need to copy any parameters to the new stack.
	  nParam = Descriptor.gate_nparam();
	  ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, Parameters, 
				 nParam*SIZE_32BIT, GetCPL());
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(SSDescriptor, newESP - nParam*SIZE_32BIT, 
				Parameters, nParam*SIZE_32BIT, GetCPL());
	  if (ExceptionPending()) return(0);

	  CPU_SS = SSSelector;
	  seg_descs[SDESC_SS] = SSDescriptor;
	  CPU_ESP = newESP - nParam;
	} else {
	  CPU_ESP = newESP;
	}
	return(1);
	break;
      }
    
      // Otherwise, CALL Gate to the same / conforming privilege. 
      if (push_return) // JMPs don't push return add, CALLs do.
	if (OP_SIZE(0) == SIZE_16BIT) {
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - SIZE_16BIT,
				&CPU_CS, SIZE_16BIT, GetCPL());
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - 2*SIZE_16BIT, 
				&CPU_IP, SIZE_16BIT, GetCPL());
	  if (ExceptionPending()) return(0);
	  CPU_ESP -= 2*SIZE_16BIT;
	} else {
	  WriteToVirtualAddress(seg_descs[SDESC_SS], CPU_ESP - SIZE_32BIT, 
				&CPU_CS, SIZE_32BIT, GetCPL());
	  if (ExceptionPending()) return(0);
	  WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP - 2*SIZE_32BIT, 
				&CPU_EIP, SIZE_32BIT, GetCPL());
	  if (ExceptionPending()) return(0);
	  CPU_ESP -= 2*SIZE_32BIT;
	}

      CPU_CS = Selector;
      seg_descs[SDESC_CS] = CSDescriptor;
      CPU_EIP = CSDescriptor.gate_offset();
      if (! CSDescriptor.use_32bit())
	CPU_EIP &= 0xFFFF;			// Only 16 bit registers.
      return(1);
      break;
    
    case DESC_TASK_GATE:
      Selector = Descriptor.gate_selector();
      if (Selector & (1<<2)) { 			// Must be in GDT
	ExceptionOccurred(EXC_INVALID_TSS, Selector); return(0);
      }
    
      if (LoadDescriptor(Descriptor, Selector, DMASK_AVAIL_TSS, 
			 true, false, false) == false) {
	return(0);
      }
      goto TSSFarCall;
      break;
  
    case DESC_16BIT_TSS_BUSY:
    case DESC_32BIT_TSS_BUSY:
      ExceptionOccurred(EXC_INVALID_TSS, Selector);
      return(0);
      break;

    case DESC_16BIT_TSS_AVAIL:
    case DESC_32BIT_TSS_AVAIL:
TSSFarCall:
      ASSERT(SwitchTasks(Selector, Descriptor));

      // make sure new (E)IP will fit...
      if(!seg_descs[SDESC_CS].offset_valid(CPU_EIP)) {
	ExceptionOccurred(EXC_INVALID_TSS, 0);
	return(0);
      }

      return(1);
      break;

    default:
      ExceptionOccurred(EXC_PROTECTION, Selector);
      return(0);
      break;
    }
  }

  return(1);  // clock count here?
}

DEFN_INST(CALL)
{
  return(CALLorJMP(inst,true) + COUNT_CLOCKS(CALL));
}
END_DEFN

DEFN_INST(CBW)
{
  if(inst.GetOperandSize() == SIZE_32BIT)
    REG16(regfile,REG_EAX_TOP) = (((CPU_AX >> 15) != 0) ? 0x0FFFF : 0); else
    CPU_AH = (((CPU_AL >> 7) != 0) ? 0x0FF : 0);

  return(COUNT_CLOCKS(CBW));
}
END_DEFN

DEFN_INST(CLC)
{
  CLEAR_BITS(CPU_EFLAGS,FLAG_CF,1);

  return(COUNT_CLOCKS(CLC));
}
END_DEFN

DEFN_INST(CLD)
{
  CLEAR_BITS(CPU_EFLAGS,FLAG_DF,1);

  return(COUNT_CLOCKS(CLD));
}
END_DEFN

DEFN_INST(CLI)
{
  // A GPF results if we are in protected mode and (CPL > IOPL)
  if(in_protected_mode() && (GetCPL() > GetIOPL()))
    ExceptionOccurred(EXC_PROTECTION,0); else
    CLEAR_BITS(CPU_EFLAGS,FLAG_IF,1);

  return(COUNT_CLOCKS(CLI));
}
END_DEFN

DEFN_INST(CLTS)
{
  if(in_protected_mode() && (GetCPL() > 0)) 
    ExceptionOccurred(EXC_PROTECTION,0); else
    CLEAR_BITS(CPU_CR0,3,1);

  return(COUNT_CLOCKS(CLTS));
}
END_DEFN

DEFN_INST(CMC)
{
  TOGGLE_BITS(CPU_EFLAGS,FLAG_CF,1);

  return(COUNT_CLOCKS(CMC));
}
END_DEFN

DEFN_INST(CMP)
{
  int32 a, b, x;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // By negating the second operand, we turn a subtraction into addition
  b = ~b;  // delay +1 til addition because it can cause carrying

  // Do sum, and _DON'T_ write results, just set flags
  x = a + b + 1;
  SetFlagsAfterOp(a,b,x,OP_SIZE(0),true);  // subtract has backwards CF and AF

  return(COUNT_CLOCKS(CMP));
}
END_DEFN

DEFN_INST(CMPS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 src, dest, diff;
  s_int32 shift_amt;

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(0);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    // load src and dest
    src = GET_OP(0);
    if(ExceptionPending()) break;

    dest = GET_OP(1);
    if(ExceptionPending()) break;

    // negate DEST so we can do an add instead of a subtract
    dest = ~dest;  // +1 delayed...

    // do the operation, and set
    diff = src + dest + 1;  // delayed +1 has correct carry results
    SetFlagsAfterOp(src,dest,diff,op_size,true);  // subtract has backwards CF and AF

    // shift (E)SI, (E)DI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_ESI += shift_amt;
      CPU_EDI += shift_amt;
    } else {
      CPU_SI += shift_amt;
      CPU_DI += shift_amt;
    }

    // count one rep
    reps++;

    // check for REPZ, REPNZ termination
    if(((rep_type == RepRepZ) && (diff != 0)) ||
       ((rep_type == RepRepNZ) && (diff == 0))) break;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      IGNORE_OSZAPC;       // on pentium pro, rep'd instructions don't set flgs
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(CMPS));
}
END_DEFN

DEFN_INST(CMPXCHG)
{
  int32 val, new_val, accum, diff;

  // First, fetch the value from memory
  val = GET_OP(0);
  if(ExceptionPending()) return(0);

  // now, get our accumulator, and the new value for memory
  new_val = GET_OP(1);
  accum = GET_OP(2);

  // figure the difference, and set flags
  diff = val + (~accum + 1);

  // write back either the old value or the new value, and set flags
  // NOTE: If diff != 0, we are writing back the same value we read from op 0,
  //  but page 25-71 explicitly states that this is the proper behavior
  if(SET_OP(0,((diff == 0) ? new_val : val))) {
    // set the accumulator on a failed match
    if(diff != 0) SET_OP(2,val);
    SetFlagsAfterOp(val,~accum + 1,diff,OP_SIZE(0),true);  // subtract has backwards CF and AF
  }

  return(COUNT_CLOCKS(CMPXCHG));
}
END_DEFN

DEFN_INST(CMPXCHG8B)
{
  int32 buffer[2];

  // First, have to fetch the 8 bytes from memory.
  if(!OP(0)->FetchBytes(*this,buffer,8)) return(0);

  // Now, compare them to EDX:EAX
  if((xtoh32(buffer[0]) == CPU_EAX) && (xtoh32(buffer[1]) == CPU_EDX)) {
    // if they are the same, write back ECX:EBX, set ZF
    buffer[0] = htox32(CPU_EBX);
    buffer[1] = htox32(CPU_ECX);
 
    if(OP(0)->WriteBytes(*this,buffer,8))
      SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1);
  } else {
    // if they're not equal load EDX:EAX with the memory value
    CPU_EAX = xtoh32(buffer[0]);
    CPU_EDX = xtoh32(buffer[1]);
    CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);
  }

  return(COUNT_CLOCKS(CMPXCHG8B));
}
END_DEFN

DEFN_INST(CPUID)
{
  switch(CPU_EAX) {
  case 0:
    CPU_EAX = 1;           // we only take CPUID requests up to EAX = 1 
    CPU_EBX = 0x0756E6547; // I think it'd be cooler if our 'vendor
    CPU_EDX = 0x049656E69; //  identification string' were something other
    CPU_ECX = 0x06C65746E; //  than "GenuineIntel", but I can't think of
                           //  anything cool right now...

    break;
  case 1:
    CPU_EAX = 0x000000525; // This is what my chip returns for model number
    CPU_EBX = 0;
    CPU_ECX = 0;
    CPU_EDX = 0x0000001BF; // We return the typical pentium feature set, even
                           //  though at this point we don't have an FPU...

    break;
  default:
    // Not an error, just do nothing
    break;
  }

  return(COUNT_CLOCKS(CPUID));
}
END_DEFN

DEFN_INST(CWD)
{
  if(inst.GetOperandSize() == SIZE_32BIT)
    CPU_EDX = (((CPU_EAX >> 31) != 0) ? 0x0FFFFFFFF : 0); else
    CPU_DX = (((CPU_AX >> 15) != 0) ? 0x0FFFF : 0);

  return(COUNT_CLOCKS(CWD));
}
END_DEFN

DEFN_INST(DAA)
{
  IGNORE_O;

  if(((CPU_AL & 0x000F) > 0x09) || 
     (BITS(CPU_EFLAGS,FLAG_AF,1) != 0))
    CPU_AL += 0x06;
    
  if(((CPU_AL & 0x00F0) > 0x0090) || 
     (BITS(CPU_EFLAGS,FLAG_CF,1) != 0)) {
    CPU_AL += 0x0060;
    SET_BITS(CPU_EFLAGS,FLAG_CF,1,1);
  }    

  return(COUNT_CLOCKS(DAA));
}
END_DEFN

DEFN_INST(DAS)
{
  int8 tmpAL;
  int tmpCF = 0;

  IGNORE_O;

  tmpAL = CPU_AL;

  if(((tmpAL & 0x0F) > 9) || (BITS(CPU_EFLAGS,FLAG_AF,1) != 0)) {
    SET_BITS(CPU_EFLAGS,FLAG_AF,1,1);
    CPU_AL -= 6;
    if(((CPU_AL & 0x0080) != 0) ||
       (BITS(CPU_EFLAGS,FLAG_CF,1) != 0)) 
      tmpCF = 1;
  }

  if((tmpAL > 0x0099) || (BITS(CPU_EFLAGS,FLAG_CF,1) != 0)) {
    CPU_AL -= 0x60;
    tmpCF = 1;
  }

  SET_BITS(CPU_EFLAGS,FLAG_CF,1,tmpCF);

  return(COUNT_CLOCKS(DAS));
}
END_DEFN

DEFN_INST(DEC)
{
  int32 val;
  int saved_cf;

  // Get the operand's value, quitting on an exception
  val = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Decrement and write, changing flags (except CF)
  if(SET_OP(0,val-1)) {
    saved_cf = BITS(CPU_FLAGS,FLAG_CF,1);
    SetFlagsAfterOp(val,0xFFFFFFFF,val-1,OP_SIZE(0),true);  // subtract has backwards CF and AF
    SET_BITS(CPU_FLAGS,FLAG_CF,1,saved_cf);
  }

  return(COUNT_CLOCKS(DEC));
}
END_DEFN

DEFN_INST(DIV)
{
  int32 quo, rem, divisor, dividend32;
  int64 dividend64;

  IGNORE_OSZAPC;

  // Common to all bit-widths is fetching the divisor
  divisor = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  switch(OP_SIZE(0)) {
  case SIZE_8BIT:
    if(CPU_AH < divisor) {
      quo = CPU_AX / divisor; // If the compiler doesn't optimize these
      rem = CPU_AX % divisor; // into a single divide, it is stupid...

      CPU_AL = quo;
      CPU_AH = rem;
    } else
      ExceptionOccurred(EXC_DIV_BY_ZERO);

    break;
  case SIZE_16BIT:
    if(CPU_DX < divisor) {
      dividend32 = (((int32) CPU_DX) << 16) | CPU_AX;

      quo = dividend32 / divisor; // If the compiler doesn't optimize these
      rem = dividend32 % divisor; // into a single divide, it is stupid...

      CPU_AX = quo;
      CPU_DX = rem;
    } else
      ExceptionOccurred(EXC_DIV_BY_ZERO);

    break;
  case SIZE_32BIT:
    if(CPU_EDX < divisor) {
      dividend64 = BUILD64(CPU_EDX,CPU_EAX);

      quo = dividend64 / divisor; // If the compiler doesn't optimize these
      rem = dividend64 % divisor; // into a single divide, it is stupid...

      CPU_EAX = quo;
      CPU_EDX = rem;
    } else
      ExceptionOccurred(EXC_DIV_BY_ZERO);

    break;
  default:
    ASSERT(0);
    break;
  };

  return(COUNT_CLOCKS(DIV));
}
END_DEFN

DEFN_INST(ENTER)
{
  int16 extra;
  int8 level;
  int8 image[(31 + 1) * 4];  // need room for old BP, up to 31 frame ptrs
  int op_size;

  // pull in the two operands first...
  extra = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  // we mod the level with 32 to make level between 0 and 31
  level = GET_OP(1) & 31;
  if(ExceptionPending()) return(0);

  op_size = inst.GetOperandSize();

  // first, stick in the old (E)BP...
  if(op_size == SIZE_32BIT)
    ((int32 *)image)[level] = htox32(CPU_EBP); else
    ((int16 *)image)[level] = htox16(CPU_BP);

  // now, if necessary, try to copy in previous stack frames
  if(level > 1) {
    if(seg_descs[SDESC_SS].use_32bit()) {
      if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],
				 CPU_EBP - ((level - 1) * op_size),
				 image+op_size,((level - 1) * op_size),
				 GetCPL()))
	return(0);
    } else {
      if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],
				 CPU_BP - ((level - 1) * op_size),
				 image+op_size,((level - 1) * op_size),
				 GetCPL()))
	return(0);
    }
  }

  // if needed (level > 0), copy in current frame ptr
  if(level > 0) {
    if(op_size == SIZE_32BIT)
      ((int32 *)image)[0] = htox32(CPU_ESP); else
      ((int16 *)image)[0] = htox16(CPU_SP);
  }

  // now, try to push the entire stack image we've built onto the stack
  if(seg_descs[SDESC_SS].use_32bit()) {
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],
			      CPU_ESP - ((level + 1) * op_size),
			      image,((level + 1) * op_size),
			      GetCPL()))
      return(0);
  } else {
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],
			      CPU_SP - ((level + 1) * op_size),
			      image,((level + 1) * op_size),
			      GetCPL()))
      return(0);
  }

  if(op_size == SIZE_32BIT) {
    CPU_EBP = CPU_ESP;
    CPU_ESP -= (extra + ((level + 1) << 2));
  } else {
    CPU_BP = CPU_SP;
    CPU_SP -= (extra + ((level + 1) << 1));
  }

  return(COUNT_CLOCKS(ENTER));
}
END_DEFN

DEFN_INST(HLT)
{
  if(in_protected_mode() && (GetCPL() > 0)) 
    ExceptionOccurred(EXC_PROTECTION,0); else
    halted = true;

  return(COUNT_CLOCKS(HLT));
}
END_DEFN

DEFN_INST(IDIV)
{
  s_int32 quo, rem, divisor, dividend32;
  s_int64 dividend64, min_quo, max_quo;

  IGNORE_OSZAPC;

  // Common to all bit-widths is fetching the divisor
  divisor = GET_OP(0);
  if(ExceptionPending()) return(0);

  switch(OP_SIZE(0)) {
  case SIZE_8BIT:
    dividend32 = CPU_AX;

    // our dividend must be between -128 and 127, so multiply to see if it'll
    //  fit
    if(divisor >= 0) {
      if(((divisor * 128) <= dividend32) || ((divisor * -129) >= dividend32)) {
	ExceptionOccurred(EXC_DIV_BY_ZERO);
	break;
      }
    } else {
      if(((divisor * 128) >= dividend32) || ((divisor * -129) <= dividend32)) {
	ExceptionOccurred(EXC_DIV_BY_ZERO);
	break;
      }
    }

    quo = dividend32 / divisor;
    rem = dividend32 - (quo * divisor);  // not as efficient, but compilers 
                                         //  tend to be weird about signed %'s

    CPU_AL = quo;
    CPU_AH = rem;
    break;
  case SIZE_16BIT:
    dividend32 = (s_int32) ((((int32) CPU_DX) << 16) | CPU_AX);

    // our dividend must be between -32768 and 32767, so multiply to see if
    //  it'll fit
    if(divisor >= 0) {
      if(((divisor * 32768) <= dividend32) || 
	 ((divisor * -32769) >= dividend32)) {
	ExceptionOccurred(EXC_DIV_BY_ZERO);
	break;
      }
    } else {
      if(((divisor * 32768) >= dividend32) || 
	 ((divisor * -32769) <= dividend32)) {
	ExceptionOccurred(EXC_DIV_BY_ZERO);
	break;
      }
    }

    quo = dividend32 / divisor;
    rem = dividend32 - (quo * divisor);  // not as efficient, but compilers 
                                         //  tend to be weird about signed %'s

    CPU_AX = quo;
    CPU_DX = rem;
    break;
  case SIZE_32BIT:
    dividend64 = (s_int64) BUILD64(CPU_EDX,CPU_EAX);

    // turns out you can specify 64-bit constants by trailing with LL, so
    //  we can check just like above...
    if(divisor >= 0) {
      min_quo = -2147483649LL;
      max_quo =  2147483648LL;
    } else {
      min_quo =  2147483648LL;
      max_quo = -2147483649LL;
    }
    if(((max_quo * divisor) <= dividend64) ||
       ((min_quo * divisor) >= dividend64)) {
      ExceptionOccurred(EXC_DIV_BY_ZERO);
      break;
    }
 
    quo = dividend64 / divisor;
    rem = dividend64 - (quo * divisor);  // not as efficient, but compilers 
                                         //  tend to be weird about signed %'s

    CPU_EAX = quo;
    CPU_EDX = rem;
    break;
  default:
    ASSERT(0);
    break;
  };

  return(COUNT_CLOCKS(IDIV));
}
END_DEFN

DEFN_INST(IMUL)
{
  s_int32 multiplicand, multiplier;
  s_int64 product;
  s_int32 product32;
  s_int16 product16;
  s_int8  product8;
  bool overflowed;

  IGNORE_SZAP;

  // Fetch multiplicand, multiplier
  multiplicand = GET_OP(1);
  if(ExceptionPending()) return(0);

  multiplier = GET_OP(2);
  if(ExceptionPending()) return(0);

  product = ((s_int64)multiplicand) * ((s_int64)multiplier);

  // check to see if result was in range
  switch(OP_SIZE(1)) {
  case SIZE_8BIT:
    product8 = product;
    overflowed = (product8 != product);
    break;
  case SIZE_16BIT:
    product16 = product;
    overflowed = (product16 != product);
    break;
  case SIZE_32BIT:
    product32 = product;
    overflowed = (product32 != product);
    break;
  default:
    ASSERT(0);
    return(0);
    break;
  }

  // This is kind of a hack - If we have a 64-bit product (going to EDX:EAX),
  //  we have OP_SIZE(0) == 8 (SIZE_64BIT), we handle the write differently
  if(OP_SIZE(0) == 8 /* SIZE_64BIT */) {
    CPU_EDX = TOP32(product);
    CPU_EAX = BOTTOM32(product);
  }

  if((OP_SIZE(0) == 8) || SET_OP(0,product)) {
    SET_BITS(CPU_EFLAGS,FLAG_CF,1,(overflowed ? 1 : 0));
    SET_BITS(CPU_EFLAGS,FLAG_OF,1,(overflowed ? 1 : 0));
  }

  return(COUNT_CLOCKS(IMUL));
}
END_DEFN

DEFN_INST(IN)
{
  int32 port;
  int32 value;

  // first, load the port number
  port = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  // check permissions
  if(!CheckIOPermissions(port,OP_SIZE(0))) {
    // if we didn't page fault, it's a gpf
    if(!ExceptionPending()) ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // now just do the read
  ReadFromIOPort(port,&value,OP_SIZE(0));
  SET_OP(0,value);

  return(COUNT_CLOCKS(IN));
}
END_DEFN

DEFN_INST(INC)
{
  int32 val;
  int saved_cf;

  // Get the operand's value, quitting on an exception
  val = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Increment and write, changing flags (except CF)
  if(SET_OP(0,val+1)) {
    saved_cf = BITS(CPU_FLAGS,FLAG_CF,1);
    SetFlagsAfterOp(val,1,val+1,OP_SIZE(0),false);
    SET_BITS(CPU_FLAGS,FLAG_CF,1,saved_cf);
  }

  return(COUNT_CLOCKS(INC));
}
END_DEFN

DEFN_INST(INS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 port, value;
  s_int32 shift_amt;

  // first, load the port number
  port = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  // check permissions
  if(!CheckIOPermissions(port,OP_SIZE(0))) {
    // if we didn't page fault, it's a gpf
    if(!ExceptionPending()) ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(0);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    ReadFromIOPort(port,&value,op_size);

    if(!SET_OP(0,value)) break;

    // shift (E)DI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_EDI += shift_amt;
    } else {
      CPU_DI += shift_amt;
    }

    // count one rep
    reps++;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(INS));
}
END_DEFN

DEFN_INST(INT)
{
  int8 int_num;

  int_num = (OP(0) ? GET_OP_Z(0) : 0x03);

#ifdef SIMOS
 if(int_num == 3) {
    /* Debugger breakpoint trap */
    if (X86SimDebug(0, 1)) return 0;
 } else if ((int_num == 0x81) && X86DoMonitorCallHack(0)) {
    /* Is a hacked monitor call */
    return 0;
 } else {
   ExceptionOccurred(int_num, 0, EXC_INT);
 }
#else
  if(int_num == 0x80) {
    _last_was_syscall = 1;
    if(SimulateLinuxSystemCall(&regfile))
      halted = true;   // halt if the syscall simulator asks us to...
  } else {
    ExceptionOccurred(int_num, 0, EXC_INT);
  }
#endif
  return(COUNT_CLOCKS(INT));
}
END_DEFN

DEFN_INST(INTO)
{
  // If Jeremy is nice, and gives us an immediate operand 0 with value 4,
  //  we can either just call INT (if OF = 1), or do nothing (OF = 0)
  return((BITS(CPU_EFLAGS,FLAG_OF,1) == 1) ?
	 INT(inst) :             // INT and INTO take the same # of clocks
	 COUNT_CLOCKS(INTO));
}
END_DEFN

DEFN_INST(INVD)
{
  if(in_protected_mode() && (GetCPL() > 0)) {
    ExceptionOccurred(EXC_PROTECTION,0);
  } else {
    // Invalidate Internal Cache?
  }

  return(COUNT_CLOCKS(INVD));
}
END_DEFN

DEFN_INST(INVLPG)
{
#ifdef notdef
//XXX might want to model a TLB for speed reasons. In which case
//this would be useful again.
  ASSERT(0);
#endif
  return(COUNT_CLOCKS(INVLPG));
}
END_DEFN

DEFN_INST(IRET)
{
  bool size32 = inst.GetOperandSize();
  
  if (in_real_mode()) {
    //  Real mode IRET(D).  Note that if Pop()ing the registers generates and
    // exception, the values of any earlier registers have been clobbered.
    if (size32) {
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, &CPU_EIP,
				 SIZE_32BIT, GetCPL()) == 0)
	return 0;
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + SIZE_32BIT,
				 &CPU_CS, SIZE_16BIT, GetCPL()) == 0)
	return 0;
      if (LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true,
			 false, false) == false)
	return 0;
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 2*SIZE_32BIT,
				 &CPU_EFLAGS, SIZE_32BIT, GetCPL()) == 0)
	return 0;
      CPU_ESP += 3*SIZE_32BIT;
    } else {
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP, &CPU_IP,
				 SIZE_16BIT, GetCPL()) == 0)
	return 0;
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + SIZE_16BIT,
				 &CPU_CS, SIZE_16BIT, GetCPL()) == 0)
	return 0;
      if (LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true,
			 false, false) == false)
	return 0;
      if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + 2*SIZE_16BIT,
				 &CPU_FLAGS, SIZE_16BIT, GetCPL()) == 0)
	return 0;
      CPU_SP += 3*SIZE_16BIT;
    }
  } else {
    // Protected Mode IRET(D).
    if (in_v86_mode()) {
      // Return from a v86 interupt handler.
      if (GetIOPL() == 3) {
	// Just like real mode.  Block copied from above.
	if (size32) {
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, &CPU_EIP,
				     SIZE_32BIT, GetCPL()) == 0)
	    return 0;
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + SIZE_32BIT,
				     &CPU_CS, SIZE_16BIT, GetCPL()) == 0)
	    return 0;
	  if (LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true,
			     false, false) == false)
	    return 0;
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP+2*SIZE_32BIT,
				     &CPU_EFLAGS, SIZE_32BIT, GetCPL()) == 0)
	    return 0;
	  CPU_ESP += 3*SIZE_32BIT;
	} else {
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP, &CPU_IP,
				     SIZE_16BIT, GetCPL()) == 0)
	    return 0;
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + SIZE_16BIT,
				     &CPU_CS, SIZE_16BIT, GetCPL()) == 0)
	    return 0;
	  if (LoadDescriptor(seg_descs[SDESC_CS], CPU_CS, DMASK_CODE, true,
			     false, false) == false)
	    return 0;
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP+2*SIZE_16BIT,
				     &CPU_FLAGS, SIZE_16BIT, GetCPL()) == 0)
	    return 0;
	  CPU_SP += 3*SIZE_16BIT;
	}
      } else {
	ExceptionOccurred(EXC_PROTECTION, 0);
	return 0;
      }
    } else {
      if (BITS(CPU_EFLAGS,FLAG_NT,1) == 1) {
	int16 RetTask;
	SegmentDescriptor RetTSS;
	// Return from a Nested Task...
	if (ReadFromVirtualAddress(seg_descs[SDESC_TR], 0, &RetTask,
				   SIZE_16BIT, GetCPL()) == 0)
	  return 0;
	if (RetTask & (1<<2)) {			// Selector must be in GDT
	  ExceptionOccurred(EXC_INVALID_TSS, RetTask);
	  return 0;
	}
	if (LoadDescriptor(RetTSS, RetTask, DMASK_BUSY_TSS, true, false,
			   false) == false)
	    return 0;				// LoadDesc() call ExcOccurred
	ASSERT(SwitchTasks(RetTask, RetTSS));
	if (seg_descs[SDESC_CS].offset_valid(CPU_EIP) == false) {
	  ExceptionOccurred(EXC_INVALID_TSS, 0);
	  return 0;
	}
      } else {
	int32 eFLAGS;

	// I'm not quite sure how this works.  It's a return to v86 Mode if
	// the Flags image on the stack has VM = 1.  But VM is Bit 17 Which
	// isn't present in a 16 Bit Flags register. So...
	if (size32) {
	  if (ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP+2*SIZE_32BIT,
				    &eFLAGS, SIZE_32BIT, GetCPL()) == false)
	    return 0;
	  if (BITS(eFLAGS, FLAG_VM, 1) == 1) {
	    int32 eIP, eSP;
	    int16 CSSelector, SSSelector;
	    SegmentDescriptor CSDesc;
	    
	    // Return to v86 Mode...
	    if (seg_descs[SDESC_SS].offset_valid(CPU_ESP + 36) == false) {
	      ExceptionOccurred(EXC_STACK_FAULT, 0);
	      return 0;
	    }
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, &eIP,
				   SIZE_32BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + SIZE_32BIT, 
				   &CSSelector, SIZE_16BIT, GetCPL());
	    if (LoadDescriptor(CSDesc, CSSelector, DMASK_CODE, true, false,
			       false) == false)
	      return 0;
	    if (CSDesc.offset_valid(eIP) == false) {
	      ExceptionOccurred(EXC_PROTECTION, 0);
	      return 0;
	    }

	    // Okay, if we made it this far, it should be safe to finish.
	    CPU_EIP = eIP;
	    CPU_EFLAGS = eFLAGS;
	    CPU_CS = CSSelector;
	    seg_descs[SDESC_CS] = CSDesc;

	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 3*SIZE_32BIT,
				   &eSP, SIZE_32BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 4*SIZE_32BIT,
				   &SSSelector, SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 5*SIZE_32BIT,
				   &CPU_ES, SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 6*SIZE_32BIT,
				   &CPU_DS, SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 7*SIZE_32BIT,
				   &CPU_FS, SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 8*SIZE_32BIT,
				   &CPU_GS, SIZE_16BIT, GetCPL());
	    CPU_ESP = eSP;
	    CPU_SS = SSSelector;
	    LoadDescriptor(seg_descs[SDESC_SS], CPU_SS, DMASK_DATA, true,
			   false, false);
	    LoadDescriptor(seg_descs[SDESC_ES], CPU_ES, DMASK_DATA, true,
			   false, false);
	    LoadDescriptor(seg_descs[SDESC_DS], CPU_DS, DMASK_DATA, true,
			   false, false);
	    LoadDescriptor(seg_descs[SDESC_FS], CPU_FS, DMASK_DATA, true,
			   false, false);
	    LoadDescriptor(seg_descs[SDESC_GS], CPU_GS, DMASK_DATA, true,
			   false, false);
	    return (COUNT_CLOCKS(IRET));
	  }
	}
      
	// Protected Mode IRET to Protected Mode Code. (Finally!)
	int16 CSSelector;
	SegmentDescriptor CSDescriptor;
	
	if (size32) {
	  if (!seg_descs[SDESC_SS].offset_valid(CPU_ESP + 3*SIZE_32BIT)) {
	    ExceptionOccurred(EXC_STACK_FAULT, 0); return 0;
	  }
	  ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + SIZE_32BIT,
				 &CSSelector, SIZE_16BIT, GetCPL());
	} else {
	  if (!seg_descs[SDESC_SS].offset_valid(CPU_SP + 3*SIZE_16BIT)) {
	    ExceptionOccurred(EXC_STACK_FAULT, 0); return 0;
	  }
	  ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + SIZE_16BIT,
				 &CSSelector, SIZE_16BIT, GetCPL());
	}

	if (BITS(CSSelector, 0, 2) < GetCPL()) {	//  RPL < CPL
	  ExceptionOccurred(EXC_PROTECTION, CSSelector); return 0;
	}

	if (BITS(CSSelector, 0, 2) == GetCPL()) {	// RPL == CPL
	  // Same privilege level far return
	  // Already checked to make sure we had enough stack space above so
	  // can skip the check at the top of RETURN-SAME-LEVEL in the
	  // psuedo-code for IRET.
	  if (!LoadDescriptor(CSDescriptor, CSSelector, DMASK_CODE, true,
	 		      false, false))	
	    return 0;

	  // If we made it this far, safe to start changing state.
	  CPU_CS = CSSelector;
	  seg_descs[SDESC_CS] = CSDescriptor;
	  if (size32) {
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, &CPU_EIP,
				  SIZE_32BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 2*SIZE_32BIT,
				  &CPU_EFLAGS, SIZE_32BIT, GetCPL());
	    CPU_ESP += 3*SIZE_32BIT;
	  } else {
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP, &CPU_IP,
				  SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + 2*SIZE_16BIT,
				  &CPU_FLAGS, SIZE_16BIT, GetCPL());
	    CPU_SP += 3*SIZE_16BIT;
	  }
	} else {					// RPL > CPL
	  int16 SSSelector, OldCSSelector;
	  SegmentDescriptor SSDescriptor;

	  // Return to higher (less privilege) level.
	  if (size32) {
	    if (seg_descs[SDESC_SS].offset_valid(CPU_ESP + 20) == false)
	      ExceptionOccurred(EXC_STACK_FAULT, 0);
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 4*SIZE_32BIT,
				   &SSSelector, SIZE_16BIT, GetCPL());
	  } else {
	    if (seg_descs[SDESC_SS].offset_valid(CPU_SP + 20) == false)
	      ExceptionOccurred(EXC_STACK_FAULT, 0);
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + 4*SIZE_16BIT,
				   &SSSelector, SIZE_16BIT, GetCPL());
	  }

	  // Okay, all of the Descriptor Loading needs to take place as if
	  // the new CSSelector determined the CPL, but we need to preserve
	  // the current CPL to be able to POP things off the stack later.
	  OldCSSelector = CPU_CS;
	  CPU_CS = CSSelector;
	  
	  if (LoadDescriptor(CSDescriptor, CSSelector, DMASK_CODE, true,
			     false, false) == false) {
	    CPU_CS = OldCSSelector; return 0;
	  }

	  // SS Selector RPL must == CS Selector RPL
	  if (BITS(CSSelector, 0, 2) != BITS(SSSelector, 0, 2)) {
	    ExceptionOccurred(EXC_PROTECTION, SSSelector);
	    CPU_CS = OldCSSelector; return 0;
	  }

	  if (LoadDescriptor(SSDescriptor, SSSelector, DMASK_WRITEABLE,
			     true, false, false) == false) {
	    CPU_CS = OldCSSelector; return 0;
	  }
	  
	  //  Okay, time to start changing state.  No further exceptions can
	  // occur.  Need to set CPL back to what it was before though
	  // or else not necessarily be privileged enough to pop.
	  CPU_CS = OldCSSelector;
	  
	  if (size32) {
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP, &CPU_EIP,
				   SIZE_32BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 2*SIZE_32BIT,
				   &CPU_EFLAGS, SIZE_32BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_ESP + 3*SIZE_32BIT,
				   &CPU_ESP, SIZE_32BIT, GetCPL());
	  } else {
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP, &CPU_IP,
				   SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + 2*SIZE_16BIT,
				   &CPU_FLAGS, SIZE_16BIT, GetCPL());
	    ReadFromVirtualAddress(seg_descs[SDESC_SS], CPU_SP + 3*SIZE_16BIT,
				   &CPU_SP, SIZE_16BIT, GetCPL());
	  }

	  // Must do this after Popping or we screw up CPL.
	  CPU_CS = CSSelector;
	  seg_descs[SDESC_CS] = CSDescriptor;
	  CPU_SS = SSSelector;
	  seg_descs[SDESC_SS] = SSDescriptor;

	  // Okay, now weird checks to make sure the other segment registers
	  // are still legal at the new privilege level.
	  if ((CPU_ES != 0) && (seg_descs[SDESC_ES].is_data() ||
	       (seg_descs[SDESC_ES].is_code() &&
		!seg_descs[SDESC_ES].is_conforming())) &&
	      (seg_descs[SDESC_ES].dpl() < GetCPL())) {
	    seg_descs[SDESC_ES].mark_invalid();
	    CPU_ES = 0;
	  }
	  if ((CPU_FS != 0) && (seg_descs[SDESC_FS].is_data() ||
	       (seg_descs[SDESC_FS].is_code() &&
		!seg_descs[SDESC_FS].is_conforming())) &&
	      (seg_descs[SDESC_FS].dpl() < GetCPL())) {
	    seg_descs[SDESC_FS].mark_invalid();
	    CPU_FS = 0;
	  }
	  if ((CPU_GS != 0) && (seg_descs[SDESC_GS].is_data() ||
	       (seg_descs[SDESC_GS].is_code() &&
		!seg_descs[SDESC_GS].is_conforming())) &&
	      (seg_descs[SDESC_GS].dpl() < GetCPL())) {
	    seg_descs[SDESC_GS].mark_invalid();
	    CPU_GS = 0;
	  }
	  if ((CPU_DS != 0) && (seg_descs[SDESC_DS].is_data() ||
	       (seg_descs[SDESC_DS].is_code() &&
		!seg_descs[SDESC_DS].is_conforming())) &&
	      (seg_descs[SDESC_DS].dpl() < GetCPL())) {
	    seg_descs[SDESC_DS].mark_invalid();
	    CPU_DS = 0;
	  }		     
	}
      }
    }
  }
#ifdef SIMOS
  X86DidIRET(0);  
#endif
  return(COUNT_CLOCKS(IRET));
}
END_DEFN

DEFN_INST(Jcc)
{
  // Operand 0 is the condition code, operand 1 is the offset

  // If the condition is met, change EIP - if not, don't
  if(check_condition(CPU_EFLAGS,GET_OP(0))) {
    CPU_EIP += GET_OP(1);

    // If we're using 16-bit code segments, EIP can't be > 0xFFFF
    if(!seg_descs[SDESC_CS].use_32bit())
      CPU_EIP &= 0x0000FFFF;
  }

  return(COUNT_CLOCKS(Jcc));
}
END_DEFN

DEFN_INST(JCXZ)
{
  // Operand 0 is the offset (decide on CX or ECX based on operand size)

  // If the condition is met, change EIP - if not, don't GGG
  if(((inst.GetOperandSize() == SIZE_32BIT) && (CPU_ECX == 0)) ||
     ((inst.GetOperandSize() == SIZE_16BIT) && (CPU_CX  == 0))) {
    CPU_EIP += GET_OP(0);

    // If we're using 16-bit code segments, EIP can't be > 0xFFFF
    if(!seg_descs[SDESC_CS].use_32bit())
      CPU_EIP &= 0x0000FFFF;
  }

  return(COUNT_CLOCKS(JCXZ));
}
END_DEFN

DEFN_INST(JMP)
{
  return(CALLorJMP(inst,false) + COUNT_CLOCKS(JMP));
}
END_DEFN

DEFN_INST(LAHF)
{
  CPU_AH = (CPU_EFLAGS & 0x000000FF);

  return(COUNT_CLOCKS(LAHF));
}
END_DEFN

DEFN_INST(LAR)
{
  int16 selector;
  SegmentDescriptor descriptor;
  
  // real and vm86 modes do not do LAR
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // first, get the specified selector
  selector = GET_OP(1);
  if(ExceptionPending()) return(0);

  // now, try to get the descriptor for it, suppressing errors
  if(LoadDescriptor(descriptor,selector,DMASK_HAS_AR_BYTE,false,false,true)) {
    // write the ar "byte" into the target, and set ZF
    if(SET_OP(0,descriptor.ar_byte()))
      SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1);
  } else
    CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);

  return(COUNT_CLOCKS(LAR));
}
END_DEFN

DEFN_INST(LDS)
{
  int16 selector;
  int32 offset;
  int op_size;

  // first, try to load the designated selector and offset
  op_size = OP_SIZE(0);

  // note that loading a 16-bit value into the lower addresses of a 32-bit
  //  variable sounds weird, but it turns out that when we convert to little
  //  endian byte order, everything gets fixed (if we ignore the (new) top
  //  bits...)
  if(!OP(1)->FetchBytes(*this,&offset,op_size) ||
     !OP(1)->FetchBytes(*this,&selector,2,op_size))
    return(0);

  // try to load the descriptor into the given segment now...  (we assume the
  //  write of the offset (which goes into a register) always works...
  if(LoadDescriptor(seg_descs[SDESC_DS],selector,DMASK_READABLE,true,true,
		    false)) {
    CPU_DS = selector;
    SET_OP(0,offset);
    seg_descs[SDESC_DS].set_accessed(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_DS],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LDS));
}
END_DEFN

DEFN_INST(LEA)
{
  // Relatively straight-forward operation...  Get EA of operand 1, and dump
  //  it in operand 0  (zero-extension of 16-bit EA into 32-bit target is
  //  handled by Fetch_EA_Offset(), and truncation of 32-bit EA into a 16-bit
  //  target should be handled automatically by SET_OP
  SET_OP(0,OP(1)->Fetch_EA_Offset(*this));

  return(COUNT_CLOCKS(LEA));
}
END_DEFN

DEFN_INST(LEAVE)
{
  int32 new_ebp;
  int op_size;

  op_size = inst.GetOperandSize();

  if(seg_descs[SDESC_SS].use_32bit()) {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_EBP,&new_ebp,op_size,
			      GetCPL()))
      CPU_ESP = CPU_EBP + op_size; else
      new_ebp = htox32(CPU_EBP);
  } else {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_BP,&new_ebp,op_size,
			      GetCPL()))
      CPU_SP = CPU_BP + op_size; else
      new_ebp = htox32(CPU_EBP);
  }

  if(op_size == SIZE_32BIT)
    CPU_EBP = xtoh32(new_ebp); else
    CPU_BP  = xtoh32(new_ebp);

  return(COUNT_CLOCKS(LEAVE));
}
END_DEFN

DEFN_INST(LES)
{
  int16 selector;
  int32 offset;
  int op_size;

  // first, try to load the designated selector and offset
  op_size = OP_SIZE(0);

  // note that loading a 16-bit value into the lower addresses of a 32-bit
  //  variable sounds weird, but it turns out that when we convert to little
  //  endian byte order, everything gets fixed (if we ignore the (new) top
  //  bits...)
  if(!OP(1)->FetchBytes(*this,&offset,op_size) ||
     !OP(1)->FetchBytes(*this,&selector,2,op_size))
    return(0);

  // try to load the descriptor into the given segment now...  (we assume the
  //  write of the offset (which goes into a register) always works...
  if(LoadDescriptor(seg_descs[SDESC_ES],selector,DMASK_READABLE,true,true,
		    false)) {
    CPU_ES = selector;
    SET_OP(0,offset);
    seg_descs[SDESC_ES].set_accessed(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_ES],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LES));
}
END_DEFN

DEFN_INST(LFS)
{
  int16 selector;
  int32 offset;
  int op_size;

  // first, try to load the designated selector and offset
  op_size = OP_SIZE(0);

  // note that loading a 16-bit value into the lower addresses of a 32-bit
  //  variable sounds weird, but it turns out that when we convert to little
  //  endian byte order, everything gets fixed (if we ignore the (new) top
  //  bits...)
  if(!OP(1)->FetchBytes(*this,&offset,op_size) ||
     !OP(1)->FetchBytes(*this,&selector,2,op_size))
    return(0);

  // try to load the descriptor into the given segment now...  (we assume the
  //  write of the offset (which goes into a register) always works...
  if(LoadDescriptor(seg_descs[SDESC_FS],selector,DMASK_READABLE,true,true,
		    false)) {
    CPU_FS = selector;
    SET_OP(0,offset);
    seg_descs[SDESC_FS].set_accessed(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_FS],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LFS));
}
END_DEFN

DEFN_INST(LGDT)
{
  int8 gdt_buffer[6];

  // check permissions first
  if(in_protected_mode() && (GetCPL() != 0)) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // load gdt_buffer, bombing if the read fails
  gdt_buffer[5] = 0;  // in case we only read 5 bytes
  if(!OP(0)->FetchBytes(*this,gdt_buffer,
			((inst.GetOperandSize() == SIZE_16BIT) ? 5 : 6)))
    return(0);

  seg_descs[SDESC_GDT].set_base_limit(xtoh32(*(int32 *)(gdt_buffer+2)),
				      xtoh16(*(int16 *)(gdt_buffer+0)),
				      PRESENT_RW);

  return(COUNT_CLOCKS(LGDT));
}
END_DEFN

DEFN_INST(LGS)
{
  int16 selector;
  int32 offset;
  int op_size;

  // first, try to load the designated selector and offset
  op_size = OP_SIZE(0);

  // note that loading a 16-bit value into the lower addresses of a 32-bit
  //  variable sounds weird, but it turns out that when we convert to little
  //  endian byte order, everything gets fixed (if we ignore the (new) top
  //  bits...)
  if(!OP(1)->FetchBytes(*this,&offset,op_size) ||
     !OP(1)->FetchBytes(*this,&selector,2,op_size))
    return(0);

  // try to load the descriptor into the given segment now...  (we assume the
  //  write of the offset (which goes into a register) always works...
  if(LoadDescriptor(seg_descs[SDESC_GS],selector,DMASK_READABLE,true,true,
		    false)) {
    CPU_GS = selector;
    SET_OP(0,offset);
    seg_descs[SDESC_GS].set_accessed(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_GS],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LGS));
}
END_DEFN

DEFN_INST(LIDT)
{
  int8 idt_buffer[6];

  // check permissions first
  if(in_protected_mode() && (GetCPL() != 0)) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // load idt_buffer, bombing if the read fails
  idt_buffer[5] = 0;  // in case we only read 5 bytes
  if(!OP(0)->FetchBytes(*this,idt_buffer,
			((inst.GetOperandSize() == SIZE_16BIT) ? 5 : 6)))
    return(0);

  seg_descs[SDESC_IDT].set_base_limit(xtoh32(*(int32 *)(idt_buffer+2)),
				      xtoh16(*(int16 *)(idt_buffer+0)),
				      PRESENT_RW);

  return(COUNT_CLOCKS(LIDT));
}
END_DEFN

DEFN_INST(LLDT)
{
  int16 selector;

  // only protected mode gets to do a LLDT
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // only cpl 0 gets to do an LLDT
  if(GetCPL() != 0) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // first, try to load the designated TSS selector
  selector = GET_OP(0);
  if(ExceptionPending()) return(0);

  // the descriptor must not point into the LDT (no cycles here, please)
  if(BITS(selector,2,1) == 1) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // try to load the descriptor into the LDT register now...
  if(LoadDescriptor(seg_descs[SDESC_LDT],selector,DMASK_LDT,true,true,false))
    CPU_LDT = selector;

  return(COUNT_CLOCKS(LLDT));
}
END_DEFN

DEFN_INST(LMSW)
{
  ASSERT(0);

  return(COUNT_CLOCKS(LMSW));
}
END_DEFN

DEFN_INST(LODS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 data;
  s_int32 shift_amt;

  // operand 0 is AL,AX, or EAX; operand 1 is DS:(E)SI

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(0);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    // for LODS we have to load in the value, and then write to AL, AX, or
    //  EAX as appropriate for the operand's size
    data = GET_OP(1);
    if(ExceptionPending()) break;

    // this should never exception, but we can check for free
    if(!SET_OP(0,data)) break;

    // shift (E)SI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_ESI += shift_amt;
    } else {
      CPU_SI += shift_amt;
    }

    // count one rep
    reps++;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(LODS));
}
END_DEFN

DEFN_INST(LOOP)
{
  bool counter_zero;

  if(inst.GetOperandSize() == SIZE_32BIT)
    counter_zero = ((--CPU_ECX) == 0); else
    counter_zero = ((--CPU_CX)  == 0);

  if(!counter_zero) {
    CPU_EIP += GET_OP(0);

    // If we're using 16-bit code segments, EIP can't be > 0xFFFF
    if(!seg_descs[SDESC_CS].use_32bit())
      CPU_EIP &= 0x0000FFFF;
  }

  return(COUNT_CLOCKS(LOOP));
}
END_DEFN

DEFN_INST(LOOPNZ)
{
  bool counter_zero;

  if(inst.GetOperandSize() == SIZE_32BIT)
    counter_zero = ((--CPU_ECX) == 0); else
    counter_zero = ((--CPU_CX)  == 0);

  if(!counter_zero && (BITS(CPU_EFLAGS,FLAG_ZF,1) == 0)) {
    CPU_EIP += GET_OP(0);

    // If we're using 16-bit code segments, EIP can't be > 0xFFFF
    if(!seg_descs[SDESC_CS].use_32bit())
      CPU_EIP &= 0x0000FFFF;
  }

  return(COUNT_CLOCKS(LOOPNZ));
}
END_DEFN

DEFN_INST(LOOPZ)
{
  bool counter_zero;

  if(inst.GetOperandSize() == SIZE_32BIT)
    counter_zero = ((--CPU_ECX) == 0); else
    counter_zero = ((--CPU_CX)  == 0);

  if(!counter_zero && (BITS(CPU_EFLAGS,FLAG_ZF,1) == 1)) {
    CPU_EIP += GET_OP(0);

    // If we're using 16-bit code segments, EIP can't be > 0xFFFF
    if(!seg_descs[SDESC_CS].use_32bit())
      CPU_EIP &= 0x0000FFFF;
  }

  return(COUNT_CLOCKS(LOOPZ));
}
END_DEFN

DEFN_INST(LSL)
{
  int16 selector;
  SegmentDescriptor descriptor;
  int32 limit;
  
  // real and vm86 modes do not do LAR
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // first, get the specified selector
  selector = GET_OP(1);
  if(ExceptionPending()) return(0);

  // now, try to get the descriptor for it, suppressing errors
  //  we need it to be present in order to know the limit
  if(LoadDescriptor(descriptor,selector,DMASK_HAS_LIMIT,true,false,true)) {
    // the limit may need to be shifted left 12 bits...
    limit = descriptor.limit();

    if(descriptor.big_granularity())
      limit = (limit << 12) | 0x0FFF;

    // write the limit into the target, and set ZF
    if(SET_OP(0,limit))
      SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1);
  } else
    CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);

  return(COUNT_CLOCKS(LSL));
}
END_DEFN

DEFN_INST(LSS)
{
  int16 selector;
  int32 offset;
  int op_size;

  // first, try to load the designated selector and offset
  op_size = OP_SIZE(0);

  // note that loading a 16-bit value into the lower addresses of a 32-bit
  //  variable sounds weird, but it turns out that when we convert to little
  //  endian byte order, everything gets fixed (if we ignore the (new) top
  //  bits...)
  if(!OP(1)->FetchBytes(*this,&offset,op_size) ||
     !OP(1)->FetchBytes(*this,&selector,2,op_size))
    return(0);

  // try to load the descriptor into the given segment now...  (we assume the
  //  write of the offset (which goes into a register) always works...
  if(LoadDescriptor(seg_descs[SDESC_SS],selector,DMASK_WRITEABLE,true,false,
		    false)) {
    CPU_SS = selector;
    SET_OP(0,offset);
    seg_descs[SDESC_SS].set_accessed(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_SS],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LSS));
}
END_DEFN

DEFN_INST(LTR)
{
  int16 selector;

  // only protected mode gets to do a LTR
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // only cpl 0 gets to do an LTR
  if(GetCPL() != 0) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // first, try to load the designated TSS selector
  selector = GET_OP(0);
  if(ExceptionPending()) return(0);

  // try to load the descriptor into the task register now...
  if(LoadDescriptor(seg_descs[SDESC_TR],selector,DMASK_AVAIL_TSS,true,false,
		    false)) {
    CPU_TR = selector;
    seg_descs[SDESC_TR].set_tss_busy(true);
    if(!WriteBackDescriptor(seg_descs[SDESC_TR],selector)) ASSERT(0);
  }

  return(COUNT_CLOCKS(LTR));
}
END_DEFN

DEFN_INST(MOV)
{
  int32 data;

  // this would be a nice simple instruction, but we need to worry about writes
  //  to segment registers, control registers, or debug registers
  data = GET_OP(1);   
  if(ExceptionPending()) return(0);

  // first, we have a little weirdness involved if we just read from DR4 or
  //  DR5 - if we did, check the DE bit (bit 3) of CR4 - if it is set, fire
  //  off an invalid opcode exception - otherwise, re-read from DR6 or DR7
  if(OP(1)->Type() == OPTYPE_REGISTER)
    switch(((RegisterOperand *) OP(1))->RegNum()) {
    case REG_DR4:
      if(BITS(CPU_CR4,3,1) == 1) {
	ExceptionOccurred(EXC_INVALID_OPCODE);
	return(0);
      } else
	data = CPU_DR6;
      break;

    case REG_DR5:
      if(BITS(CPU_CR4,3,1) == 1) {
	ExceptionOccurred(EXC_INVALID_OPCODE);
	return(0);
      } else
	data = CPU_DR7;
      break;

    default:
      break;
    }

  if(OP(0)->Type() == OPTYPE_REGISTER)
    switch(((RegisterOperand *) OP(0))->RegNum()) {
    case REG_DS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_DS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_DS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_DS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_ES:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_ES],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_ES].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_ES],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_FS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_FS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_FS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_FS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_GS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_GS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_GS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_GS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_SS:  // stack register must be able to load a writeable desc.
      if(LoadDescriptor(seg_descs[SDESC_SS],data,DMASK_WRITEABLE,true,false,
			false)) {
	seg_descs[SDESC_SS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_SS],data)) ASSERT(0);

	// suppress IF temporarily?
      } else
	return(0);
      break;

    case REG_CR0:
      // writes to control registers require cpl 0
      // other rule for CR0 is that PG (bit 31) cannot be set if PE (bit 0)
      //  is off
      if((GetCPL() > 0) ||
	 ((data & 0x080000001) == 0x080000000)) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }
      break;

    case REG_CR1:  // CR1 just plain doesn't exist
      ExceptionOccurred(EXC_INVALID_OPCODE);
      return(0);
      break;

    case REG_CR2:
    case REG_CR3:
      // writes to control registers require cpl 0
      if(GetCPL() > 0) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }
      break;

    case REG_CR4:
      // writes to control registers require cpl 0
      // other rule for CR4 is that we can't write 1's to reserved spots
      if((GetCPL() > 0) ||
	 ((data & 0x0FFFFFFA0) != 0)) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }
      break;

    case REG_DR0:
    case REG_DR1:
    case REG_DR2:
    case REG_DR3:
    case REG_DR6:
    case REG_DR7:
      // writes to debug registers require cpl 0
      if(GetCPL() > 0) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }
      break;

    case REG_DR4:
      // if the DE bit (bit 3) of CR4 is set, DR4 doesn't exist
      if(BITS(CPU_CR4,3,1) == 1) {
	ExceptionOccurred(EXC_INVALID_OPCODE);
	return(0);
      }

      // writes to debug registers require cpl 0
      if(GetCPL() > 0) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }

      // write through to DR6 also...
      CPU_DR6 = data;
      break;

    case REG_DR5:
      // if the DE bit (bit 3) of CR4 is set, DR5 doesn't exist
      if(BITS(CPU_CR4,3,1) == 1) {
	ExceptionOccurred(EXC_INVALID_OPCODE);
	return(0);
      }

      // writes to debug registers require cpl 0
      if(GetCPL() > 0) {
	ExceptionOccurred(EXC_PROTECTION,0);
	return(0);
      }

      // write through to DR7 also...
      CPU_DR7 = data;
      break;

    default:
      // we hope this is a general purpose register, and do nothing...
      break;
    };

  SET_OP(0,data);

  return(COUNT_CLOCKS(MOV));
}
END_DEFN

DEFN_INST(MOVS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 data;
  s_int32 shift_amt;

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(0);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    // load from src...
    data = GET_OP(0);
    if(ExceptionPending()) break;

    // and write to dest...
    if(!SET_OP(1,data)) break;

    // shift (E)SI, (E)DI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_ESI += shift_amt;
      CPU_EDI += shift_amt;
    } else {
      CPU_SI += shift_amt;
      CPU_DI += shift_amt;
    }

    // count one rep
    reps++;
 
    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(MOVS));
}
END_DEFN

DEFN_INST(MOVSX)
{
  int32 data;

  // nice simple instruction, since we can't move into segregs or CRx/DRx
  data = GET_OP(1);   // auto sign-extend
  if(ExceptionPending()) return(0);

  SET_OP(0,data);

  return(COUNT_CLOCKS(MOVSX));
}
END_DEFN

DEFN_INST(MOVZX)
{
  int32 data;

  // nice simple instruction, since we can't move into segregs or CRx/DRx
  data = GET_OP_Z(1);   // zero extend value
  if(ExceptionPending()) return(0);

  SET_OP(0,data);

  return(COUNT_CLOCKS(MOVZX));
}
END_DEFN

DEFN_INST(MUL)
{
  int32 multiplicand, multiplier;
  int64 product;
  bool overflowed;

  IGNORE_SZAP;

  // Fetch multiplicand, multiplier
  multiplicand = GET_OP(1);
  if(ExceptionPending()) return(0);

  multiplier = GET_OP(2);
  if(ExceptionPending()) return(0);

  product = ((int64)multiplicand) * ((int64)multiplier);

  // based on operand size, write result to AX/EAX/EDX:EAX, check for overflow
  switch(OP_SIZE(1)) {
  case SIZE_8BIT:
    CPU_AX = product;
    overflowed = ((product >> 8) != 0);
    break;
  case SIZE_16BIT:
    CPU_AX = product & 0x0FFFF;
    CPU_DX = (product >> 16);
    overflowed = ((product >> 16) != 0);
    break;
  case SIZE_32BIT:
    CPU_EAX = product & 0x0FFFFFFFF;
    CPU_EDX = (product >> 32);
    overflowed = ((product >> 32) != 0);
    break;
  default:
    ASSERT(0);
    return(0);
    break;
  }

  SET_BITS(CPU_EFLAGS,FLAG_CF,1,(overflowed ? 1 : 0));
  SET_BITS(CPU_EFLAGS,FLAG_OF,1,(overflowed ? 1 : 0));

  return(COUNT_CLOCKS(MUL));
}
END_DEFN

DEFN_INST(NEG)
{
  int32 val, neg_val;

  val = GET_OP(0);
  if(ExceptionPending()) return(0);

  neg_val = ~val + 1;

  if(SET_OP(0,neg_val)) {
    SetFlagsAfterOp(0,~val,neg_val,OP_SIZE(0),true);

    // carry flag is special under NEG - ACTUALLY, NO, it's just a 0-val
    // SET_BITS(CPU_EFLAGS,FLAG_CF,1,((val == 0) ? 0 : 1));
  }

  return(COUNT_CLOCKS(NEG));
}
END_DEFN

DEFN_INST(NOP)
{
  // No Operation...

  return(COUNT_CLOCKS(NOP));
}
END_DEFN

DEFN_INST(NOT)
{
  int32 val;

  val = GET_OP(0);

  if(!ExceptionPending()) SET_OP(0,~val);

  return(COUNT_CLOCKS(NOT));
}
END_DEFN

DEFN_INST(OR)
{
  int32 a, b, x;

  IGNORE_A;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // Do operation and write value, and (if no exceptions occur) set flags
  x = a | b;
  if(SET_OP(0,x))
    SetArithFlags(0  /* OF = 0 */,
		  ((x >> 31) == 1),
		  (x == 0),
		  0  /* AF = 0 (value undefined in spec) */,
		  even_parity(x),
		  0  /* CF = 0 */);

  return(COUNT_CLOCKS(OR));
}
END_DEFN

DEFN_INST(OUT)
{
  int32 port;
  int32 value;

  // first, load the port number, and data to write
  port = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  value = GET_OP(1);
  if(ExceptionPending()) return(0);

  // check permissions
  if(!CheckIOPermissions(port,OP_SIZE(1))) {
    // if we didn't page fault, it's a gpf
    if(!ExceptionPending()) ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // now just do the write
  WriteToIOPort(port,&value,OP_SIZE(1));

  return(COUNT_CLOCKS(OUT));
}
END_DEFN

DEFN_INST(OUTS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 port, value;
  s_int32 shift_amt;

  // first, load the port number
  port = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  // check permissions
  if(!CheckIOPermissions(port,OP_SIZE(1))) {
    // if we didn't page fault, it's a gpf
    if(!ExceptionPending()) ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(1);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    value = GET_OP(1);
    if(ExceptionPending()) break;

    WriteToIOPort(port,&value,op_size);

    // shift (E)SI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_ESI += shift_amt;
    } else {
      CPU_SI += shift_amt;
    }

    // count one rep
    reps++;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(OUTS));
}
END_DEFN

DEFN_INST(POP)
{
  int32 data;
  int op_size;

  // While it might make sense to use the size of operand 0 as our data size,
  //  there is a problem in that when an 8-bit quantity is being pushed, we
  //  need to use the default operand size for the instruction (and 16- and
  //  32-bit operands always match the default operand size).
  // In the case of smaller values, we'll read too much, but only write as
  //  much as we need to (taken care of automatically (almost) by the Operand
  //  code (segment loading needs special whacking).
  op_size = inst.GetOperandSize();

  data = 0;   // this clears the top bytes in case of a 2-byte read

  // first, try to pop the new flags (double)word, but we won't increment
  //  (E)SP 'til the end...
  if(seg_descs[SDESC_SS].use_32bit()) {
    if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,&data,
			       op_size,GetCPL())) return(0);
  } else {
    if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_SP,&data,
			       op_size,GetCPL())) return(0);
  }

  // now fix the byte order
  data = xtoh32(data);

  // something of a hack: if we're in 32-bit mode, but read only a 16-bit
  //   value, zero out the top 16 bits now, in case we need to use 'data' to
  //   load a segment selector.
  if((op_size == SIZE_32BIT) && (OP_SIZE(0) == SIZE_16BIT))
    data &= 0x0FFFF;

  // check to make sure we're not popping into segment registers
  if(OP(0)->Type() == OPTYPE_REGISTER)
    switch(((RegisterOperand *) OP(0))->RegNum()) {
    case REG_DS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_DS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_DS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_DS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_ES:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_ES],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_ES].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_ES],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_FS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_FS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_FS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_FS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_GS:  // data registers must be able to load the desc (or null)
      if(LoadDescriptor(seg_descs[SDESC_GS],data,DMASK_READABLE,true,true,
			false)) {
	seg_descs[SDESC_GS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_GS],data)) ASSERT(0);
      } else
	return(0);
      break;

    case REG_SS:  // stack register must be able to load a writeable desc.
      if(LoadDescriptor(seg_descs[SDESC_SS],data,DMASK_WRITEABLE,true,false,
			false)) {
	seg_descs[SDESC_SS].set_accessed(true);
	if(data && !WriteBackDescriptor(seg_descs[SDESC_SS],data)) ASSERT(0);

	// suppress IF temporarily?
      } else
	return(0);
      break;

    default:
      break;
    }

  // finally, write popped value (pre-decrementing (E)SP in case write is
  //  to (E)SP or something based on (E)SP...)
  if(seg_descs[SDESC_SS].use_32bit())
    CPU_ESP += op_size; else
    CPU_SP += op_size;

  if(!SET_OP(0,data)) {
    // put back (E)SP...
    if(seg_descs[SDESC_SS].use_32bit())
      CPU_ESP -= op_size; else
      CPU_SP -= op_size;
  }

  return(COUNT_CLOCKS(POP));
}
END_DEFN

DEFN_INST(POPA)
{
  int8 reg_buffer[32];
  int op_size;

  op_size = inst.GetOperandSize();

  // try to rip out of the stack enough to hold 8 * op_size
  if(seg_descs[SDESC_SS].use_32bit()) {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,reg_buffer,
			      (op_size << 3),GetCPL()))
      CPU_ESP += (op_size << 3); else
      return(0);
  } else {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_SP,reg_buffer,
			      (op_size << 3),GetCPL()))
      CPU_SP += (op_size << 3); else
      return(0);
  }

  if(op_size == SIZE_32BIT) {
    CPU_EAX = xtoh32(((int32 *)reg_buffer)[7]);
    CPU_ECX = xtoh32(((int32 *)reg_buffer)[6]);
    CPU_EDX = xtoh32(((int32 *)reg_buffer)[5]);
    CPU_EBX = xtoh32(((int32 *)reg_buffer)[4]);
    /* don't mess with ESP any more */
    CPU_EBP = xtoh32(((int32 *)reg_buffer)[2]);
    CPU_ESI = xtoh32(((int32 *)reg_buffer)[1]);
    CPU_EDI = xtoh32(((int32 *)reg_buffer)[0]);
  } else {
    CPU_AX = xtoh16(((int16 *)reg_buffer)[7]);
    CPU_CX = xtoh16(((int16 *)reg_buffer)[6]);
    CPU_DX = xtoh16(((int16 *)reg_buffer)[5]);
    CPU_BX = xtoh16(((int16 *)reg_buffer)[4]);
    /* don't mess with SP any more */
    CPU_BP = xtoh16(((int16 *)reg_buffer)[2]);
    CPU_SI = xtoh16(((int16 *)reg_buffer)[1]);
    CPU_DI = xtoh16(((int16 *)reg_buffer)[0]);
  }

  return(COUNT_CLOCKS(POPA));
}
END_DEFN

DEFN_INST(POPF)
{
  int32 new_flags, old_flags;
  int op_size;

  // vm86 mode, when iopl = 3 causes a GPF on a POPF
  if(in_v86_mode() && (GetIOPL() == 3)) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  op_size = inst.GetOperandSize();

  // first, try to pop the new flags (double)word, but we won't increment
  //  (E)SP 'til the end...
  if(seg_descs[SDESC_SS].use_32bit()) {
    if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,&new_flags,
			       op_size,GetCPL())) return(0);
  } else {
    if(!ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_SP,&new_flags,
			       op_size,GetCPL())) return(0);
  }

  // now fix the byte order, and knock out flags that don't exist
  new_flags = (xtoh32(new_flags) & FLAG_DEFINED) | FLAG_UNDEFVAL;
  old_flags = CPU_EFLAGS;

  // now, we never alter VM, RF, VIF, or VIP
  SET_BITS(new_flags,FLAG_VM, 1,BITS(old_flags,FLAG_VM,1));
  SET_BITS(new_flags,FLAG_RF, 1,BITS(old_flags,FLAG_RF,1));
  SET_BITS(new_flags,FLAG_VIF,1,BITS(old_flags,FLAG_VIF,1));
  SET_BITS(new_flags,FLAG_VIP,1,BITS(old_flags,FLAG_VIP,1));

  // we don't alter IOPL if CPL > 0
  if(GetCPL() > 0)
    SET_BITS(new_flags,FLAG_IOPL,2,BITS(old_flags,FLAG_IOPL,2));

  // we don't alter IF if CPL > IOPL
  if(GetCPL() > GetIOPL())
    SET_BITS(new_flags,FLAG_IF,1,BITS(old_flags,FLAG_IF,1));

  if(op_size == SIZE_32BIT)
    CPU_EFLAGS = new_flags; else
    CPU_FLAGS = new_flags;

  if(seg_descs[SDESC_SS].use_32bit())
    CPU_ESP += op_size; else
    CPU_SP += op_size;

  return(COUNT_CLOCKS(POPF));
}
END_DEFN

DEFN_INST(PUSH)
{
  int32 value;
  int32 op_size;

  // first, get the value we'll be pushing...
  // we use a 32-bit byte-swap for byte order, even though we may have a 16-
  //  bit value, because that would swap the bottom two bytes into the lower
  //  address-ed two bytes, which is what we need...
  value = htox32(GET_OP(0));
  if(ExceptionPending()) return(0);

  // While it might make sense to use the size of operand 0 as our data size,
  //  there is a problem in that when an 8-bit quantity is being pushed, we
  //  need to use the default operand size for the instruction (and 16- and
  //  32-bit operands always match the default operand size).
  op_size = inst.GetOperandSize();

  // to push, pre-decrement the stack, then try to write, and it if fails,
  //  put back the old (E)SP...
  if(seg_descs[SDESC_SS].use_32bit()) {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_ESP > 0) && (CPU_ESP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_ESP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,&value,op_size,
			      GetCPL()))
      CPU_ESP += op_size;
  } else {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_SP > 0) && (CPU_SP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_SP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_SP,&value,op_size,
			      GetCPL()))
      CPU_SP += op_size;
  }

  return(COUNT_CLOCKS(PUSH));
}
END_DEFN

DEFN_INST(PUSHA)
{
  int8 reg_buffer[32];
  int32 op_size;

  // first, build the buffer...  note that registers are in backwards order
  if((op_size = inst.GetOperandSize()) == SIZE_32BIT) {
    ((int32 *)reg_buffer)[7] = htox32(CPU_EAX);
    ((int32 *)reg_buffer)[6] = htox32(CPU_ECX);
    ((int32 *)reg_buffer)[5] = htox32(CPU_EDX);
    ((int32 *)reg_buffer)[4] = htox32(CPU_EBX);
    ((int32 *)reg_buffer)[3] = htox32(CPU_ESP);
    ((int32 *)reg_buffer)[2] = htox32(CPU_EBP);
    ((int32 *)reg_buffer)[1] = htox32(CPU_ESI);
    ((int32 *)reg_buffer)[0] = htox32(CPU_EDI);
  } else {
    ((int16 *)reg_buffer)[7] = htox16(CPU_AX);
    ((int16 *)reg_buffer)[6] = htox16(CPU_CX);
    ((int16 *)reg_buffer)[5] = htox16(CPU_DX);
    ((int16 *)reg_buffer)[4] = htox16(CPU_BX);
    ((int16 *)reg_buffer)[3] = htox16(CPU_SP);
    ((int16 *)reg_buffer)[2] = htox16(CPU_BP);
    ((int16 *)reg_buffer)[1] = htox16(CPU_SI);
    ((int16 *)reg_buffer)[0] = htox16(CPU_DI);
  }

  // adjust op_size to represent the size of the entire register buffer
  op_size <<= 3;

  // now try to smoosh them into the stack...  real and vm86 modes allow the
  //  stupid possibility of wrapping around the stack limit which, as long as
  //  it doesn't cut a single operand in half, is ok (WHY?  These people are
  //  weird!)
  // to push, pre-decrement the stack, then try to write, and it if fails,
  //  put back the old (E)SP...
  if(seg_descs[SDESC_SS].use_32bit()) {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_ESP > 0) && (CPU_ESP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_ESP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,reg_buffer,op_size,
			      GetCPL()))
      CPU_ESP += op_size;
  } else {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_SP > 0) && (CPU_SP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_SP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_SP,reg_buffer,op_size,
			      GetCPL()))
      CPU_SP += op_size;
  }

  ASSERT(0);

  return(COUNT_CLOCKS(PUSHA));
}
END_DEFN

DEFN_INST(PUSHF)
{
  int32 value;
  int32 op_size;

  // vm86 mode, when iopl = 3 causes a GPF on a PUSHF
  if(in_v86_mode() && (GetIOPL() == 3)) {
    ExceptionOccurred(EXC_PROTECTION,0);
    return(0);
  }

  // first, get the value we'll be pushing...
  if((op_size = inst.GetOperandSize()) == SIZE_32BIT)
    value = CPU_EFLAGS & 0x0FCFFFF; else
    value = CPU_FLAGS;

  // we use a 32-bit byte-swap for byte order, even though we may have a 16-
  //  bit value, because that would swap the bottom two bytes into the lower
  //  address-ed two bytes, which is what we need...
  value = htox32(value);

  // to push, pre-decrement the stack, then try to write, and it if fails,
  //  put back the old (E)SP...
  if(seg_descs[SDESC_SS].use_32bit()) {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_ESP > 0) && (CPU_ESP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_ESP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,&value,op_size,
			      GetCPL()))
      CPU_ESP += op_size;
  } else {
    // real an vm86 modes can cause shutdowns...
    if((!in_protected_mode() || in_v86_mode()) &&
       (CPU_SP > 0) && (CPU_SP < op_size)) {
      DEBUG(DEBUG_MEMORY,"Shutdown caused by stack underflow on PUSH!");
      halted = true;
      return(0);
    }

    CPU_SP -= op_size;
    if(!WriteToVirtualAddress(seg_descs[SDESC_SS],CPU_SP,&value,op_size,
			      GetCPL()))
      CPU_SP += op_size;
  }

  return(COUNT_CLOCKS(PUSHF));
}
END_DEFN

DEFN_INST(RCL)
{
  int32 val, count;
  int width, new_cf;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // Technically, we are supposed to mod the count with 32 unless we're in
  //  vm86 mode, but since we're always (for performance) going to do at
  //  least that much with the following line, we leave it out.
  // On second thought, we have the weird quirk that a ROL AX,16 is identical
  //  to a ROL AX,0, except the ,0 doesn't set the stupid CF.
  if(!in_v86_mode()) count &= 31;
  
  // since our rotate now includes the carry flag, we don't have to special
  //  case a full rotation (it is now just a NOP)
  count %= (width+1);   // size is 1 bit wider to account for CF

  // If the count is in fact 0, skip everything
  if(count != 0) {
    // since we're going to lose the old CF, figure the new one now
    new_cf = (val >> (width-count)) & 1;
    
    // do the rotate left
    val = ((val << count) | 
	   (BITS(CPU_EFLAGS,FLAG_CF,1) << (count-1)) |
	   (val >> (width+1-count)));

    // if we can write back the value, CF is the (new) bottom bit of 'val'
    if(SET_OP(0,val))
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,new_cf);
  }

  if(count != 1) IGNORE_O;

  return(COUNT_CLOCKS(RCL));
}
END_DEFN

DEFN_INST(RCR)
{
  int32 val, count;
  int width, new_cf;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // Technically, we are supposed to mod the count with 32 unless we're in
  //  vm86 mode, but since we're always (for performance) going to do at
  //  least that much with the following line, we leave it out.
  // On second thought, we have the weird quirk that a ROL AX,16 is identical
  //  to a ROL AX,0, except the ,0 doesn't set the stupid CF.
  if(!in_v86_mode()) count &= 31;
  
  // since our rotate now includes the carry flag, we don't have to special
  //  case a full rotation (it is now just a NOP)
  count %= (width+1);   // size is 1 bit wider to account for CF

  // If the count is in fact 0, skip everything
  if(count != 0) {
    // since we're going to lose the old CF, figure the new one now
    new_cf = (val >> (count-1)) & 1;
    
    // do the rotate right
    val = ((val >> count) | 
	   (BITS(CPU_EFLAGS,FLAG_CF,1) << (width-count)) |
	   (val << (width+1-count)));

    // if we can write back the value, CF is the (new) bottom bit of 'val'
    if(SET_OP(0,val))
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,new_cf);
  }

  if(count != 1) IGNORE_O;

  return(COUNT_CLOCKS(RCR));
}
END_DEFN

DEFN_INST(RDMSR)
{
  ASSERT(0);

  return(COUNT_CLOCKS(RDMSR));
}
END_DEFN

DEFN_INST(RDTSC)
{
  if(!in_protected_mode() ||
     (!in_v86_mode() && ((GetCPL() == 0) || (BITS(CPU_CR4,2,1) == 0)))) {
    CPU_EDX = CPU_TSC_HI;
    CPU_EAX = CPU_TSC_LO;
  } else
    ExceptionOccurred(EXC_PROTECTION,0);

  return(COUNT_CLOCKS(RDTSC));
}
END_DEFN

DEFN_INST(RETfar)
{
  ASSERT(0);

  return(COUNT_CLOCKS(RETfar));
}
END_DEFN

DEFN_INST(RETnear)
{
  int32 new_eip, old_esp;
  int32 extra;

  new_eip = 0;  // so that if we only load 16 bits, we're ok

  extra = OP(0) ? GET_OP_Z(0) : 0;

  old_esp = CPU_ESP;  // save the entire ESP, so that even if we just
                      // modify SP, we can fix it

  // try to pop new (E)IP
  if(seg_descs[SDESC_SS].use_32bit()) {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_ESP,&new_eip,
			       inst.GetOperandSize(),GetCPL()))
      CPU_ESP += inst.GetOperandSize() + extra; else
      return(0);
  } else {
    if(ReadFromVirtualAddress(seg_descs[SDESC_SS],CPU_SP,&new_eip,
			       inst.GetOperandSize(),GetCPL()))
      CPU_SP += inst.GetOperandSize() + extra; else
      return(0);
  }

  new_eip = xtoh32(new_eip);  // get the correct byte order

  // check the new eip against the code segment's limit
  if(seg_descs[SDESC_CS].offset_valid(new_eip)) {
    CPU_EIP = new_eip;   // just slam in the new EIP...
  } else {
    CPU_ESP = old_esp;   // restore the old ESP
    ExceptionOccurred(EXC_PROTECTION,0);
  }

  return(COUNT_CLOCKS(RETnear));
}
END_DEFN

DEFN_INST(ROL)
{
  int32 val, count;
  int width;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // Technically, we are supposed to mod the count with 32 unless we're in
  //  vm86 mode, but since we're always (for performance) going to do at
  //  least that much with the following line, we leave it out.
  // On second thought, we have the weird quirk that a ROL AX,16 is identical
  //  to a ROL AX,0, except the ,0 doesn't set the stupid CF.
  if(!in_v86_mode()) count &= 31;
  
  // If the count is in fact 0, skip everything
  if(count != 0) {
    count %= width;
    
    // do the rotate left
    val = ((val << count) | (val >> (width-count)));

    // if we can write back the value, CF is the (new) bottom bit of 'val'
    if(SET_OP(0,val))
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,(val & 1));
  }

  if(count != 1) IGNORE_O;

  return(COUNT_CLOCKS(ROL));
}
END_DEFN

DEFN_INST(ROR)
{
  int32 val, count;
  int width;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // Technically, we are supposed to mod the count with 32 unless we're in
  //  vm86 mode, but since we're always (for performance) going to do at
  //  least that much with the following line, we leave it out.
  // On second thought, we have the weird quirk that a ROL AX,16 is identical
  //  to a ROL AX,0, except the ,0 doesn't set the stupid CF.
  if(!in_v86_mode()) count &= 31;
  
  // If the count is in fact 0, skip everything
  if(count != 0) {
    count %= width;

    // do the rotate right
    val = ((val >> count) | (val << (width-count)));

    // if we can write back the value, CF is the (new) top bit of 'val'
    if(SET_OP(0,val))
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((val >> (width-1)) & 1));
  }

  if(count != 1) IGNORE_O;

  return(COUNT_CLOCKS(ROR));
}
END_DEFN

DEFN_INST(RSM)
{
  ASSERT(0);

  return(COUNT_CLOCKS(RSM));
}
END_DEFN

DEFN_INST(SAHF)
{
  CPU_EFLAGS = (CPU_EFLAGS & 0xFFFFFF2A) | (CPU_AH & 0x00D5);

  return(COUNT_CLOCKS(SAHF));
}
END_DEFN

DEFN_INST(SAR)
{
  int32 val, new_val, count;
  int8 width;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  if(count == 1)
    IGNORE_A; else
    IGNORE_OA;
  
  width = OP_SIZE(0) << 3;

  // We are supposed to mod the count with 32 unless we're in vm86 mode...
  if(!in_v86_mode()) count &= 31;
  
  // We have four different cases: shift by 0, shift by 1, shift larger than
  //  the field width, and everything else
  if(count > width) {
    // if the original number was positive, the answer is 0, else -1
    new_val = (((val >> (width-1)) != 0) ? 0x0FFFFFFFF : 0);
    
    if(SET_OP(0,new_val)) {
      SET_SZP(0,OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,(new_val & 1));
    }
  } else
  if(count == 1) {
    new_val = (val >> 1) | (val & (1 << (width-1)));

    if(SET_OP(0,new_val)) {
      SET_SZP(new_val,OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,(val & 1));
      CLEAR_BITS(CPU_EFLAGS,FLAG_OF,1);
    }
  } else
  if(count != 0) {
    // if the original number was positive, start with 0, else -1
    new_val = (((val >> (width-1)) != 0) ? 0x0FFFFFFFF : 0);

    // now basically shift 0's (or 1's) up, and val down
    new_val = (new_val << (width-count)) | (val >> count);

    if(SET_OP(0,new_val)) {
      SET_SZP(new_val,OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((val >> (count-1)) & 1));
    }
  }

  if(count == 1)
    IGNORE_A; else
    IGNORE_OA;
  
  return(COUNT_CLOCKS(SAR));
}
END_DEFN

DEFN_INST(SBB)
{
  int32 a, b, x;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // I can't decide if a - (b+1) is equivalent to (a-b), with a 1 subtracted
  //  in a manner that prevents overflow (see ADC), but page 25-281 
  //  specifically calls for the former...
  // By inverting the second operand, we turn +b into -(b+1)
  b = ~b;  // delay +1 til addition because it can cause carrying

  // Do sum and write value, and (if no exceptions occur) set flags
  x = a + b + (BITS(CPU_FLAGS,FLAG_CF,1) ? 0 : 1);
  if(SET_OP(0,x)) SetFlagsAfterOp(a,b,x,OP_SIZE(0),true);  // subtract has backwards CF and AF

  return(COUNT_CLOCKS(SBB));
}
END_DEFN

DEFN_INST(SCAS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  int32 data, compare_to, diff;
  s_int32 shift_amt;

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  // SCAS has the Y{b,l} operand in operand #1
  reps = 0;
  op_size = OP_SIZE(1);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  // first, load the value we'll be comparing against
  switch(op_size) {
  case SIZE_8BIT:
    compare_to = CPU_AL;
    break;
  case SIZE_16BIT:
    compare_to = CPU_AX;
    break;
  case SIZE_32BIT:
    compare_to = CPU_EAX;
    break;
  default:
    ASSERT(0);
    break;
  }

  while(reps < max_reps) {
    // SCAS is weird...  we actually load from the destination register...
    //  also, we need to load zero-extended because the 'compare_to' value
    //  was zero-extended, and we're just going to compare them
    data = GET_OP_Z(1);
    if(ExceptionPending()) break;

    // Do the subtraction and set the flags
    diff = compare_to + (~data + 1);  /* +1 isn't really part of op b
					 (just like subtraction, the +1 can
					 cause a carry...) */
    SetFlagsAfterOp(compare_to,~data,diff,op_size,true);
                                        // subtract has backwards CF and AF

    // shift (E)DI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_EDI += shift_amt;
    } else {
      CPU_DI += shift_amt;
    }

    // count one rep
    reps++;

    // check for REPZ, REPNZ termination
    if(((rep_type == RepRepZ) && (diff != 0)) ||
       ((rep_type == RepRepNZ) && (diff == 0))) break;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      IGNORE_OSZAPC;       // on pentium pro, rep'd instructions don't set flgs
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(SCAS));
}
END_DEFN

DEFN_INST(SETcc)
{
  // Operand 0 is the condition code, operand 1 is the 8-bit target

  // If the condition is met, set the target to 1 - if not, set to 0
  if(check_condition(CPU_EFLAGS,GET_OP(0)))
    SET_OP(1,1); else
    SET_OP(1,0);

  return(COUNT_CLOCKS(SETcc));
}
END_DEFN

DEFN_INST(SGDT)
{
  int8 gdt_buffer[6];

  *(int16 *)(gdt_buffer+0) = htox16(seg_descs[SDESC_GDT].limit());
  *(int32 *)(gdt_buffer+2) = htox32(seg_descs[SDESC_GDT].base());
  if(inst.GetOperandSize() == SIZE_16BIT) gdt_buffer[5] = 0;

  OP(0)->WriteBytes(*this,gdt_buffer,6);

  return(COUNT_CLOCKS(SGDT));
}
END_DEFN

DEFN_INST(SHL)
{
  int32 val, count;
  int8 width;

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // We are supposed to mod the count with 32 unless we're in vm86 mode...
  if(!in_v86_mode()) count &= 31;
  
  // We have four different cases: shift by 0, shift by 1, shift larger than
  //  the field width, and everything else
  if(count > width) {
    // just write a 0 back out, and clear CF
    if(SET_OP(0,0)) {
      SET_SZP(0,OP_SIZE(0));
      CLEAR_BITS(CPU_EFLAGS,FLAG_CF,1);
    }
  } else
  if(count == 1) {
    if(SET_OP(0,val << 1)) {
      SET_SZP((val << 1),OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,(val >> (width-1)));
      if(BITS(val,width-1,1) != BITS(val,width-2,1))
	SET_BITS(CPU_EFLAGS,FLAG_OF,1,1); else
	CLEAR_BITS(CPU_EFLAGS,FLAG_OF,1);
    }
  } else
  if(count != 0) {
    if(SET_OP(0,val << count)) {
      SET_SZP((val << count),OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((val >> (width-count)) & 1));
    }
  }

  if(count == 1)
    IGNORE_A; else
    IGNORE_OA;
  
  return(COUNT_CLOCKS(SHL));
}
END_DEFN

DEFN_INST(SHLD)
{
  int32 new_dest, dest, src, count;
  int8 width;

  IGNORE_OA;

  // Load operands, watching for exceptions, making sure we've got 0's in
  //  upper bits
  dest = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  src = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  // count must be mod 32 (and 31 is faster)
  count = GET_OP_Z(2) & 31;
  if(ExceptionPending()) return(0);

  // get the width of our operands, in bits
  width = OP_SIZE(0) << 3;

  // technically, the results of a SHLD count >= width is undefined, so we
  //  just won't do anything in that case
  if((count > 0) && (count < width)) {
    new_dest = (dest << count) | (src >> (width-count));

    if(SET_OP(0,new_dest)) {
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((dest >> (width-count)) & 1));
      if(count == 1) {
	if(BITS(dest,width-1,1) != BITS(dest,width-2,1))
	  SET_BITS(CPU_EFLAGS,FLAG_OF,1,1); else
	  CLEAR_BITS(CPU_EFLAGS,FLAG_OF,1);
      }
      SET_SZP(new_dest,OP_SIZE(0));
      /*
      Set_SZP_Flags(((new_dest >> (width-1)) == 1),
		    ((new_dest & ~(0xFFFFFFFF << width)) == 0),
		    even_parity(new_dest)); */
    }	  
  }

  return(COUNT_CLOCKS(SHLD));
}
END_DEFN

DEFN_INST(SHR)
{
  int32 val, count;
  int8 width;

  SET_BITS(CPU_EFLAGS,FLAG_AF,1,1); // my CPU likes this - SJT

  // First, fetch the value and rotate count, so we can quit on an exception
  //  Have to use zero-extend to keep 1's out of bits shifted in
  val = GET_OP_Z(0);
  if(ExceptionPending()) return(0);
  
  // count is always positive, so use zero-extend
  count = GET_OP_Z(1);
  if(ExceptionPending()) return(0);
  
  width = OP_SIZE(0) << 3;

  // We are supposed to mod the count with 32 unless we're in vm86 mode...
  if(!in_v86_mode()) count &= 31;
  
  // We have four different cases: shift by 0, shift by 1, shift larger than
  //  the field width, and everything else
  if(count > width) {
    // just write a 0 back out, and clear CF
    if(SET_OP(0,0)) {
      SET_SZP(0,OP_SIZE(0));
      CLEAR_BITS(CPU_EFLAGS,FLAG_CF,1);
    }
  } else
  if(count == 1) {
    if(SET_OP(0,val >> 1)) {
      SET_SZP((val >> 1),OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,(val & 1));
      SET_BITS(CPU_EFLAGS,FLAG_OF,1,(val >> (width-1)));
    }
  } else
  if(count != 0) {
    if(SET_OP(0,val >> count)) {
      SET_SZP((val >> count),OP_SIZE(0));
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((val >> (count-1)) & 1));
    }
  }

  if(count == 1)
    IGNORE_A; else
    IGNORE_OA;
  
  return(COUNT_CLOCKS(SHR));
}
END_DEFN

DEFN_INST(SHRD)
{
  int32 new_dest, dest, src, count;
  int8 width;

  IGNORE_OA;

  // Load operands, watching for exceptions, making sure we've got 0's in
  //  upper bits
  dest = GET_OP_Z(0);
  if(ExceptionPending()) return(0);

  src = GET_OP_Z(1);
  if(ExceptionPending()) return(0);

  // count must be mod 32 (and 31 is faster)
  count = GET_OP_Z(2) & 31;
  if(ExceptionPending()) return(0);

  // get the width of our operands, in bits
  width = OP_SIZE(0) << 3;

  // technically, the results of a SHRD count >= width is undefined, so we
  //  just won't do anything in that case
  if((count > 0) && (count < width)) {
    new_dest = (dest >> count) | (src << (width-count));

    if(SET_OP(0,new_dest)) {
      SET_BITS(CPU_EFLAGS,FLAG_CF,1,((dest >> (count-1)) & 1));
      if(count == 1) {
	if(BITS(dest,width-1,1) != BITS(src,0,1))
	  SET_BITS(CPU_EFLAGS,FLAG_OF,1,1); else
	  CLEAR_BITS(CPU_EFLAGS,FLAG_OF,1);
      }
      SET_SZP(new_dest,OP_SIZE(0));
      /*
      Set_SZP_Flags(((new_dest >> (width-1)) == 1),
		    ((new_dest & ~(0xFFFFFFFF << width)) == 0),
		    even_parity(new_dest));*/
    }	  
  }

  return(COUNT_CLOCKS(SHRD));
}
END_DEFN

DEFN_INST(SIDT)
{
  int8 idt_buffer[6];

  *(int16 *)(idt_buffer+0) = htox16(seg_descs[SDESC_IDT].limit());
  *(int32 *)(idt_buffer+2) = htox32(seg_descs[SDESC_IDT].base());
  if(inst.GetOperandSize() == SIZE_16BIT) idt_buffer[5] = 0;

  OP(0)->WriteBytes(*this,idt_buffer,6);

  return(COUNT_CLOCKS(SIDT));
}
END_DEFN

DEFN_INST(SLDT)
{
  // just dump the LDTR into the operand
  SET_OP(0,CPU_LDT);

  return(COUNT_CLOCKS(SLDT));
}
END_DEFN

DEFN_INST(SMSW)
{
  ASSERT(0);

  return(COUNT_CLOCKS(SMSW));
}
END_DEFN

DEFN_INST(STC)
{
  SET_BITS(CPU_EFLAGS,FLAG_CF,1,1);

  return(COUNT_CLOCKS(STC));
}
END_DEFN

DEFN_INST(STD)
{
  SET_BITS(CPU_EFLAGS,FLAG_DF,1,1);

  return(COUNT_CLOCKS(STD));
}
END_DEFN

DEFN_INST(STI)
{
  // A GPF results if we are in protected mode and (CPL > IOPL)
  if(in_protected_mode() && (GetCPL() > GetIOPL())) 
    ExceptionOccurred(EXC_PROTECTION,0); else
    SET_BITS(CPU_EFLAGS,FLAG_IF,1,1);

  return(COUNT_CLOCKS(STI));
}
END_DEFN

DEFN_INST(STOS)
{
  int rep_type, op_size;
  int32 max_reps, reps;
  s_int32 shift_amt;

  if((rep_type = inst.GetRepeat()) != RepNone)
    max_reps = ((inst.GetAddressSize() == SIZE_32BIT) ? CPU_ECX : CPU_CX); else
    max_reps = 1;

  reps = 0;
  op_size = OP_SIZE(0);
  shift_amt = ((BITS(CPU_EFLAGS,FLAG_DF,1) == 0) ? op_size : (-op_size));

  while(reps < max_reps) {
    // for STOS we can cheat - we should be writing AL, AX, or EAX, depending
    //  on the operand size, but we can just always write EAX, and let the
    //  writing code chop off the top bits
    if(!SET_OP(0,CPU_EAX)) break;  // stop on an exception

    // shift (E)DI
    if(inst.GetAddressSize() == SIZE_32BIT) {
      CPU_EDI += shift_amt;
    } else {
      CPU_DI += shift_amt;
    }

    // count one rep
    reps++;

    // only do one loop on single-stepping
    if(single_stepping && (reps < max_reps)) {
      CPU_EIP = prev_eip;  // so we come back and do the same instr next time
      break;
    }
  }

  // adjust (E)CX for the number of repetitions completed successfully
  if((rep_type = inst.GetRepeat()) != RepNone)
    if(inst.GetAddressSize() == SIZE_32BIT)
      CPU_ECX = (max_reps - reps); else
      CPU_CX = (max_reps - reps);

  return(COUNT_CLOCKS(STOS));
}
END_DEFN

DEFN_INST(STR)
{
  // just dump the TR into the operand
  SET_OP(0,CPU_TR);

  return(COUNT_CLOCKS(STR));
}
END_DEFN

DEFN_INST(SUB)
{
  int32 a, b, x;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // By negating the second operand, we turn a subtraction into addition
  b = ~b;  // delay +1 til addition because it can cause carrying

  // Do sum and write value, and (if no exceptions occur) set flags
  x = a + b + 1;
  if(SET_OP(0,x)) 
    SetFlagsAfterOp(a,b,x,OP_SIZE(0),true); // subtract has backwards CF and AF

  return(COUNT_CLOCKS(SUB));
}
END_DEFN

DEFN_INST(TEST)
{
  int32 and_result;

  IGNORE_A;

  and_result = GET_OP(0) & GET_OP(1);

  // assuming no exceptions occurred on the fetch, set flags accordingly
  if(!ExceptionPending())
    SetArithFlags(0  /* OF = 0 */,
		  ((and_result >> 31) == 1),
		  (and_result == 0),
		  0  /* AF = 0 (value undefined in spec) */,
		  even_parity(and_result),
		  0  /* CF = 0 */);

  return(COUNT_CLOCKS(TEST));
}
END_DEFN

DEFN_INST(VERR)
{
  int16 selector;
  SegmentDescriptor descriptor;
  
  // real and vm86 modes do not do LAR
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // first, get the specified selector
  selector = GET_OP(0);
  if(ExceptionPending()) return(0);

  // now, try to get the descriptor for it, suppressing errors
  //  we just ask for only readable segments, and if we get one, we're done
  if(LoadDescriptor(descriptor,selector,DMASK_READABLE,false,false,true))
    SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1); else
    CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);

  return(COUNT_CLOCKS(VERR));
}
END_DEFN

DEFN_INST(VERW)
{
  int16 selector;
  SegmentDescriptor descriptor;
  
  // real and vm86 modes do not do LAR
  if(!in_protected_mode() || in_v86_mode()) {
    ExceptionOccurred(EXC_INVALID_OPCODE);
    return(0);
  }

  // first, get the specified selector
  selector = GET_OP(0);
  if(ExceptionPending()) return(0);

  // now, try to get the descriptor for it, suppressing errors
  //  we just ask for only writeable segments, and if we get one, we're done
  if(LoadDescriptor(descriptor,selector,DMASK_WRITEABLE,false,false,true))
    SET_BITS(CPU_EFLAGS,FLAG_ZF,1,1); else
    CLEAR_BITS(CPU_EFLAGS,FLAG_ZF,1);

  return(COUNT_CLOCKS(VERW));
}
END_DEFN

DEFN_INST(WAIT)
{
  // WAIT results in a #NM (int 7) if both MP and TS are set in CR0
  if((CPU_CR0 & 0x0000000A) == 0x0000000A)
    ExceptionOccurred(EXC_NO_DEVICE);

  // Now just check for the exception
  CheckForFPUException();

  return(COUNT_CLOCKS(WAIT));
}
END_DEFN

DEFN_INST(WBINVD)
{
  if(in_protected_mode() && (GetCPL() > 0)) {
    ExceptionOccurred(EXC_PROTECTION,0);
  } else {
    // Write Back, Invalidate Internal Cache, External Cache?
  }

  return(COUNT_CLOCKS(WBINVD));
}
END_DEFN

DEFN_INST(WRMSR)
{
  ASSERT(0);

  return(COUNT_CLOCKS(WRMSR));
}
END_DEFN

DEFN_INST(XADD)
{
  int32 val1, val2, sum;

  // Fetch both values, and check for exceptions (we need two temporaries,
  //  because we have to be able to undo the change to the first if the
  //  write to the second causes an exception)
  val1 = GET_OP(0);
  val2 = GET_OP(1);
  if(ExceptionPending()) return(0);

  sum = val1 + val2;

  if(SET_OP(0,sum)) {
    if(SET_OP(1,val1)) {
      SetFlagsAfterOp(val1,val2,sum,OP_SIZE(0),false);
      return(COUNT_CLOCKS(XADD));
    } else
      SET_OP(0,val1);   // undo the first change
  }

  // we only get here if one of the SET_OPs failed...
  return(0);
}
END_DEFN

DEFN_INST(XCHG)
{
  int32 val1, val2;

  // Fetch both values, and check for exceptions (we need two temporaries,
  //  because we have to be able to undo the change to the first if the
  //  write to the second causes an exception)
  val1 = GET_OP(0);
  val2 = GET_OP(1);
  if(ExceptionPending()) return(0);

  if(SET_OP(0,val2)) {
    if(SET_OP(1,val1)) 
      return(COUNT_CLOCKS(XCHG)); else
      SET_OP(0,val1);   // undo the first change
  }

  // we only get here if one of the SET_OPs failed...
  return(0);
}
END_DEFN

DEFN_INST(XLAT)
{
  int8 new_al;

  // do the table lookup
  new_al = GET_OP_DISP(0,CPU_AL);

  // if it succeeded, write it
  if(!ExceptionPending()) CPU_AL = new_al;

  return(COUNT_CLOCKS(XLAT));
}
END_DEFN

DEFN_INST(XOR)
{
  int32 a, b, x;

  IGNORE_A;

  // Get first operand, watching for exceptions
  a = GET_OP(0);
  if(ExceptionPending()) return(0);

  // Get second operand, watching for exceptions
  b = GET_OP(1);
  if(ExceptionPending()) return(0);

  // Do operation and write value, and (if no exceptions occur) set flags
  x = a ^ b;
  if(SET_OP(0,x))
    SetArithFlags(0  /* OF = 0 */,
		  ((x >> 31) == 1),
		  (x == 0),
		  0  /* AF = 0 (value undefined in spec) */,
		  even_parity(x),
		  0  /* CF = 0 */);

  return(COUNT_CLOCKS(XOR));
}
END_DEFN
