/* -*- C++ -*-
 * $Id: x86_instr.h,v 1.4 1998/04/29 04:55:02 nikos Exp $
 * 
 * RCS Log:
 * $Log: x86_instr.h,v $
 * Revision 1.4  1998/04/29 04:55:02  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:54  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:39  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:13  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:13  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: x86_instr.h,v 1.4 1998/04/29 04:55:02 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86_instr.h,v 1.4 1998/04/29 04:55:02 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:32  bugnion
 * stanford update (x86)
 *
 * Revision 1.8  1997/10/18  00:39:45  sean
 * added instruction counting code, and rigged it so some system calls
 *   (e.g. exit) can cause the CPU to halt, effectively terminating simulation
 *
 * Revision 1.7  1997/09/27 18:35:31  sean
 * added END_DEFN macro to make x86_instr.cc and fpu_instr.cc cleaner
 *   (balanced brackets make emacs happier)
 *
 * Revision 1.6  1997/09/24 08:17:20  sean
 * 1) header of operation methods tweaked to return string name of operation
 *    to sufficiently whacked callers (see instruction.cc)
 * 2) SCAS fixed to be proper subtraction
 * 3) shift functions fixed to set AF (Pentium book says undefined, by my
 *    CPU apparently likes to always set AF for shifts, so I did it to make
 *    debugging easier.)
 *
 * Revision 1.5  1997/02/28 09:12:25  sean
 * split RET into RETnear and RETfar
 *
 * Revision 1.4  1997/02/04 22:38:11  sean
 * made operation functions return number of clock cycles spent (sorta - all
 * return 1 right now), and SAL is a synonym for SHL
 *
 * Revision 1.3  1997/02/04 02:38:43  sjt
 * cosmetic changes
 *
 */

#ifndef _X86_INSTR_H
#define _X86_INSTR_H

// This is a list of all the instructions supported by the simulator
//
//  This file is embedded inside of the 'class CPU' declaration, so declaring
// classes and such inside this file may not do what you think it does...

// We'll use this to declare instructions in the CPU class declaration
#define DECL_INST(mnemonic)         int64 mnemonic(const Instruction& inst);

// We'll use this to define the actual instructions in x86_instr.cc
#define DEFN_INST(mnemonic)         \
  int64 CPU::mnemonic(const Instruction& inst)  \
  {  \
    if(&inst == 0) return((int64) #mnemonic); \
    if(instr_counter) instr_counter->AddInstruction(#mnemonic);

#define END_DEFN   }

// Instruction synonyms go here...
#define CDQ       CWD
#define CWDE      CBW
#define CMPSB     CMPS
#define CMPSD     CMPS
#define CMPSW     CMPS
#define INSB      INS
#define INSD      INS
#define INSW      INS
#define IRETD     IRET
#define JECXZ     JCXZ
#define LODSB     LODS
#define LODSD     LODS
#define LODSW     LODS
#define LOOPE     LOOPZ
#define LOOPNE    LOOPNZ
#define MOVSB     MOVS
#define MOVSD     MOVS
#define MOVSW     MOVS
#define OUTSB     OUTS
#define OUTSD     OUTS
#define OUTSW     OUTS
#define SAL       SHL
#define SCASB     SCAS
#define SCASD     SCAS
#define SCASW     SCAS
#define STOSB     STOS
#define STOSD     STOS
#define STOSW     STOS

// Instructions follow below
DECL_INST(AAA)
DECL_INST(AAD)
DECL_INST(AAM)
DECL_INST(AAS)
DECL_INST(ADC)
DECL_INST(ADD)
DECL_INST(AND)
DECL_INST(ARPL)
DECL_INST(BOUND)
DECL_INST(BSF)
DECL_INST(BSR)
DECL_INST(BSWAP)
DECL_INST(BT)
DECL_INST(BTC)
DECL_INST(BTR)
DECL_INST(BTS)
DECL_INST(CALL)
DECL_INST(CBW)
DECL_INST(CLC)
DECL_INST(CLD)
DECL_INST(CLI)
DECL_INST(CLTS)
DECL_INST(CMC)
DECL_INST(CMP)
DECL_INST(CMPS)
DECL_INST(CMPXCHG)
DECL_INST(CMPXCHG8B)
DECL_INST(CPUID)
DECL_INST(CWD)
DECL_INST(DAA)
DECL_INST(DAS)
DECL_INST(DEC)
DECL_INST(DIV)
DECL_INST(ENTER)
DECL_INST(HLT)
DECL_INST(IDIV)
DECL_INST(IMUL)
DECL_INST(IN)
DECL_INST(INC)
DECL_INST(INS)
DECL_INST(INT)
DECL_INST(INTO)
DECL_INST(INVD)
DECL_INST(INVLPG)
DECL_INST(IRET)
DECL_INST(Jcc)
DECL_INST(JCXZ)
DECL_INST(JMP)
DECL_INST(LAHF)
DECL_INST(LAR)
DECL_INST(LDS)
DECL_INST(LEA)
DECL_INST(LEAVE)
DECL_INST(LES)
DECL_INST(LFS)
DECL_INST(LGDT)
DECL_INST(LGS)
DECL_INST(LIDT)
DECL_INST(LLDT)
DECL_INST(LMSW)
DECL_INST(LODS)
DECL_INST(LOOP)
DECL_INST(LOOPNZ)
DECL_INST(LOOPZ)
DECL_INST(LSL)
DECL_INST(LSS)
DECL_INST(LTR)
DECL_INST(MOV)
DECL_INST(MOVS)
DECL_INST(MOVSX)
DECL_INST(MOVZX)
DECL_INST(MUL)
DECL_INST(NEG)
DECL_INST(NOP)
DECL_INST(NOT)
DECL_INST(OR)
DECL_INST(OUT)
DECL_INST(OUTS)
DECL_INST(POP)
DECL_INST(POPA)
DECL_INST(POPF)
DECL_INST(PUSH)
DECL_INST(PUSHA)
DECL_INST(PUSHF)
DECL_INST(RCL)
DECL_INST(RCR)
DECL_INST(RDMSR)
DECL_INST(RDTSC)
DECL_INST(RETfar)
DECL_INST(RETnear)
DECL_INST(ROL)
DECL_INST(ROR)
DECL_INST(RSM)
DECL_INST(SAHF)
DECL_INST(SAR)
DECL_INST(SBB)
DECL_INST(SCAS)
DECL_INST(SETcc)
DECL_INST(SGDT)
DECL_INST(SHL)
DECL_INST(SHR)
DECL_INST(SHLD)
DECL_INST(SHRD)
DECL_INST(SIDT)
DECL_INST(SLDT)
DECL_INST(SMSW)
DECL_INST(STC)
DECL_INST(STD)
DECL_INST(STI)
DECL_INST(STOS)
DECL_INST(STR)
DECL_INST(SUB)
DECL_INST(TEST)
DECL_INST(VERR)
DECL_INST(VERW)
DECL_INST(WAIT)
DECL_INST(WBINVD)
DECL_INST(WRMSR)
DECL_INST(XADD)
DECL_INST(XCHG)
DECL_INST(XLAT)
DECL_INST(XOR)

// Helper functions come below...
int64 CALLorJMP(const Instruction& inst, bool push_return);

#endif
