/*
 * $Id: x86_operand.cc,v 1.4 1998/04/29 04:55:02 nikos Exp $
 *
 * RCS Log:
 * $Log: x86_operand.cc,v $
 * Revision 1.4  1998/04/29 04:55:02  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:54  nikos
 * atf_tlb_fixed
 *
// Revision 1.3  1998/03/13  17:12:39  nikos
// Fix vaddr->physaddr translation bug and tlb behavior in ATF.
// Changes affect only people who use ATF for their traces.
//
 * Revision 1.3  1998/03/13 16:37:14  nikos
 * .
 *
// Revision 1.2  1998/02/16  23:20:13  nikos
// Fixed AranaTrace command which was broken with the last commit.
// Macros in external_trace.h are set such that the name "arana"
// appears only in arana_trace.[ch] and no-where else!!!
// In the previous revision, kappa.c has not been cleaned up.
//
// If you want to use the AranaTrace command,
// compile SimOS with the following variables set :
//
// EXTERNAL_TRACE=ARANA_TRACE
// ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
// ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
// ARANA_TRACE_OBJS=arana_trace.o
// ARANA_TRACE_HEADER=arana_trace.h
//
// This commit reports a bunch of modified files which are not
// really modified (only some rcs fields like $Id: x86_operand.cc,v 1.4 1998/04/29 04:55:02 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86_operand.cc,v 1.4 1998/04/29 04:55:02 nikos Exp $, etc).
//
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
// Revision 1.1  1997/12/16  01:28:32  bugnion
// stanford update (x86)
//
// Revision 1.19  1997/09/27  18:34:43  sean
// Fixed IMUL decoding to always return three operands (even if first two are
//   the same)
//
 * Revision 1.18  1997/03/14 08:03:34  yoel
 * Added decoders for the encodings unique to the FPU.  Notably a lot of
 * oddly sized M encodings and the stack encodings.
 *
 * Revision 1.17  1997/03/14 04:03:35  sean
 * added AH_IB and EAX_E?? operand decoding functions
 *
 * Revision 1.16  1997/03/09 21:14:29  sean
 * fixed call to MemoryOperand copy constructor in Ma(?) (it needed a cast)
 *
 * Revision 1.15  1997/03/09 08:36:57  yoel
 * Removed the nDecoded from the prototypes.
 *
 * Revision 1.14  1997/03/07 05:45:46  sean
 * fixed Y's, did GV_M, GV_Ma, and GV_Mp
 *
 * Revision 1.13  1997/03/06 19:57:18  yoel
 * SIB and DO_E stuff are now all inlines nad happy for debugging.
 *
 * Revision 1.12  1997/03/06 10:06:05  yoel
 * Added casts to all the ReadNBytes() since the numbers are signed, but the
 * data stream reads unsigned info.
 *
 * Revision 1.11  1997/03/06 09:34:56  yoel
 * Added the final necessary functions from the various GrpN opcodes.  Need
 * to do the SIB stuff and the M encodings, but I think the M's are just Es.
 *
 * Revision 1.10  1997/03/06 07:42:07  yoel
 * Added in the hooks for the rest of the Eb, Ev, Ew calls, checking in for
 * Sean now.
 *
 * Revision 1.9  1997/03/06 07:30:38  yoel
 * Added all of the Eb, Ev, Ew as first operand check in so Sean can use it.
 *
 * Revision 1.8  1997/03/03 05:09:16  yoel
 * Added support for a few more.  At this point, All I have left are M and E
 * types (i.e. Memory with the ModR?m
 * Err ModR/M which is evil.
 *
 * Revision 1.7  1997/03/02 00:46:09  yoel
 * Added support for a number more operands.  The only ones left are the ones
 * starting with A, E, J, M, X, and Y.  T too, except that T does not actually
 * appear to exist (read A-2 and then find an example of a T operand)
 *
 * Revision 1.6  1997/02/28 09:24:18  yoel
 * Expanded massively the prototype for the decoders so that they can
 * take the seg register, and the sizes for both addr's and immediates.
 *
 * Revision 1.5  1997/02/27 08:48:03  yoel
 * Added immediate operand support and #defined generic IB, IV, and IW's s
 * that I don't incur the overhead of an extra function call in IB_IW and
 * other multi-operand calls...
 *
 * Revision 1.4  1997/02/20 05:52:17  yoel
 * Removed a bunch of #includes that were redundant and causing the statics
 * to be multiply defined in object files.  Also fixed the :'s that should
 * have been ;'s and tweaked the prototype to *ops[] from ops[].
 *
 * Revision 1.3  1997/02/20 05:38:48  yoel
 * Revised the Register decoding to be slightly shorter.  Note it depends
 * upon the fact that REG_EAX == REG_AX and in general REG_Exx == REG_xx.
 *
 * Revision 1.2  1997/02/19 06:51:59  yoel
 * Whee, all the register operands will work fine once Sean provides me a
 * means of saying which CPU's registers we're using.  Everything else
 * is stubbed in such a way as to compile and (hopefully) be straightforward
 * to fill in.
 *
 *
 */

static char rcsid[] __attribute__((unused))
        = "$Id: x86_operand.cc,v 1.4 1998/04/29 04:55:02 nikos Exp $";
	
//  File x86_operand.cc,  Author Jeremy Sugerman (yoel@stanford.edu)
//
//  implementation for all of the Operand decoding functions in the table in
// decode.h

#include <stdio.h>
#include <stdlib.h>
#include "data_stream.h"
#include "instruction.h"
#include "registers.h"
#include "modrm.h"

#define DECL_OPER(mnem) \
    bool Instruction::mnem(DataStream &instr_data, int opSize, int addrSize, \
			   int segReg, Operand *ops[])

#define DO_IB(opNum, iByte) \
  if (instr_data.ReadOneByte((int8&) iByte))               \
    ops[opNum] = new ImmediateOperand(SIZE_8BIT, iByte);   \
  else                                                     \
    return false;

#define DO_IV(opNum, iWord, iDblWord) \
  if (opSize == SIZE_16BIT)                                   \
    if (instr_data.ReadTwoBytes((int16&) iWord))              \
      ops[opNum] = new ImmediateOperand(SIZE_16BIT, iWord);   \
    else                                                      \
      return false;                                           \
  else                                                        \
    if (instr_data.ReadFourBytes((int32&) iDblWord))          \
      ops[opNum] = new ImmediateOperand(SIZE_32BIT, iDblWord);\
    else                                                      \
      return false;

#define DO_IW(opNum, iWord) \
  if (instr_data.ReadTwoBytes((int16&)iWord))                \
    ops[opNum] = new ImmediateOperand(SIZE_16BIT, iWord);    \
  else                                                       \
    return false;

#define DO_OB(opNum, aWord, aDblWord) \
  if (addrSize == SIZE_16BIT)                                               \
    if (instr_data.ReadTwoBytes((int16&) aWord))                            \
      ops[opNum] = new MemoryOperand(SIZE_8BIT, addrSize, segReg, REG_NONE, \
				     REG_NONE, SCALE_BY_1, aWord);          \
    else                                                                    \
      return false;                                                         \
  else                                                                      \
    if (instr_data.ReadFourBytes((int32&) aDblWord))                        \
      ops[opNum] = new MemoryOperand(SIZE_8BIT, addrSize, segReg, REG_NONE, \
                                     REG_NONE, SCALE_BY_1, aDblWord);       \
    else                                                                    \
      return false;

#define DO_OV(opNum, aWord, aDblWord) \
  if (addrSize == SIZE_16BIT)                                               \
    if (instr_data.ReadTwoBytes((int16&) aWord))                            \
      ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_NONE,    \
				     REG_NONE, SCALE_BY_1, aWord);          \
    else                                                                    \
      return false;                                                         \
  else                                                                      \
    if (instr_data.ReadFourBytes((int32&) aDblWord))                        \
      ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_NONE,    \
                                     REG_NONE, SCALE_BY_1, aDblWord);       \
    else                                                                    \
      return false;

#define DO_XB(opNum) \
  ops[opNum] = new MemoryOperand(SIZE_8BIT, addrSize, segReg, REG_SI);

#define DO_XV(opNum) \
  ops[opNum] = new MemoryOperand(opSize, addrSize, segReg, REG_SI);

// YB _ALWAYS_ uses ES (no overriding allowed)
#define DO_YB(opNum) \
  ops[opNum] = new MemoryOperand(SIZE_8BIT, addrSize, REG_ES, REG_DI);

// YV _ALWAYS_ uses ES (no overriding allowed)
#define DO_YV(opNum) \
  ops[opNum] = new MemoryOperand(opSize, addrSize, REG_ES, REG_DI);

//  Below are all of the functions needed for one byte operands...

//  Easy.  Op[0] is AH.
DECL_OPER(AH)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AH);
  return true;
}

DECL_OPER(AL_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  DO_IB(1, iByte);
  return true;
}

//  Easy, this is just the AL register.
DECL_OPER(AL)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  return true;
}

DECL_OPER(AL_DX)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  ops[1] = new RegisterOperand(SIZE_16BIT, REG_DX);
  return true;
}

DECL_OPER(AL_OB)
{
  s_int16 aWord;
  s_int32 aDblWord;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  DO_OB(1, aWord, aDblWord);
  return true;
}

DECL_OPER(AL_XB)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  DO_XB(1);
  return true;
}

DECL_OPER(AL_YB)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AL);
  DO_YB(1);
  return true;
}

DECL_OPER(AP)
{
  s_int16 iWord;
  s_int32 iDblWord;

  DO_IV(0, iWord, iDblWord);
  DO_IW(1, iWord);
  return true;
}

//  BH register.
DECL_OPER(BH)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_BH);
  return true;
}

//  BL register.
DECL_OPER(BL)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_BL);
  return true;
}

//  CH register
DECL_OPER(CH)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_CH);
  return true;
}

// CL Register
DECL_OPER(CL)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_CL);
  return true;
}

DECL_OPER(CS)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_CS);
  return true;
}

//  DH Register
DECL_OPER(DH)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_DH);
  return true;
}

//  DL Register
DECL_OPER(DL)
{
  ops[0] = new RegisterOperand(SIZE_8BIT, REG_DL);
  return true;
}

DECL_OPER(DS)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_DS);
  return true;
}

DECL_OPER(DX_AL)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_DX);
  ops[1] = new RegisterOperand(SIZE_16BIT, REG_AL);
  return true;
}

DECL_OPER(DX_EAX)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_DX);
  ops[1] = new RegisterOperand(opSize, REG_AX);
  return true;
}

DECL_OPER(DX_XB)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_DX);
  DO_XB(1);
  return true;
}

DECL_OPER(DX_XV)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_DX);
  DO_XV(1);
  return true;
}

//  EAX register / AX Register depending on size.
DECL_OPER(EAX)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  return true;
}

DECL_OPER(EAX_DX)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(SIZE_16BIT, REG_DX);
  return true;
}

DECL_OPER(EAX_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(opSize, REG_EAX);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(EAX_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_EAX);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EAX_OV)
{
  s_int16 aWord;
  s_int32 aDblWord;
  
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  DO_OV(1, aWord, aDblWord);
  return true;
}

DECL_OPER(EAX_XV)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  DO_XV(1);
  return true;
}

DECL_OPER(EAX_YV)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  DO_YV(1);
  return true;
}

DECL_OPER(EB)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops);
}

DECL_OPER(EB_CL)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops)) return false;
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_CL);
  return true;
}

DECL_OPER(EB_GB)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops)) return false;
  DO_GB(1, modRM);
  return true;
}

DECL_OPER(EB_IB)
{ 
  s_int8 iByte;
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops)) return false;
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(EB_OB)
{
  int8 modRM;
  s_int16 aWord;
  s_int32 aDblWord;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops)) return false;
  DO_OB(1, aWord, aDblWord);
  return true;
}

DECL_OPER(EBP)
{
  ops[0] = new RegisterOperand(opSize, REG_EBP);
  return true;
}

DECL_OPER(EBX)
{
  ops[0] = new RegisterOperand(opSize, REG_EBX); 
  return true;
}

DECL_OPER(ECX)
{
  ops[0] = new RegisterOperand(opSize, REG_ECX);
  return true;
}

DECL_OPER(EDI)
{
  ops[0] = new RegisterOperand(opSize, REG_EDI);
  return true;
}

DECL_OPER(EDX)
{
  ops[0] = new RegisterOperand(opSize, REG_EDX);
  return true;
}

DECL_OPER(ES)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_ES);
  return true;
}

DECL_OPER(ESI)
{
  ops[0] = new RegisterOperand(opSize, REG_ESI);
  return true;
}

DECL_OPER(ESP)
{
  ops[0] = new RegisterOperand(opSize, REG_ESP);
  return true;
}

DECL_OPER(EV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops);
}

DECL_OPER(EV_CL)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_CL);
  return true;
}

DECL_OPER(EV_GV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_GV(1, modRM);
  return true;
}

DECL_OPER(EV_IB)
{
  s_int8 iByte;
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(EV_IV)
{
  int8 modRM;
  s_int16 iWord;
  s_int32 iDblWord;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EW_GW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops)) return false;
  DO_GW(1, modRM);
  return true;
}

DECL_OPER(EW_SW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops)) return false;
  DO_SW(1, modRM);
  return true;
}

DECL_OPER(FV)
{
  ops[0] = new RegisterOperand(opSize, REG_FLAGS);
  return true;
}

DECL_OPER(GB_EB)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GB(0, modRM);
  return DO_E(1, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops);
}

DECL_OPER(GV_EV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, opSize, addrSize, segReg, ops);
}

DECL_OPER(GV_EV_IB)
{
  s_int8 iByte;
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  if (!DO_E(1, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_IB(2, iByte);
  return true;
}

DECL_OPER(GV_EV_IV)
{
  int8 modRM;
  s_int16 iWord;
  s_int32 iDblWord;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  if (!DO_E(1, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_IV(2, iWord, iDblWord);
  return true;
}

// A GV_GV_EV is just like a GV_EV, except the same operand appears for 0 and 1
DECL_OPER(GV_GV_EV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  DO_GV(1, modRM);
  return DO_E(2, instr_data, modRM, opSize, addrSize, segReg, ops);
}

// A GV_M is never going to actually look at the memory, so call it an Mb
DECL_OPER(GV_M)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops, true);
}

// An Ma is twice the size of an Mv...
DECL_OPER(GV_MA)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, (opSize * 2), addrSize, segReg, ops, true);
}

// An Mp is the size of a Mv + an Mw
DECL_OPER(GV_MP)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, (opSize + 2), addrSize, segReg, ops, true);
}

DECL_OPER(IB)
{
  s_int8 iByte;

  DO_IB(0, iByte);
  return true;
}

DECL_OPER(IB_AL)
{
  s_int8 iByte;

  DO_IB(0, iByte);
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_AL);
  return true;
}

DECL_OPER(IB_EAX)
{
  s_int8 iByte;

  DO_IB(0, iByte);
  ops[1] = new RegisterOperand(opSize, REG_EAX);
  return true;
}

DECL_OPER(IV)
{
  s_int16 iWord;
  s_int32 iDblWord;
  
  DO_IV(0, iWord, iDblWord);
  return true;
}

DECL_OPER(IW)
{
  s_int16 iWord;

  DO_IW(0, iWord);
  return true;
}

DECL_OPER(IW_IB)
{
  s_int16 iWord;
  s_int8 iByte;

  DO_IW(0, iWord);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(JB)
{
  s_int8 iByte;

  DO_IB(0, iByte);
  return true;
}

DECL_OPER(JV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  DO_IV(0, iWord, iDblWord);
  return true;
}

DECL_OPER(OB_AL)
{
  s_int16 aWord;
  s_int32 aDblWord;

  DO_OB(0, aWord, aDblWord);
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_AL);
  return true;
}

DECL_OPER(OV_EAX)
{
  s_int16 aWord;
  s_int32 aDblWord;

  DO_OV(0, aWord, aDblWord);
  ops[1] = new RegisterOperand(opSize, REG_AX);
  return true;
}

DECL_OPER(SS)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_SS);
  return true;
}

DECL_OPER(SW_EW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_SW(0, modRM);
  if (!DO_E(1, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops)) return false;
  return true;
}

DECL_OPER(XB_YB)
{
  DO_XB(0);
  DO_YB(1);
  return true;
}

DECL_OPER(XV_YV)
{
  DO_XV(0);
  DO_YV(1);
  return true;
}

DECL_OPER(YB_AL)
{
  DO_YB(0);
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_AL);
  return true;
}

DECL_OPER(YB_DX)
{
  DO_YB(0);
  ops[1] = new RegisterOperand(SIZE_16BIT, REG_DX);
  return true;
}

DECL_OPER(YV_DX)
{
  DO_YV(0);
  ops[1] = new RegisterOperand(SIZE_16BIT, REG_DX);
  return true;
}

DECL_OPER(YV_EAX)
{
  DO_YV(0);
  ops[1] = new RegisterOperand(opSize, REG_EAX);
  return true;
}

//  Now add the extras for Two Byte Opcodes..

DECL_OPER(CD_RD)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_CD(0, modRM);
  DO_RD(1, modRM);
  return true;
}

DECL_OPER(DD_RD)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_DD(0, modRM);
  DO_RD(1, modRM);
  return true;
}

DECL_OPER(EV_GV_CL)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_GV(1, modRM);
  ops[2] = new RegisterOperand(SIZE_8BIT, REG_CL);
  return true;
}

DECL_OPER(EV_GV_IB)
{
  s_int8 iByte;
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  DO_GV(1, modRM);
  DO_IB(2, iByte);
  return true;
}

DECL_OPER(FS)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_FS);
  return true;
}

DECL_OPER(GS)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_GS);
  return true;
}

DECL_OPER(GV_EB)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops);
}

DECL_OPER(GV_EW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_GV(0, modRM);
  return DO_E(1, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops);
}

DECL_OPER(MP)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (opSize == SIZE_16BIT)
    return DO_E(0, instr_data, modRM, SIZE_32BIT, addrSize, segReg, ops, true);
  else
    return DO_E(0, instr_data, modRM, SIZE_48BIT, addrSize, segReg, ops, true);
}

DECL_OPER(RD_CD)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_RD(0, modRM);
  DO_CD(1, modRM);
  return true;
}

DECL_OPER(RD_DD)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  DO_RD(0, modRM);
  DO_DD(1, modRM);
  return true;
}


//  Time to add a whole bunch.  Apparently, all the MOV immediate to
// register instructions weren't documented in the book as using immediates
// so now I get to add all of these to make it happy.

DECL_OPER(CL_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_CL);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(DL_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_DL);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(BL_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_BL);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(AH_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_AH);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(CH_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_CH);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(DH_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_DH);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(BH_IB)
{
  s_int8 iByte;

  ops[0] = new RegisterOperand(SIZE_8BIT, REG_BH);
  DO_IB(1, iByte);
  return true;
}

DECL_OPER(ECX_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_ECX);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EDX_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_EDX);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EBX_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_EBX);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(ESP_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_ESP);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EBP_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_EBP);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(ESI_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_ESI);
  DO_IV(1, iWord, iDblWord);
  return true;
}

DECL_OPER(EDI_IV)
{
  s_int16 iWord;
  s_int32 iDblWord;

  ops[0] = new RegisterOperand(opSize, REG_EDI);
  DO_IV(1, iWord, iDblWord);
  return true;
}

//  Grp6 Opcodes require EW encoding
DECL_OPER(EW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops);
}

//  Ep is a sick concept.  It's 32 bit for a SIZE_16 and 48 bit for a
// SIZE_32.  So, I'm going to encode it as two memory operands, the first
// either size 16 or 32 and the second of size 16.  Luckily there's a copy
// constructor for memory operands here :)
DECL_OPER(EP)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);

  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  //  The second argument is just the first one plus opSize.  Its size,
  // however is always SIZE_16BIT.
  ops[1] = new MemoryOperand(*(MemoryOperand *)ops[0], opSize, SIZE_16BIT); 
  return true;
}

DECL_OPER(EB_AL)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops)) return false;
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_AL);
  return true;
}

DECL_OPER(EV_EAX)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  if (!DO_E(0, instr_data, modRM, opSize, addrSize, segReg, ops)) return false;
  ops[1] = new RegisterOperand(opSize, REG_EAX);
  return true;
}

DECL_OPER(MS)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_48BIT, addrSize, segReg, ops, true);
}

DECL_OPER(MQ)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_64BIT, addrSize, segReg, ops, true);
}

// Required for XCHGs...
DECL_OPER(EAX_ECX)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_ECX);
  return true;
}

DECL_OPER(EAX_EDX)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_EDX);
  return true;
}

DECL_OPER(EAX_EBX)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_EBX);
  return true;
}

DECL_OPER(EAX_ESP)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_ESP);
  return true;
}

DECL_OPER(EAX_EBP)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_EBP);
  return true;
}

DECL_OPER(EAX_ESI)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_ESI);
  return true;
}

DECL_OPER(EAX_EDI)
{
  ops[0] = new RegisterOperand(opSize, REG_EAX);
  ops[1] = new RegisterOperand(opSize, REG_EDI);
  return true;
}

// A couple sickos for IMUL
DECL_OPER(AX_AL_EB)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);

  ops[0] = new RegisterOperand(SIZE_16BIT, REG_AX);
  ops[1] = new RegisterOperand(SIZE_8BIT, REG_AL);

  return DO_E(2, instr_data, modRM, SIZE_8BIT, addrSize, segReg, ops);
}

DECL_OPER(EDXEAX_EAX_EV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);

  ops[0] = new RegisterOperand(opSize << 1, REG_EDXEAX);
  ops[1] = new RegisterOperand(opSize, REG_EAX);

  return DO_E(2, instr_data, modRM, opSize, addrSize, segReg, ops);
}

//  Needed for FPU decoding:
DECL_OPER(AX)
{
  ops[0] = new RegisterOperand(SIZE_16BIT, REG_AX);
  return true;
}

DECL_OPER(MW)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_16BIT, addrSize, segReg, ops, true);
}

DECL_OPER(MD)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_32BIT, addrSize, segReg, ops, true);
}

DECL_OPER(M80)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, SIZE_80BIT, addrSize, segReg, ops, true);
}

DECL_OPER(MENV)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, opSize * 7, addrSize, segReg, ops, true);
}

DECL_OPER(MSTATE)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  return DO_E(0, instr_data, modRM, opSize*7+80, addrSize, segReg, ops, true);
}

DECL_OPER(N)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  ops[0] = new ImmediateOperand(SIZE_8BIT, modRM & 0x7);
  return true;
}

DECL_OPER(N_ST)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  ops[0] = new ImmediateOperand(SIZE_8BIT, modRM & 0x7);
  ops[1] = new ImmediateOperand(SIZE_8BIT, 0);
  return true;
}

DECL_OPER(ST_N)
{
  int8 modRM;

  GET_MODRM(instr_data, modRM);
  ops[0] = new ImmediateOperand(SIZE_8BIT, 0);
  ops[1] = new ImmediateOperand(SIZE_8BIT, modRM & 0x7);
  return true;
}


