/* -*- C++ -*-
 * $Id: x86_operand.h,v 1.4 1998/04/29 04:55:03 nikos Exp $
 *
 * RCS Log:
 * $Log: x86_operand.h,v $
 * Revision 1.4  1998/04/29 04:55:03  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:55  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:40  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:14  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:14  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: x86_operand.h,v 1.4 1998/04/29 04:55:03 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86_operand.h,v 1.4 1998/04/29 04:55:03 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:33  bugnion
 * stanford update (x86)
 *
 * Revision 1.13  1997/09/27  18:34:43  sean
 * Fixed IMUL decoding to always return three operands (even if first two are
 *   the same)
 *
 * Revision 1.12  1997/03/14 08:03:34  yoel
 * Added prototypes for all the FPU operand decoders.
 *
 * Revision 1.11  1997/03/14 04:03:35  sean
 * added AH_IB and EAX_E?? operand decoding functions
 *
 * Revision 1.10  1997/03/09 08:36:57  yoel
 * Not sure what I changed, I think removed the nDecoded from the prot.
 * err prototype.
 *
 * Revision 1.9  1997/03/06 09:34:56  yoel
 * Added the final necessary functions from the various GrpN opcodes.  Need
 * to do the SIB stuff and the M encodings, but I think the M's are just Es.
 *
 * Revision 1.8  1997/03/06 07:30:38  yoel
 * Added all of the Eb, Ev, Ew as first operand check in so Sean can use it.
 *
 * Revision 1.7  1997/02/28 23:53:47  sean
 * removed prototype for AI_IB decoder (which doesn't exist)
 *
 * Revision 1.6  1997/02/28 09:24:18  yoel
 * Expanded prototype to include all the new parameters (segREg, addrSize)
 *
 * Revision 1.5  1997/02/27 08:48:03  yoel
 * Made the prototype return a bool now so I can indicate failure.  Also
 * removed a few functions that shouldn't exist where I read an I as an l.
 *
 * Revision 1.4  1997/02/20 05:52:17  yoel
 * Change prototype slightly to *ops[] from ops[].
 *
 * Revision 1.3  1997/02/19 06:51:59  yoel
 * Bleah.  Now it all compiles and has a consistent prototype.  Hopefully there
 * will be no more massive overhauls.
 *
 * Revision 1.2  1997/02/19 02:56:45  yoel
 * There, file should be relatively stable now.  Maybe the little $Log $ will
 * hold now.
 *
 *
 */

//  File x86_operand.h,  Author Jeremy Sugerman (yoel@stanford.edu)
//
//  prototypes for all of the Operand decoding functions in the table in
// decode.h

#ifndef __x86_OPERAND_H__
#define __x86_OPERAND_H__

#include "instruction.h"

#define PROTO_OPER(mnem) \
    bool mnem(DataStream &instr_data, int opSize, int addrSize, int segReg, \
	      Operand *ops[])

//  Necessary for OneByte opcodes
static PROTO_OPER(AH);
static PROTO_OPER(AL);
static PROTO_OPER(AL_DX);
static PROTO_OPER(AL_IB);
static PROTO_OPER(AL_OB);
static PROTO_OPER(AL_XB);
static PROTO_OPER(AL_YB);
static PROTO_OPER(AP);
static PROTO_OPER(BH);
static PROTO_OPER(BL);
static PROTO_OPER(CH);
static PROTO_OPER(CL);
static PROTO_OPER(CS);
static PROTO_OPER(DH);
static PROTO_OPER(DL);
static PROTO_OPER(DS);
static PROTO_OPER(DX_AL);
static PROTO_OPER(DX_EAX);
static PROTO_OPER(DX_XB);
static PROTO_OPER(DX_XV);
static PROTO_OPER(EAX);
static PROTO_OPER(EAX_DX);
static PROTO_OPER(EAX_IB);
static PROTO_OPER(EAX_IV);
static PROTO_OPER(EAX_OV);
static PROTO_OPER(EAX_XV);
static PROTO_OPER(EAX_YV);
static PROTO_OPER(EB);
static PROTO_OPER(EB_CL);
static PROTO_OPER(EB_GB);
static PROTO_OPER(EB_IB);
static PROTO_OPER(EB_OB);
static PROTO_OPER(EBP);
static PROTO_OPER(EBX);
static PROTO_OPER(ECX);
static PROTO_OPER(EDI);
static PROTO_OPER(EDX);
static PROTO_OPER(ES);
static PROTO_OPER(ESI);
static PROTO_OPER(ESP);
static PROTO_OPER(EV);
static PROTO_OPER(EV_CL);
static PROTO_OPER(EV_GV);
static PROTO_OPER(EV_IB);
static PROTO_OPER(EV_IV);
static PROTO_OPER(EW_GW);
static PROTO_OPER(EW_SW);
static PROTO_OPER(FV);
static PROTO_OPER(GB_EB);
static PROTO_OPER(GV_EV);
static PROTO_OPER(GV_EV_IB);
static PROTO_OPER(GV_EV_IV);
static PROTO_OPER(GV_GV_EV);
static PROTO_OPER(GV_M);
static PROTO_OPER(GV_MA);
static PROTO_OPER(GV_MP);
static PROTO_OPER(IB);
static PROTO_OPER(IB_AL);
static PROTO_OPER(IB_EAX);
static PROTO_OPER(IV);
static PROTO_OPER(IW);
static PROTO_OPER(IW_IB);
static PROTO_OPER(JB);
static PROTO_OPER(JV);
static PROTO_OPER(OB_AL);
static PROTO_OPER(OV_EAX);
static PROTO_OPER(SS);
static PROTO_OPER(SW_EW);
static PROTO_OPER(XB_YB);
static PROTO_OPER(XV_YV);
static PROTO_OPER(YB_AL);
static PROTO_OPER(YB_DX);
static PROTO_OPER(YV_DX);
static PROTO_OPER(YV_EAX);

//  Extras needed for Two Byte Opcodes
static PROTO_OPER(CD_RD);
static PROTO_OPER(DD_RD);
static PROTO_OPER(EV_GV_CL);
static PROTO_OPER(EV_GV_IB);
static PROTO_OPER(FS);
static PROTO_OPER(GS);
static PROTO_OPER(GV_EB);
static PROTO_OPER(GV_EW);
static PROTO_OPER(MP);
static PROTO_OPER(RD_CD);
static PROTO_OPER(RD_DD);

//  Extras to deal with the implicit immediates of MOV instructions (B0-BF)
static PROTO_OPER(CL_IB);
static PROTO_OPER(DL_IB);
static PROTO_OPER(BL_IB);
static PROTO_OPER(AH_IB);
static PROTO_OPER(CH_IB);
static PROTO_OPER(DH_IB);
static PROTO_OPER(BH_IB);
static PROTO_OPER(ECX_IV);
static PROTO_OPER(EDX_IV);
static PROTO_OPER(EBX_IV);
static PROTO_OPER(ESP_IV);
static PROTO_OPER(EBP_IV);
static PROTO_OPER(ESI_IV);
static PROTO_OPER(EDI_IV);

// Grp6 requires an EW
static PROTO_OPER(EW);

// Grp5 needs and Ep
static PROTO_OPER(EP);

//  The other Grp instructions
static PROTO_OPER(EB_AL);
static PROTO_OPER(EV_EAX);
static PROTO_OPER(MS);
static PROTO_OPER(MQ);

// Required for XCHGs...
static PROTO_OPER(EAX_ECX);
static PROTO_OPER(EAX_EDX);
static PROTO_OPER(EAX_EBX);
static PROTO_OPER(EAX_ESP);
static PROTO_OPER(EAX_EBP);
static PROTO_OPER(EAX_ESI);
static PROTO_OPER(EAX_EDI);

// A couple sickos for IMUL
static PROTO_OPER(AX_AL_EB);
static PROTO_OPER(EDXEAX_EAX_EV);

// For FPU decoding...
static PROTO_OPER(AX);
static PROTO_OPER(MW);
static PROTO_OPER(MD);
static PROTO_OPER(M80);
static PROTO_OPER(MENV);
static PROTO_OPER(MSTATE);
static PROTO_OPER(N);
static PROTO_OPER(N_ST);
static PROTO_OPER(ST_N);

#endif
