/* -*- C++ -*-
 * $Id: x86sim.h,v 1.4 1998/04/29 04:55:04 nikos Exp $
 * 
 * RCS Log:
 * $Log: x86sim.h,v $
 * Revision 1.4  1998/04/29 04:55:04  nikos
 * Added new AranaTrace command "AranaTrace  filter_thru_L1" that dumps
 * only instructions with references (pc + ea) that miss in the L1 cache.
 * MBs are always dumped.
 *
 * Revision 1.4  1998/03/25 19:16:56  nikos
 * atf_tlb_fixed
 *
 * Revision 1.3  1998/03/13  17:12:41  nikos
 * Fix vaddr->physaddr translation bug and tlb behavior in ATF.
 * Changes affect only people who use ATF for their traces.
 *
 * Revision 1.3  1998/03/13 16:37:15  nikos
 * .
 *
 * Revision 1.2  1998/02/16  23:20:15  nikos
 * Fixed AranaTrace command which was broken with the last commit.
 * Macros in external_trace.h are set such that the name "arana"
 * appears only in arana_trace.[ch] and no-where else!!!
 * In the previous revision, kappa.c has not been cleaned up.
 *
 * If you want to use the AranaTrace command,
 * compile SimOS with the following variables set :
 *
 * EXTERNAL_TRACE=ARANA_TRACE
 * ARANA_TRACE_INC_DIR=/perf_sym/ATF/atf
 * ARANA_TRACE_LIB=/perf_sym/ATF/atf/libatf.a
 * ARANA_TRACE_OBJS=arana_trace.o
 * ARANA_TRACE_HEADER=arana_trace.h
 *
 * This commit reports a bunch of modified files which are not
 * really modified (only some rcs fields like $Id: x86sim.h,v 1.4 1998/04/29 04:55:04 nikos Exp $, $Header: /wrl/proj/simos/cvsroot/simulation/cpus-x86/x86sim/x86sim.h,v 1.4 1998/04/29 04:55:04 nikos Exp $, etc).
 *
 * Revision 1.2  1998/02/11 14:24:55  nikos
 * .
 *
 * Revision 1.1  1997/12/16  01:28:33  bugnion
 * stanford update (x86)
 *
 * Revision 1.5  1997/10/29  22:44:55  sean
 * added instruction counter module, option (-C) to enable it
 *
 * Revision 1.4  1997/09/14 20:11:09  yoel
 * Revised usage message, alphabetized, cleaned up the no-argument flags, added
 * -L to the usage line as well as the help message.
 *
 * Revision 1.3  1997-09-05 17:39:51-07  yoel
 * Added LockStep support and some display.
 *
 * Revision 1.2  1997/03/18 20:47:49  sean
 * added command-line option handling and fixed CreateMemory() for mmap()
 *   implementations that do not support MAP_ANON (all but Linux?)
 *
 * Revision 1.1  1997/03/18 19:12:28  sean
 * Initial revision
 *
 */

#ifndef _X86SIM_H
#define _X86SIM_H

#define X86SIM_OPTIONS  "S:P:CLd:i:cb:vh"

#define X86SIM_VERSION \
"			  .PY.  .pPYo. .oPYo.  o                \n"     \
"			  8  8  8      8                        \n"     \
"		  `o  o' .oPYo. 8oPYo. `Yooo. o8 ooYoYo.        \n"     \
"		   `bd'  8'  `8 8'  `8     `8  8 8' 8  8        \n"     \
"		   d'`b  8.  .P 8.  .P      8  8 8  8  8        \n"     \
"		  o'  `o `YooP' `YooP' `YooP'  8 8  8  8        \n"     \
"	  A Software Simulator for the Pentium(tm) Processor    \n\n"   \
"                             Version 1.0                       \n"     \
"                 Jeremy Sugerman (yoel@cs.Stanford.EDU)        \n"     \
"                  Sean Treichler (sjt@cs.Stanford.EDU)         \n"     \
"     (RCS %s)\n"

#define X86SIM_USAGE \
"Usage: %s [-S sim_stack_size] [-P program_stack_size] [-d [+-]debug flags]\n"\
"         [-i initial_instr_count] [-b initial_breakpoint] [-CLchv]\n"        \
"         <elf_program> [<args> ...]\n\n"                                     \
"        -L                     - run in lockstep on the real CPU as well\n"  \
"        -C                     - count executions of each instruction\n"     \
"        -S sim_stack_size      - sets size of simulator stack (native\n"     \
"                                   only) - may be specified in decimal,\n"   \
"                                   octal (leading 0), or hex (leading 0x)\n" \
"        -P program_stack_size  - sets size of program stack (non-native\n"   \
"                                   only) - may be specified in decimal,\n"   \
"                                   octal (leading 0), or hex (leading 0x)\n" \
"        -b initial_breakpoint  - execute until 'initial_breakpoint' is\n"    \
"                                   reached\n"                                \
"        -c                     - execute continuously (never stop)\n"        \
"        -d [+-]debug_flags     - turns on or off debugging flags (turns on\n"\
"                                   with leading + or nothing, turns off\n"   \
"                                   with leading -\n"                         \
"        -h                     - print this message\n"                       \
"        -i initial_instr_count - executes 'initial_instr_count' instructions\n"\
"                                   before stopping the first time (default\n"\
"                                   is 0)\n"                                  \
"        -v                     - show version information\n"


#ifdef NATIVE_X86_HOST
// this is where Linux stacks usually start
#define DEFAULT_LINUX_STACK_TOP    0xC0000000U

// This is the place we'll ask to put our simulator stack (Linux will move it
//  if it feels like - we don't care too much where it is
#define NEW_SIM_STACK_START        0x28000000

// This the is the default size of our simulator's stack - can be overridden 
//  with the -S option
#define NEW_SIM_STACK_SIZE         0x00100000
#else
// If we don't possess the ability to move our own stack, we'll have to 
//  allocate one for the program we simulate - this is the default size for it,
//  although it can be overridden with the -P option
#define DEFAULT_PROGRAM_STACK_SIZE 0x00100000
#endif

#endif
