/*
 * Copyright (C) 1997 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * embra_globals.h
 *
 * This file contains definitions for embra's globals. They are
 * stored in a per-thread structure.
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:01 $
 *****************************************************************/

#ifndef _EMBRA_GLOBAL_H_
#define _EMBRA_GLOBAL_H_

#include <setjmp.h>
#include <sys/time.h>
#include "cpu_state.h"
#include "embra.h"

/* Embra Global Variables */

typedef struct EmbraGlobal {
   int id;			/* thread identifier */
   EmbraState *curEmp;		/* current CPU */
   int cpuCount;		/* number of CPU's in my run queue */
   int stalledCount;		/* number of stalled CPU's */
   jmp_buf Embra_Run_Setjmp;	/* setjmp for reentry to emulator */
   int exceptionDuringBackdoor;  /* see mem_ref() */
   int tcGenNumber;		/* see mem_ref(), Em_EXCEPTION */
   long long lastBarrier;	/* DEBUGGING: last barrier time */
} EmbraGlobal;

extern EmbraGlobal EmbraGlobals[];

/* Per-emulated CPU state */
extern EmbraState* EMP;

/* This is necessary to allow the cp0 module to fool the rest of us */
/* into thinking that ASID 0 never runs */
/* It is set in cp0.c */
extern int quick_ASID[SIM_MAXCPUS];

/* Real, shared globals */
extern EmbraParams embra;		/* Simulation parameters */

extern struct timeval run_start;	/* Time when we enter TC (driver.c) */
extern struct timeval run_end;		/* Time when we exit TC (simos_interface.c */

extern int callout_interval;


#endif /* _EMBRA_GLOBAL_H_ */
