/*
 * Copyright (C) 1997 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * parallel.h
 *
 * This file contains interfaces which handle thread-level
 * parallelism in embra. Note that since we use macros (our
 * own implementation of the ANL macros) this file actually
 * needs to be run through m4 as part of the build process.
 *
 * We provide a level of indirection so that we can change
 * the parallel implementation as desired.
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:04 $
 *****************************************************************/

#ifndef _PARALLEL_H_
#define _PARALLEL_H_

#include "embra_globals.h"

#define MAX_EMBRA_THREADS 32

/*** Exported routines ***/

void InitParallel(void);			/* Initialize parallel stuff */
void CreateThread(void (*start)(void *), EmbraGlobal *global);	/* Create a thread */
void ExitThread(int cpu);
void SetGlobals (int cpu, EmbraGlobal *globals);		/* thread-local globals */
EmbraGlobal *GetGlobals(void);
void CalloutLock(int cpu, int where);			/* Acquire callout lock */
void CalloutUnlock(int cpu, int where);			/* Release callout lock */
void TCLock(int cpu, int where);		/* Acquire TC index lock */
void TCUnlock(int cpu, int where);		/* Release TC index lock */
void TransLock(int cpu, int where);		/* Acquire translator lock */
void TransUnlock(int cpu, int where);		/* Release translator lock */
void EventLock(int cpu);
void EventUnlock(int cpu);
void MMULock(void);
void MMUUnlock(void);
void PeriodicBarrier(int numProcs);		/* Wait at barrier */
void IncActiveCpuCount(void);			/* keep track of active CPU's */

extern int activeCpuCount;			/* active cpu count */
/* probably a bad name; activethreadcount? who knows... -BL */

extern int parallel_running;			/* are we parallel??!! */

#endif /* _PARALLEL_H */
