/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


/*
 *  mc_opc.h  -  Opcode definitions for the MXS compiler and simulator
 *
 *	Jim Bennett
 *	1993, 1994, 1995
 */


/*
 *  INST  -  The format of all instructions for MXS
 */

typedef struct s_reg_info {
   int reg;
   int preg;
   int fpreg;
   int flags;
} RegInfo;

#define REG_INT    0x0001

#define REG_FPS    0x0010
#define REG_FPD    0x0020
#define REG_FPW    0x0040
#define REG_FPL    0x0080

#define REG_FPU    0x00f0

#define REG_STORE  0x0100   /* Register is the source of a store */

#define MAX_SRC_REGS 3
#define MAX_DST_REGS 2

struct mxs_op_desc;

typedef struct {
   /* Opcode information structure */
   struct mxs_op_desc *opcode;
   /* Opcode source and destination registers */
   RegInfo src[MAX_SRC_REGS];
   RegInfo dst[MAX_DST_REGS];
   /* Number of destination registers */
   int     ndst;
   /* The instruction itself */
   Reg32_s instr;
   /* Branch/jump immediate target PC */
   Reg     target;
   /* Miscellaneous instruction flags (see below) */
   int     flags;
   /* Floating point exception causes set by instruction */
   int     fp_exc;
} INST;

/* 
 * Instruction flags 
 */

/* Load/store flags */
#define INST_LOAD         0x00000001
#define INST_STORE        0x00000002
#define INST_PREFETCH     0x00000004
#define INST_SC           0x00000008

/* Both load/store and system */
#define INST_CACHE        0x00000010

/* System operation flags */
#define INST_SYNC         0x00000020
#define INST_CP0          0x00000040
#define INST_ERET         0x00000080
#define INST_FPCTL        0x00000100

#define INST_TRAP         0x00000200
#define INST_BREAK        0x00000400
#define INST_SYSCALL      0x00000800
#define INST_BDOOR        0x00001000

/* Branch flags */
#define INST_BRANCH       0x00002000
#define INST_LIKELY       0x00004000
#define INST_CALL         0x00008000
#define INST_RETURN       0x00010000

#define INST_INDIRECT     0x00020000
#define INST_CONDITIONAL  0x00040000

/* Miscellaneous flags */
#define INST_ILLEGAL      0x00100000
#define INST_HALF_DEF     0x00200000
#define INST_LDHACK       0x00400000
#define INST_SDHACK       0x00800000

/* Aggregate flags for instruction groups */
#define INST_LDST         0x0000000f
#define INST_SYS          0x00001ff8

/* Handy macros for decoding CPU instructions, and	*/
/* accessing the CPU registers			*/

#define	RSRI(i) (((i)>>21)&0x1f)		/* Get register index	*/
#define RTRI(i) (((i)>>16)&0x1f)
#define RDRI(i) (((i)>>11)&0x1f)

#define FRRI(i) ((((i)>>21)&0x1f)+FPREG)	/* Get register index	*/
#define FTRI(i) ((((i)>>16)&0x1f)+FPREG)
#define FSRI(i) ((((i)>>11)&0x1f)+FPREG)
#define FDRI(i) ((((i)>> 6)&0x1f)+FPREG)
