/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


	/*
	 *  ms_cache.h  -  Header file for load/store buffer and MXS
	 *	cache simulator.
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#ifndef _SYS_TYPES_H

typedef	unsigned int	uint;
typedef	unsigned short	ushort;
typedef	unsigned char	uchar;

#endif


	/* Define the load/store buffer	*/

struct s_ldst_buffer {
   struct s_ldst_buffer *next;
   struct s_ldst_buffer *prev;
   struct s_lsq *ls;
   char *dataPtr;
   void *missTag;
   int cacheStallReason;
   char data[8];
   char validBytes[8];
   PA paddr;
   int size;
   int invalidated;
   struct s_ldst_buffer *loadhit;
};

struct	s_lsq
	{
	int	status;		/* Status bits and lstypes defined below */
	int	lstype;
	VA	addr;
	PA	paddr;
	int     preg;
        int     preg2;	/* Source reg for LWL/LWR, dst reg for SC */
        int     fpreg;  /* Hack for handling register pairs */
	int	inum;			/* Inum causing reference	*/
	struct s_ldst_buffer *ldst_ent;	/* Entry in ldst_buffer		*/
	};


	/* Load/store status bits	*/

#define	LS_ST_PEND	0x0001		/* Waiting for data to be ready	*/
#define	LS_ST_UNCACHED	0x0002		/* If is an uncached operation	*/
#define	LS_ST_CACHED	0x0004		/* If already sent to cache	*/
#define	LS_ST_CONFLICT	0x0008		/* If conflicts with earlier entry */
#define	LS_ST_DONE	0x0010		/* Operation has been completed	*/
#define	LS_ST_FAILED	0x0020		/* Cache rejected operation	*/
#define	LS_ST_STALL	0x0040		/* Cache will call back		*/
#define	LS_ST_PC_DLY	0x0080		/* Primary cache delay (on hit)	*/
#define LS_ST_BDOOR_FUNC 0x0100
#define LS_ST_BDOOR_DATA 0x0200

	/* Load/store types		*/

#define	LD_DELETED	-1		/* Invalid entry	*/

#define	LD_UNSIGNED_B	0		/* Unsigned byte	*/
#define	LD_UNSIGNED_H	1		/* Unsigned half word	*/
#define	LD_UNSIGNED_L	2		/* Unsigned left word	*/
#define	LD_UNSIGNED_R	3		/* Unsigned right word	*/

/* MIPS III */
#define	LD_UNSIGNED_W	4		/* Unsigned word	*/
#define	LD_UNSIGNED_DL	5		/* Unsigned left doubleword	*/
#define	LD_UNSIGNED_DR	6		/* Unsigned right doubleword	*/

#define	LD_INTEGER_B	7		/* Integer byte		*/
#define	LD_INTEGER_H	8		/* Integer half word	*/
#define	LD_INTEGER_W	9		/* Integer word		*/
#define	LD_INTEGER_D	10		/* Integer double		*/
#define	LD_FLOAT	11		/* Floating point	*/
#define	LD_DOUBLE	12		/* Double word (FP)	*/
#define	LD_INTEGER_LL	13		/* Integer load linked  */

/* MIPS III */
#define	LD_INTEGER_LLD	14		/* Doubleword load linked  */

#define	ST_INTEGER_B	15		/* Store byte		*/
#define	ST_INTEGER_H	16		/* Store half word	*/
#define	ST_INTEGER_W	17		/* Store word		*/
#define	ST_INTEGER_L	18		/* Store word left	*/
#define	ST_INTEGER_R	19		/* Store word right	*/

/* MIPS III */
#define	ST_INTEGER_DL	20		/* Store doubleword left	*/
#define	ST_INTEGER_DR	21		/* Store doubleword right	*/

#define	ST_INTEGER_D	22		/* Store double (int)	*/
#define	ST_FLOAT	23		/* Store floating point	*/
#define	ST_DOUBLE	24		/* Store double (FP)	*/
#define ST_INTEGER_SC	25		/* Store conditional word */

/* MIPS III */
#define ST_INTEGER_SCD	26		/* Store conditional doubleword */

#define PREFETCH        27
#define CACHE_INST      28
#define	IsStore(t)	((t)>=ST_INTEGER_B && (t)<PREFETCH)
#define   IsPrefetch(t) ((t)==PREFETCH)
#define IsCacheInst(t)  ((t)==CACHE_INST)
#define	IsLoad(t)	((t)<ST_INTEGER_B)
#define	IsDouble(t)	(((t) == LD_DOUBLE) || ((t) == ST_DOUBLE))
#define	IsSC(t)		(((t) == ST_INTEGER_SC) || ((t) == ST_INTEGER_SCD))
