/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

#ifndef MS_FPU_H
#define MS_FPU_H

#include "ms_opcodes.h"
#include "simtypes.h"

extern void   MXSDecodeC1Instruction(INST *, Reg);
extern int    MXSExecuteC1Instruction(struct s_cpu_state *, INST *, THREAD *);

/* Return values from executing */
#define C1_SUCCESS SUCCESS
#define C1_FAILURE FAILURE

static Result CompareOp(struct s_cpu_state *st, int cond, int fcc, int fpcsr);

#define ROUND_MODE(_st) ((((CPUState*)((_st)->mipsyPtr))->FCR[31]) & 0x03)

#if defined(SIM_MIPS32)
#define MXS_NFRB     0x1
#else
#define MXS_NFRB     (((CPUState*)(st->mipsyPtr))->notFRbit)
#endif

/* If we're using register pairs, we need to know which half we're accessing */
/* We encode that information in fpreg and decode it in the FPR_? access macros */
#ifdef HOST_BIG_ENDIAN
#define REG_FPREG(_reg, _preg) (((_preg) << 1) | (((_reg) & MXS_NFRB) ^ 0x1))
#else
#define REG_FPREG(_reg, _preg) (((_preg) << 1) | (((_reg) & MXS_NFRB) ^ 0x1))
#endif

#define REG_PAIR(_reg) ((_reg) ^ 0x1)
#define PREG(_fpreg) ((_fpreg) >> 1)
#define PREG_IDX(_fpreg) ((_fpreg) & 0x1)

#define FPR_W(_r1)   (st->fpr[PREG(_r1)].fgrInt[PREG_IDX(_r1)])
#define FPR_S(_r1)   (st->fpr[PREG(_r1)].fgr[PREG_IDX(_r1)])
#define FPR_L(_r1)   (st->fpr[PREG(_r1)].fgrLL)
#define FPR_D(_r1)   (st->fpr[PREG(_r1)].fpr)

#define SP_NEAR_ONE     0.999999
#define DP_NEAR_ONE     0.9999999999999

/* Floating Point format field decoding (Page B-8) */
#define FMT_S       0x10 /* Single - binary floating pt. */
#define FMT_D       0x11 /* Double - binary floating pt. */
#define FMT_W       0x14 /* Single - binary fixed pt. */
#define FMT_L       0x15 /* Double - binary fixed pt. */

#define MXS_FPE_I 0x01
#define MXS_FPE_U 0x02
#define MXS_FPE_O 0x04
#define MXS_FPE_Z 0x08
#define MXS_FPE_V 0x10
#define MXS_FPE_E 0x20

#endif
