/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

	/*
	 *  ms_inst  -  All the code for executing instructions for the
	 *		R4000 like processor
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#include <stdlib.h>
#include <math.h>
#include "ms.h"

#ifdef MIPSY_MXS
#include "cp0.h"
#include "simmisc.h"
#include "sim_error.h"
#include "sim.h"
#define EXC_FPE EXC_CODE(15)   /* Need for R4000 */
#else
typedef	int	bool;
#endif

#define	SP_NEAR_ONE	0.999999
#define	DP_NEAR_ONE	0.9999999999999

	/*
	 *  CHECKFP handles the exception that all floating point
	 *	operations can potentially cause:  the coprocessor
	 *	unusable exception.
	 */

define(CHECKFP, `
	if (!FPusable(st)) {   
	    int _inum = ip - st->iwin; 
	    st->iwin_except [_inum] = GetLastException(st); 
	    st->iwin_flags [_inum] |= IWIN_FAULT; 
	    CheckSquash (st,ip); 
	    th->stall_except = 1; 
	    UpdateStallFetch (th); 
	    return; 
	} ' )


	/*
	 *  LS_EXC - is the common code for handling exceptions in load
	 *	and store instructions.
	 */

define(LS_EXC, ` {
#ifdef MIPSY_MXS
	CPUState *P = (CPUState *)st->mipsyPtr;
	RECORD_EXCEPTION(P,$1,E_VEC, st->iwin_addr[inum], 
				P->CP0[C0_TLBHI],P->CP0[C0_CTXT],
                             P->CP0[C0_XCTXT]);
    st->iwin_except [inum] = GetLastException(st);
	st->iwin_flags[inum] |= IWIN_FAULT;
#endif
	} ' )


	/*
	 *  opBRANCH and opCALL define the processing common to branch
	 *  and call instructions.  The only difference between most of
	 *  these instructions is the condition under which the branch
	 *  is taken.  The first argument to the macro is the condition,
	 *  the second is a flag indicating if it's a floating point
	 *  branch, and the third is a flag indicating if it's a branch
	 *  likely instruction.
	 */

define(opBRANCH, ` {
	int	inum, ix;
	BrTREE	*br, *right;
	WorkDecls;

	if ($2) { CHECKFP }
	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	right = &st->branch_tree [br->rchild];

	ix = bp_pc_to_index(st->iwin_pc [inum]);
	if ($1)
		{
		st->bp_bits[ix]++;
		if (st->bp_bits[ix] > BP_MAX_VAL)
			st->bp_bits[ix] = BP_MAX_VAL;
		if (PredictTaken (right->condition, $3))
			br->resolution = PRUNE_LEFT;
		else
			br->resolution = PRUNE_RIGHT;
		st->iwin_branch_pc [inum] = ip->imm;
                st->iwin_flags[inum] |= IWIN_TAKENBR;
		if (br->lchild == -1)
			{
			th->pc = ip->imm;
			th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
			if (jmpbrk && (th->pc == jmpbrk))
			    ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
			}
		}
	else
		{
		st->bp_bits[ix]--;
		if (st->bp_bits[ix] < 0)
			st->bp_bits[ix] = 0;
		if (PredictTaken (right->condition, $3))
			br->resolution = PRUNE_RIGHT;
		else
			br->resolution = PRUNE_LEFT;
		st->iwin_branch_pc [inum] = st->iwin_pc [inum] +
						(PC_INC*(1+BRANCH_SLOTS));
		}

#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
	} ')


define(opCALL, ` {
	int	inum, ix;
	BrTREE	*br, *right;
	WorkDecls;
	
	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	right = &st->branch_tree [br->rchild];

	ix = bp_pc_to_index(st->iwin_pc [inum]);
	if ($1)
		{
		Ireg(ip->r1) = cadr_to_addr (st, th->returnpc);
		st->bp_bits[ix]++;
		if (st->bp_bits[ix] > BP_MAX_VAL)
			st->bp_bits[ix] = BP_MAX_VAL;
		if (PredictTaken (right->condition, $3))
			br->resolution = PRUNE_LEFT;
		else
			br->resolution = PRUNE_RIGHT;
		st->iwin_branch_pc [inum] = ip->imm;
#ifdef PROCSTATS
		push_stats (st, th->pc);
#endif
		if (br->lchild == -1)
			{
			th->pc = ip->imm;
			th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
			if (jmpbrk && (th->pc == jmpbrk))
			    ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
			}
		}
	else
		{
		st->bp_bits[ix]--;
		if (st->bp_bits[ix] < 0)
			st->bp_bits[ix] = 0;
		if (PredictTaken (right->condition, $3))
			br->resolution = PRUNE_RIGHT;
		else
			br->resolution = PRUNE_LEFT;
		st->iwin_branch_pc [inum] = st->iwin_pc [inum] +
						(PC_INC*(1+BRANCH_SLOTS));
		}
#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_call, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
	} ')


	/*
	 *  LoadStats  -  Common routine in loads for updating statistics
	 *		  related information.
	 */

define(LoadStats, ` if (!st->stall_issue)
#if PC_LATENCY > 1
	{ st->reg_excuse[ip->r1>>1] = ST_PC_DLY;
	  st->new_excuse[ip->r1>>1] = ST_CACHE_LSQ;
	  Add_to_worklist (st, PC_LATENCY, update_excuse, (void *)(ip->r1)); }
#else
	st->reg_excuse[ip->r1>>1] = ST_CACHE_LSQ;
	st->new_excuse[ip->r1>>1] = ST_NO_EXCUSE;
#endif
	' )


static void clear_fpc_stall(struct s_cpu_state *st, THREAD *th, int inum);
static int alls_quiet(struct s_cpu_state *st, INST *ip, int excuse);


void opILL (struct s_cpu_state *st, INST *ip, THREAD *th) {
    int inum = ip - st->iwin;
#ifdef MIPSY_MXS
    CPUState *P = (CPUState *)st->mipsyPtr;

    RECORD_EXCEPTION(P,EXC_II,E_VEC, P->CP0[C0_BADVADDR],
				P->CP0[C0_TLBHI],P->CP0[C0_CTXT],P->CP0[C0_XCTXT]);
    st->iwin_except [inum] = GetLastException(st);
    st->iwin_flags[inum] |= IWIN_FAULT;
#endif

    CheckSquash (st,ip);
    th->stall_except = 1;
    UpdateStallFetch (th);

	}

void opJ (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum;
	WorkDecls;

	th->pc = ip->imm;
#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (void *)ip);
#else
	inum = ip - st->iwin;
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
	th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
	if (jmpbrk && (th->pc == jmpbrk))
		ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
	}

void opJAL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum;
	BrTREE	*br;
	WorkDecls;

	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	Ireg(ip->r1) = cadr_to_addr (st, th->returnpc);
	br->resolution = PRUNE_LEFT;
	st->iwin_branch_pc [inum] = ip->imm;

#ifdef PROCSTATS
	push_stats (st, th->pc);
#endif
	if (br->lchild == -1)
		{
		th->pc = ip->imm;
		th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
		if (jmpbrk && (th->pc == jmpbrk))
			ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
		}

#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_call, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
	}

void opBEQ (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) == Ireg(ip->r3),0,0)

void opBNE (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) != Ireg(ip->r3),0,0)

void opBLEZ (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) <= 0,0,0)

void opBGTZ (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) > 0,0,0)

void opBLTZ (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) < 0,0,0)

void opBGEZ (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) >= 0,0,0)

void opBC1F (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH((Ireg(ip->r2) & CONDBIT) == 0,1,0)

void opBC1T (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) & CONDBIT,1,0)

void opBEQL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) == Ireg(ip->r3),0, 1)

void opBNEL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) != Ireg(ip->r3),0, 1)

void opBLEZL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) <= 0,0, 1)

void opBGTZL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) > 0,0, 1)

void opBLTZL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) < 0,0, 1)

void opBGEZL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) >= 0,0, 1)

void opBC1FL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH((Ireg(ip->r2) & CONDBIT) == 0,1, 1)

void opBC1TL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opBRANCH(Ireg(ip->r2) & CONDBIT,1, 1)

void opBLTZAL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opCALL(Ireg(ip->r2) < 0,0,0)

void opBGEZAL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opCALL(Ireg(ip->r2) >= 0,0,0)

void opBLTZALL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opCALL(Ireg(ip->r2) < 0,0,1)

void opBGEZALL (struct s_cpu_state *st, INST *ip, THREAD *th)
	opCALL(Ireg(ip->r2) >= 0,0,1)

void opADDI (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) + ip->imm;
	}

void opADDIU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ureg(ip->r1) = Ureg(ip->r2) + ip->imm;
	}

void opSLTI (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = (Ireg(ip->r2) < ip->imm ? 1 : 0);
	}

void opSLTIU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = (Ureg(ip->r2) < 
			(unsigned)ip->imm ? 1 : 0);
	}

void opANDI (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) & ip->imm;
	}

void opORI (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) | ip->imm;
	}

void opXORI (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) ^ ip->imm;
	}

void opHCPY (struct s_cpu_state *st, INST *ip, THREAD *th) {
	if (ip->op == OPHCPYC1) {
		CHECKFP()
	}
	if ((ip->r1 & (~0x01)) != ip->r3)
		{
		Dreg(ip->r1) = Dreg(ip->r3);
		}
	Ireg(ip->r1) = Ireg(ip->r2);
	}

void opLB (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;	    
	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_INTEGER_B, ip - st->iwin);
	LoadStats()
	}

void opPREF (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				PREFETCH, ip - st->iwin);
	LoadStats()
	}

void opLH (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x1) {
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_INTEGER_H, ip - st->iwin);
	LoadStats()
	}

void opLWL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum = ip - st->iwin;
	WorkDecls;

	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, ip->r3,
				LD_UNSIGNED_L, ip - st->iwin);
	LoadStats()
	}

void opLW (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x3) {	    
#ifdef R3KLOCKSUPPORT
	    extern void *lockCAShackAddr;
	    if ((ip->imm == 1) && 
		(st->iwin_pc[inum] == *(int *) lockCAShackAddr)) {
		goto out;
	    }
#endif
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}
#ifdef R3KLOCKSUPPORT
  out:
#endif
	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_INTEGER_W, ip - st->iwin);
	LoadStats()
	}

void opLBU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum = ip - st->iwin;
	WorkDecls;

	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_UNSIGNED_B, ip - st->iwin);
	LoadStats()
	}

void opLHU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x1) {
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_UNSIGNED_H, ip - st->iwin);
	LoadStats()
	}

void opLWR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;

	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, ip->r3,
				LD_UNSIGNED_R, ip - st->iwin);
	LoadStats()
	}

void opSB (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	inum = ip - st->iwin;
	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_INTEGER_B, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSH (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x1) {
	    LS_EXC(EXC_WADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_INTEGER_H, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSWL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	inum = ip - st->iwin;

	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_INTEGER_L, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSW (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x3) {
	    LS_EXC(EXC_WADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_INTEGER_W, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSWR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	inum = ip - st->iwin;

	if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_INTEGER_R, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opLWC1 (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	CHECKFP()
	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x3) {
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_FLOAT, ip - st->iwin);
	LoadStats()
	}

void opLDC1 (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	CHECKFP()
	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x7) {
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
				LD_DOUBLE, ip - st->iwin);
	LoadStats()
	}

void opSWC1 (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	CHECKFP()
	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x3) {
	    LS_EXC(EXC_WADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}


	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_FLOAT, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSDC1 (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;

	CHECKFP()	
	inum = ip - st->iwin;
	if (st->iwin_addr[inum] & 0x7) {
	    LS_EXC(EXC_WADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}

	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;
	}


	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, 0,
				ST_DOUBLE, ip - st->iwin);
	if (st->stall_issue) return;
#ifdef BREAKPOINT
	if (membrk && ((Ireg(ip->r2) + ip->imm) == membrk))
		ms_break (st, &st->iwin[st->iwin_curi], "MEMBRK");
#endif
	}

void opSLL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) << ip->imm;
	if ((ip->r1 == 0) && (ip->r2 == 0) && (ip->imm == 0))
		{
		IncStat(ST_NOP);
		}
	}

void opSRL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ureg(ip->r1) = Ureg(ip->r2) >> ip->imm;
	}

void opSRA (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) >> ip->imm;
	}

void opSLLV (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) <<
			(Ireg(ip->r3) & 0x1f);
	}

void opSRLV (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ureg(ip->r1) = Ureg(ip->r2) >>
			(Ireg(ip->r3) & 0x1f);
	}

void opSRAV (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) >>
			(Ireg(ip->r3) & 0x1f);
	}

void opJR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum, ix;
	BrTREE	*br, *right;
	WorkDecls;

	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	right = &st->branch_tree [br->rchild];

	ix = bp_pc_to_index(st->iwin_pc [inum]);
	th->pc = addr_to_cadr (st, Ireg(ip->r2));
	st->bp_targets [ix]= th->pc;
	br->resolution =
	    (right->condition == th->pc ? PRUNE_LEFT : PRUNE_RIGHT);
	st->iwin_branch_pc [inum] = Ireg(ip->r2);
#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
	th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
	if (jmpbrk && (th->pc == jmpbrk))
		ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
	}

void opJRET (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum;
	BrTREE	*br, *right;
	WorkDecls;

	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	right = &st->branch_tree [br->rchild];

	th->pc = addr_to_cadr (st, Ireg(ip->r2));
	br->resolution =
	    (right->condition == th->pc ? PRUNE_LEFT : PRUNE_RIGHT);
	st->iwin_branch_pc [inum] = Ireg(ip->r2);
#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_branch, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
#ifdef PROCSTATS
	pop_stats (st);
#endif
	th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
	if (jmpbrk && (th->pc == jmpbrk))
		ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
	}

void opJALR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum, ix;
	BrTREE	*br, *right;
	WorkDecls;

	inum = ip - st->iwin;
	br = &st->branch_tree [st->iwin_br_node[inum]];
	right = &st->branch_tree [br->rchild];

	ix = bp_pc_to_index(st->iwin_pc [inum]);
	Ireg(ip->r1) = cadr_to_addr (st, th->returnpc);
	th->pc = addr_to_cadr (st, Ireg(ip->r2));
	st->bp_targets [ix]= th->pc;
	br->resolution =
	    (right->condition == th->pc ? PRUNE_LEFT : PRUNE_RIGHT);
	st->iwin_branch_pc [inum] = Ireg(ip->r2);
#if BRANCH_LATENCY > 1
	Add_to_worklist (st, BRANCH_LATENCY, finish_call, (void *)ip);
#else
	th->stall_branch = 0;
	UpdateStallFetch (th);
	prune_branch (st, st->iwin_br_node[inum]);
#endif
#ifdef PROCSTATS
	push_stats (st, th->pc);
#endif
	th->debugpc = cadr_to_addr (st, th->pc);
#ifdef BREAKPOINT
	if (jmpbrk && (th->pc == jmpbrk))
		ms_break (st, &st->iwin[st->iwin_curi], "JMPBRK");
#endif
	}

void opSYSCALL (struct s_cpu_state *st, INST *ip, THREAD *th) {

#ifdef MIPSY_MXS
	int	inum;
	CPUState *P = (CPUState *)st->mipsyPtr;
	RECORD_EXCEPTION(P, EXC_SYSCALL, E_VEC, 0,
				P->CP0[C0_TLBHI],P->CP0[C0_CTXT],P->CP0[C0_XCTXT]);
	inum = ip - st->iwin;
	st->iwin_except [inum] = GetLastException(st);
	st->iwin_flags [inum] |= IWIN_FAULT;
	CheckSquash (st, ip);
	th->stall_except = 1;
	UpdateStallFetch (th);
#else
	if (!alls_quiet (st, ip, ST_SYSCALL))
		return;
	opsysc (st, th);
	th->stall_sys = 0;
	UpdateStallFetch (th);
#endif
	}

void opBREAK (struct s_cpu_state *st, INST *ip, THREAD *th) {
	fprintf (stderr, "Unimplemented opcode brk\n");
        CPUError("Unimplemented opcode brk\n");
	exit (1);
	}

void opADD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) + Ireg(ip->r3);
	}

void opADDU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ureg(ip->r1) = Ureg(ip->r2) + Ureg(ip->r3);
	}

void opSUB (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) - Ireg(ip->r3);
	}

void opSUBU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ureg(ip->r1) = Ureg(ip->r2) - Ureg(ip->r3);
	}

void opAND (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) & Ireg(ip->r3);
	}

void opOR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) | Ireg(ip->r3);
	}

void opXOR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = Ireg(ip->r2) ^ Ireg(ip->r3);
	}

void opNOR (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = ~(Ireg(ip->r2) | Ireg(ip->r3));
	}

void opSLT (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = (Ireg(ip->r2) < Ireg(ip->r3) ?
					1 : 0);
	}

void opSLTU (struct s_cpu_state *st, INST *ip, THREAD *th) {
	Ireg(ip->r1) = (Ureg(ip->r2) < Ureg(ip->r3) ?
					1 : 0);
	}

void opCPY (struct s_cpu_state *st, INST *ip, THREAD *th) {
	if (ip->op == OPCPYC1) {
		CHECKFP()
	}
	Ireg(ip->r1) = Ireg(ip->r2);
	}

void opCVTSW (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if ((ip->r1 & (~0x01)) != ip->r3)
		Dreg(ip->r1) = Dreg(ip->r3);

	Freg(ip->r1) = Ireg(ip->r2);
#if FPCVTSW_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVTSW_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCVTDW (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = Ireg(ip->r2);
#if FPCVTDW_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVTDW_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFADDS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Freg(ip->r1) = Freg(ip->r2) + Freg(ip->r3);
#if FPADD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPADD;
	Add_to_worklist (st, FPADD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFSUBS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Freg(ip->r1) = Freg(ip->r2) - Freg(ip->r3);
#if FPADD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPADD;
	Add_to_worklist (st, FPADD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFMULS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Freg(ip->r1) = Freg(ip->r2) * Freg(ip->r3);
#if FPMULS_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPMUL;
	Add_to_worklist (st, FPMULS_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFDIVS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r3) == 0.0)
		{
		int	inum = ip - st->iwin;
#ifdef MIPSY_MXS
		CPUState *P = (CPUState *)st->mipsyPtr;
		RECORD_EXCEPTION (P, EXC_FPE, E_VEC, 0,
				P->CP0[C0_TLBHI],P->CP0[C0_CTXT],P->CP0[C0_XCTXT]);

		st->iwin_except [inum] = GetLastException(st);
#else
		st->reg_rstat[ip->r1 >> 1].reg_status |= REG_ERROR;
#endif
		st->iwin_flags[inum] |= IWIN_FAULT;
		CheckSquash (st, ip);
		th->stall_except = 1;
		UpdateStallFetch (th);
		return;
		}
	Freg(ip->r1) = Freg(ip->r2) / Freg(ip->r3);
#if FPDIVS_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPDIV;
	Add_to_worklist (st, FPDIVS_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFSQRTS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
#ifdef MIPSY_MXS
	fprintf(stderr, "Hit a FSQRTS inst\n");
#else
 	Freg(ip->r1) = sqrtf (Freg(ip->r2));
#endif
#if FPSQRTS_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPSQRT;
	Add_to_worklist (st, FPSQRTS_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFABSS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Freg(ip->r1) = (Freg(ip->r2) >= 0 ? Freg(ip->r2) : -Freg(ip->r2) );
#if FPABS_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPABS;
	Add_to_worklist (st, FPABS_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFNEGS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Freg(ip->r1) = -Freg(ip->r2);
#if FPNEG_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPNEG;
	Add_to_worklist (st, FPNEG_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFROUNDS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = (Freg(ip->r2) > 0.0 ? Freg(ip->r2) + 0.5 :
						Freg(ip->r2) - 0.5 );
	}

void opFTRUNCS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = Freg(ip->r2);
	}

void opFCEILS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = (Freg(ip->r2) >= 0.0 ? Freg(ip->r2) + SP_NEAR_ONE :
						Freg(ip->r2) );
	}

void opFFLOORS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = (Freg(ip->r2) >= 0.0 ? Freg(ip->r2) :
					Freg(ip->r2) - SP_NEAR_ONE );
	}

void opCVTDS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = Freg(ip->r2);
#if FPCVT_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVT_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCVTWS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if ((ip->r1 & (~0x01)) != ip->r3)
		Dreg(ip->r1) = Dreg(ip->r3);

	switch (st->round_mode)
		{
		case 0:			/* RN: Round to nearest		*/
			Ireg(ip->r1) = (Freg(ip->r2) > 0.0 ?
						Freg(ip->r2) + 0.5 :
						Freg(ip->r2) - 0.5 );
			break;

		case 1:			/* RZ: Round toward 0		*/
			Ireg(ip->r1) = Freg(ip->r2);
			break;

		case 2:			/* RP: Round to +infinity	*/
			Ireg(ip->r1) = (Freg(ip->r2) >= 0.0 ?
						Freg(ip->r2) + SP_NEAR_ONE :
						Freg(ip->r2) );
			break;

		case 3:			/* RM: Round to -infinity	*/
			Ireg(ip->r1) = (Freg(ip->r2) >= 0.0 ?
						Freg(ip->r2) :
						Freg(ip->r2) - SP_NEAR_ONE );
			break;
		}

#if FPCVT_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVT_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCUNS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) &= ~CONDBIT;
	}

void opCUEQS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) == Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCULTS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) < Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCULES (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) <= Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

/* Trap if unordered, but otherwise the same as above	*/

void opCNGLES (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) &= ~CONDBIT;
	}

void opCNGLS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) == Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCNGES (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) < Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCNGTS (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Freg(ip->r2) <= Freg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFADDD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = Dreg(ip->r2) + Dreg(ip->r3);
#if FPADD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPADD;
	Add_to_worklist (st, FPADD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFSUBD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = Dreg(ip->r2) - Dreg(ip->r3);
#if FPADD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPADD;
	Add_to_worklist (st, FPADD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFMULD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = Dreg(ip->r2) * Dreg(ip->r3);
#if FPMULD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPMUL;
	Add_to_worklist (st, FPMULD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFDIVD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r3) == 0.0)
		{
		int	inum = ip - st->iwin;
#ifdef MIPSY_MXS
		CPUState *P = (CPUState *)st->mipsyPtr;
          CPUWarning("MXS:ms_inst:: EXC_FPE (division by zero) at PC=0x%x\n",
                  P->PC);
		RECORD_EXCEPTION (P, EXC_FPE, E_VEC, 0,
				P->CP0[C0_TLBHI],P->CP0[C0_CTXT],P->CP0[C0_XCTXT]);

		st->iwin_except [inum] = GetLastException(st);
#else
		st->reg_rstat[ip->r1 >> 1].reg_status |= REG_ERROR;
#endif
		st->iwin_flags[inum] |= IWIN_FAULT;
		CheckSquash (st, ip);
		th->stall_except = 1;
		UpdateStallFetch (th);

		return;
		}
	Dreg(ip->r1) = Dreg(ip->r2) / Dreg(ip->r3);
#if FPDIVD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPDIV;
	Add_to_worklist (st, FPDIVD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFSQRTD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
#ifdef MIPSY_MXS
	fprintf(stderr, "Hit a FSQRTS inst\n");
#else
	Dreg(ip->r1) = sqrt (Dreg(ip->r2));
#endif
#if FPSQRTD_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPSQRT;
	Add_to_worklist (st, FPSQRTD_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFABSD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = (Dreg(ip->r2) >= 0 ? Dreg(ip->r2) : -Dreg(ip->r2) );
#if FPABS_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPABS;
	Add_to_worklist (st, FPABS_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFMOVD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Dreg(ip->r1) = Dreg(ip->r2);
	}

void opFNEGD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	Dreg(ip->r1) = -Dreg(ip->r2);
#if FPNEG_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPNEG;
	Add_to_worklist (st, FPNEG_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opFROUNDD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = (Dreg(ip->r2) > 0.0 ? Dreg(ip->r2) + 0.5 :
						Dreg(ip->r2) - 0.5 );
	}

void opFTRUNCD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = Dreg(ip->r2);
	}

void opFCEILD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = (Dreg(ip->r2) >= 0.0 ? Dreg(ip->r2) + DP_NEAR_ONE :
						Dreg(ip->r2) );
	}

void opFFLOORD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP();
	Ireg(ip->r1) = (Dreg(ip->r2) >= 0.0 ? Dreg(ip->r2) :
					Dreg(ip->r2) - DP_NEAR_ONE );
	}

void opCVTSD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if ((ip->r1 & (~0x01)) != ip->r3)
		Dreg(ip->r1) = Dreg(ip->r3);
	Freg(ip->r1) = Dreg(ip->r2);
#if FPCVT_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVT_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCVTWD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if ((ip->r1 & (~0x01)) != ip->r3)
		Dreg(ip->r1) = Dreg(ip->r3);

	switch (st->round_mode)
		{
		case 0:			/* RN: Round to nearest		*/
			Ireg(ip->r1) = (Dreg(ip->r2) > 0.0 ?
						Dreg(ip->r2) + 0.5 :
						Dreg(ip->r2) - 0.5 );
			break;

		case 1:			/* RZ: Round toward 0		*/
			Ireg(ip->r1) = Dreg(ip->r2);
			break;

		case 2:			/* RP: Round to +infinity	*/
			Ireg(ip->r1) = (Dreg(ip->r2) >= 0.0 ?
						Dreg(ip->r2) + DP_NEAR_ONE :
						Dreg(ip->r2) );
			break;

		case 3:			/* RM: Round to -infinity	*/
			Ireg(ip->r1) = (Dreg(ip->r2) >= 0.0 ?
						Dreg(ip->r2) :
						Dreg(ip->r2) - DP_NEAR_ONE );
			break;
		}

#if FPCVT_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCVT;
	Add_to_worklist (st, FPCVT_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCUEQD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) == Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCULTD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) < Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCULED (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) <= Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

/* Trap if unordered, but otherwise the same as above	*/

void opCNGLD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) == Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCNGED (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) < Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

void opCNGTD (struct s_cpu_state *st, INST *ip, THREAD *th) {
	WorkDecls;
	CHECKFP()
	if (Dreg(ip->r2) <= Dreg(ip->r3))
		Ireg(ip->r1) |= CONDBIT;
	else
		Ireg(ip->r1) &= ~CONDBIT;
#if FPCMP_LATENCY > 1
	st->reg_excuse[ip->r1>>1] = ST_FPCMP;
	Add_to_worklist (st, FPCMP_LATENCY,
		reg_writeback, (void *)(ip->r1));
#endif
	}

			/* Set floating point unit control word		*/
void opCTL_FPC (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int	inum;

	CHECKFP()
	if (!alls_quiet (st, ip, ST_FPC))
		return;
	inum = ip - st->iwin;
	Ireg(ip->r1) = Ireg(ip->r2);
	st->regs[th->regnames[CNDREG]] = CONDBIT & Ireg(ip->r2);
	st->round_mode = Ireg(ip->r2) & 0x03;
	set_rm (st->round_mode);
	clear_fpc_stall(st,th,inum);

	}
static void clear_fpc_stall(struct s_cpu_state *st,THREAD *th, int inum) {
	BrTREE	*br;
			/* Now can start fetching new instructions	*/
	th->stall_fpc = 0;
	UpdateStallFetch (th);

			/* Unstall children, too			*/

	br = &st->branch_tree [st->iwin_br_node[inum]];
	if (br->rchild >= 0)
		{
		br = &st->branch_tree [br->rchild];
		th = &st->threads [br->thread];
		th->stall_fpc = 0;
		UpdateStallFetch (th);
		}
	}

			/* Read floating point unit control word	*/
void opCTL_CFC(struct s_cpu_state *st, INST *ip, THREAD *th) {
	CHECKFP()
	Ireg(ip->r1) = Ireg(ip->r2) | (CONDBIT & Ireg(ip->r3));
	}


			/* Control instruction: branch delay cycle	*/
void opCTL_DLY (struct s_cpu_state *st, INST *ip, THREAD *th) {
	IncStat(ST_BR_NOP);
	}

void opCP0 (struct s_cpu_state *st, INST *ip, THREAD *th) {
    uint srcreg, dstreg;
    bool hack, err;
    int inum = ip - st->iwin;

    if (ip->op == OPCP0) { 
	if (!alls_quiet (st, ip, ST_SYSCALL))
	   return;
#ifdef PRECISE
	if (st->iwin_headgrad != (ip - st->iwin)) {
   	    st->stall_type = ST_SYSCALL;
	    st->stall_issue = 1;
	    return;
	}
#endif
    }

    srcreg = (ip->r2 > 0) ? Ureg(ip->r2) : 0;

    err = DoPrivInst(st, ip->imm, srcreg, &dstreg, &hack);
    if (err) {
	st->iwin_except [inum] = GetLastException(st);
	st->iwin_flags[inum] |= IWIN_FAULT;
	CheckSquash (st, ip);
	th->stall_except = 1;
	UpdateStallFetch (th);
	return;
    }
#ifndef R3K
    if (hack) {
	CPUState *P = (CPUState *)st->mipsyPtr;
        st->iwin_branch_pc[inum] = P->PC;
        th->pc = P->PC;
    }        
#endif
    if (ip->r1 > 0) {
	Ureg(ip->r1) = dstreg;
    }
    if (ip->op == OPCP0) { 
	th->stall_cp0 = 0;	
	clear_fpc_stall(st,th, inum);
    }
}
void opBDOOR (struct s_cpu_state *st, INST *ip, THREAD *th) {

	if (!alls_quiet (st, ip, ST_SYSCALL))
		return;

#ifdef MIPSY_MXS
	{
	   /* Assume this is a call. Execute it on the real CPU. This */
	   /* assumes that all backdoor calls take less than 4 */
	   /* arguments. The prom stuff is a way for the simulated cpu to */
	   /* actually run the code in simbootprom.c */
	   CPUState *P = (CPUState *)st->mipsyPtr;
	   int64 bdoorRetval; 
           PA pAddr;
           Result ret;

           ret = TranslateVirtualNoSideeffect(P, 
                                             (VA) th->pc-4,
                                             &pAddr);
           ASSERT(ret==SUCCESS);
/* BASEM watch this! */
	   bdoorRetval = ((int64 (*)(int,int,int,int))(pAddr))
		(st->regs[th->regnames[4]], st->regs[th->regnames[5]], 
		st->regs[th->regnames[6]], st->regs[th->regnames[7]]);
	   st->regs[th->regnames[2]] = bdoorRetval >> 32;
	   st->regs[th->regnames[3]] = bdoorRetval & 0xffffffff;
	   th->pc = st->regs[th->regnames[31]];
	}
#endif
	th->stall_sys = 0;
	clear_fpc_stall(st,th, ip - st->iwin);
	}

void opLL (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;

	inum = ip - st->iwin;
	if ((st->iwin_addr[inum] & 0x3) || 
	    (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
	    LS_EXC(EXC_RADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}
	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    UpdateStallFetch (th);
	    return;	    
	}	
	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r1, 0,
					LD_INTEGER_LL, inum);
	LoadStats()
	}

void opSC (struct s_cpu_state *st, INST *ip, THREAD *th) {
	int inum;
	WorkDecls;
	
	inum = ip - st->iwin;
	if ((st->iwin_addr[inum] & 0x3) || 
	    (st->iwin_flags[inum] & IWIN_UNCACHED)) {	    
	    LS_EXC(EXC_WADE)
	} else if (st->iwin_flags[inum] & IWIN_TLBFAULT) {
	    st->iwin_flags[inum] |= IWIN_FAULT;
	}
	if (st->iwin_flags[inum] & IWIN_FAULT) {
	    CheckSquash (st,ip);
	    th->stall_except = 1;
	    th->stall_fpc = 0;
	    UpdateStallFetch (th);
	    return;
	}	

	ms_lsq (st, Ireg(ip->r2) + ip->imm, ip->r3, ip->r1,
					ST_INTEGER_SC, inum);
	LoadStats()
	}

/*
 * Hacked LD and SD instruction execution. 
 */

void opLDHACK (struct s_cpu_state *st, INST *ip, THREAD *th) {

        uint tlbFlavor;
        uint vAddr, pAddr = 0;
        int inst, rt, rt2, rs, imm, ret;
        CPUState *P = (CPUState *)st->mipsyPtr;
        uint64 val;
        int err;
        void *bdoor = 0;

        /*
         * Need to single thread the processor for these hacked 
         * isntructions. 
         */
	if (!alls_quiet (st, ip, ST_SYSCALL))
		return;

        inst = ip->imm; 

        tlbFlavor = TLB_READING;

        rs = th->regnames[(inst>>21)&0x1f];
        rt = th->regnames[(inst>>16)&0x1f];
        rt2 = th->regnames[((inst>>16)&0x1f)+1];
        imm = ((inst&0x8000)?(inst&0x0000ffff)-0x10000:(inst&0x0000ffff));

        vAddr = Ireg(rs) + imm;
     
        ret = TranslateVirtual(P, vAddr, &pAddr, &tlbFlavor, &bdoor);        
        if (ret != SUCCESS) {
#ifdef HWBCOPY
           if (tlbFlavor & TLB_MSG_SPACE) err = SUCCESS;
           else {
#else 
           {
#endif
              ms_break (st, &st->iwin[st->iwin_curi], "LDHACK");  
              err = FAILURE;
           }
        } else {
           /* 64bit load into registers rt and rt+1 */
           err = DoUncachedRead(st, vAddr, pAddr, 8, &val);
           if (err == 0) {
              Ureg(rt) = val>>32;
              Ureg(rt2) = (uint)val;
           }
        }
        if (err == 0) {
            /* Worked - clear stall and continue */
            th->stall_sys = 0;
	    clear_fpc_stall(st,th, ip - st->iwin);
        } else {
           /* Retry instruction until the memory system can take it. */
  	   st->stall_type = ST_SYSCALL;
	   st->stall_issue = 1;
	   return;
        }
	}

void opSDHACK (struct s_cpu_state *st, INST *ip, THREAD *th) {

        uint tlbFlavor;
        uint vAddr, pAddr = 0;
        int inst, rt, rt2, rs, imm, ret;
        CPUState *P = (CPUState *)st->mipsyPtr;
        uint64 val;
        int err;
        void *bdoor = 0;

        /*
         * Need to single thread the processor for these hacked 
         * isntructions. 
         */
	if (!alls_quiet (st, ip, ST_SYSCALL))
		return;

        inst = ip->imm; 
        tlbFlavor = TLB_WRITING;

        rs = th->regnames[(inst>>21)&0x1f];
        rt = th->regnames[(inst>>16)&0x1f];
        rt2 = th->regnames[((inst>>16)&0x1f)+1];
        imm = ((inst&0x8000)?(inst&0x0000ffff)-0x10000:(inst&0x0000ffff));

        vAddr = Ireg(rs) + imm;
     
        ret = TranslateVirtual(P, vAddr, &pAddr, &tlbFlavor, &bdoor);        
        if (ret != SUCCESS) {
#ifdef HWBCOPY
           if (tlbFlavor & TLB_MSG_SPACE) err = SUCCESS;
           else {
#else 
           {
#endif
              ms_break (st, &st->iwin[st->iwin_curi], "SDHACK");  
              err = FAILURE;
           }
        } else {
           /* 64bit store from registers rt and rt+1 */
           val = Ureg(rt);
           val <<= 32;
           val |= Ureg(rt2);
           err = DoUncachedWrite(st,vAddr, pAddr, 8,
                              tlbFlavor & TLB_ACCELERATED, &val);
        }
        if (err == 0) {
            /* Worked - clear stall and continue */
            th->stall_sys = 0;
	    clear_fpc_stall(st,th, ip - st->iwin);
        } else {
           /* Retry instruction until the memory system can take it. */
  	   st->stall_type = ST_SYSCALL;
	   st->stall_issue = 1;
	   return;
        }
	}


	/*
	 *  alls_quiet  -  A utility routine for instructions that
	 *			need to wait until all outstanding
	 *			instructions are completed.
	 *
	 *	Returns TRUE if everything is quiet,
	 *	otherwise returns FALSE.
	 */

static int alls_quiet(struct s_cpu_state *st, INST *ip, int excuse)
	{
	int	i;

		/* Force in order issue around this instruction	*/
		/* by stalling until all work has been done	*/
		/* and all registers have been written.		*/

	if (st->work_head) goto return_false;
	if ((st->ldst_head) || (st->ldst_nextReserved)) goto return_false;
#ifndef MIPSY_MXS
	if (ms_bus_busy ()) goto return_false;
#endif

	for (i=0; i<MAX_PREG/2; i++)
		{
		if (st->reg_rstat[i].reg_status & REG_IN_WIN)
			{
			if ((ip->r1 < 0) || (i != (ip->r1>>1)))
				goto return_false;
			}
		}
	return (1);

return_false:
	st->stall_type = excuse;
	st->stall_issue = 1;
	return (0);
	}
