/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/****************************************************************
 * opcodes.h
 * 
 * $Author: verghese $
 * $Date: 1999/09/27 22:08:25 $
 *****************************************************************/

#ifndef MS_OPCODES_H
#define MS_OPCODES_H

typedef enum { 
  MXS_NO_UNIT, MXS_INT_UNIT, MXS_BRANCH_UNIT, 
  MXS_LDST_UNIT, MXS_FP_UNIT, MXS_NUM_UNITS
} MXSFuncUnit;

Result bdoorEX(struct s_cpu_state *, INST *, THREAD *);

typedef Result (*MXSOpcodeEXFPtr)(struct s_cpu_state *, INST *, THREAD *);
typedef void   (*MXSOpcodeIDFPtr)(INST *, Reg);

typedef struct mxs_op_desc {
  char            *opname;
  MXSOpcodeIDFPtr  decode;
  MXSOpcodeEXFPtr  execute;
  uint             num;
  MXSFuncUnit      unit;
  uint             latency;
  uint             repeat;
} MXSOpcodeDesc;

/* Routines needed by both ms_opcodes.c and ms_fpu.c */
void BranchOp(struct s_cpu_state *st, INST *ip,
              THREAD *th, int cond, int likely);
int  AllsQuiet(struct s_cpu_state *st, INST *ip, int excuse);
void ClearFPCStall(struct s_cpu_state *st, THREAD *th, int inum);

extern MXSOpcodeDesc mxsOpcodes[];

/*
 * EX_RECORD_EXCEPTION -- common code for recording an exception
 *                        occurring during instruction execution
 */
#define EX_RECORD_EXCEPTION(_st, _inum, _exc, _addr)            \
{                                                               \
   CPUState *_P = (CPUState *)(_st)->mipsyPtr;                  \
   RECORD_EXCEPTION(_P, _exc, E_VEC, _addr, _P->CP0[C0_TLBHI],  \
                    _P->CP0[C0_CTXT], _P->CP0[C0_XCTXT]);       \
   (_st)->iwin_except[_inum] = GetLastException(_st);           \
   (_st)->iwin_flags[_inum] |= IWIN_FAULT;                      \
}

/*
 * EX_HANDLE_EXCEPTION -- common code for executing a faulting instruction
 */
#define EX_HANDLE_EXCEPTION(_st, _ip, _th)      \
{                                               \
   CheckSquash(_st, _ip);                       \
   (_th)->stall_except = 1;                     \
   UpdateStallFetch(_th);                       \
}

/*
 * ENABLE_64BIT() - Check to see if 64bit instructions are permitted
 */
#if defined(SIM_MIPS64)
#define ENABLE_64BIT(_st)                              \
   (!(((CPUState*)(_st->mipsyPtr))->is32bitMode) ||     \
    IS_KERNEL_MODE((CPUState*)(_st)->mipsyPtr))
#else
#define ENABLE_64BIT(_st) 0
#endif

/*
 * ENABLE_FPU() - Check to see if FP instructions are permitted
 */
#define ENABLE_FPU(_st)                                         \
      ((((CPUState*)(_st)->mipsyPtr)->CP0[C0_SR]) & SR_CU1)

/*
 * ENABLE_MIPS4() - Check to see if MIPS IV opcodes are permitted
 */
#if defined(SIM_MIPS64)
#define ENABLE_MIPS4(_st)                                         \
      (((((CPUState*)(_st)->mipsyPtr)->CP0[C0_SR]) & SR_CU3) ||   \
       IS_KERNEL_MODE((CPUState*)(_st)->mipsyPtr))
#else
#define ENABLE_MIPS4(_st) 0
#endif

#endif
