/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */



	/*
	 *  param_ss1pl  -  Parameters file for superscalar model 1,
	 *			with precise interrupts.
	 *
	 *	This models a superscalar processor with a short pipeline.
	 *	Cache accesses are single cycle, and FP operations take
	 *	generally 2 to 3 cycles.
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#ifndef MS_PARAM_H
#define	MS_PARAM_H

/*
 *  Number and types of registers in the CPU
 */

/* HI, LO, and FP control regs are appended to the end of the GPRs */
#define MXS_REG_HI 32
#define MXS_REG_LO 33
#define MXS_REG_FPCSR 34        /* Control/status register */
#define MXS_REG_FPIRR 35        /* Implementation/revision register */

#define	N_IREG	36		/* Number of integer registers		*/
#define	N_FREG	32		/* Number of floating point registers	*/
				/* (in single precision)		*/
#define	FPREG	N_IREG		/* Offset to FP registers		*/
#define	TOT_REG	(N_IREG+N_FREG)	/* Total number of registers	*/

/*
 *  Instruction window size,
 *  load/store buffer size, and the nature of the instruction
 *  reordering.
 */

extern int mxsIwinSize;
extern int mxsLdstBufferSize;
extern int mxsLdstBufferMask;

#define MAX_IWIN_SIZE   64
#define	IWIN_SIZE	(mxsIwinSize)	/* Instruction window size		*/
#define	LOAD_BYPASS	1	/* Bypass values from stores to loads	*/
#define	PRECISE		1	/* Support precise interrupts		*/

#define MAX_LDST_BUFFER_SIZE    32
#define	LDST_BUFFER_SIZE	(mxsLdstBufferSize)
#define	LDST_BUFFER_MASK	(mxsLdstBufferMask)


/*
 *  Bandwidth parameters
 */

extern int mxsFetchWidth;
extern int mxsIssueWidth;
extern int mxsCacheWidth;
extern int mxsWbackWidth;
extern int mxsGradWidth;
extern int mxsThreadWidth;

#define	FETCH_WIDTH	 (mxsFetchWidth)
#define MAX_ISSUE_WIDTH  4
#define	ISSUE_WIDTH	 (mxsIssueWidth)
#define	CACHE_WIDTH	 (mxsCacheWidth)
#define	WBACK_WIDTH	 (mxsWbackWidth)
#define	GRAD_WIDTH	 (mxsGradWidth)
/* Max number of threads to follow - Supports
 * up to four predicted branch and one active one. */
#define MAX_THREAD_WIDTH 5
#define	THREAD_WIDTH	 (mxsThreadWidth)

#define	MAX_ACT_THREADS	 1	/* (during speculative execution)	*/

/*
 *  Branch prediction parameters
 */

#define	BP_TABLE_SIZE    	1024
#define	BP_RETURN_STACK		 32

#define	BRANCH_LIKELY	1	/* Support branch likely instructions	*/


	/*
	 *  Latencies of the various operations of the functional units
	 */

#ifdef ONE_PHASE_LS
#define	PC_LATENCY	 2	/* Latency of primary data cache access	*/
#else
#define	PC_LATENCY	 1
#endif

#define	BRANCH_SLOTS	 1	/* Number of branch delay slots		*/
#define	BRANCH_LATENCY	 1	/* Latency of branch instructions	*/

#define	TAKEN_LATENCY	 1	/* Latency for taken branch prediction	*/
#define	FALLTHRU_LATENCY 1	/* Latency for fall-thru branch pred.	*/

#endif /* MS_PARAM_H */
