/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */

/*****************************************************************
 * simmisc.c
 *
 * Miscellaneous SimOS stuff which had no better place to go.
 * Dan Teodosiu, 07/96
 *
 *****************************************************************/
#include <sys/signal.h>
#ifndef i386
#include <sys/sysinfo.h>
#endif
#include <sys/times.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>

#include "syslimits.h"
#include "sim.h"
#include "machine_params.h"
#include "simstats.h"
#include "cpu_interface.h"
#include "cpu_state.h"
#include "sim_error.h"
#include "cpu_interface.h"
#include "registry.h"
#include "expect.h"
#include "file_formats.h"
#include "console.h"
#include "simmisc.h"
#include "machine_defs.h"
#include "../../caches/memref.h"
#include "../../cpus/mipsy/mipsy_simos.h"
#include "../../cpus/embra/embra_interface.h"

#include "linux_init.h"

#define MAX_SCNS 48


#ifdef IRIX6_4
#define PromWord Reg64
#else
#define PromWord Reg32
#endif


/* Storage for sim_misc catch-all structure */
SimMisc sim_misc;

/* Boot prom file name */
char *FPromFile;

/* Shared array of CPUStates */
extern CPUState* SBase;

#define CPUlaunchRoutine(i) (sim_misc.launchAddr[i])
#define CPUlaunched(i)  (CPUlaunchRoutine(i) != 0)

static int LoadFPROMImage(char* pathname, VA promAddr,  char * promSimAddr,
                          VA ramAddr, char * ramSimAddr);

static int LoadCoffHeaders(char *pathname, 
			   FILHDR *fileHeader, 
			   AOUTHDR *aoutHeader, 
			   SCNHDR secHeaders[MAX_SCNS], 
			   int *fd, VA *entry);

static int LoadElf32Headers(char *pathname,
                 Elf32_Ehdr *ehdr,
                 FILHDR *fileHeader,
                 AOUTHDR *aoutHeader,
                 SCNHDR secHeaders[MAX_SCNS],
                 int *fd, VA *entry);

static int LoadElf64Headers(char *pathname,
                 Elf64_Ehdr *ehdr,
                 FILHDR *fileHeader,
                 AOUTHDR *aoutHeader,
                 SCNHDR secHeaders[MAX_SCNS],
                 int *fd, VA* entry);

static void InitMemLayout(void);



/*****************************************************************
 * SimulatorEnter - 
 * This guy is called on the initial switch into a detailed cpu
 * simulator. All switches after this go through SimulatorSwitch.
 *****************************************************************/

static struct {
   jmp_buf jmpEnv;
   int defined;
} simulatorSwitch;

void
SimulatorEnter(CPUType simulator, int restoreCkpt, int swtch)
{
  int cpu;

  sim_misc.enterThisCPU = simulator;

  for (cpu = 0; cpu < TOTAL_CPUS; cpu++) {
    /* Print where we are */
    CPUPrint("CPU %d is at 0x%llx %s\r\n", cpu,
	     (Reg64)SBase[cpu].PC, simName[simulator]);
    if ((swtch || restoreCkpt) && (SBase[cpu].cpuStatus == cpu_not_booted)) {
       /* don't change state for failed cpus */
    } else if ((!CPUlaunched(cpu)) && (!restoreCkpt) && (!swtch) ) {
      SBase[cpu].outOfSlaveLoop = 0;
      SBase[cpu].cpuStatus      = cpu_not_booted;
    } else {
      CPUWarning("SimulatorEnter: Launching CPU %d\n", cpu);
      SBase[cpu].outOfSlaveLoop = 1;
      SBase[cpu].cpuStatus      = cpu_running;
    } 
    sim_misc.myCPUType = simulator;
  }
  simosCPUType = simulator;
  
  switch (simulator) {
  case NO_CPU: 
     exit(0);
     break;
  case EMBRA_PAGE:
  case EMBRA_CACHE:
#ifndef NO_EMBRA
    EmbraEnter(swtch);
#else
    CPUError("Can't start Embra - simulator compiled -DNO_EMBRA\n");
#endif
    break;
    
  case MIPSY:
    MipsyEnter(swtch);
    break; 
    
  default:
    ASSERT(0);
    break;
  } 
}
  
/*****************************************************************
 * SimulatorSwitch
 * 
 * For now, the simulators just call this function to switch to each
 * other... if we get into serious sampling, we can't have the stack
 * grow that much.
 *****************************************************************/




void 
SimulatorSwitch(CPUType oldSimulator, CPUType newSimulator)
{

   if (newSimulator != NO_CPU) {
      Sim_Warning("SimOS: SWITCHING from %s to %s\n",
                  simName[oldSimulator], simName[newSimulator]);
   } else {
      Sim_Warning("SimOS: exit\n");
   }

  ASSERT(sim_misc.myCPUType == oldSimulator);


  sim_misc.myCPUType = newSimulator;
  simosCPUType = newSimulator;

  /* XXX Blow away call stack */
  if (!simulatorSwitch.defined) {
     setjmp(simulatorSwitch.jmpEnv);
     simulatorSwitch.defined = 1;
  } else { 
     longjmp(simulatorSwitch.jmpEnv,1);
  }

  /* becasue cpu switch causes cyclecount to drop to 0, need to adjust callbacks */
  /* xxx AdjustEventCallbackTime(); */
  SimulatorEnter(sim_misc.myCPUType, 0, 1);

}

/* ****************************************************************
 * Restrictions on Printf:
 * Only the first argument (s) can be a string pointer!
 * Additional arguments must be integers (no K0_TO_MEMADDR)
 * At most 3 integer  arguments (unless if you use int64s)
 * ****************************************************************/

void Printf(char *s,...)
{
#ifndef __alpha
    char buf[1024];
    va_list ap;
    int cpu = CPUVec.CurrentCpuNum();

    va_start(ap, s);
    /* we do not support %s */
    vsprintf(buf, K0_TO_MEMADDR(M_FROM_CPU(cpu), s), ap);
    LogEntry("printf", cpu, "%s\n", buf);
    CPUPrint(buf);
    va_end(ap);
#endif

}



/*****************************************************************
 * Initialize (some of the) values in sim_misc.
 *****************************************************************/
void
SimMiscInit(int debug_intr)
{
  int i;

  sim_misc.enterThisCPU = EMBRA_PAGE;

  /* load the FProm if specified */
  if (strlen(FPromFile) != 0) {

     /* they have to be at least a page to copy over the FRAM cache
	error handler */
     ASSERT(FPROM_SIZE > SIM_PAGE_SIZE);
     ASSERT(FRAM_SIZE > SIM_PAGE_SIZE);

     /* allocate space to hold the fprom and fram.
      * NOTE: we only need one copy of the fprom.
      */
     sim_misc.fprom[0] = malloc(FPROM_SIZE);
     ASSERT(sim_misc.fprom);
     for (i = 0; i < TOTAL_CPUS; i++) {
       sim_misc.fram[i] = calloc(1, FRAM_SIZE);
       ASSERT(sim_misc.fram[i]);
       sim_misc.fprom[i] = sim_misc.fprom[0]; /* share copy */
     }

     /* load fprom image */
     if (LoadFPROMImage(FPromFile,FPROM_BASE, sim_misc.fprom[0],
			FRAM_BASE + SIM_PAGE_SIZE,
			sim_misc.fram[0] + SIM_PAGE_SIZE))
       CPUError("LoadFPROMImage failed for FPROM (%s)\n", FPromFile);

     /* copy the first page of fprom to fram to install
	the cache err handler */
     memcpy(sim_misc.fram[0], sim_misc.fprom[0], SIM_PAGE_SIZE);

     /* duplicate cpu 0's fram to other cpus */
     for (i = 1; i < TOTAL_CPUS; i++) {
        memcpy(sim_misc.fram[i], sim_misc.fram[0], FRAM_SIZE);
     }

     /* note: fprom is already shared by all cpus */
  } else {
     for (i = 0; i < TOTAL_CPUS; i++) {
        sim_misc.fram[i] = NULL;
	sim_misc.fprom[i] = NULL;
     }
  }

  /* array with 1 bit per page.  0 => coherent.  1 => incoherent */
  sim_misc.incoherentPages = calloc(1,MEM_SIZE(0)/SIM_PAGE_SIZE/sizeof(char) + 1);

  sim_misc.debug_intr = debug_intr;

#ifdef IRIX6_4
  InitMemLayout();
#endif
}

/*****************************************************************
 *
 * Support for loading and starting a kernel image.
 *
 *****************************************************************/



static int
LoadHeaders(char      *pathname, 
	    HeaderBuf *header,
	    FILHDR    *fileHeader, 
	    bool      *elfFormat,
	    AOUTHDR   *aoutHeader, 
	    SCNHDR    secHeaders[MAX_SCNS], 
	    int       *fd,
            VA        *entry)
{
  ssize_t retval;


  if ((*fd = open(pathname, O_RDONLY, 0)) < 0) {
    Sim_Warning("Problem opening kernel named %s\n", pathname);
    retval = errno;
    goto error;
  }

  retval = read(*fd,(char *)header, sizeof(HeaderBuf));
  if ((retval != sizeof(HeaderBuf))) {
    Sim_Warning("LoadHeaders:  Bad file header\n");
    retval = ENOEXEC;
    goto error;
  }

  if (!IS_ELF(header->hb.elfHeader)) {
    *elfFormat = FALSE;
    return (LoadCoffHeaders(pathname, &(header->hb.coffHeader), aoutHeader, 
			    secHeaders, fd,entry));
  } else {
    *elfFormat = TRUE;
    if (header->hb.elfHeader.e_ident[EI_CLASS] == ELFCLASS32) {
       return (LoadElf32Headers(pathname, &(header->hb.elfHeader), fileHeader, 
			     aoutHeader, secHeaders, fd,entry ));
    } else if (header->hb.elfHeader.e_ident[EI_CLASS] == ELFCLASS64) {
       return (LoadElf64Headers(pathname, &(header->hb.elf64Header), fileHeader, 
			     aoutHeader, secHeaders, fd,entry ));
    } else {
     Sim_Warning("Unknown elf class %d\n",header->hb.elfHeader.e_ident[EI_CLASS]);
     retval = ENOEXEC;
    }
  }

 error:
  (void) close(*fd);
  return (int) retval;

}

  


/*****************************************************************
 *
 * Support for loading and starting a linux kernel image
 * (assumes elf format)
 *
 *****************************************************************/

#define MAX_STRINGLEN 512
static int
LoadElf32Headers(char *pathname, 
		 Elf32_Ehdr *ehdr,
		 FILHDR *fileHeader, 
		 AOUTHDR *aoutHeader, 
		 SCNHDR secHeaders[MAX_SCNS], 
		 int *fd,
                 VA *entry)
{

  ssize_t retval;
  int i;
  char strs[MAX_STRINGLEN];
  char *secName;
  Elf32_Shdr shdr[MAX_SCNS];
  Elf32_Shdr *strSecHdr;

  if (BE2HO_2(ehdr->e_shnum) >= MAX_SCNS) {
    Sim_Warning("Not enough section headers in Elf32LoadHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  fileHeader->f_nscns = BE2HO_2(ehdr->e_shnum);

  *entry = BE2HO_2(ehdr->e_entry);

  if (lseek(*fd, BE2HO_4(ehdr->e_shoff), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to section headers in ElfLoadHeaders\n");
     goto error;
  }
  retval = read(*fd,(char *)shdr, (BE2HO_2(ehdr->e_shnum)*sizeof(Elf32_Shdr)));
  if (retval < 0) {
    retval = errno;
    Sim_Warning("Could not read section headers in Elf32LoadHeaders\n");
    goto error;
  }
  if (retval != sizeof(Elf32_Shdr) * BE2HO_2(ehdr->e_shnum)) {
    Sim_Warning("Read too few section headers in Elf32LoadHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  strSecHdr = &shdr[BE2HO_2(ehdr->e_shstrndx)];

  if (lseek(*fd, BE2HO_2(strSecHdr->sh_offset), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to strings in ElfLoadHeaders\n");
     goto error;
  }

  if ( BE2HO_2(strSecHdr->sh_size) >= sizeof(strs)) {
     Sim_Warning("String table too large if ElfLoadHeaders\n");
     retval = ENOEXEC;
     goto error;
  }
  retval = read(*fd,(char*)strs, BE2HO_2(strSecHdr->sh_size));
  if (retval !=  BE2HO_2(strSecHdr->sh_size)){
     if (retval < 0) { 
        retval = errno;
        Sim_Warning("Could not reads strings in ElfLoadHeaders\n");
     } else {
        retval = ENOEXEC;
        Sim_Warning("Read too few strings in Elf32LoadHeaders\n");
     }
     goto error;
  }     
  
  for (i=0; i<BE2HO_2(ehdr->e_shnum); i++) {
    secName = strs + BE2HO_4(shdr[i].sh_name);
    memcpy(secHeaders[i].s_name,secName,8);
    secHeaders[i].s_vaddr = BE2HO_4(shdr[i].sh_addr);
    secHeaders[i].s_size = BE2HO_2(shdr[i].sh_size);
    secHeaders[i].s_scnptr = BE2HO_2(shdr[i].sh_offset);
  }
  
  return 0;

 error:
  (void) close(*fd);
  return (int) retval;


}

static int
LoadElf64Headers(char *pathname, 
		 Elf64_Ehdr *ehdr,
		 FILHDR *fileHeader, 
		 AOUTHDR *aoutHeader, 
		 SCNHDR secHeaders[MAX_SCNS], 
		 int *fd,
                 VA *entry)
{

  ssize_t retval;
  int i;
  char strs[MAX_STRINGLEN];
  char *secName;
  Elf64_Shdr shdr[MAX_SCNS];
  Elf64_Shdr *strSecHdr;


  if (BE2HO_2(ehdr->e_shnum) >= MAX_SCNS) {
    CPUError("Not enough section headers in Elf64LoadHeaders %d>%d\n",
            BE2HO_2(ehdr->e_shnum),MAX_SCNS  );
    retval = ENOEXEC;
    goto error;
  }

  fileHeader->f_nscns = BE2HO_2(ehdr->e_shnum);

#if defined(TORNADO)
  *entry = BE2HO_8(ehdr->e_entry);
#else
  if (IS_XKPHYS(BE2HO_8(ehdr->e_entry))) {
     *entry = XKPHYS_TO_CKSEG0(BE2HO_8(ehdr->e_entry));
  } else if (IS_XKSEG(BE2HO_8(ehdr->e_entry))) {
     *entry = XKPHYS_FOUR_PAGE_OFFSET(BE2HO_8(ehdr->e_entry)) + CKSEG0_START_ADDR;
  } else {
     /* Some of the entry addresses, e.g., that of the boot prom, are
        not in xkphys. */
     *entry = BE2HO_8(ehdr->e_entry);
  }
#endif /* TORNADO */

  if (lseek(*fd, (off_t)BE2HO_8(ehdr->e_shoff), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to section headers in Elf64LoadHeaders\n");
     goto error;
  }
  retval = read(*fd,(char *)shdr, (BE2HO_2(ehdr->e_shnum)*sizeof(Elf64_Shdr)));
  if (retval < 0) {
    retval = errno;
    Sim_Warning("Could not read section headers in Elf64LoadHeaders\n");
    goto error;
  }
  if (retval != sizeof(Elf64_Shdr) * BE2HO_2(ehdr->e_shnum)) {
    Sim_Warning("Read too few section headers in Elf32LoadHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  strSecHdr = &shdr[BE2HO_2(ehdr->e_shstrndx)];

  if (lseek(*fd, BE2HO_8(strSecHdr->sh_offset), SEEK_SET) < 0) {
     retval = errno;
     Sim_Warning("Could not lseek to strings in Elf64LoadHeaders\n");
     goto error;
  }

  if ( BE2HO_8(strSecHdr->sh_size) >= sizeof(strs)) {
     Sim_Warning("String table too large if Elf64LoadHeaders\n");
     retval = ENOEXEC;
     goto error;
  }
  retval = read(*fd,(char*)strs, BE2HO_8(strSecHdr->sh_size));
  if (retval !=  BE2HO_8(strSecHdr->sh_size)){
     if (retval < 0) { 
        retval = errno;
        Sim_Warning("Could not reads strings in Elf64LoadHeaders\n");
     } else {
        retval = ENOEXEC;
        Sim_Warning("Read too few strings in Elf64LoadHeaders\n");
     }
     goto error;
  }     
  
  for (i=0; i<BE2HO_2(ehdr->e_shnum); i++) {
    secName = strs + BE2HO_4(shdr[i].sh_name);
    memcpy(secHeaders[i].s_name,secName,8);
#if defined(TORNADO) 
    secHeaders[i].s_vaddr = BE2HO_8(shdr[i].sh_addr);
#else
    if (IS_XKSEG(BE2HO_8(shdr[i].sh_addr))) {
       /* IRIX 6.4 is linked to mapped XKSEG but the kernel seems to 
        * assume the prom just reads the kernel into low 16MB. */
       secHeaders[i].s_vaddr =  XKPHYS_TO_CKSEG0(
            (BE2HO_8(shdr[i].sh_addr)-XKSEG_START_ADDR) % (16*1024*1024));
    } else if (IS_XKPHYS(BE2HO_8(shdr[i].sh_addr))) {
       secHeaders[i].s_vaddr = XKPHYS_TO_CKSEG0(BE2HO_8(shdr[i].sh_addr));
    } else if (BE2HO_8(shdr[i].sh_addr) == 0) {
       /* IRIX6.4 has some sections that are not used in loading with addr == 0,
        and they need to be given a valid s_vaddr for later on */
       secHeaders[i].s_vaddr = XKPHYS_TO_CKSEG0(BE2HO_8(shdr[i].sh_addr));
    } else {
       /* Some of the entry addresses, e.g., that of the boot prom, are
          not in xkphys. */
      secHeaders[i].s_vaddr = BE2HO_8(shdr[i].sh_addr);
    }
#endif
    secHeaders[i].s_size = BE2HO_8(shdr[i].sh_size);
    secHeaders[i].s_scnptr = BE2HO_8(shdr[i].sh_offset);
  }
  
  return 0;

 error:
  (void) close(*fd);
  return (int) retval;


}

/*****************************************************************
 *
 * Support for loading and starting a kernel image.
 *
 *****************************************************************/

static int
LoadCoffHeaders(char *pathname, 
		FILHDR *fileHeader, 
		AOUTHDR *aoutHeader, 
		SCNHDR secHeaders[MAX_SCNS], 
		int *fd, VA *entry)
{
  ssize_t retval;

  if (!IS_MIPSEBMAGIC(BE2HO_2(fileHeader->f_magic))) {
    Sim_Warning("Bad file header in LoadCoffHeaders\n");
    retval = ENOEXEC;
    goto error;
  }
  
  lseek(*fd, sizeof(FILHDR), SEEK_SET);
  retval = read(*fd,(char *)aoutHeader, sizeof(AOUTHDR));
  if (retval < 0) {
    retval = errno;
    Sim_Warning("Bad header read in LoadCoffHeaders\n");
    goto error;
  }
  
  if (BE2HO_2(aoutHeader->magic) != OMAGIC) {
    Sim_Warning("Bad header magic number in LoadCoffHeaders\n");
    retval = ENOEXEC;
    goto error;
  }
  
  if (BE2HO_2(fileHeader->f_nscns) >= MAX_SCNS) {
    Sim_Warning("Not enough section headers in LoadCoffHeaders\n");
    retval = ENOEXEC;
    goto error;
  }
  
  retval = read(*fd,(char *)secHeaders, sizeof(SCNHDR) * 
                BE2HO_2(fileHeader->f_nscns));
  if (retval < 0) {
    retval = errno;
    Sim_Warning("Could not read section headers in LoadCoffHeaders\n");
    goto error;
  }
  if (retval != sizeof(SCNHDR) * BE2HO_2(fileHeader->f_nscns)) {
    Sim_Warning("Read too few section headers in LoadCoffHeaders\n");
    retval = ENOEXEC;
    goto error;
  }

  *entry = (VA)_SEXT(BE2HO_4(aoutHeader->entry));

  return 0;

error:
  (void) close(*fd);
  return (int) retval;
}


/*
 * This function allows loading a boot image directly from the disk of
 * the host system.
 * Return value is 0 if success, errno if failure.
 */

#define XFER_BUF_SIZE 4096

static void
write_consistency(int machine, VA startvaddr, int nbytes)
{
  if (CPUVec.CycleCount && (CPUVec.CycleCount(0) > 0)) {
    /* system has already started running.  Need to ensure these
     * writes are consistent w.r.t. the simulated caches (eg. mipsy
     * with data handling) and w.r.t. the cached state in simos (eg.
     * embra TC). Can only do page at a time because that is the limit
     * for the putmemory functions.
     */
    while (nbytes > XFER_BUF_SIZE) {
      Result res = CPUVec.PutMemory(0, startvaddr, XFER_BUF_SIZE, 
				    K0_TO_MEMADDR(machine, startvaddr));
      ASSERT(res == SUCCESS);
      nbytes     -= XFER_BUF_SIZE;
      startvaddr += XFER_BUF_SIZE;
    }
    if (nbytes > 0) {
      Result res = CPUVec.PutMemory(0, startvaddr, nbytes, 
				    K0_TO_MEMADDR(machine, startvaddr));
      ASSERT(res == SUCCESS);
    }
  }
}

typedef enum {
  S_TEXT = 0,
  S_DATA,     
  S_RODATA,
  S_BSS,
  S_GOT,
  S_SZ
} SegIx;

typedef struct SegInfo {
  VA  segStart;
  int segSize;
} SegInfo;

#define IN_SEG(_x, _s) (si[_s].segStart <= (_x) &&                 \
			(_x) < (si[_s].segStart + si[_s].segSize))

static int
LoadImage(char* pathname,   /* boot image filename */
          int   machine,    /* machine */
	  VA    img_addr,   /* image start address */
	  VA    ld_addr,    /* load start addres (if diff than above, reloc) */
	  int   nbytes,     /* image length (max.) */
	  bool  reloc_data, /* relocate data segments? */
	  bool* elfFormat,  /* out: ELF file? */
	  VA*   entry       /* out: entry point */
	  )
{
  int       fd, i;
  int       retval;
  HeaderBuf header;
  FILHDR    fileHeader;
  AOUTHDR   aoutHeader;
  SCNHDR    secHeaders[MAX_SCNS];
  FILHDR*   hdrPtr;

  SegInfo   si[S_SZ]; /* info on segments */

  for (i = 0; i < S_SZ; i++) si[i].segSize = 0;

  retval = LoadHeaders(pathname, &header, &fileHeader, elfFormat, 
		       &aoutHeader, secHeaders, &fd, entry);
  if (retval) return retval;

  if (!*elfFormat && img_addr != ld_addr) {
    CPUError("LoadImage: cannot relocate COFF\n");
    retval = EFAULT;
    goto error;
  }

  /* relocate entry point, if necessary */
  *entry += (ld_addr - img_addr);

  hdrPtr = (*elfFormat) ? &fileHeader : &(header.hb.coffHeader);
  ASSERT( BE2HO_2(hdrPtr->f_nscns) < MAX_SCNS);

  for (i = 0; i < BE2HO_2(hdrPtr->f_nscns); i++ ) {

    VA  sec_addr_k0  = (VA)_SEXT(BE2HO_4(secHeaders[i].s_vaddr));
    VA  read_addr_k0 = ld_addr + (sec_addr_k0 - img_addr);
    int size         = BE2HO_4(secHeaders[i].s_size);
    MA  read_addr_ma = K0_TO_MEMADDR(machine, read_addr_k0);
    
    if (!strcmp(secHeaders[i].s_name, ".text")) {
      si[S_TEXT].segStart = sec_addr_k0;   si[S_TEXT].segSize = size;
    } else if (!strcmp(secHeaders[i].s_name, ".data")) {
      si[S_DATA].segStart = sec_addr_k0;   si[S_DATA].segSize = size;
    } else if (!strcmp(secHeaders[i].s_name, ".rodata")) {
      si[S_RODATA].segStart = sec_addr_k0; si[S_RODATA].segSize = size;
    } else if (!strcmp(secHeaders[i].s_name, ".bss")) {
      si[S_BSS].segStart = sec_addr_k0;    si[S_BSS].segSize = size;
    } else if (!strcmp(secHeaders[i].s_name, ".got")) {
      si[S_GOT].segStart = sec_addr_k0;    si[S_GOT].segSize = size;
    }

    if ((strcmp(secHeaders[i].s_name, ".text")  == 0)  ||
	(strcmp(secHeaders[i].s_name, ".data")  == 0)  ||
	(strcmp(secHeaders[i].s_name, ".sdata") == 0)  ||
        (strcmp(secHeaders[i].s_name, ".rodata") == 0) ||
        (strcmp(secHeaders[i].s_name, ".srdata") == 0) ||
        (strcmp(secHeaders[i].s_name, ".lit8") == 0)   ||
	(strcmp(secHeaders[i].s_name, ".rdata") == 0)  ||
        (strcmp(secHeaders[i].s_name, ".got") == 0) ) {
      
      if (sec_addr_k0 < img_addr || (sec_addr_k0+size) >= img_addr+nbytes) {
	CPUError("LoadImage: section %s (va %llx nbytes %x)\r\n"
		 "           outside limit (va %x nbytes %x)\n",
		 secHeaders[i].s_name, (uint64)secHeaders[i].s_vaddr,
		 size, img_addr, nbytes);
	retval = EFAULT;
	goto error;
      }
      
      if (lseek(fd, BE2HO_4(secHeaders[i].s_scnptr), SEEK_SET) !=
	  BE2HO_4(secHeaders[i].s_scnptr)) {
	retval = errno;
	CPUError("Bad section seek in LoadImage\n");
	goto error;
      }
      
      if (strcmp(secHeaders[i].s_name, ".text") != 0 &&
	  img_addr != ld_addr && !reloc_data)
	/* don't even load data in the case of replicated text
	 * segments sharing the same data.
	 */
	continue;
      
      if (read(fd, read_addr_ma, size) != size) {
	retval = errno;
	CPUError("Bad section read in LoadImage - "
		 "size = %d, hold = %#x errno %d\n",
		 size, K0_TO_MEMADDR(machine, sec_addr_k0), errno);
	CPUError("s_vaddr = 0x%08x\n", sec_addr_k0);
	goto error;
      }
      write_consistency(machine, sec_addr_k0, size);

    } else if ((strcmp(secHeaders[i].s_name, ".bss") == 0) ||
	       (strcmp(secHeaders[i].s_name, ".sbss") == 0)) {

      if (sec_addr_k0 < img_addr || (sec_addr_k0+size) >= img_addr+nbytes) {
	CPUError("LoadImage: section %s (va %llx nbytes %x)\r\n"
		 "           outside limit (va %x nbytes %x)\n",
		 secHeaders[i].s_name, (uint64)secHeaders[i].s_vaddr,
		 size, img_addr, nbytes);
	retval = EFAULT;
	goto error;
      }
      
      if (img_addr != ld_addr && !reloc_data)
	/* don't even load data in the case of replicated text
	 * segments sharing the same data.
	 */
	continue;

      memset(read_addr_ma, 0, size);
      write_consistency(machine, sec_addr_k0, size);

    }

  }

  if (img_addr != ld_addr) {
    VA  got_start_k0 = si[S_GOT].segStart;
    VA* got          = (VA*)K0_TO_MEMADDR(machine, got_start_k0);
    int got_size     = si[S_GOT].segSize;

    /* must relocate -- can do this only for PIC */
    if (!got_size) {
      CPUError("LoadImage: must relocate but this is not PIC (no GOT).\n");
      retval = EFAULT;
      goto error;
    }
    /* go over GOT and perform relocation */
    for (i = 0; i < got_size/sizeof(char*); i++) {
      if (IN_SEG(got[i], S_TEXT)) {
	got[i] += ld_addr - img_addr;
	continue;
      }
      if (!reloc_data)
	continue;
      if (IN_SEG(got[i], S_DATA) ||
	  IN_SEG(got[i], S_RODATA) ||
	  IN_SEG(got[i], S_BSS)) {
	got[i] += ld_addr - img_addr;
	continue;
      }
      CPUError("LoadImage: bad address in GOT 0x%08x.\n", got[i]);
      retval = EFAULT;
      goto error;
    }
    /* make sure GOT is consistent in the caches */
    write_consistency(machine, got_start_k0, got_size);
  }

  retval = 0;
  goto done;
  
error:
done:
  (void) close(fd);
  return retval;
}


/*
 * This function is very similar to LoadImage, excet for range checking and placement
 */
static int
LoadFPROMImage(char* pathname, VA promAddr, char * promSimAddr,
               VA ramAddr, char * ramSimAddr)
{
  int fd, i;
  int retval;
  HeaderBuf header;
  FILHDR  fileHeader;
  AOUTHDR aoutHeader;
  SCNHDR  secHeaders[MAX_SCNS];
  bool    elfFormat;
  FILHDR *hdrPtr;
  VA      entry;
  
  retval = LoadHeaders(pathname, &header, &fileHeader, &elfFormat, 
		       &aoutHeader, secHeaders, &fd,&entry);
  if (retval != 0) {
     return retval;
  }

  if (elfFormat) {
    hdrPtr = &fileHeader;
  } else {
    hdrPtr = &(header.hb.coffHeader);
  }
  
  for (i = 0; i < hdrPtr->f_nscns; i++ ) {
     
     if ((strcmp(secHeaders[i].s_name, ".text") == 0) ||
         (strcmp(secHeaders[i].s_name, ".data") == 0) ||
	 (strcmp(secHeaders[i].s_name, ".sdata") == 0)||
         (strcmp(secHeaders[i].s_name, ".rdata") == 0)||
	 (strcmp(secHeaders[i].s_name, ".bss") == 0)  ||
	 (strcmp(secHeaders[i].s_name, ".sbss") == 0)) {

        VA    startaddr;
        char *destAddr;
        int   nbytes;

	VA    s_vaddr = (VA)secHeaders[i].s_vaddr;
	int   s_size  = (int)secHeaders[i].s_size;

        if (strcmp(secHeaders[i].s_name, ".text") == 0) {
           startaddr = promAddr;
           nbytes    = FPROM_SIZE;
           destAddr  = promSimAddr;
        } else {
           /* data or bss */
           startaddr = ramAddr;
           nbytes    = FRAM_SIZE;
           destAddr  = ramSimAddr;
        }
        
        if (s_vaddr < startaddr || (s_vaddr+s_size) >= startaddr + nbytes) {
           CPUError("LoadFPROMImage: section %s (va %x nbytes %x)\r\n"
                       "                outside limit (va %x nbytes %x)\n",
                       secHeaders[i].s_name, s_vaddr, s_size,
                       startaddr, nbytes);
           retval = EFAULT;
           goto error;
        }

	if (!secHeaders[i].s_scnptr)
	  /* empty - skip */
	  continue;

        if (lseek(fd, secHeaders[i].s_scnptr, SEEK_SET) !=
	      secHeaders[i].s_scnptr) {
           retval = errno;
           Sim_Warning("Bad section seek in LoadFPROMImage\n");
           goto error;
        }

        if (read(fd, destAddr + (s_vaddr - startaddr), s_size) != s_size) {
           retval = errno;
           Sim_Warning("Bad section read in LoadFPROMImage\n");
           Sim_Warning("s_vaddr = 0x%08x\n", s_vaddr);
           goto error;
        }
     } else if (strcmp(secHeaders[i].s_name, ".comment") == 0) {
        /* ignore */
     } else {
        /* Elf has a lot more section types that are not listed here. */
        if (!elfFormat) {
           CPUError("Bad section type %s in FPROM.\n", secHeaders[i].s_name);
        }
     }
  }
  
  retval = 0;
  
error:
  (void) close(fd);
  return retval;
}

/*
 * Copy bootline arguments (for master CPU) into kernel memory so that
 * kernel will be able to read them.
 */
static void
store_mem_word(int cpu, VA addr, PromWord val)
{
  PromWord valBE;


  if (sizeof(PromWord) == 4) {
     valBE = HO2BE_4(val);
  } else {
     valBE = HO2BE_8(val);
  }
   
  ASSERT(IS_KSEG0(addr));
  if (sim_misc.myCPUType == MIPSY) {
     MemRefDebugWriteData(cpu,addr,K0_TO_PHYS(addr),(char*)&valBE,sizeof(PromWord));
  } else {
    *((PromWord*)K0_TO_MEMADDR(M_FROM_CPU(cpu), addr)) = valBE;
  }
}

static void
store_mem_byte(int cpu, VA addr, char c)
{
  ASSERT(IS_KSEG0(addr));
  if (sim_misc.myCPUType == MIPSY) {
     MemRefDebugWriteData(cpu,addr,K0_TO_PHYS(addr),(char*)&c,sizeof(char));
  } else {
     *((char*)K0_TO_MEMADDR(M_FROM_CPU(cpu), addr)) = c;
  }
}

static void
copy_bootline_args(int machine, int argc, char** argv,
		   VA* argcVA, VA* argvVA)
{
    int cpu  = FIRST_CPU(machine);
    VA argz  = (K0BASE + ARGZONE_OFFS);
    VA strp  = argz + (argc + 2)*sizeof(VA);
    VA argvp = argz + sizeof(VA);
    int i, j;
    
    *argcVA = argz;
    *argvVA = argvp;

    /* first word: argc */
    store_mem_word(cpu, argz, (PromWord)argc);
    
    for (i = 0; i < argc; i++) {
        /* store pointer in argv zone */
        store_mem_word(cpu, argvp + (sizeof(PromWord)*i), (PromWord)strp);
      
        /* copy argument string */
        for (j = 0; argv[i][j]; j++) {
            store_mem_byte(cpu, strp, argv[i][j]); 
            strp++;
        }
        store_mem_byte(cpu, (VA)strp, 0); 
        strp++;
    }
    /* store final 0 in argv */
    store_mem_word(cpu, argvp+sizeof(PromWord)*i, 0); 
}


/*
 * Enter reset vector in boot prom.
 */
static void
JumpToResetVec(void)
{
  int            firstCPU;
  VA             entry=FPROM_BASE;
  int            machine;
  
  for (machine = 0; machine < NUM_MACHINES; machine++) {

      /* This deterministically enters detailed mode so that every */
      /* instruction is emulated */
      firstCPU = FIRST_CPU(machine);
      SBase[firstCPU].R[4] = 0;
      SBase[firstCPU].R[5] = 0;
      SBase[firstCPU].R[6] = 0;
      
      SBase[firstCPU].PC   = entry;
      
      CPUWarning( "Fprom image machine=%d pc=%llx \n", machine,
                  (uint64)SBase[firstCPU].PC);
  
      SIM_DEBUG(('g', "Fprom image machine=%d pc=%x \n", machine,
                 SBase[firstCPU].PC));
      
  }
  
  SimulatorEnter(simosCPUType, 0, 0);
  ASSERT(0);
}


/*
 * Read and start the program.
 */
static void
LoadNgo(char *pathname, int argc, char **argv)
{
  int            retval;
  int            firstCPU;
  VA             entry=0;
  VA             argcVA, argvVA;
  TagList        tags;
  VA             endaddr;
  bool           elfFormat = FALSE;
  int            machine;
  
  for (machine = 0; machine < NUM_MACHINES; machine++) {

      retval = LoadImage(pathname,
                         machine, 
                         (VA)K0BASE, /* img addr */
			 (VA)K0BASE, /* ld addr: no relocation */
/* this used to be mem_size/num_cells, which is still valid for SIM32 but not
   for SIM64 sparse addresses.  But this should still work for SIM32. - KG */
                         MEM_SIZE(machine) / NUM_CPUS(machine),
			 FALSE,
                         &elfFormat,
                         &entry);
  
#ifdef SIM_LINUX
      if (elfFormat) {
          
          retval = 0;
          
          TagInit(machine, &tags);
          
          /* copy tag data to end of vmem --ejs */
          endaddr = (VA)K0BASE +
              (MEM_SIZE(machine) / NUM_CELLS(machine));
          if (memcpy(K0_TO_MEMADDR(machine,
                                   BE2HO_4(endaddr - sizeof(TagList))), &tags,
                     sizeof(TagList)) == NULL) {
              retval = errno;
              Sim_Warning("Bad tags copy in LoadNgo\n");
              Sim_Warning("vaddr = 0x%08x\n", BE2HO_4(endaddr - sizeof(TagList)));
              perror(pathname);
              return;
          }
      }
#endif
      
      if (retval != 0) {
          perror(pathname);
          return;
      }
      
      /* Must copy argc / argv to simulated memory so that kernel
       * can read the arguments. 
       */
      copy_bootline_args(machine, argc, argv, &argcVA, &argvVA);
      
      /* CPUWarning("Boot args copied to 0x%08x 0x%08x\n", argcVA, argvVA); */
      
      /* This deterministically enters detailed mode so that every */
      /* instruction is emulated */
      firstCPU = FIRST_CPU(machine);
      SBase[firstCPU].R[4] = argc;
      SBase[firstCPU].R[5] = argvVA;
      SBase[firstCPU].R[6] = 0;
      
      SBase[firstCPU].PC   = entry;
      
      CPUWarning( "Load image machine=%d pc=%llx argc=%x argv = %x \n", machine,
                  (uint64)SBase[firstCPU].PC, SBase[firstCPU].R[4], SBase[firstCPU].R[5]);
  
      SIM_DEBUG(('g', "Load image machine=%d pc=%x argc=%x argv = %x \n", machine,
                 SBase[firstCPU].PC, SBase[firstCPU].R[4], SBase[firstCPU].R[5]));
      
  }
  
  SimulatorEnter(simosCPUType, 0, 0);
  ASSERT(0);
}


/*
 * Launch slave cpu :  can get here through two different paths:
 *    from simmagic.c:MAGIC_PPC_OP_STARTSLAVENODE  for setting the
 *               start address of a slave cpu
 *    from tcl_init.c:hive launchslave  for passing arguments to
 *               a cell when booting it
 *
 * Returns:
 *  0 - success
 *  1 - failure, this CPU has already been launched.
 */

int
LaunchSlave(int cpu, VA pc,
	    Reg a0, Reg a1,Reg a2, Reg a3)
{
   ASSERT(0 <= cpu && cpu < TOTAL_CPUS);
#if !defined(IRIX6_4)
   ASSERT(IS_KSEG0(pc));
#endif

   if (SBase[cpu].cpuStatus == cpu_running ||
       SBase[cpu].outOfSlaveLoop != 0)
     return 1; /* already running */

   CPUPrint("LaunchSlave: cpu %d PC 0x%08x "
    	       "a0 0x%08x a1 0x%08x a2 0x%08x a3 0x%08x\n", 
    	       cpu, pc, a0, a1, a2, a3);
#define EMBRA_PARALLEL 1

#if !(EMBRA_PARALLEL)
   SBase[cpu].cpuStatus      = cpu_running;    /* mipsy polls this each cycle */
   SBase[cpu].PC             = pc;
   SBase[cpu].nPC            = 4 + pc;
#endif
   
#if !defined(SIM_MIPS64)
   SBase[cpu].R[4]           = (int)a0;
   SBase[cpu].R[5]           = (int)a1;
   SBase[cpu].R[6]           = (int)a2;
   SBase[cpu].R[7]           = (int)a3;
#else
   SBase[cpu].R[4]           = a0;
   SBase[cpu].R[5]           = a1;
   SBase[cpu].R[6]           = a2;
   SBase[cpu].R[7]           = a3;
   SBase[cpu].CP0[C0_SR] = SR_KX; /* 64bit kernel mode with interrupt disabled */
   SBase[cpu].cpuMode = KERNEL_MODE;
   SBase[cpu].is32bitMode = FALSE;
#endif

#if EMBRA_PARALLEL
   /* set launch address!! */
   sim_misc.launchAddr[cpu] =  pc;
#else
 /* For now, we don't want to do this for parallel embra??!! */
   SBase[cpu].outOfSlaveLoop = 1;              /* embra polls this each context switch */
#endif

   return 0;
}



int
ResetCPUs(int firstcpu, int lastcpu)
{
   /* still to be done: reset interrupts, pending sips, etc. */
   int err = 0;
   if (lastcpu != firstcpu) {
      Sim_Warning("SIMOS: reset cpus %d through %d\n", firstcpu, lastcpu);
   } else {
      Sim_Warning("SIMOS: reset cpu %d\n", firstcpu);
   }
   for  (;firstcpu <= lastcpu; firstcpu++) {
      if ((err = CPUVec.ResetCPU(firstcpu)))
         break;
      SBase[firstcpu].cpuStatus = cpu_not_booted;   /* mipsy now skips this cpu */
      SBase[firstcpu].outOfSlaveLoop = 0;           /* embra now skips this cpu */
   }
   return err;
}


void
SimHalt(void)
{
  exit(0);
}


#define MAX_IMG_NAME 256

/* Tcl interface into hive-specifics. */
int
HiveCmd(ClientData clientData, Tcl_Interp *interp,
        int argc, char *argv[])
{
   if (argc <= 2) {
     Tcl_AppendResult(interp, "hive function args...\n", NULL);
     return TCL_ERROR;  
   }

   if (!strcmp(argv[1], "loadimage")) {
     char*          img_name;
     VA             start = 0;
     unsigned       len   = 0;
     bool           format;
     VA             entryAddr;
     char           result[32];

     if (argc != 5) {
       Tcl_AppendResult(interp, "hive loadimage name start len\n", NULL);
       return TCL_ERROR;  
     }
     img_name = argv[2];
     Tcl_GetInt(interp, argv[3], (int*)&start);
     Tcl_GetInt(interp, argv[4], (int*)&len);

     if (!IS_KSEG0(start) || !IS_KSEG0(start+len)) {
       Tcl_AppendResult(interp,
			"hive loadimage bad address range ",
			argv[3], " ", argv[4], "\n", NULL);
       return TCL_ERROR;
     }

     /* XXX 64bit */
     if (LoadImage(img_name,
		   0,     /* machine */
		   start, /* img addr */
		   start, /* ld addr */
		   len,
		   FALSE, /* data reloc */
		   &format,
		   &entryAddr) != 0) {
       Tcl_AppendResult(interp,
			"hive loadimage can't load img ",img_name,"\n",NULL);
       return TCL_ERROR;
     }
    
     /* return entrypoint address (in hex) */
     sprintf(result, "0x%08llx", (Reg64)entryAddr);
     Tcl_AppendResult(interp, result, NULL);
     return TCL_OK;

   } else if (!strcmp(argv[1], "launchslave")) {
     int cpu   = 0;
     VA  entry = 0;
     VA  argvp = 0;
     VA  env   = 0;

     if (argc != 6) {
       Tcl_AppendResult(interp, "hive launchslave cpu entry argv env\n", NULL);
       return TCL_ERROR;  
     }
     Tcl_GetInt(interp, argv[2], (int*)&cpu);
     Tcl_GetInt(interp, argv[3], (int*)&entry);
     Tcl_GetInt(interp, argv[4], (int*)&argvp);
     Tcl_GetInt(interp, argv[5], (int*)&env);

     if (!IS_KSEG0(entry) || !IS_KSEG0(argvp) ||
	 (!IS_KSEG0(env) && env)) {
       Tcl_AppendResult(interp,
			"hive launchslave bad address ",
			argv[3]," ",argv[4]," ",argv[5],"\n",NULL);
       return TCL_ERROR;
     }

     if (LaunchSlave(cpu, (VA)PTR_TO_UINT64(entry), 0, 
                    argvp, env, 0) != 0) {
       Tcl_AppendResult(interp, "1", NULL);    /* failure */
     } else {
       Tcl_AppendResult(interp, "0", NULL);    /* success */
     }

     return TCL_OK;

   } else if (!strcmp(argv[1], "resetcpus")) {

     uint firstcpu   = 0;
     uint lastcpu    = 0;

     if (argc != 4) {
       Tcl_AppendResult(interp, "hive resetcpus firstcpu lastcpu\n", NULL);
       return TCL_ERROR;  
     }
     Tcl_GetInt(interp, argv[2], (int*)&firstcpu);
     Tcl_GetInt(interp, argv[3], (int*)&lastcpu);
     
     if (firstcpu > lastcpu
         || lastcpu >= TOTAL_CPUS) {
       Tcl_AppendResult(interp,
			"hive resetcpus bad cpupair ",
			argv[2]," ",argv[3], "\n",NULL);
       return TCL_ERROR;
     }

     if (ResetCPUs(firstcpu, lastcpu) != 0) {
       Tcl_AppendResult(interp, "1", NULL);    /* failure */
     } else {
       Tcl_AppendResult(interp, "0", NULL);    /* success */
     }
     return TCL_OK;

   } else {
     Tcl_AppendResult(interp, "bad hive function\n", NULL);
     return TCL_ERROR;  
   }
}

/*
 * Main loop for master processor startup
 */
void
SimPromEnter(void)
{
  char line[MAXBOOTARGBYTES];
  int  retval;
  int  i;
  char prompt[64];
  static int prompt_no = 1;
  int console = 0;
  char usage[]    = "\n\rBoot PROM commands\n\r"
                    "------------------\n\r"
                    "b pathname - Boot the kernel\n\r"
                    "r - Reset\n\r"
                    "q - Quit SimOS\n\r"
                    "h - Halt the machine\n\r"
                    "d - Enter the debugger\n\r"
                    "? - This help listing\r\n\n";

  /* indicate that this cpu has launched for other people to check */
  for (i = 0; i < NUM_MACHINES; i++) {
      sim_misc.launchAddr[FIRST_CPU(i)] =  (Reg) &SimPromEnter; /* BOGUS */
  }

  while (1) {
      sprintf(prompt, "\rsimosboot (%d)> ", prompt_no++);
      sim_console_write(console, prompt, (uint)strlen(prompt));
      
      for (i = 0; i < MAXBOOTARGBYTES; i++) {
         /*
           * Expect hooks
           *  - expect characters take priority over user typed chars
           */
          if ((line[i] = ExpectGetChar(console))) {
              sim_console_write(console, &line[i], 1);
          } else {
              line[i] = retval = sim_console_in(console); /* hmmm */
              if (retval == -1) {
#ifdef sgi
                  sginap(0);
#elif sun
                  sleep(1);
#endif
                  i--;
                  continue;
              }
              sim_console_write(console, &line[i], 1);
          }
          if (line[i] == '\n' || line[i] == '\r') {
              i++;
              sim_console_write(console, "\r\n", 2);
              break;
          }
          if (line[i] == 'C'-'@') {
              sim_console_write(console, "\r\n", 2);
              exit(0);
          }
      }
      
      line[i] = '\000';
      
      switch (line[0]) {
      case 'b':
      case 'B': {    /* B pathname  - Boot the specified kernel */
          char *arg, *file;
          char* argv[MAXBOOTARGS+1]; /* list of arguments */
          int   argc;                /* argcount (for kernel) */
          
          for (arg = line + 1; *arg == ' '; arg++) ; /* skip spaces */
          if (!*arg) { 
              sim_console_write(console, "Must specify kernel file\n", 25);
              break;
          } else {
              file = arg;
              while (*arg && !isspace(*arg)) arg++;
              argc = 0;
              if (*arg) {
                  *arg = 0; /* 0-terminate file name */
                  arg++;
                  for(argc = 0; argc < MAXBOOTARGS; argc++) {
                      for ( ; isspace(*arg); arg++) ; /* skip spaces */
                      if (!*arg) break;
                      argv[argc] = arg;
                      while (*arg && !isspace(*arg)) arg++;
                      if (*arg) { *arg = 0; arg++; }
                  }
              }
          }
          
          LoadNgo(file, argc, argv);
          
          break;
      }
      
      case 'r':
      case 'R': {  /* Reset - Execute from the reset vector. */
	  sim_console_write(console, "Executing from the reset vector\n", 32);
          JumpToResetVec();
          break;
      }

      case 'h':
      case 'H': {    /* H - Halt the simulated machine. */
          SimHalt();
          break;
      }
      case 'd':
      case 'D': {
          kill((int)getpid(), SIGUSR1);
          break;
      }
      case '?': {
          sim_console_write(console, usage, sizeof(usage)-1);
          break;
      }
      case 'q':
      case 'Q':
          sim_console_write(console, "\r\n", 2);
          exit(0);
          break;
      default:
          sim_console_write(console, "Unknown command\n", 16);
      }
  }
}

#ifdef IRIX6_4
static PA nodeMemSize[MAX_MACHINES][SIM_MAXCPUS];
static MA nodeMemStartMA[MAX_MACHINES][SIM_MAXCPUS];
static PA memPerNode[MAX_MACHINES];

static void 
InitMemLayout(void)
{
   int i, machine;
   /* XXX for now assume each node has same amount of memory
      and that it's contiguous in one bank.  This will be changed soon */
   for (machine=0; machine<NUM_MACHINES; machine++) {
      MA startMA = SIM_MEM_ADDR(machine);
#ifdef SIM_ORIGIN
      int numnodes = (NUM_CPUS(machine) + 1)/ 2;
#else
      int numnodes = NUM_CPUS(machine);
#endif
      memPerNode[machine] = MEM_SIZE(machine)/numnodes;

      for (i=0; i<numnodes; i++) {
         nodeMemSize[machine][i] = MEM_SIZE(machine)/numnodes;
         /* Make sure the memory size is a multiple of the page size */
         ASSERT(nodeMemSize[machine][i] % 16384 == 0);
         nodeMemStartMA[machine][i] = startMA;
         startMA += MEM_SIZE(machine)/numnodes;
      }
   }
   
}

int
IsValidPA(int machNo, PA pAddr)
{
   int node = pAddr >> MAGIC_NODE_SIZE_BITS;
   PA nodeOffset = pAddr & MAGIC_NODE_OFFSET_MASK;
   if (nodeOffset >= nodeMemSize[machNo][node]) {
      return 0;
   } else {
      return 1;
   }
}

MA
PhysToMemAddr(int machNo, PA pAddr)
{
   int node = pAddr >> MAGIC_NODE_SIZE_BITS;
   PA nodeOffset = pAddr & MAGIC_NODE_OFFSET_MASK;

   ASSERT(node < SIM_MAXCPUS);
   /* NOTE that '>' is used instead of '>=' because this function is 
      also called to get the end addr */
   if (nodeOffset > nodeMemSize[machNo][node]) {
      ASSERT(0);
      return 0; /* for compiler warning */
   } else {
      return nodeMemStartMA[machNo][node] + nodeOffset;
   }
}

PA
MemToPhysAddr(int machNo, MA mAddr)
{
   PA machineOffset = (PA)(mAddr - SIM_MEM_ADDR(machNo));
   PA node = machineOffset / memPerNode[machNo];
   PA nodeOffset = (PA)(mAddr - nodeMemStartMA[machNo][node]);

   ASSERT(IS_VALID_MA(machNo, mAddr));
   
   return (node << MAGIC_NODE_SIZE_BITS) | nodeOffset;
}

#endif /* IRIX6_4 */
